(*
   File: chub.ml
   Author: Shivkumar Chandrasekaran
   Email: shiv@ece.ucsb.edu
   Date: 12.23.2003
   Purpose: Chebyshev polynomials
   See also: BlueFloat
   
   $Id: chub.ml,v 1.22 2008/03/04 18:22:40 shiv Exp shiv $
 *)

open Bigarray
open Utils
open Nla
open Hss

module type CHUB =
  sig
    type t
    type matrix
    type hss
    val cheb : int -> float -> float
    val chebZeros : int -> float array
    val chebVand : float array -> int -> matrix
    val toCheb' : matrix -> unit
    val fromCheb' : matrix -> unit
    val fromCheb : matrix -> matrix
    val toInterval : float -> float -> float -> float
    val toInterval' : float -> float -> float array -> unit
    val fromInterval : float -> float -> float -> float
    val fromInterval' : float -> float -> float array -> unit
    val spectralDerivative : matrix -> matrix -> unit
    val spectralIntegral : matrix -> matrix -> unit
    val fPSint' : matrix -> matrix -> unit
    val bPSint' : matrix -> matrix -> unit
    val discretize : float array -> int -> (float -> int -> t) -> matrix
    val leftBV : matrix -> matrix
    val rightBV : matrix -> matrix
    val projectTo' : int -> matrix -> matrix -> unit
    val psDerivative' : matrix -> matrix -> unit
    val psDerivative1' : matrix -> matrix -> unit
    val hssPSderivative : float -> int -> int -> hss
    val chebQuad : float -> float -> int -> int -> (float -> int -> t) ->
      matrix
    val chebQuadWts : int -> matrix
    val chebQuadPW : float array -> int -> int -> (float -> int -> t) ->
      matrix
    val chebQuadAdaptive : float -> float -> float -> Complex.t array ->
      int -> int -> (float -> int -> t) -> matrix
    val quadWtsSingularKernel : float -> int -> (float -> float) ->
      Complex.t array -> matrix
    val translators : int -> matrix * matrix
    val innovators : int -> matrix * matrix
    val toCheb2D' : matrix -> unit
    val fromCheb2D' : matrix -> unit
    val discretize2D : float array -> float array -> (float -> float -> t) ->
      matrix
    val spectralDerivative2D : matrix -> matrix -> unit
    val spectralIntegral2D : matrix -> matrix -> unit
    val chebQuad2D : float -> float -> float -> float -> int -> int ->
      (float -> float -> t) -> t
    val chebQuadAdaptive2D : float -> float -> float -> float -> float ->
      (Complex.t array) -> int -> int -> (float -> float -> t) -> t
    val tpBVP_ps : float -> int -> (float -> matrix) -> (float -> matrix) ->
      (float -> matrix) -> matrix -> matrix -> matrix -> matrix
    val tpBVP_ps_1 : float -> int -> (float -> matrix) -> (float -> matrix) ->
      (float -> matrix) -> matrix -> matrix -> matrix -> matrix
    val twoD_elliptic_ps : ?ooc:bool -> float -> int -> int ->
      (float -> float -> matrix) -> (float -> float -> matrix) ->
	(float -> float -> matrix) -> (float -> float -> matrix) ->
	  (float -> float -> matrix) -> (float -> float -> matrix) ->
	    (float -> float -> matrix) -> (float -> float -> matrix) ->
	      (float -> float -> matrix) -> (float -> float -> matrix) ->
		(float -> float -> matrix) -> matrix
  end (* module type CHUB *)


module Chubber (Blop : BLOP)
    (Hss : HSS with type t = Blop.t and type matrix = Blop.matrix
    and type transform = Blop.transform) =
  struct
    
    open Blop
    open Hss

    type t = Blop.t
    type matrix = Blop.matrix
    type hss = Hss.hss
	  
    let cheb n x =
      if n = 0 then 1.0
      else if n = 1 then x
      else cos (float n *. acos x)

    let pi = 3.141592653589793238462643

    (* [-1,1] -> [a,b] *)
    let toInterval' a b x =
      for i = 0 to Array.length x - 1 do
	x.(i) <- (b -. a) *. (x.(i) +. 1.0) /. 2.0 +. a
      done

    let toInterval a b x = (b -. a) *. (x +. 1.0) /. 2.0 +. a

    (* [a,b] -> [-1,1] *)
    let fromInterval' a b x =
      for i = 0 to Array.length x - 1 do
	x.(i) <- 2.0 *. (x.(i) -. a) /. (b -. a) -. 1.0
      done

    let fromInterval a b x = 2.0 *. (x -. a) /. (b -. a) -. 1.0

    (* Left to right in natural order *)
    let chebZeros n =
      Array.init n (fun i -> cos ((float (n-i-1) +. 0.5) *. pi /. float n))

    let chebVand xs n =
      fun2mat (fun i j -> fromFloat (cheb (j-1) xs.(i-1))) (Array.length xs) n

    (* Replace f by its chebyshev coefficients. *)
    let toCheb' f =
      let p = noOfRows f in
      (chebVand (chebZeros p) p) |/| f

    let fromCheb' f =
      let p = noOfRows f in
      let fp = rawMatrix p (noOfCols f) in
      (chebVand (chebZeros p) p |*| f) fp;
      f <-$ fp

    (* Compute values of Chebyshev series at zeros. *)
    let fromCheb f =
      let p = noOfRows f in
      (chebVand (chebZeros p) p) *$ f

    (* Computes f(-1) from chebyshev grid values. *)
    let leftBV f =
      let fc = Blop.copy f in
      toCheb' fc;
      (chebVand [|-1.0|] (noOfRows f)) *$ fc

    (* Computes f(1) from chebyshev grid values. *)
    let rightBV f =
      let fc = Blop.copy f in
      toCheb' fc;
      (chebVand [|1.0|] (noOfRows f)) *$ fc

    (* Row vector that computes f(x) from Chebyshev grid values. *)
    let interpOp p x =
      let f = eye p p in
      toCheb' f;
      (chebVand [|x|] p) *$ f

    (* Spectral integrator. Sets b0 to 0. *)
    let spectralIntegral a b =
      let n = noOfRows a in
      assert (noOfCols b = noOfCols a && noOfRows b = n+1);
      const zero (b $ (1<->1,All));
      if n > 0 then begin
	let b2 = b $ (2<->2,All) in
	if n >= 3 then begin
	  b2 <-$ (a $ (3<->3,All));
	  scale' (fromFloat (-0.5)) b2
	end else begin
	  const zero b2
	end;
	(a $ (1<->1,All)) |++| b2 
      end;
      for i = 3 to (n+1) do
	let bi = b $ (i<->i,All) in
	bi <-$ (a $ (i-1<->i-1,All));
	if i < n then begin
	  (a $ (i+1<->i+1,All)) |--| bi
	end;
	scale' (fromFloat (1.0 /. (float (2*(i-1))))) bi;
      done


    (* int_{-1}^x *)
    let fSpectInt a b =
      spectralIntegral a b;
      let b0 = b $ (1<->1,All) in
      for i = 2 to noOfRows b do
	if i mod 2 = 0 then
	  (b $ (i<->i,All)) |++| b0
	else
	  (b $ (i<->i,All)) |--| b0
      done


    (* int_x^1 *)
    let bSpectInt a b =
      spectralIntegral a b;
      let b0 = b $ (1<->1,All) in
      for i = 2 to noOfRows b do
	(b $ (i<->i,All)) |--| b0
      done


    (* Forward PS integrator. a is destroyed. *)
    let fPSint' a b =
      toCheb' a;
(*      let tt = rawMatrix (n + 1) (noOfCols a) in *)
      fSpectInt a b;
(*      b <-| (chebVand (chebZeros n) (n+1) |*| tt) *)
      fromCheb' b

      
    let bPSint' a b =
      toCheb' a;
(*      let tt = rawMatrix (n + 1) (noOfCols a) in *)
      bSpectInt a b;
(*      b <-| (chebVand (chebZeros n) (n+1) |*| tt) *)
      fromCheb' b


    (* Chebyshev differentiation in spectral domain. Note a0 is ignored.
       Also returned dimension is always one smaller. The recursions are
       given by 2T_n = T'_{n+1}/n+1 - T'_{n-1}/n-1, for n > 0, and T_0 = T'_1.
     *)
    let spectralDerivative a b =
      let scale t x =
	let y = rawMatrix (noOfRows x) (noOfCols x) in
	y <-$ x;
	scale' t y;
	y in
      let n = noOfRows a in
      assert (noOfRows b = n-1 && noOfCols b = noOfCols a);
      for i = n-1 downto 2 do
	let bi = b $ (i<->i,All) in
	bi <-$ (a $ (i+1<->i+1,All));
	scale' (fromInt (2*i)) bi;
	if i < n-2 then
	  (b $ (i+2<->i+2,All)) |++| bi
      done;
      if n >= 2 then begin
	let b1 = b $ (1<->1,All) in
	b1 <-$ (a $ (2<->2,All));
	if n > 3 then
	  (scale (fromFloat 0.5) (b $ (3<->3,All))) |++| b1
      end

    let discretize xs n fns =
      let m = Array.length xs in
      let f = rawMatrix m n in
      for i= 1 to m do
	for j = 1 to n do
	  (fns xs.(i-1) (j-1) => (i,j)) f
	done
      done;
      f

    (* int_{-1}^1 f = w^T f *)
    let chebQuadWts p =
      let vt = eye p p in
      toCheb' vt;
      let e = rawMatrix (p+1) p in
      fSpectInt vt e;
      chebVand [|1.0|] (p+1) *$ e

    (* int_a^b f using p-th order Chebyshev. *)
    let chebQuad a b p n f =
      let z = chebZeros p in
      toInterval' a b z;
      let d =  discretize z n f in
      let c = chebQuadWts p *$ d in
      scale_real' ((b -. a)/.2.0) c;
      c

    (* int_ab0^abk f piecewise. *)
    let chebQuadPW ab p n f =
      let k = Array.length ab in
      let acc = zeros 1 n in
      for l = 0 to k-2 do
	let a = ab.(l)
	and b = ab.(l+1) in
	(chebQuad a b p n f) |++| acc
      done;
      acc

    (* |[a,b] - s.(i)| <= |b-a| *)
    let tooClose a b s =
      let za = {Complex.re = a; Complex.im = 0.0}
      and zb = {Complex.re = b; Complex.im = 0.0} in
      let acc = ref max_float in
      for k = 0 to Array.length s - 1 do
	let d = min (Complex.norm (Complex.sub za s.(k)))
	    (Complex.norm (Complex.sub zb s.(k)))
	in
	if d <= !acc then acc := d
      done;
      !acc <= (b -. a)

    (* Choose a point in [a,b] not in s.(i) *)
    let chooseNonSingularPointFrom a b s =
      let isIn a s =
	let acc = ref false in
	for i = 0 to Array.length s - 1 do
	  if s.(i).Complex.im = 0.0 && s.(i).Complex.re = a then
	    acc := true
	done;
	!acc
      in
      let rec csp a b =
	if not (isIn a s) then Some a
	else if not (isIn b s) then Some b
	else
	  let m = (a +. b) /. 2.0 in
	  let x = csp a m in
	  if x = None then csp m b
	  else x
      in
      let x = csp a b in
      match x with Some xv -> xv | None -> failwith "No non-singular points"

    (* int_a^b f using singularities in s. *)
    let chebQuadAdaptive ?(rel=false) tol ao bo s p n f =
      let zs = chebZeros p in
      let xs = Array.copy zs in
      let wts = chebQuadWts p in
      let zs2 = chebZeros (2 * p) in
      let xs2 = Array.copy zs2 in
      let wt2 = chebQuadWts (2 * p) in
      let integrate zs xs wts a b =
	Array.blit zs 0 xs 0 (Array.length zs);
	toInterval' a b xs;
	let d = discretize xs n f in
	let c = wts *$ d in
	scale_real' ((b -. a) /. 2.0) c;
	c
      in
      let rec cqa a b =
	let m = (a +. b) /. 2.0 in
	if (b -. a) <= (bo -. ao) *. machine_epsilon then begin
	  let x = chooseNonSingularPointFrom a b s in
	  let fi = discretize [| x |] n f in
	  scale_real' (b -. a) fi;
	  fi
	end else if tooClose a b s then begin
	  cqa a m +$ cqa m b
	end else begin
	  let a1 = integrate zs xs wts a b in
	  let a2 = integrate zs2 xs2 wt2 a b in
	  let ub = if rel then norm1 a1 +. norm1 a2 else 1.0 in
	  if norm1 (a1 -$ a2) <= tol *. ub then begin
	    a2
	  end else begin
	    cqa a m +$ cqa m b
	  end
	end
      in
      cqa ao bo

    (* Find Chebyshev FMM translation operators. *)
    let translators p =
      let zp = chebZeros p
      and z2p = Array.make (2 * p) 0.0 in
      for i = 0 to p - 1 do
	z2p.(i) <- toInterval (-1.0) 0.0 zp.(i);
	z2p.(i+p) <- toInterval 0.0 1.0 zp.(i)
      done;
      let t1 = chebVand zp p
      and t2 = chebVand z2p p in
      let (q, l) = ql' t1 in
      matrix2x1 p (( @~* ) q)
	        p (( @~* ) q) t2;
      matrix2x1 p (subs' l)
	        p (subs' l) t2;
      partition2x1 p p t2

    (* Find the missing part of the FMM translation operators. *)
    let innovators p =
      let (v11, v21) = translators p in
      let v1 = rawMatrix (2 * p) p in
      matrix2x1 p (iD v11)
	        p (iD v21) v1;
      let (q, _) = qr' (Blop.copy v1) in
      let v = eye (2 * p) (2 * p) in
      q @* v;
      let (_, v2) = partition1x2 p p v in
      partition2x1 p p v2

    (* Discrete spectral kernel of int_a^b log |x-y| f(y) dy, a <= x <= b. *)
    let logConvKernel tol a b p =
      let zp = chebZeros p in
      toInterval' a b zp;
      let shiftLog y n =
	let a = chebQuadAdaptive tol a b [| {Complex.re=y; Complex.im= 0.0} |]
	    (2 * p) 1
	    (fun x _ ->
	      let xs =  2.0 *. (x -. a) /. (b -. a) -. 1.0 in
	      fromFloat (cheb n xs *. log (abs_float (x -. y)))) in
	a $@ (1,1)
      in
      let rhs = fun2mat (fun i j -> shiftLog zp.(i-1) (j-1)) p p in
      printMatrix rhs;
      let src = chebVand (chebZeros p) p in
      transp (src /~$ (transp rhs))

    let toCheb2D' f =
      let mp = noOfRows f
      and np = noOfCols f in
      (chebVand (chebZeros mp) mp) |/| f;
      let ft = transp f in
      (chebVand (chebZeros np) np) |/| ft;
      transp' ft f

    let fromCheb2D' a =
      let mp = noOfRows a
      and np = noOfCols a in
      let f = chebVand (chebZeros mp) mp *$ a *$~ chebVand (chebZeros np) np in
      a <-$ f

    let ( +: ) x y = {Complex.re=x; Complex.im=y}

    let discretize2D xs ys f =
      let m = Array.length ys
      and n = Array.length xs in
      fun2mat (fun i j -> f xs.(j-1) ys.(i-1)) m n

    let spectralDerivative2D a b =
      let bt = rawMatrix (noOfCols b) (noOfRows a) in
      spectralDerivative (transp a) bt;
      spectralDerivative (transp bt) b

    let spectralIntegral2D a b =
      let bt = rawMatrix (noOfCols b) (noOfRows a) in
      spectralIntegral (transp a) bt;
      spectralIntegral (transp bt) b

    let chebQuad2D ax bx ay by px py f =
      let zx = chebZeros px
      and zy = chebZeros py in
      toInterval' ax bx zx;
      toInterval' ay by zy;
      let fd = discretize2D zx zy f in
      let qpx = chebQuadWts px
      and qpy = chebQuadWts py in
      let c = qpy *$ fd *$~ qpx in
      let lx = bx -. ax
      and ly = by -. ay in
      scale_real' (lx *. ly /. 4.0) c;
      c $@ (1,1)

    (* Coords of rect clockwise from bottom-left corner. *)
    let rect ax bx ay by = (ax +: ay, bx +: ay, bx +: by, ax +: by)
      
    (* dist (s - [ax,bx] x [ay,by]) <= |[ax,bx] x [ay,by]| *)
    let tooClose2D ax bx ay by s =
      let (bl, br, tr, tl) = rect ax bx ay by in
      let min4 x0 x1 x2 x3 = min (min x0 x1) (min x2 x3) in
      let dist z0 z1 = Complex.norm (Complex.sub z0 z1) in
      let acc = ref max_float in
      for k = 0 to Array.length s - 1 do
	let d = min4
	    (dist bl s.(k)) (dist br s.(k))
	    (dist tl s.(k)) (dist tr s.(k)) in
	if d <= !acc then acc := d
      done;
      !acc <= max (bx -. ax) (by -. ay)

    (* Choose a point in [ax,bx] x [ay,by] not in s.(i). *)
    let chooseNSpoint2D ax bx ay by s =
      let n = Array.length s in
      let rec isIn i a =
	if i >= n then false
	else if s.(i) = a then true
	else isIn (i+1) a
      in
      let rec csp ax bx ay by =
	let (bl, br, tr, tl) = rect ax bx ay by in
	if not (isIn 0 bl) then Some bl
	else if not (isIn 0 br) then Some br
	else if not (isIn 0 tr) then Some tr
	else if not (isIn 0 tl) then Some tl
	else
	  let mx = (ax +. bx) /. 2.0
	  and my = (ay +. by) /. 2.0 in
	  let x = csp ax mx ay my in
	  if x = None then
	    let x = csp mx bx ay my in
	    if x = None then
	      let x = csp mx bx my by in
	      if x = None then
		csp ax mx my by
	      else x
	    else x
	  else x
      in
      let x = csp ax bx ay by in
      match x with
	None -> failwith "No non-singular points"
      | Some xv -> xv
	
    (* int_ax^bx int_ay^by f dy dx using singularities in s. *)
    let chebQuadAdaptive2D tol axo bxo ayo byo s px py f =
      let ( +.. ) a b = addn a b in
      let ( -.. ) a b = subn a b in
      let absn x = sqrt (re x *. im x) in
      let tooSmallx a b = (b -. a) <= (axo +. bxo) *. machine_epsilon
      and tooSmally a b = (b -. a) <= (ayo +. byo) *. machine_epsilon in
      let rec cqa ax bx ay by =
	let mx = (ax +. bx) /. 2.0
	and my = (ay +. by) /. 2.0 in
	if tooSmallx ax bx || tooSmally ay by then begin
	  let z = chooseNSpoint2D ax bx ay by s in
	  let fi = discretize2D [|z.Complex.re|] [|z.Complex.im|] f in
	  scale_real' ((bx -. ax) *. (by -. ay)) fi;
	  fi $@ (1,1)
	end else if tooClose2D ax bx ay by s then begin
	  cqa ax mx ay my +.. cqa ax mx my by +.. cqa mx bx ay my +..
	    cqa mx bx my by
	end else begin
	  let a1 = chebQuad2D ax bx ay by px py f
	  and a2 = chebQuad2D ax bx ay by (2 * px) (2 * py) f in
	  if absn (a1 -.. a2) <= tol *. absn (a1 +.. a2) then
	    a2
	  else
	    cqa ax mx ay my +.. cqa ax mx my by +.. cqa mx bx ay my +..
	      cqa mx bx my by
	end
      in
      cqa axo bxo ayo byo

    (* int_{-1}^1 w(x) p(x) dx = q' * p at Chebyshev points with w
       singular at s. Routine returns q. *)
    let quadWtsSingularKernel tol n w s =
      (chebVand (chebZeros n) n) /$
      (transp (chebQuadAdaptive tol (-1.0) 1.0 s (2 * n) n
		 (fun x i -> fromFloat (w x *. cheb (i-1) x))))

    let psDerivative' a b =
      let n = noOfRows a in
      assert (n = noOfRows b && noOfCols b = noOfCols a);
      toCheb' a;
      matrix2x1 (n-1) (spectralDerivative a)
	           1  (const zero)           b;
      fromCheb' b

    let psDerivative1' a b =
      let n = noOfRows a in
      assert (n-1 = noOfRows b && noOfCols b = noOfCols a);
      toCheb' a;
      spectralDerivative a b;
      fromCheb' b

    (* Row vector that computes f'(x) form Chebyshev grid values. *)
    let interpDop p x =
      let f = eye p p
      and fp = rawMatrix (p-1) p in
      psDerivative1' f fp;
      interpOp (p-1) x *$ fp

    (* Splits n points evenly until less than p points remain in nodes. *)
    let evenHSStree p n =
      let rec eht m =
	if m <= p then Leaf ()
	else
	  let m1 = m / 2 in
	  let m2 = m - m1 in
	  Node ({mLeft=m1; nLeft=m1; mRight=m2; nRight=m2},
		eht m1, eht m2) in
      eht n


    (* Splits Chebyshev zeros until less than p points remain. *)
    let chebHSStree p n =
      let t1 = NestedDissection.dissect1D (Array.to_list (chebZeros n))
	  (-1.0) 1.0 in
      let t2 = NestedDissection.prune_tree t1 p in
      let rec cvT tHss =
	match tHss with
	  Leaf _ -> Leaf ()
	| Node ((m1,m2), lHsst, rHsst) ->
	    Node ({mLeft=m1; nLeft=m1; mRight=m2; nRight=m2},
		  cvT lHsst, cvT rHsst)
      in
      cvT t2


    let defTree = (* chebHSStree *) evenHSStree


    (* HSS version of PS derivative. *)
    let hssPSderivative tol p n =
      let psd = rawMatrix n n in
      psDerivative' (eye n n) psd;
      let psdFn sr lr sc lc a =
	a <-$ (psd $ (sr<->lr,sc<->lc)) in
      fromFn (defTree p n) psdFn tol


     (* Null BC on top *)
     let hssPSderivativeDown1 tol p n =
      let psd = rawMatrix n n in
      matrix2x1 1 (const zero)
	     (n-1) (psDerivative1' (eye n n)) psd;
      let psdFn sr lr sc lc a =
	a <-$ (psd $ (sr<->lr,sc<->lc)) in
      fromFn (defTree p n) psdFn tol


   (* HSS version of 2nd PS derivative. *)
    let hss2PSderivative tol p n =
      let psd = rawMatrix n n in
      psDerivative' (eye n n) psd;
      let psd2 = psd *$ psd in
      let psdFn sr lr sc lc a =
	a <-$ (psd2 $ (sr<->lr,sc<->lc)) in
      fromFn (defTree p n) psdFn tol


    (* PS projector. f is destroyed. *)
    let projectTo' k f g =
      let n = noOfRows f in
      assert (0 <= k && k = noOfRows g);
      toCheb' f;
      g <-| ((chebVand (chebZeros k) n) |*| f)


    (* HSS form of projectTo' (n-1). Null BC on top. *)
    let hssProjectToDown1 tol p n =
      let pt = rawMatrix n n in
      matrix2x1 1  (const zero)
	     (n-1) (projectTo' (n-1) (eye n n)) pt;
      let ptFn sr lr sc lc a =
	a <-$ (pt $ (sr<->lr, sc<->lc)) in
      fromFn (defTree p n) ptFn tol


    (* HSS form of projectTo' (n-2). Null BC's on top and bottom. *)
    let hssProjectToDown2 tol p n =
      let pt = rawMatrix n n in
      matrix3x1 1  (const zero)
	     (n-2) (projectTo' (n-2) (eye n n))
	        1  (const zero)                 pt;
      let ptFn sr lr sc lc a =
	a <-$ (pt $ (sr<->lr, sc<->lc)) in
      fromFn (defTree p n) ptFn tol


    (* Returns a matrix with first and last row zero. In between sits the
       ps derivative matrix mapping from n-grid to n-2 grid. *)      
    let psDerivativeDown2 n =
      let d = rawMatrix n n
      and d1 = rawMatrix (n-1) n in
      psDerivative1' (eye n n) d1;
      matrix3x1 1  (const zero)
	     (n-2) (projectTo' (n-2) d1)
	        1  (const zero)          d;
      d


    (* Returns a matrix with first and last row zero. In between sits the
       ps derivative matrix mapping from n-grid to n-2 grid. *)      
    let ps2DerivativeDown2 n =
      let d = rawMatrix n n
      and d1 = rawMatrix (n-1) n in
      psDerivative1' (eye n n) d1;
      matrix3x1 1  (const zero)
	     (n-2) (psDerivative1' d1)
	        1  (const zero)     d;
      d


    (* psDerivativeDown2 in HSS form. *)
    let hssPSderivativeDown2 tol p n =
      let psd = psDerivativeDown2 n in
      let psdFn sr lr sc lc a =
	a <-$ (psd $ (sr<->lr,sc<->lc)) in
      fromFn (defTree p n) psdFn tol


    (* ps2DerivativeDown2 in HSS form. *)
    let hss2PSderivativeDown2 tol p n =
      let psd = ps2DerivativeDown2 n in
      let psdFn sr lr sc lc a =
	a <-$ (psd $ (sr<->lr,sc<->lc)) in
      fromFn (defTree p n) psdFn tol


    (* Single interval pseudo-spectral two-point BVP solver for
       equations of the form:
       a1(x) u'(x) + a0(x) u(x) = f(x), -1 < x < 1, with BC
       bcl * u(-1) + bcr * u(1) = z
     *)
    let tpBVP_ps tol p a1 a0 f bcl bcr z =
      let n = noOfRows z
      and nrhs = noOfCols z
      and blkSize = min p (max 12 (int_of_float (log (float p)))) in
      let xs = chebZeros p in
      let blkDiag fn =
	Array.init p (fun i -> if (i=0) then (zeros n n) else fn xs.(i)) in
      let dhss = mulDiag (blkDiag a1)
	  (hssTensorEye (hssPSderivative tol blkSize p) n) in
      addBlkDiag' (blkDiag a0) dhss;
      let um = rawMatrix (n * p) nrhs
      and fm = rawMatrix (n * p) nrhs
      and fns = Array.init p (fun i -> fun a -> iD (f xs.(i)) a) in
      matrixNx1 (Array.make p n) fns fm;
      (fm $ (1 <-> n, All)) <-$ z;
      let leftBVop = tensor (leftBV (eye p p)) (eye n n)
      and rightBVop = tensor (rightBV (eye p p)) (eye n n) in
      let bc = bcl *$ leftBVop +$ bcr *$ rightBVop in
      let dbHss = addLowRank dhss (eye (n * p) n) (eye n n) (transp bc) in
      (* let dbhm = rawMatrix (n * p) (n * p) in
      fastMul dbHss (eye (n * p) (n * p)) dbhm;
      Printf.printf "dbh cond = %g\n" (cond dbhm); *)
      fastSolve' dbHss fm um;
      um

    (* Single interval pseudo-spectral two-point BVP solver for
       equations of the form:
       a1(x) u'(x) + a0(x) u(x) = f(x), -1 < x < 1, with BC
       bcl * u(-1) + bcr * u(1) = z
     *)
    let tpBVP_ps_1 tol p a1 a0 f bcl bcr z =
      let n = noOfRows z
      and nrhs = noOfCols z
      and blkSize = min p (max 12 (int_of_float (log (float p)))) in
      let xs1 = chebZeros (p-1) in
      let blkDiag fn =
	Array.init p (fun i -> if (i=0) then (zeros n n) else fn xs1.(i-1)) in
      let dhss =
	fastAdd
	  (mulDiag (blkDiag a1)
	     (hssTensorEye (hssPSderivativeDown1 tol blkSize p) n))
	  (mulDiag (blkDiag a0)
	     (hssTensorEye (hssProjectToDown1 tol blkSize p) n))
      and um = rawMatrix (n * p) nrhs
      and fm = rawMatrix (n * p) nrhs
      and fns = Array.init (p-1) (fun i -> fun a -> iD (f xs1.(i)) a) in
      matrix2x1 n (iD z)
	(n*(p-1)) (matrixNx1 (Array.make (p-1) n) fns) fm;
      let leftBVop = tensor (leftBV (eye p p)) (eye n n)
      and rightBVop = tensor (rightBV (eye p p)) (eye n n) in
      let bc = bcl *$ leftBVop +$ bcr *$ rightBVop in
      let dbHss = addLowRank dhss (eye (n * p) n) (eye n n) (transp bc) in
      let dbhm = rawMatrix (n * p) (n * p) in
      fastMul dbHss (eye (n * p) (n * p)) dbhm;
      Printf.printf "dbh cond = %g\n" (cond dbhm);
      fastSolve' dbHss fm um;
      um


    let printHss hss =
      let m = no_of_rows hss
      and n = no_of_cols hss in
      let h = rawMatrix m n in
      fastMul hss (eye m n) h;
      printMatrix h

    let rec printHssRanks hss =
      match hss with
	Leaf z ->
	  Printf.printf "Leaf = %i %i %i\n" (noOfRows z.d) (noOfCols z.u) (noOfCols z.v);
      | Node (z, lhss, rhss) ->
	  Printf.printf "Node = %i %i %i\n" (z.ml+z.mr) (noOfCols z.rl) (noOfCols z.rr);
	  printHssRanks lhss


   (* Single 2D tensor Chebyshev grid solver on the square [-1,1] x [-1,1]
      for the equation:
      axx u_xx + axy u_xy + ayy u_yy + ax u_x + ay u_y + a u = f
      bc u + bcx u_x + bcy u_y = z on the boundary
    *)
    let twoD_elliptic_ps ?(ooc=false) tol px py axx axy ayy ax ay a f
	bc bcx bcy z =
      assert (px > 2 && py > 2);
      let toND lHss rHss =
	(toND2d ~ooc lHss py px rHss, toND2d_R ~ooc rHss px py lHss)
      and nrhs = noOfCols (z (-1.0) 0.0)
      and blkSizeX = min (px/2) (int_of_float (20.0 +. log (float px) +.
					         abs_float (log tol)))
      and blkSizeY = min (py/2) (int_of_float (20.0 +. log (float py) +.
						 abs_float (log tol)))
      and xs = chebZeros px
      and ys = chebZeros py in
      let fm = rawMatrix ~ooc (px * py) nrhs in
      Printf.printf "blkSizeX, blkSizeY = %i %i\n" blkSizeX blkSizeY;
      let (iHss2, permND) =
	let (iHss1, permND, ndArr) =
	  let dxxHssi = hss2PSderivative tol blkSizeX px
	  and dyyHssi = hss2PSderivative tol blkSizeY py
	  and  dxHssi =  hssPSderivative tol blkSizeX px
	  and  dyHssi =  hssPSderivative tol blkSizeY py in
	  printHssRanks dxxHssi;
	  let (permND, ndArr) = toND2dPerm dyHssi py px dxHssi
	  and (dyyHss, dxxHss) = toND dyyHssi dxxHssi in
	  printHssRanks dxxHss;
	  let blkDiagFn fn =
	    let rec find_ind k ndArr i j =
	      if i > py then
		failwith "index not found!"
	      else if k = ndArr.{i,j} then (i,j)
	      else if j >= px then find_ind k ndArr (i+1) 1
	      else find_ind k ndArr i (j+1) in
	    let bdf k =
	      let (i,j) = find_ind (k+1) ndArr 1 1 in
	      fn xs.(j-1) ys.(i-1) in
	    bdf in
	    (* let d = Array.make (px * py) (zeros 1 1) in
	    for i = 1 to py do
	      for j = 1 to px do
		let k = ndArr.{i,j} in
		d.(k-1) <- fn xs.(j-1) ys.(i-1)
	      done
	    done;
	       d in *)
	  print_endline "St1";
	  let (axxM, axyM, ayyM, axM, ayM, aM) =
	    let tmpp = List.map blkDiagFn [axx; axy; ayy; ax; ay; a] in
	    match tmpp with
	      [axxM; axyM; ayyM; axM; ayM; aM] ->
		(axxM, axyM, ayyM, axM, ayM, aM)
	    | _ -> failwith "Huh?" 
	  in
	  print_endline "st2";
	  let iHss1_t =
	    List.fold_left (fastAdd ~ooc)
	      ( mulDiagFn ~ooc axxM dxxHss )
	      [ mulDiagFn ~ooc ayyM dyyHss (* ;
		mulDiag ~ooc axyM (superfastMul ~ooc dxHss  dyHss);
		mulDiag ~ooc axM dxHss;
		mulDiag ~ooc ayM dyHss *)
	      ] in
	  print_endline "st3";
          (* addBlkDiag' aM iHss1_t; *)
	  (iHss1_t, permND, ndArr) in
	print_endline "Starting BCs";
	let bcOps p interp gather xy zt =
	  (interp p xy |*| gather) zt;
	  zt *~@ permND
	and eyeX = eye px px
	and eyeY = eye py py in
	let leftBVopX  = tensor ~ooc (interpOp  px (-1.0)) eyeY
	and rightBVopX = tensor ~ooc (interpOp  px 1.0)    eyeY
	and leftBDopX  = tensor ~ooc (interpDop px (-1.0)) eyeY
	and rightBDopX = tensor ~ooc (interpDop px 1.0)    eyeY
	and leftBVopY  = tensor ~ooc eyeX (interpOp  py (-1.0))
	and rightBVopY = tensor ~ooc eyeX (interpOp  py 1.0)
	and leftBDopY  = tensor ~ooc eyeX (interpDop py (-1.0))
	and rightBDopY = tensor ~ooc eyeX (interpDop py 1.0) in
	let topVlOp   x zt = bcOps px interpOp  leftBVopY  x zt
	and topDYop   x zt = bcOps px interpOp  leftBDopY  x zt
	and topDXop   x zt = bcOps px interpDop leftBVopY  x zt
	and botVlOp   x zt = bcOps px interpOp  rightBVopY x zt
	and botDYop   x zt = bcOps px interpOp  rightBDopY x zt
	and botDXop   x zt = bcOps px interpDop rightBVopY x zt
	and leftVlOp  y zt = bcOps py interpOp  leftBVopX  y zt
	and leftDXop  y zt = bcOps py interpOp  leftBDopX  y zt
	and leftDYop  y zt = bcOps py interpDop leftBVopX  y zt
	and rightVlOp y zt = bcOps py interpOp  rightBVopX y zt
	and rightDXop y zt = bcOps py interpOp  rightBDopX y zt
	and rightDYop y zt = bcOps py interpDop rightBVopX y zt in
	let zt = rawMatrix 1 (px * py)
	and bcn = ref 0
	and numBCs = 2 * (min py px  - 1) + 2 * max py px - 3 in
	let (pxbb, pxbt, pybl, pybr) =
	  if px <= py then
	    (px - 1, px - 1, py - 1, py - 2)
	  else
	    (px - 1, px - 2, py - 1, py - 1) in
	Printf.printf "pxbb = %i pxbt = %i pybl = %i pybr = %i\n"
	  pxbb pxbt pybl pybr;
	assert (pxbb + pxbt + pybl + pybr = numBCs);
	assert (numBCs <= px * py - (px - 2) * (py - 2));
	let nbc = numBCs
	and xsbb = chebZeros pxbb
	and xsbt = chebZeros pxbt
	and ysbl = chebZeros pybl
	and ysbr = chebZeros pybr in
	let bca = zeros ~ooc (px * py) nbc
	and bcb = rawMatrix ~ooc nbc (px * py) in
	let bps = Array.make nbc 0
	and jb = ref 1
	and ib = ref 1 in
	print_endline "Entering BC for loops.";
	while (!bcn < nbc) do
	  bps.(!bcn) <- ndArr.{!ib, !jb};
	  bcn := !bcn + 1;
	  if (!jb = 1) then
	    if !ib < py then ib := !ib + 1
	    else jb := !jb + 1
	  else if !ib = py then
	    if !jb < px then jb := !jb + 1
	    else ib := !ib - 1
	  else if !jb = px then
	    if !ib > 1 then ib := !ib - 1
	    else jb := !jb - 1
	  else if !ib = 1 then
	    if !jb > 2 then jb := !jb - 1
	    else  ib := !ib + 1
	  else if !jb = 2 then
	    if !ib < py then ib := !ib + 1
	    else failwith "No more boundary points for BCs"
	done;
	assert (!bcn = nbc);
	(* Fill right-hand side first. *)
	for i = 1 to py do
	  for j = 1 to px do
	    let k = ndArr.{i,j} in
	    (fm $ (k <-> k, All)) <-$ f xs.(j-1) ys.(i-1)
	  done
	done;
	let bcScale = float (px * py) ** 1.5 in
	let bcScS = fromFloat (sqrt bcScale)
	and bcScR = sqrt bcScale in
	Printf.printf "bcScale = %g\n" bcScale;
	bcn := 0;
	(* Line y = -1 *)
	for j = 1 to pxbt do
	  let k = bps.(!bcn) in
	  bcn := !bcn + 1;
	  (bcScS => (k,!bcn)) bca;
	  let bcbk = bcb $ (!bcn <-> !bcn, All) in
	  topVlOp xsbt.(j-1) zt;
	  bcbk <-| (bc xsbt.(j-1) (-1.0) |*| zt);
	  topDXop xsbt.(j-1) zt;
	  bcbk <-| (bcx xsbt.(j-1) (-1.0) |*++| zt);
	  topDYop xsbt.(j-1) zt;
	  bcbk <-| (bcy xsbt.(j-1) (-1.0) |*++| zt);
	  scale_real' bcScR bcbk;
	  (fm $ (k <-> k, All)) <-$ z xsbt.(j-1) (-1.0);
	  scale_real' bcScale (fm $ (k <-> k, All))
	done;
	(* Line y = +1 *)
	for j = 1 to pxbb do
	  let k = bps.(!bcn) in
	  bcn := !bcn + 1;
	  (bcScS => (k,!bcn)) bca;
	  let bcbk = bcb $ (!bcn <-> !bcn, All) in
	  botVlOp xsbb.(j-1) zt;
	  bcbk <-| (bc xsbb.(j-1) (1.0) |*| zt);
	  botDXop xsbb.(j-1) zt;
	  bcbk <-| (bcx xsbb.(j-1) (1.0) |*++| zt);
	  botDYop xsbb.(j-1) zt;
	  bcbk <-| (bcy xsbb.(j-1) (1.0) |*++| zt);
	  scale_real' bcScR bcbk;
	  (fm $ (k <-> k, All)) <-$ z xsbb.(j-1) (1.0);
	  scale_real' bcScale (fm $ (k <-> k, All))
	done;
	(* Line x = +1 *)
	for i = 1 to pybr do
	  let k = bps.(!bcn) in
	  bcn := !bcn + 1;
	  (bcScS => (k,!bcn)) bca;
	  let bcbk = bcb $ (!bcn <-> !bcn, All) in
	  rightVlOp ysbr.(i-1) zt;
	  bcbk <-| (bc 1.0 ysbr.(i-1) |*| zt);
	  rightDXop ysbr.(i-1) zt;
	  bcbk <-| (bcx 1.0 ysbr.(i-1) |*++| zt);
	  rightDYop ysbr.(i-1) zt;
	  bcbk <-| (bcy 1.0 ysbr.(i-1) |*++| zt);
	  scale_real' bcScR bcbk;
	  (fm $ (k <-> k, All)) <-$ z 1.0 ysbr.(i-1);
	  scale_real' bcScale (fm $ (k <-> k, All))
	done; 
	(* Line x = -1 *)
	for i = 1 to pybl do
	  let k = bps.(!bcn) in
	  bcn := !bcn + 1;
	  (bcScS => (k,!bcn)) bca;
	  let bcbk = bcb $ (!bcn <-> !bcn, All) in
	  leftVlOp ysbl.(i-1) zt;
	  bcbk <-| (bc (-1.0) ysbl.(i-1) |*| zt);
	  leftDXop ysbl.(i-1) zt;
	  bcbk <-| (bcx (-1.0) ysbl.(i-1) |*++| zt);
	  leftDYop ysbl.(i-1) zt;
	  bcbk <-| (bcy (-1.0) ysbl.(i-1) |*++| zt);
	  scale_real' bcScR bcbk;
	  (fm $ (k <-> k, All)) <-$ z (-1.0) ysbl.(i-1);
	  scale_real' bcScale (fm $ (k <-> k, All))
	done;
	assert (!bcn = nbc);
	let bcHss = fromLowRank' ~ooc (toHSSspine iHss1) bca
	    (transp ~ooc bcb) (tol) in
	print_endline "bcs";
	printHssRanks bcHss;
	(* let bcab = rawMatrix (px * py) (px * py) in
      fastMul bcHss (eye (px * py) (px * py)) bcab;
	   Printf.printf "bcberr = %g\n" (norm1 (bcab -$ bca *$ bcb)); *)
	zeroRows' (Array.to_list bps) iHss1;
	print_endline "iHss1";
	printHssRanks iHss1;
	(fastAdd ~ooc iHss1 bcHss, permND) in
      let um = rawMatrix ~ooc (px * py) nrhs in
      (* let bb = rawMatrix (px * py) (px * py) in
      fastMul iHss2 (eye (px*py) (px*py)) bb;
      let s = singValues' bb in
      (* for i = 1 to px * py do
	 Printf.printf "%i = %g\n" i s.{i}
	 done; *)
      Printf.printf "cond, norm, tol = %g %g %g\n"
	(s.{1} /. s.{px*py}) s.{1} tol; *)
      print_endline "iHss2";
      printHssRanks iHss2;
      let t1 = Sys.time () in
      fastSolve' iHss2 fm um;
      let t2 = Sys.time () -. t1 in
      Printf.printf "time = %g\n" t2;
      permND @~* um;
      um


  end (* module Chubber (Blop : BLOP) *)


module DChub = Chubber (DBlop) (DHss)
module SChub = Chubber (SBlop) (SHss)
module ZChub = Chubber (ZBlop) (ZHss)
module CChub = Chubber (CBlop) (CHss)

