(*
   File: fd.ml
   Author: Shivkumar Chandrasekaran
   Email: shiv@ece.ucsb.edu
   Date: 6.09.2004
   Purpose: finite difference codes.

   $Id$
*)

open Nla
open Sparse

module Fder (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix) =
  struct
    
    open Blop
    type t = Blop.t
    type matrix = Blop.matrix
    type sparse = Sparse.sparse
	  
    let fivePt m n =
      let a = Sparse.empty ((m-2) * (n-2)) ((m-2) * (n-2))
      and grid = Array.make_matrix m n (-1) in
      for y = 1 to m-2 do
	for x = 1 to n-2 do
	  grid.(y).(x) <- (y-1) * (n-2) + (x-1)
	done
      done;
      for y = 1 to m-2 do
	for x = 1 to n-2 do
	  let i = (y - 1) * (n - 2)  + (x - 1) in
	  Sparse.set a i i (Some (consts (fromFloat (-4.0)) 1 1));
	  if x > 1 then
	    Sparse.set a i (grid.(y).(x-1)) (Some (consts (fromFloat 1.0) 1 1));
	  if x < n-2 then
	    Sparse.set a i (grid.(y).(x+1)) (Some (consts (fromFloat 1.0) 1 1));
	  if y > 1 then
	    Sparse.set a i (grid.(y-1).(x)) (Some (consts (fromFloat 1.0) 1 1));
	  if y < m-2 then
	    Sparse.set a i (grid.(y+1).(x)) (Some (consts (fromFloat 1.0) 1 1))
	done
      done;
      a

    let sscond1' sp =
      let m = Sparse.no_of_rows sp
      and n = Sparse.no_of_cols sp
      and nrhs = 5
      and spnrm = Sparse.blkNorm1 sp in
      let b = randomMatrix (max m n) nrhs in
      let x = Sparse.linear_solve' sp (Blop.copy b) in
      let nrm = ref 0.0 in
      for i = 1 to nrhs do
	let cnrm = norm1 (x $ (All, i <-> i)) /. norm1 (b $ (All, i <-> i)) in
	if cnrm > !nrm then
	  nrm := cnrm
      done;
      (spnrm, sqrt (float m) *. !nrm /. sqrt (float nrhs))


    let triDiag n a b c =
      fun2mat (fun i j ->
	if i = j then a
	else if j=i+1 then c
	else if j = i-1 then b
	else zero) n n

    let blkTriDiag diag subDiag supDiag =
      let m = Array.length diag in
      let a = Sparse.empty m m in
      for i = 0 to m-1 do
	Sparse.set a i i (Some diag.(i));
	if i > 0 then
	  Sparse.set a i (i-1) (Some subDiag.(i-1));
	if i < m-1 then
	  Sparse.set a i (i+1) (Some supDiag.(i))
      done;
      a


    let ninePtMatrix =
      let xs = [| -1.0; 0.0; 1.0 |] in
      let ys = xs in
      fun2mat (fun i j ->
	let ix = (i-1) / 3
	and iy = (i-1) mod 3 in
	if j = 1 then one
	else if j = 2 then fromFloat xs.(ix)
	else if j = 3 then fromFloat ys.(iy)
	else if j = 4 then fromFloat (xs.(ix) *. xs.(ix))
	else if j = 5 then fromFloat (ys.(iy) *. ys.(iy))
	else (assert (j=6); fromFloat (xs.(ix) *. ys.(iy))))
	9 6

      
  end


module DFd = Fder (DBlop) (DSparse)
module SFd = Fder (SBlop) (SSparse)
module ZFd = Fder (ZBlop) (ZSparse)
module CFd = Fder (CBlop) (CSparse)

