(*
   File: fmm.ml
   Author: Shivkumar Chandrasekaran
   Date: 6.12.2006
   Purpose: Super fast algorithms for FMM matrices

   $Id: fmm.ml,v 1.3 2006/06/22 21:29:36 shiv Exp shiv $
*)

open Bigarray
open Utils
open Nla
open Sparse
open Hss

module type FMM =
  sig
    type t
    type matrix
    type transform

    type sMatrix =
      |	Exp of matrix (* expansion coefficient *)
      | Dns of matrix (* dense matrix *)
      | Sup2x2 of superMatrix2x2
      | Sup2x1 of superMatrix2x1
      | Sup1x2 of superMatrix1x2
    and superMatrix2x2 = {oo : sMatrix; oi : sMatrix;
			  io : sMatrix; ii : sMatrix}
    and superMatrix2x1 = {ov : sMatrix;
			  iv : sMatrix}
    and superMatrix1x2 = {ho : sMatrix; hi : sMatrix}
    and fmm_node = {bu : sMatrix; bl : sMatrix;
		     nl : int; nr : int; wl : matrix; wr : matrix;
		     ml : int; mr : int; rl : matrix; rr : matrix}
    and fmm_leaf = {u : matrix; v : matrix; d : matrix}
    and fmm = (fmm_leaf, fmm_node) btree
    val no_of_rows : fmm -> int
    val no_of_cols : fmm -> int
    val fastMul : fmm -> matrix -> matrix -> unit
    type odiag_split =
      | NoSplitExp
      | NoSplitDns
      | Split2x2 of odiag_split * odiag_split * odiag_split * odiag_split
      | Split2x1 of odiag_split * odiag_split
      | Split1x2 of odiag_split * odiag_split
    type leafSpine = { m : int; n : int }
    type nodeSpine = { mL : int; mR : int; nL : int; nR : int;
		       ou : odiag_split; ol : odiag_split }
    type fmmSpine = (leafSpine, nodeSpine) btree
    val fmmSize : fmmSpine -> int * int
    val fromFn : float -> fmmSpine ->
      (int -> int -> int -> int -> matrix -> unit) -> fmm
    type point = float array
    type nestedDissectionTree =
	(point * point * point list, point * point) btree
    val nestedDissection : int -> point list -> nestedDissectionTree
    val fmmSpineFromND : nestedDissectionTree -> fmmSpine
    val pointsFromNDtree : nestedDissectionTree -> point list
    val ppFmmRanks : fmm -> unit
    val ppBs : odiag_split -> unit

  end (* module type FMM *)


module Fmmer (Blop : BLOP) =
  struct

    open Blop
    type t = Blop.t
    type matrix = Blop.matrix
    type transform = Blop.transform

    type sMatrix =
      |	Exp of matrix
      | Dns of matrix
      | Sup2x2 of superMatrix2x2
      | Sup2x1 of superMatrix2x1
      | Sup1x2 of superMatrix1x2
    and superMatrix2x2 = {oo : sMatrix; oi : sMatrix;
			  io : sMatrix; ii : sMatrix}
    and superMatrix2x1 = {ov : sMatrix;
			  iv : sMatrix}
    and superMatrix1x2 = {ho : sMatrix; hi : sMatrix}
    and  fmm_node = {bu : sMatrix; bl : sMatrix;
		     nl : int; nr : int; wl : matrix; wr : matrix;
		     ml : int; mr : int; rl : matrix; rr : matrix}
    and  fmm_leaf = {u : matrix; v : matrix; d : matrix}
    and  fmm = (fmm_leaf, fmm_node) btree

    let no_of_rows fmm =
      match fmm with
      | Leaf a -> noOfRows a.d
      | Node (a, _, _) -> a.ml + a.mr

    let no_of_cols fmm =
      match fmm with
      | Leaf a -> noOfCols a.d
      | Node (a, _, _) -> a.nl + a.nr


    let add_opt a b =
      match a, b with
      | None, None -> None
      | Some _, None -> a
      | None, Some _ -> b
      | Some a, Some b -> Some (a +$ b)

    let rec add_f fu fl =
      match fu, fl with
      | Leaf fu, Leaf fl ->
	  Leaf (add_opt fu fl)
      | Node (fu, ful, fur), Node (fl, fll, flr) ->
	  Node (add_opt fu fl, add_f ful fll, add_f fur flr)
      | Leaf fu, Node (fl, fll, flr) ->
	  Node (add_opt fu fl, fll, flr)
      | Node (fu, ful, fur), Leaf fl ->
	  Node (add_opt fu fl, ful, fur)

    let rec b_x_g b g =
      match b with
      | Exp    b -> Leaf (Some (b *$ getNodeData g))
      | Dns    d -> Leaf None
      | Sup2x2 b ->
	  begin match g with
	  | Leaf _           -> failwith "Bad fmm g tree"
	  | Node (_, gl, gr) ->
	      let fu = add_f (b_x_g b.oo gl) (b_x_g b.oi gr)
	      and fl = add_f (b_x_g b.io gl) (b_x_g b.ii gr) in
	      Node (None, fu, fl)
	  end
      | Sup2x1 b ->
	  let fu = b_x_g b.ov g
	  and fl = b_x_g b.iv g in
	  Node (None, fu, fl)
      | Sup1x2 b ->
	  begin match g with
	  | Leaf _ -> failwith "Bad g tree"
	  | Node (_, gl, gr) ->
	      add_f (b_x_g b.ho gl) (b_x_g b.hi gr)
	  end

    (* The actual FMM multiply. *)
    let fastMul fmm x b =
      let nrhs = noOfCols b in
      let rec mulOffDiag lFmm ob rFmm x b =
	match ob with
	| Exp _ -> ()
	| Dns d -> b <-| (d |*++| x)
	| Sup2x2 bs ->
	    begin match lFmm, rFmm with
	    | Node (l, uFmm, bFmm), Node (r, lFmm, rFmm) ->
		let b1, b2 = partition2x1 l.ml l.mr b
		and x1, x2 = partition2x1 r.nl r.nr x in
		mulOffDiag uFmm bs.oo lFmm x1 b1;
		mulOffDiag uFmm bs.oi rFmm x2 b1;
		mulOffDiag bFmm bs.io lFmm x1 b2;
		mulOffDiag bFmm bs.ii rFmm x2 b2
	    | _, _ -> failwith "Fmm trees do not match b"
	    end
	| Sup2x1 bs ->
	    begin match lFmm with
	    | Node (l, uFmm, bFmm) ->
		let b1, b2 = partition2x1 l.ml l.mr b in
		mulOffDiag uFmm bs.ov rFmm x b1;
		mulOffDiag bFmm bs.iv rFmm x b2
	    | Leaf _ -> failwith "Bad lFmm tree"
	    end
	| Sup1x2 bs ->
	    begin match rFmm with
	    | Node (r, fmmL, fmmR) ->
		let x1, x2 = partition2x1 r.nl r.nr x in
		mulOffDiag lFmm bs.ho fmmL x1 b;
		mulOffDiag lFmm bs.hi fmmR x2 b
	    | Leaf _ -> failwith "Bad rFmm tree"
	    end
      in
      let rec mulDns fmm x b =
	match fmm with
	| Leaf a -> b <-| (a.d |*| x)
	| Node (a, lFmm, rFmm) ->
	    let b1, b2 = partition2x1 a.ml a.mr b
	    and x1, x2 = partition2x1 a.nl a.nr x in
	    mulDns lFmm x1 b1;
	    mulDns rFmm x2 b2;
	    mulOffDiag lFmm a.bu rFmm x2 b1;
	    mulOffDiag rFmm a.bl lFmm x1 b2
      in
      let rec upSweep fmm x =
	match fmm with
	| Leaf a -> Leaf (a.v *~$ x)
	| Node (a, lFmm, rFmm) ->
	    let (xl, xr) = partition2x1 a.nl a.nr x in
	    let gl = upSweep lFmm xl
	    and gr = upSweep rFmm xr in
	    Node (a.wl *~$ (getNodeData gl) +$ a.wr *~$ (getNodeData gr),
		  gl, gr)
      in
      let rec downSweep fmm f gs x b =
	match fmm with
	| Leaf a ->
	    begin match f with
	    | Leaf (Some f) -> b <-| (a.u |*++| f)
	    | Leaf None     -> ()
	    | Node _        -> failwith "Bad fmm tree"
	    end
	| Node (a, lFmm, rFmm) ->
	    match gs with
	    | Leaf _           -> failwith "Bad fmm g tree"
	    | Node (_, gl, gr) -> begin
		let fu = b_x_g a.bu gr
		and fl = b_x_g a.bl gl
		and (bl, br) = partition2x1 a.ml a.mr b
		and (xl, xr) = partition2x1 a.nl a.nr x in
		let (foL, foR) =
		  match f with
		  | Leaf (Some f) -> (Leaf (Some (a.rl *$ f)),
				      Leaf (Some (a.rr *$ f)))
		  | Leaf None     -> (f, f)
		  | Node (Some f, fL, fR) ->
		      (add_f (Leaf (Some (a.rl *$ f))) fL,
		       add_f (Leaf (Some (a.rr *$ f))) fR)
		  | Node (None, fL,fR) -> (fL, fR)
		in
		downSweep lFmm (add_f foL fu) gl xl bl;
		downSweep rFmm (add_f foR fl) gr xr br
	    end
      in
      assert (noOfCols x = nrhs);
      mulDns fmm x b;
      downSweep fmm (Leaf None) (upSweep fmm x) x b


    type odiag_split =
      | NoSplitExp
      | NoSplitDns
      | Split2x2 of odiag_split * odiag_split * odiag_split * odiag_split
      | Split2x1 of odiag_split * odiag_split
      | Split1x2 of odiag_split * odiag_split
    and leafSpine = { m : int; n : int }
    and nodeSpine = { mL : int; mR : int; nL : int; nR : int;
		      ou : odiag_split; ol : odiag_split }
    and fmmSpine = (leafSpine, nodeSpine) btree

    let rec fmmSize fmmSpine =
      match fmmSpine with
      | Leaf a -> (a.m, a.n)
      | Node (a, lSpine, rSpine) ->
	  let l = fmmSize lSpine
	  and r = fmmSize rSpine in
	  (fst l + fst r, snd l + snd r)

    let rec fillFmmSpine fmmSpine =
      let rec loop fmmSpine =
	match fmmSpine with
	| Leaf a -> (fmmSpine, a.m, a.n)
	| Node (a, lSpine, rSpine) ->
	    let (lSpine', mL, nL) = loop lSpine
	    and (rSpine', mR, nR) = loop rSpine in
	    (Node ({a with mL = mL; mR = mR; nL = nL; nR = nR},
		   lSpine', rSpine'),
	     mL+mR, nL+nR)
      in
      let (fmmSpine', _, _) = loop fmmSpine in
      fmmSpine'

    (* numerical-rank SVD *)
    let svdE' ?(ooc=false) tol a =
      let k = min (noOfRows a) (noOfCols a) in
      let (u, s, vt) = svd' ~ooc a in
      let r = ref 0 in
      while !r < k && s.{!r+1} > tol do
	r := !r + 1
      done;
      assert (!r <= k && (!r = 0 || s.{!r} > tol));
      (u $ (All, 1 <-> !r), Array1.sub s 1 !r, vt $ (1 <-> !r, All))


    (* acc is a disjoint union of intervals U_i (a_i, b_i) where a_i
	is the first integer of interval i, and b_i is length of interval. *)
    let mergeIntervals bs acc =
      let rec mergeInterval ((kf, k) as k0) acc =
	match acc with
	| [] -> [k0]
	| (nf, n) :: ns ->
	    let lk = kf + k - 1
	    and ln = nf + n - 1 in
	    if lk < nf then
	      (kf, k) :: acc
	    else if kf <= nf then
	      if lk <= ln then
		(kf, ln - kf + 1) :: ns
	      else
		mergeInterval (kf, k) ns
	    else if kf <= ln then
	      if lk <= ln then
		acc
	      else
		mergeInterval (nf, lk - nf + 1) ns
	    else
	      (nf, n) :: mergeInterval (kf, k) ns
      in
      List.fold_right mergeInterval bs acc

    (* We need a formal mechanism to hold the G and H matrices for the
       construction algorithm. These matrices are block rows/columns of
       the original matrix from which some sub-blocks have been stripped
       out. We also need to be able to merge two such rows/columns
       together. Compact storage is a plus as SVD's need to be
       computed. *)
    type strippedBlock =
      | Row of (int * int) list * matrix
      | Col of (int * int) list * matrix

    let checkStrippedBlock strpBlk =
      let rec loop ok nl n intervals =
	if ok then begin
	  match intervals with
	  | [] -> (ok, n)
	  | (kf, k) :: ks when k > 0 ->
	      if kf > nl then
		loop ok (kf + k - 1) (n + k) ks
	      else
		(false, n)
	  | _ -> (false, n)
	end else
	  (false, n)
      in
      match strpBlk with
      | Row (intervals, block) ->
	  let (ok, n) = loop true 0 0 intervals in
	  ok && n = noOfCols block
      | Col (intervals, block) ->
	  let (ok, m) = loop true 0 0 intervals in
	  ok && m = noOfRows block

    (* Strip the given intervals. Use mergeIntervals before stripping
       to reduce number of unwarranted copies. This can be improved
       dramatically if necessary. *)
    let stripIntervals strpBlk intervals =
      let stripRowInterval strpBlk (kf, k) =
	let kl = kf + k - 1 in
	let rec loop pl g blkIntervals acc =
	  match blkIntervals with
	  | [] -> failwith "Interval outside range"
	  | (nf, n) :: ns ->
	      let nl = nf + n - 1 in
	      assert (kf >= nf);
	      if kf > nl then
		loop (pl + n) g ns ((nf, n) :: acc)
	      else begin
		assert (kl <= nl);
		let ng = noOfCols g in
		let gnew = rawMatrix (noOfRows g) (ng - k) in
		matrix1x2 (pl+kf-nf) (ng-k-pl-kf+nf)
		  (iD (g $ (All, 1 <-> pl+kf-nf)))
		  (iD (g $ (All, pl+kf-nf+k+1 <-> ng))) gnew;
		if kf > nf then
		  if kl < nl then
		    (gnew, List.rev_append acc
		       ((nf, kf-nf) :: (kf+k, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ((nf, kf - nf) :: ns))
		else
		  if kl < nl then
		    (gnew, List.rev_append acc ((kl+1, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ns)
	      end
	in
	let (gnew, newBlkIntervals) =
	  match strpBlk with
	  | Row (blkIntervals, block) ->
	      if k > 0 then
		loop 0 block blkIntervals []
	      else
		(block, blkIntervals)
	  | Col _ -> failwith "Wrong stripping"
	in
	Row (List.filter (fun ints -> snd ints > 0) newBlkIntervals, gnew)
      in
      let stripColInterval strpBlk (kf, k) =
	let kl = kf + k - 1 in
	let rec loop pl g blkIntervals acc =
	  match blkIntervals with
	  | [] -> failwith "Interval outside range"
	  | (nf, n) :: ns ->
	      let nl = nf + n - 1 in
	      assert (kf >= nf);
	      if kf > nl then
		loop (pl + n) g ns ((nf, n) :: acc)
	      else begin
		assert (kl <= nl);
		let mg = noOfRows g in
		let gnew = rawMatrix (mg - k) (noOfCols g) in
		matrix2x1
		       (pl+kf-nf) (iD (g $ (1 <-> pl+kf-nf, All)))
		  (mg-k-pl-kf+nf) (iD (g $ (pl+kf-nf+k+1 <-> mg, All))) gnew;
		if kf > nf then
		  if kl < nl then
		    (gnew, List.rev_append acc
		       ((nf, kf-nf) :: (kf+k, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ((nf, kf - nf) :: ns))
		else
		  if kl < nl then
		    (gnew, List.rev_append acc ((kl+1, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ns)
	      end
	in
	let (gnew, newBlkIntervals) =
	  match strpBlk with
	  | Col (blkIntervals, block) ->
	      if k > 0 then
		loop 0 block blkIntervals []
	      else
		(block, blkIntervals)
	  | Row _ -> failwith "Wrong stripping"
	in
	Col (List.filter (fun ints -> snd ints > 0) newBlkIntervals, gnew)
      in
      match strpBlk with
      | Row _ ->
	  let newStrpBlk = List.fold_left stripRowInterval strpBlk intervals in
	  assert (checkStrippedBlock newStrpBlk);
	  newStrpBlk
      | Col _ ->
	  let newStrpBlk = List.fold_left stripColInterval strpBlk intervals in
	  assert (checkStrippedBlock newStrpBlk);
	  newStrpBlk

    (* From the given row strip block get the desired column blocks. *)
    let getRowIntervalBlks strpBlk intervals =
      let getInterval strpBlk (kf, k) =
	assert (k >= 0);
	let kl = kf + k - 1 in
	(* pl is the last column in the storage of strpBlk *)
	let rec loop pl g blkIntervals =
	  match blkIntervals with
	  | [] -> failwith
		(Printf.sprintf "(%i, %i) Interval not available" kf k)
	  | (nf, n) :: ns ->
	      let nl = nf + n - 1 in
	      assert (kf >= nf);
	      if kf > nl then
		loop (pl + n) g ns
	      else begin
		assert (kl <= nl);
		let bnew = rawMatrix (noOfRows g) k in
		bnew <-$ (g $ (All, pl+1+kf-nf <-> pl+kf-nf+k));
		bnew
	      end
	in
	match strpBlk with
	| Row (blkIntervals, block) ->
	    if k = 0 then
	      rawMatrix (noOfRows block) 0
	    else
	      loop 0 block blkIntervals
	| Col _ -> failwith "Wrong stripping"
      in
      List.map (getInterval strpBlk) intervals


    (* Merge two stripped blocks. Columns/rows must line up! *)
    let mergeStrips strpBlk1 strpBlk2 =
      match strpBlk1, strpBlk2 with
      | Row (intervals1, block1), Row (intervals2, block2) ->
	  assert (intervals1 = intervals2);
	  let m1 = noOfRows block1
	  and m2 = noOfRows block2 in
	  let nBlk = rawMatrix (m1+m2) (noOfCols block1) in
	  matrix2x1 m1 (iD block1)
	            m2 (iD block2) nBlk;
	  Row (intervals1, nBlk)
      | Col (intervals1, block1), Col (intervals2, block2) ->
	  assert (intervals1 = intervals2);
	  let n1 = noOfCols block1
	  and n2 = noOfCols block2 in
	  let nBlk = rawMatrix (noOfRows block1) (n1+n2) in
	  matrix1x2  n1         n2
	        (iD block1) (iD block2) nBlk;
	  Col (intervals1, nBlk)
      | _ -> failwith "Incompatible row/column blocks"


    (* Pass down a list of row-blocks and column-blocks to each node,
       for computing U and V at that level. For each such block B is
       computed. The R and W follow from the U and V computation. Here
       fn fstRow noOfRows fstCol noOfCols is what is assumed. *)
    let fromFn tol fmmSpine fn =
      let (m, n) = fmmSize fmmSpine in
      let rec addBlks lBlks rBlks =
	match lBlks, rBlks with
	| Leaf bs, Leaf cs           -> Leaf (bs @ cs)
	| Leaf bs, Node (cs, rL, rR) -> Node (bs @ cs, rL, rR)
	| Node (bs, lL, lR), Leaf cs -> Node (bs @ cs, lL, lR)
	| Node (bs, lL, lR), Node (cs, rL, rR) ->
	    Node (bs @ cs, addBlks lL rL, addBlks lR rR)
      in
      (* Not in linear order, but tree order *)
      let rec stripeRowBlks nl od rSpine =
	match od, rSpine with
	| NoSplitExp, Leaf a         -> Leaf [(nl+1, a.n)]
	| NoSplitExp, Node (a, _, _) -> Leaf [(nl+1, a.nL+a.nR)]
	| NoSplitDns, Leaf _         -> Leaf []
	| NoSplitDns, Node _ -> failwith "Leaf-node dense block"
	| Split2x2 (od00, od01, od10, od11), Node (a, lSpine, rSpine) ->
	    Node ([],
		  addBlks (stripeRowBlks nl od00 lSpine)
		    (stripeRowBlks (nl+a.nL) od01 rSpine),
		  addBlks (stripeRowBlks nl od10 lSpine)
		    (stripeRowBlks (nl+a.nL) od11 rSpine))
	| Split1x2 (od00, od01), Node (a, lSpine, rSpine) ->
	    addBlks (stripeRowBlks nl od00 lSpine)
	      (stripeRowBlks (nl+a.nL) od01 rSpine)
	| Split2x1 (od00, od10), _ ->
	    Node ([],
		  stripeRowBlks nl od00 rSpine,
		  stripeRowBlks nl od10 rSpine)
	| _ -> failwith "Bad fmm/B"
      in
      (* Not in linear order, but tree order *)
      let rec stripeRowDnsBlks nl od rSpine =
	match od, rSpine with
	| NoSplitDns, Leaf a -> Leaf [(nl+1, a.n)]
	| NoSplitDns, Node _ -> failwith "Leaf-node dense block"
	| NoSplitExp, _      -> Leaf []
	| Split2x2 (od00, od01, od10, od11), Node (a, lSpine, rSpine) ->
	    Node ([],
		  addBlks (stripeRowDnsBlks nl od00 lSpine)
		    (stripeRowDnsBlks (nl+a.nL) od01 rSpine),
		  addBlks (stripeRowDnsBlks nl od10 lSpine)
		    (stripeRowDnsBlks (nl+a.nL) od11 rSpine))
	| Split1x2 (od00, od01), Node (a, lSpine, rSpine) ->
	    addBlks (stripeRowDnsBlks nl od00 lSpine)
	      (stripeRowDnsBlks (nl+a.nL) od01 rSpine)
	| Split2x1 (od00, od10), _ ->
	    Node ([],
		  stripeRowDnsBlks nl od00 rSpine,
		  stripeRowDnsBlks nl od10 rSpine)
	| _ -> failwith "Bad fmm/B"
      in
      let rec unStripeRowBlks bvs ds od =
	match bvs, ds, od with
	| Leaf bs, _, NoSplitExp ->
	    (Exp (List.hd bs), Leaf (List.tl bs), ds)
	| Node (bs, lBvs, rBvs), _, NoSplitExp ->
	    (Exp (List.hd bs), Node (List.tl bs, lBvs, rBvs), ds)
	| _, Leaf ds, NoSplitDns ->
	    (Dns (List.hd ds), bvs, Leaf (List.tl ds))
	| _, Node (ds, lDs, rDs), NoSplitDns ->
	    (Dns (List.hd ds), bvs, Node (List.tl ds, lDs, rDs))
	| Node (bs, lBvs, rBvs), Node (ds, lDs, rDs),
	    Split2x2 (od00, od01, od10, od11) ->
	      let (b00, lBvs1, lDs1) = unStripeRowBlks lBvs  lDs  od00 in
	      let (b01, lBvs2, lDs2) = unStripeRowBlks lBvs1 lDs1 od01 in
	      let (b10, rBvs1, rDs1) = unStripeRowBlks rBvs  rDs  od10 in
	      let (b11, rBvs2, rDs2) = unStripeRowBlks rBvs1 rDs1 od11 in
	      (Sup2x2 {oo = b00; oi = b01; io = b10; ii = b11},
	       Node (bs, lBvs2, rBvs2), Node (ds, lDs2, rDs2))
	| Node (bs, lBvs, rBvs), Node (ds, lDs, rDs), Split2x1 (od00, od10) ->
	      let (b00, lBvs1, lDs1) = unStripeRowBlks lBvs  lDs  od00 in
	      let (b10, rBvs1, rDs1) = unStripeRowBlks rBvs  rDs  od10 in
	      (Sup2x1 {ov = b00; iv = b10},
	       Node (bs, lBvs1, rBvs1), Node (ds, lDs1, rDs1))
	| _, _, Split1x2 (od00, od01) ->
	    let (b00, bvs1, ds1) = unStripeRowBlks bvs  ds  od00 in
	    let (b01, bvs2, ds2) = unStripeRowBlks bvs1 ds1 od01 in
	    (Sup1x2 {ho = b00; hi = b01}, bvs2, ds2)
	| _ -> failwith "Bad bvs/od"
      in
      let rec getBs bvs vs =
	match bvs, vs with
	| Exp bv, Leaf v         -> Exp (bv *$ v)
	| Exp bv, Node (v, _, _) -> Exp (bv *$ v)
	| Dns _, Leaf _          -> bvs
	| Dns _, Node _ -> failwith "Leaf-node dense block"
	| Sup2x2 bv, Node (_, vsL, vsR) ->
	    Sup2x2 { oo = getBs bv.oo vsL; oi = getBs bv.oi vsR;
		     io = getBs bv.io vsL; ii = getBs bv.ii vsR }
	| Sup1x2 bv, Node (_, vsL, vsR) ->
	    Sup1x2 { ho = getBs bv.ho vsL; hi = getBs bv.hi vsR }
	| Sup2x1 bv, _ ->
	    Sup2x1 { ov = getBs bv.ov vs;
		     iv = getBs bv.iv vs }
	| _ -> failwith "Sup/Leaf in getBs"
      in
      (* Not in linear order, but tree order *)
      let rec stripeColBlks ml od lSpine =
	match od, lSpine with
	| NoSplitExp, Leaf a         -> Leaf [(ml+1, a.m)]
	| NoSplitExp, Node (a, _, _) -> Leaf [(ml+1, a.mL+a.mR)]
	| NoSplitDns, Leaf _         -> Leaf []
	| NoSplitDns, Node _ -> failwith "Node-leaf dense block"
	| Split2x2 (od00, od01, od10, od11), Node (a, lSpine, rSpine) ->
	    Node ([],
		  addBlks (stripeColBlks ml od00 lSpine)
		    (stripeColBlks (ml+a.mL) od10 rSpine),
		  addBlks (stripeColBlks ml od01 lSpine)
		    (stripeColBlks (ml+a.mL) od11 rSpine))
	| Split1x2 (od00, od01), _ ->
	    Node ([],
		  stripeColBlks ml od00 lSpine,
		  stripeColBlks ml od01 lSpine)
	| Split2x1 (od00, od10), Node (a, lSpine, rSpine) ->
	    addBlks (stripeColBlks ml od00 lSpine)
	      (stripeColBlks (ml+a.mL) od10 rSpine)
	| _ -> failwith "Bad fmm/B"
      in
      let rec stripeColDnsBlks ml od lSpine =
	match od, lSpine with
	| NoSplitDns, Leaf a -> Leaf [(ml+1, a.m)]
	| NoSplitDns, Node _ -> failwith "Node-leaf dense block"
	| NoSplitExp, _      -> Leaf []
	| Split2x2 (od00, od01, od10, od11), Node (a, lSpine, rSpine) ->
	    Node ([],
		  addBlks (stripeColDnsBlks ml od00 lSpine)
		    (stripeColDnsBlks (ml+a.mL) od10 rSpine),
		  addBlks (stripeColDnsBlks ml od01 lSpine)
		    (stripeColDnsBlks (ml+a.mL) od11 rSpine))
	| Split1x2 (od00, od01), _ ->
	    Node ([],
		  stripeColDnsBlks ml od00 lSpine,
		  stripeColDnsBlks ml od01 lSpine)
	| Split2x1 (od00, od10), Node (a, lSpine, rSpine) ->
	    addBlks (stripeColDnsBlks ml od00 lSpine)
	      (stripeColDnsBlks (ml+a.mL) od10 rSpine)
	| _ -> failwith "Bad fmm/B"
      in
      (* ml, nl, are last previous row and column *)
      let rec split fmmSpine ml nl rBlks cBlks dBlks cdBlks =
	match fmmSpine with
	| Leaf a ->
	    let d = rawMatrix a.m a.n in
	    d <-| fn (ml+1) a.m (nl+1) a.n;
	    let (u, h, bvs, ds) =
	      let h = rawMatrix a.m (n - a.n) in
	      matrix1x2 nl (n - a.n - nl)
		(fn (ml+1) a.m 1 nl) (fn (ml+1) a.m (nl+a.n+1) (n-a.n-nl)) h;
	      let strpH = Row ([(1, nl); (nl+a.n+1, n-a.n-nl)], h) in
	      let (ds, exclD) =
		match dBlks with
		| Leaf ds -> (Leaf (getRowIntervalBlks strpH ds),
			      mergeIntervals ds [])
		| _ -> failwith "Incorrect level in dBlks"
	      in
	      let (uh, vhtBlk) =
		match stripIntervals strpH exclD with
		| Row (cols, blk) ->
		    let (uh, sh, vht) = svdE' tol blk in
		    rowScale_real' sh vht;
		    (uh, Row (cols, vht))
		| _ -> failwith "Got column block!"
	      in
	      let (bvs, exclR) =
		match rBlks with
		| Leaf bs -> (Leaf (getRowIntervalBlks vhtBlk bs),
			      mergeIntervals bs [])
		| _    -> failwith "Incorrect level in rBlks"
	      in
	      let hR = stripIntervals vhtBlk exclR in
	      (uh, hR, bvs, ds)
	    in
	    let (g, v) =
	      let g = rawMatrix (m - a.m) a.n in
	      matrix2x1    ml  (fn 1 ml (nl+1) a.n)
		(m - a.m - ml) (fn (ml+a.m+1) (m-a.m-ml) (nl+1) a.n) g;
	      let strpG = Col ([(1, ml); (ml+a.m+1, m-a.m-ml)], g) in
	      let exclD =
		match cdBlks with
		| Leaf ds -> mergeIntervals ds []
		| _ -> failwith "Incorrect cdBlks"
	      in
	      let (ugStrpd, vgt) =
		match stripIntervals strpG exclD with
		| Col (rows, blk) ->
		    let (ug, sg, vgt) = svdE' tol blk in
		    colScale_real' ug sg;
		    (Col (rows, ug), vgt)
		| _ -> failwith "Got row block!"
	      in
	      let exclC =
		match cBlks with
		| Leaf bs -> mergeIntervals bs []
		| Node _  -> failwith "Incorrect level in cBlks"
	      in
	      let gR = stripIntervals ugStrpd exclC in
	      (gR, transp vgt)
	    in
	    (Leaf {d = d; u = u; v = v}, h, g, bvs, Leaf v, ds)    
	| Node (a, lSpine, rSpine) ->
	    let (bBlksT, bBlksD) =
	      let bBlksT = stripeRowBlks (nl+a.nL) a.ou rSpine
	      and bBlksD = stripeRowBlks nl a.ol lSpine in
	      match rBlks with
	      | Leaf _ -> (bBlksT, bBlksD)
	      | Node (_, rBlksT, rBlksD) ->
		  (addBlks bBlksT rBlksT, addBlks bBlksD rBlksD)
	    in
	    let (bBlksL, bBlksR) =
	      let bBlksL = stripeColBlks (ml+a.mL) a.ol rSpine
	      and bBlksR = stripeColBlks ml a.ou lSpine in
	      match cBlks with
	      | Leaf _ -> (bBlksL, bBlksR)
	      | Node (_, cBlksL, cBlksR) ->
		  (addBlks bBlksL cBlksL, addBlks bBlksR cBlksR)
	    in
	    let (dBlksT, dBlksD) =
	      let dBlksT = stripeRowDnsBlks (nl+a.nL) a.ou rSpine
	      and dBlksD = stripeRowDnsBlks nl a.ol lSpine in
	      match dBlks with
	      | Leaf _ -> (dBlksT, dBlksD)
	      | Node (_, dBlksOT, dBlksOD) ->
		  (addBlks dBlksT dBlksOT, addBlks dBlksD dBlksOD)
	    in
	    let (cdBlksL, cdBlksR) =
	      let cdBlksL = stripeColDnsBlks (ml+a.mL) a.ol rSpine
	      and cdBlksR = stripeColDnsBlks ml a.ou lSpine in
	      match cdBlks with
	      | Leaf _ -> (cdBlksL, cdBlksR)
	      | Node (_, cdBlksOL, cdBlksOR) ->
		  (addBlks cdBlksL cdBlksOL, addBlks cdBlksR cdBlksOR)
	    in
	    let (fmmL, svhtL, usgL, bvT, vL, dsT) =
	      split lSpine ml nl bBlksT bBlksL dBlksT cdBlksL in
	    let (fmmR, svhtR, usgR, bvD, vR, dsD) =
	      split rSpine (ml+a.mL) (nl+a.nL) bBlksD bBlksR dBlksD cdBlksR in
	    let (buvs, buvs1, uDs1) = unStripeRowBlks bvT dsT a.ou
	    and (blvs, blvs1, lDs1) = unStripeRowBlks bvD dsD a.ol in
	    let bu = getBs buvs vR
	    and bl = getBs blvs vL in
	    let (rl, rr, h, bvs) =
	      let sRows a =
		match a with
		| Row (_, a) -> noOfRows a
		| _ -> failwith "Row expected"
	      in
	      let svht = mergeStrips svhtL svhtR in
	      let (hInts, h) =
		match svht with
		| Row (hInts, h) -> (hInts, h)
		| _ -> failwith "huh?"
	      in
	      let (r, s, vt) = svdE' tol h in
	      rowScale_real' s vt;
	      let (rl, rr) = partition2x1 (sRows svhtL) (sRows svhtR) r in
	      let svtnew = Row (hInts, vt) in
	      let (bvs, exclR) =
		let bs = getNodeData rBlks in
		(Node (getRowIntervalBlks svtnew bs, buvs1, blvs1),
		 mergeIntervals bs [])
	      in
	      (rl, rr, stripIntervals svtnew exclR, bvs)
	    in
	    let (wl, wr, g, v) =
	      let sCols a =
		match a with
		| Col (_, a) -> noOfCols a
		| _ -> failwith "Col expected"
	      in
	      let usg = mergeStrips usgL usgR in
	      let (gInts, gb) =
		match usg with
		| Col (gInts, g) -> (gInts, g)
		| _ -> failwith "huh?"
	      in
	      let (u, s, wt) = svdE' tol gb in
	      colScale_real' u s;
	      let usgnew = Col (gInts, u) in
	      let (wlt, wrt) = partition1x2 (sCols usgL) (sCols usgR) wt in
	      let v = rawMatrix (a.nL + a.nR) (noOfRows wt) in
	      matrix2x1 a.nL (getNodeData vL |*~| wlt)
		        a.nR (getNodeData vR |*~| wrt) v;
	      let exclC = mergeIntervals (getNodeData cBlks) [] in
	      (transp wlt, transp wrt, stripIntervals usgnew exclC,
	       Node (v, vL, vR))
	    in
	    (Node ({ml = a.mL; mr = a.mR; nl = a.nL; nr = a.nR;
		    bu = bu; bl = bl;
		    rl = rl; rr = rr; wl = wl; wr = wr},
		   fmmL, fmmR),
	     h, g, bvs, v, Node ([], uDs1, lDs1))
      in
      let (fmm, _, _, _, _, _) = split (fillFmmSpine fmmSpine) 0 0
	  (Leaf []) (Leaf []) (Leaf []) (Leaf []) in
      fmm

    let arrayWith a i ai =
      let b = Array.copy a in
      b.(i) <- ai;
      b

    type point = float array
    type nestedDissectionTree =
	(point * point * point list, point * point) btree

    let nestedDissection p xs : nestedDissectionTree =
      let d = Array.length (List.hd xs) in
      assert (List.for_all (fun x -> Array.length x = d) xs);
      let bl = Array.init d (fun i -> List.fold_left (fun m x -> min m x.(i))
	  infinity xs) in
      let tr = Array.init d (fun i -> List.fold_left (fun m x -> max m x.(i))
	  neg_infinity xs) in
      let rec nd bl tr xs =
	if List.length xs <= p then begin
	  Leaf (bl, tr, xs)
	end else begin
	  let i =
	    let mv = ref neg_infinity
	    and m = ref (-1) in
	    Array.iteri (fun j blj -> if tr.(j) -. blj > !mv then begin
	      mv := tr.(j) -. blj; m := j end) bl;
	    !m
	  in
	  let mi = (tr.(i) +. bl.(i)) /. 2.0 in
	  let (xs1, xs2) = List.partition
	      (fun x -> assert (bl.(i) <= x.(i) && x.(i) <= tr.(i));
		x.(i) <= mi) xs
	  in
	  Node ((bl, tr),
		nd bl (arrayWith tr i mi) xs1, nd (arrayWith bl i mi) tr xs2)
	end
      in
      nd bl tr xs

    let rec pointsFromNDtree ndTree =
      match ndTree with
      | Leaf (_, _, a) -> a
      | Node (_, lNDtree, rNDtree) ->
	  pointsFromNDtree lNDtree @ pointsFromNDtree rNDtree

    let tooClose (bl1, tr1) (bl2, tr2) =
      let dim = Array.length bl1 in
      let l = Array.make dim 0.0
      and r = Array.make dim 0.0 in
      for i = 0 to dim-1 do
	if tr1.(i) <= bl2.(i) then begin
	  l.(i) <- tr1.(i);
	  r.(i) <- bl2.(i)
	end else if tr2.(i) <= bl1.(i) then begin
	  l.(i) <- bl1.(i);
	  r.(i) <- tr2.(i)
	end else begin
	  l.(i) <- 0.0;
	  r.(i) <- 0.0
	end
      done;
      let dist = ref 0.0 in
      for i = 0 to dim-1 do
	dist := !dist +. (l.(i) -. r.(i)) ** 2.0
      done;
      dist := sqrt !dist;
      let tc = ref false
      and i = ref 0 in
      while not !tc && !i < dim do
	tc := tr1.(!i) -. bl1.(!i) > !dist || tr2.(!i) -. bl2.(!i) > !dist;
	incr i
      done;
      !tc

    let rec makeBs lNDtree rNDtree =
      match lNDtree, rNDtree with
      | Leaf (blL, trL, _), Leaf (blR, trR, _) ->
	  if tooClose (blL, trL) (blR, trR) then
	    NoSplitDns
	  else
	    NoSplitExp
      | Node (lRect, ll, lr), Node (rRect, rl, rr) ->
	  if tooClose lRect rRect then
	    Split2x2 (makeBs ll rl, makeBs ll rr,
		      makeBs lr rl, makeBs lr rr)
	  else
	    NoSplitExp
      | Leaf (blL, trL, _) as l, Node (rRect, rl, rr) ->
	  if tooClose (blL, trL) rRect then
	    Split1x2 (makeBs l rl, makeBs l rr)
	  else
	    NoSplitExp
      | Node (lRect, ll, lr), (Leaf (blR, trR, _) as r) ->
	  if tooClose lRect (blR, trR) then
	    Split2x1 (makeBs ll r,
		      makeBs lr r)
	  else
	    NoSplitExp

    let rec fmmSpineFromND ndTree =
      match ndTree with
      | Leaf (bl, tr, xs) -> Leaf { m = List.length xs; n = List.length xs }
      | Node ((bl, tr), lNDtree, rNDtree) ->
	  let lFmmSpine = fmmSpineFromND lNDtree
	  and rFmmSpine = fmmSpineFromND rNDtree in
	  Node ({mL= -1; mR= -1; nL= -1; nR= -1;
		 ou = makeBs lNDtree rNDtree;
		 ol = makeBs rNDtree lNDtree},
		lFmmSpine, rFmmSpine)

    let ppFmmRanks fmm =
      ppBtree
	(fun a -> pp2Tuple Format.print_int Format.print_int
	    (noOfCols a.u, noOfCols a.v))
	(fun a -> pp2Tuple Format.print_int Format.print_int
	    (noOfCols a.rl, noOfCols a.wl))
	fmm

    let rec ppBs b =
      match b with
      | NoSplitExp ->
	  Format.open_box 0;
	  Format.print_string "NoSplitExp";
	  Format.close_box ()
      | NoSplitDns ->
	  Format.open_box 0;
	  Format.print_string "NoSplitDns";
	  Format.close_box ()
      | Split2x2 (oo, oi, io, ii) ->
	  Format.open_box 0;
	  Format.print_string "Split2x2";
	  Format.print_space ();
	  Format.open_vbox 0;
	  Format.print_char '(';
	  Format.open_box 0;
	  ppBs oo;
	  Format.print_char ',';
	  Format.print_space ();
	  ppBs oi;
	  Format.print_char ',';
	  Format.close_box ();
	  Format.print_space ();
	  Format.open_box 0;
	  ppBs io;
	  Format.print_char ',';
	  Format.print_space ();
	  ppBs ii;
	  Format.close_box ();
	  Format.print_char ')';
	  Format.close_box ()
      | Split2x1 (oo, io) ->
	  Format.open_box 0;
	  Format.print_string "Split2x1";
	  Format.print_space ();
	  Format.open_vbox 0;
	  Format.print_char '(';
	  Format.open_box 0;
	  ppBs oo;
	  Format.print_char ',';
	  Format.close_box ();
	  Format.print_space ();
	  Format.open_box 0;
	  ppBs io;
	  Format.close_box ();
	  Format.print_char ')';
	  Format.close_box ()
      | Split1x2 (oo, oi) ->
	  Format.open_box 0;
	  Format.print_string "Split1x2";
	  Format.print_space ();
	  Format.open_vbox 0;
	  Format.print_char '(';
	  Format.open_box 0;
	  ppBs oo;
	  Format.print_char ',';
	  Format.print_space ();
	  ppBs oi;
	  Format.close_box ();
	  Format.print_char ')';
	  Format.close_box ()


    type wChiLeaf = { w : transform; chi : matrix }
    and  wChiNode = { wN : transform; chiN : matrix; nm0 : int; nm1 : int }

    let lqEliminate d v dvs =
      let lq a =
	let (l, q) = lq' a in
	(copyL l, q)
      in
      let m = noOfRows d in
      assert (m = noOfCols d);
      (* let (d, q1) = lq d in *)
      let rec lqe d v dvs qs vs =
      match dvs with
      | [] -> (d, qs, v, List.rev vs)
      | (d1, v1) :: dvs1 ->
	  let n1 = noOfCols d1 in
	  let dc = rawMatrix m (m + n1) in
	  matrix1x2 m      n1
	         (iD d) (iD d1) dc;
	  let (l1, q1) = lq dc in
	  let m0 = noOfRows v
	  and n0 = noOfCols v
	  and m1 = noOfRows v1
	  and n1 = noOfCols v1 in
	  let vc = rawMatrix (m0+m1) (n0+n1) in
	  matrix2x2 n0      n1
	    m0 (iD v)  (const zero)
	    m1 (const zero) (iD v1) vc;
	  q1 @* vc;
	  let (v, v1) = partition2x1 m0 m1 vc in
	  lqe l1 v dvs1 (q1 :: qs) (v1 :: vs)
      in
      lqe d v dvs [(*q1*)] []

    let fastSolve' fmm b x =
      let nrhs = noOfCols b in
      assert (nrhs = noOfCols x);
      let eliminateLeaf dvs a b =
	(* Sparse part *)
	let d, ws, v, vs = 
	  match dvs with
	  | Node _ -> failwith "Bad d tree"
	  | Leaf dv -> let (l, ws, v, vs) = lqEliminate a.d a.v dv in
	    (l, ws, v, Leaf vs)
	in
	(* Low-rank part *)
	if noOfCols a.u >= noOfRows a.u then (* Nothing to compress *)
	  (Leaf { w = List.rev (eyeTransform :: ws); chi = zeros 0 nrhs },
	   Leaf (zeros (noOfCols v) nrhs), (* g *)
	   b,                              (* remaining b *)
	   a.u,                            (* remaining u *)
	   v,                              (* remaining v *)
	   d,                              (* remaining d *)
	   List.rev ws,                    (* ws to be applied *)
	   vs                              (* vs go with off-diag d's *)
	  )
	else (* Compress *)
	  let ku = noOfCols a.u
	  and m  = noOfRows a.u
	  and n  = noOfCols d
	  and (q, l_u) = ql' a.u in
	  q @~* b;
	  q @~* d;
	  let (b0, b1) = partition2x1 (m-ku) ku b
	  and (d0, d1) = partition2x1 (m-ku) ku d in
	  let (d00, w) = lq' d0 in
	  d1 *~@ w;
	  w @* v;
	  subs' d00 b0;
	  let (d10, d11) = partition1x2 (m-ku) (n-m+ku) d1
	  and (v00, v01) = partition2x1 (m-ku) (n-m+ku) v
	  and chi = b0 in
	  ( Leaf { w = List.rev (w :: ws); chi = chi },
	    Leaf (v00 *~$ chi), (* g *)
	    b1 -$ d10 *$ chi,   (* remaining b *)
	    copyL l_u,          (* remaining u *)
	    v01,                (* remaining v *)
	    d11,                (* remaining d *)
	    List.rev ws,        (* ws to be applied *)
	    vs                  (* vs go with off-diag d's *)
	  )
      in
      let rec ulSweep dvs fmm b f =
	match fmm with
	| Leaf a ->
	    let bnew =
	      match f with
	      | Leaf f -> b -$ a.u *$ f
	      | Node _ -> failwith "Bad f tree"
	    in
	    eliminateLeaf dvs a bnew
	| Node (a, lFmm, rFmm) ->
	    let (b0, b1) = partition2x1 a.ml a.mr b in
	    let (lWchiTree, gl, bl, ul, vl, dl) = ulSweep lFmm b0
		(a.rl *$ f) in
	    let (rWchiTree, gr, br, ur, vr, dr) = ulSweep rFmm b1
		(a.bl *$ gl +$ a.rr *$ f) in
	    let m0 = noOfRows dl
	    and n0 = noOfCols dl
	    and m1 = noOfRows dr
	    and n1 = noOfCols dr in
	    let d = rawMatrix (m0+m1) (n0+n1)
	    and u = rawMatrix (m0+m1) (noOfCols a.rl)
	    and v = rawMatrix (n0+n1) (noOfCols a.wl)
	    and bm = rawMatrix (m0+m1) nrhs in
	    matrix2x2   n0         n1
	      m0 (iD dl) (ul |*| (a.bu *$~ vr))
	      m1 (ur |*| (a.bl *$~ vl)) (iD dr) d;
	    matrix2x1 m0 (ul |*| a.rl)
	      m1 (ur |*| a.rr) u;
	    matrix2x1 n0 (vl |*| a.wl)
	      n1 (vr |*| a.wr) v;
	    matrix2x1 m0 (bl |-| ul *$ a.bu *$ gr)
 	              m1 (iD br                  ) bm;
	    let retLeaf = eliminateLeaf {u = u; v = v; d = d} bm in
	    match retLeaf with
	      (Leaf wCh, g, bmr, umr, vmr, dmr) ->
		(Node ({wN = wCh.w; chiN = wCh.chi; nm0 = n0; nm1 = n1},
		       lWchiTree, rWchiTree),
		 g +$ a.wl *~$ gl +$ a.wr *~$ gr,
		 bmr,
		 umr,
		 vmr,
		 dmr)
	    | _ -> failwith "eliminateLeaf returned Node!"
      in
      let rec vSweep hss wChiTree xp x =
	match (hss, wChiTree) with
	  (Leaf _, Leaf {w=w; chi=chi}) ->
	    matrix2x1 (noOfRows chi) (iD chi)
	      (noOfRows xp ) (iD xp ) x;
	    w @~* x
	| (Node (a, lHss, rHss),
	   Node ({wN=w;chiN=chi;nm0=n0;nm1=n1}, lWchi, rWchi)) ->
	     let mc = noOfRows chi
	     and mx = noOfRows xp in
	     let chip = rawMatrix (mc+mx) nrhs in
	     matrix2x1 mc (iD chi)
	       mx (iD xp ) chip;
	     w @~* chip;
	     let (xp0, xp1) = partition2x1 n0 n1 chip in
	     matrix2x1 a.nl (vSweep lHss lWchi xp0)
	       a.nr (vSweep rHss rWchi xp1) x
	| _ -> failwith "fastSolve': Mismatched trees."
      in
      match hss with
	Leaf a ->
	  x <-$ b;
	  a.d |/| x
	| _ ->
	    let (wChiTree, _, _, _, _, _) = ulSweep hss b
		(zeros 0 nrhs) in
	    vSweep hss wChiTree (zeros 0 nrhs) x
	      
  end (* module Fmmer (Blop) (Sparse) *)


module DFmm = Fmmer (DBlop)
module SFmm = Fmmer (SBlop)
module ZFmm = Fmmer (ZBlop)
module CFmm = Fmmer (CBlop)
