(*
  File: fmm1D.ml
  Author: Shivkumar Chandrasekaran
  Email: shiv@ece.ucsb.edu
  Date: 7.24.2005
  Purpose: Fast algorithms for 1D FMM

  $Id: fmm1D.ml,v 1.7 2005/08/02 11:59:29 shiv Exp shiv $
*)

open Utils
open Nla


module type FMM1D =
  sig
    type t
    type matrix
    type u_exp_coeffs = { bu11 : matrix; bu12 : matrix;
			  bu21 : upper_exp_coeffs; bu22 : matrix }
    and upper_exp_coeffs =
	Su of u_exp_coeffs
      |	Pu of matrix
    type l_exp_coeffs = { bl11 : matrix; bl12 : lower_exp_coeffs;
			  bl21 : matrix; bl22 : matrix }
    and lower_exp_coeffs =
	Sl of l_exp_coeffs
      |	Pl of matrix
    type leaf = {
	u : matrix; v : matrix; d : matrix
      }
    type node = {
	ml : int; mr : int;
	nl : int; nr : int;
	rl : matrix; rr : matrix;
	wl : matrix; wr : matrix;
	bu : upper_exp_coeffs;
	bl : lower_exp_coeffs
      }
    type fmm1D = (leaf, node) btree
    val no_of_rows : fmm1D -> int
    val no_of_cols : fmm1D -> int
    val fastMul : fmm1D -> matrix -> matrix -> unit
    val fastSolve' : fmm1D -> matrix -> matrix -> unit
  end (* module type FMM1D *)


module Fmm1Der (Blop : BLOP) =
  struct

    open Blop
    type t = Blop.t
    type matrix = Blop.matrix

    (* All edges are between nodes at the same level even in a skewed tree.
       To obtain this we might have to use lots of empty nodes. *)
    type u_exp_coeffs = { bu11 : matrix; bu12 : matrix;
			  bu21 : upper_exp_coeffs; bu22 : matrix }
    and upper_exp_coeffs =
	Su of u_exp_coeffs
      |	Pu of matrix

    type l_exp_coeffs = { bl11 : matrix; bl12 : lower_exp_coeffs;
			  bl21 : matrix; bl22 : matrix }
    and lower_exp_coeffs =
	Sl of l_exp_coeffs
      |	Pl of matrix

    type leaf = {
	u : matrix; v : matrix; d : matrix
      }

    type node = {
	ml : int; mr : int;
	nl : int; nr : int;
	rl : matrix; rr : matrix;
	wl : matrix; wr : matrix;
	bu : upper_exp_coeffs;
	bl : lower_exp_coeffs
      }

    type fmm1D = (leaf, node) btree


    let no_of_rows fmm1D =
      match fmm1D with
	Leaf a -> noOfRows a.d
      |	Node (a, _, _) -> a.ml + a.mr
	    

    let no_of_cols fmm1D =
      match fmm1D with
	Leaf a -> noOfCols a.d
      |	Node (a, _, _) -> a.nl + a.nr


    (* Aux. routine to get data from one level down on a btree of tuples *)
    let getNodeData nl =
      match nl with
	Leaf (d, _) -> d
      | Node (d, _, _) -> d

    (* Aux. routine to add 2 BG trees *)
    let addBG a b =
      let rec mergeBG fn bg nbg =
	match (bg, nbg) with
	  (Leaf abg, Leaf nabg) -> Leaf (fn abg nabg)
	| (Leaf abg, Node (nabg, lbg, rbg)) -> Node (fn abg nabg, lbg, rbg)
	| (Node (abg, lbg, rbg), Leaf nabg) -> Node (fn abg nabg, lbg, rbg)
	| (Node (abg, lbg, rbg), Node (nabg, nlbg, nrbg)) ->
	    Node (fn abg nabg, mergeBG fn lbg nlbg, mergeBG fn rbg nrbg)
      in
      let optAdd a b =
	match a, b with
	  None, None -> None
	| Some a, None -> Some a
	| None, Some a -> Some a
	| Some a, Some b -> Some (a +$ b)
      in
      mergeBG optAdd a b

    (* Aux. routine to multiply bu with g Tree *)
    let rec rBG bu rgTree =
      match bu, rgTree with
	Pu bu, Leaf (gr, _) -> Leaf (Some (bu *$ gr))
      | Su bu, Node (_, lrgTree, rrgTree) ->
	  let gl = getNodeData lrgTree
	  and gr = getNodeData rrgTree in
	  let bll = bu.bu11 *$ gl +$ bu.bu12 *$ gr
	  and blr = bu.bu22 *$ gr
	  and blrTree = rBG bu.bu21 lrgTree in
	  Node (None, Leaf (Some bll), addBG (Leaf (Some blr)) blrTree)
      | _ -> failwith "Mismatch in bu tree and g tree"

    (* Aux. routine to multiplly bl with g tree *)
    let rec lBG bl lgTree =
      match bl, lgTree with
	Pl bl, Leaf (gl, _) -> Leaf (Some (bl *$ gl))
      | Sl bl, Node (_, llgTree, rlgTree) ->
	  let gl = getNodeData llgTree
	  and gr = getNodeData rlgTree in
	  let brl = bl.bl11 *$ gl
	  and brr = bl.bl21 *$ gl +$ bl.bl22 *$ gr
	  and brlTree = lBG bl.bl12 rlgTree in
	  Node (None, addBG (Leaf (Some brl)) brlTree, Leaf (Some brr))
      |  _ -> failwith "Mismatch in bl and g tree"

    (* Fast fmm1D-vector multiply. *)
    let fastMul fmm1D x z =
      let rec upSweep fmm1D x =
	match fmm1D with
	  Leaf a -> Leaf (a.v *~$ x, a.d *$ x)
	| Node (a, lfmm, rfmm) ->
	    let (xl, xr) = partition2x1 a.nl a.nr x in
	    let lgTree = upSweep lfmm xl
	    and rgTree = upSweep rfmm xr in
	    let gl = getNodeData lgTree
	    and gr = getNodeData rgTree in
	    Node (a.wl *~$ gl +$ a.wr *~$ gr, lgTree, rgTree)
      in
      let rec downSweep fmm1D gTree bgTree f z =
	match fmm1D, gTree with
	  Leaf a, Leaf (_, dx) ->
	    z <-$ dx;
	    begin
	      match bgTree with
		Leaf (Some bg) -> z <-| (a.u |*++| (f +$ bg))
	      | Leaf None      -> z <-| (a.u |*++|  f       )
	      | Node _         -> failwith "Bad bgTree"
	    end
	| Node (a, lfmm, rfmm), Node (_, lgTree, rgTree) ->
	    let lbg = rBG a.bu rgTree
	    and rbg = lBG a.bl lgTree
	    and (aBG, olbg, orbg) =
	      match bgTree with
		Leaf aBG -> (aBG, Leaf None, Leaf None)
	      | Node (aBG, olbg, orbg) -> (aBG, olbg, orbg)
	    in
	    let nlbg = addBG lbg olbg
	    and nrbg = addBG rbg orbg in
	    let nf =
	      match aBG with
		None -> f
	      | Some bg -> f +$ bg
	    in
	    matrix2x1 a.ml (downSweep lfmm lgTree nlbg (a.rl *$ nf))
	              a.mr (downSweep rfmm rgTree nrbg (a.rr *$ nf)) z
	| _ -> failwith "fmm and g tree do not match"
      in
      match fmm1D with
	Leaf a -> z <-| (a.d |*| x)
      |	_ -> downSweep fmm1D (upSweep fmm1D x) (Leaf None)
	    (zeros 0 (noOfCols x)) z


    (* Aux. routine to merge bu *)
    let rec mergeBU bu lFmm rFmm =
      let rec mbu11 bu11 rFmm =
	match rFmm with
	  Leaf _ -> bu11
	| Node (a, lrFmm, rrFmm) ->
	    let m = noOfRows bu11
	    and n1 = noOfRows a.wl
	    and n2 = noOfRows a.wr in
	    let bu11' = rawMatrix m (n1+n2) in
	    matrix1x2         n1                  n2
	      (mbu11 bu11 lrFmm |*~| a.wl) (bu11 |*~| a.wr) bu11';
	    bu11'
      in
      let rec mbu22 bu22 lFmm =
	match lFmm with
	  Leaf _ -> bu22
	| Node (a, llFmm, rlFmm) ->
	    let n = noOfCols bu22
	    and m1 = noOfRows a.rl
	    and m2 = noOfRows a.rr in
	    let bu22' = rawMatrix (m1+m2) n in
	    matrix2x1 m1 (a.rl |*| bu22)
	              m2 (a.rr |*| mbu22 bu22 rlFmm) bu22';
	    bu22
      in
      match bu, lFmm, rFmm with
	Pu bu, Leaf _, Leaf _ -> bu
      | Su bu, Node (al, llFmm, rlFmm), Node (ar, lrFmm, rrFmm) ->
	  let m1 = noOfRows bu.bu12
	  and n2 = noOfCols bu.bu12
	  and bu11' = mbu11 bu.bu11 lrFmm
	  and bu22' = mbu22 bu.bu22 rlFmm
	  and bu21' = mergeBU bu.bu21 rlFmm lrFmm in
	  let m2 = noOfRows bu21'
	  and n1 = noOfCols bu21' in
	  let bu' = rawMatrix (m1+m2) (n1+n2) in
	  matrix2x2 n1         n2
	    m1 (iD bu11') (iD bu.bu12)
	    m2 (iD bu21')   (iD bu22') bu';
	  bu'
      | Su bu, Leaf _, Node (ar, lrFmm, rrFmm) ->
	  let m = noOfRows bu.bu22
	  and n2 = noOfCols bu.bu22
	  and bu21' = mergeBU bu.bu21 lFmm lrFmm in
	  let n1 = noOfCols bu21' in
	  let bu' = rawMatrix m (n1+n2) in
	  matrix1x2 n1          n2
	        (iD bu21') (iD bu.bu22) bu';
	  bu'
      | Su bu, Node (al, llFmm, rlFmm), Leaf _ ->
	  let n = noOfCols bu.bu11
	  and m1 = noOfRows bu.bu11
	  and bu21' = mergeBU bu.bu21 rlFmm rFmm in
	  let m2 = noOfRows bu21' in
	  let bu' = rawMatrix (m1+m2) n in
	  matrix2x1 m1 (iD bu.bu11)
	            m2 (iD bu21')   bu';
	  bu'
      | _ -> failwith "Mismatched bu and fmm trees"

    (* Aux. routine to merge bl *)
    let rec mergeBL bl lFmm rFmm =
      let rec mbl11 bl11 rFmm =
	match rFmm with
	  Leaf _ -> bl11
	| Node (a, lrFmm, rrFmm) ->
	    let n = noOfCols bl11
	    and m1 = noOfRows a.rl
	    and m2 = noOfRows a.rr in
	    let bl11' = rawMatrix (m1+m2) n in
	    matrix2x1 m1 (a.rl |*| mbl11 bl11 lrFmm)
	              m2 (a.rr |*| bl11)            bl11';
	    bl11'
      in
      let rec mbl22 bl22 lFmm =
	match lFmm with
	  Leaf _ -> bl22
	| Node (a, llFmm, rlFmm) ->
	    let m = noOfRows bl22
	    and n1 = noOfRows a.wl
	    and n2 = noOfRows a.wr in
	    let bl22' = rawMatrix m (n1+n2) in
	    matrix1x2 n1                      n2
	      (bl22 |*~| a.wl) (mbl22 bl22 rlFmm |*~| a.wr) bl22';
	    bl22'
      in
      match bl, lFmm, rFmm with
	Pl bl, Leaf _, Leaf _ -> bl
      | Sl bl, Node (al, llFmm, rlFmm), Node (ar, lrFmm, rrFmm) ->
	  let m2 = noOfRows bl.bl21
	  and n1 = noOfCols bl.bl21
	  and bl11' = mbl11 bl.bl11 lrFmm
	  and bl22' = mbl22 bl.bl22 rlFmm
	  and bl12' = mergeBL bl.bl12 rlFmm lrFmm in
	  let m1 = noOfRows bl11'
	  and n2 = noOfCols bl22' in
	  let bl' = rawMatrix (m1+m2) (n1+n2) in
	  matrix2x2 n1          n2
	    m1 (iD bl11')   (iD bl12')
	    m2 (iD bl.bl21) (iD bl22') bl';
	  bl'
      | Sl bl, Leaf _, Node (ar, lrFmm, rrFmm) ->
      | Sl bl, Node (al, llFmm, rlFmm), Leaf _ ->
      | _ -> failwith "Mismatched bl and fmm trees"


    (* Data structure to hold V factor of ULV *)
    type wChiLeaf = {w : transform; chi : matrix}
    type wChiNode = {wN : transform; chiN : matrix; nm0 : int; nm1 : int}

    (* fmm1D is destroyed in the process. *)
    let fastSolve' fmm1D b x =
      let nrhs = noOfCols b in
      assert (nrhs = noOfCols x);
      let eliminateLeaf a b =
	print_endline "eliminating leaf";
	if noOfCols a.u >= noOfRows a.u then (* Nothing to compress *)
	  (Leaf {w = eyeTransform; chi = zeros 0 nrhs},
	   Leaf (zeros (noOfCols a.v) nrhs, ()), (* g *)
	   b, (* remaining b *)
	   a) (* remaining u, v, d *)
	else (* Compress *)
	  let ku = noOfCols a.u
	  and m = noOfRows a.u
	  and n = noOfCols a.d
	  and (q, l_u) = ql' a.u in (* u is destroyed *)
	  print_endline "Compressing leaf";
	  q @~* b;                   (* b is destroyed *)
	  q @~* a.d;
	  let (b0, b1) = partition2x1 (m-ku) ku b
	  and (d0, d1) = partition2x1 (m-ku) ku a.d in
	  let (d00, w) = lq' d0     (* d is destroyed *) in
	  d1 *~@ w;
	  w @* a.v;
	  subs' d00 b0;              (* b is destroyed *)
	  let (d10, d11) = partition1x2 (m-ku) (n-m+ku) d1
	  and (v0, v1)   = partition2x1 (m-ku) (n-m+ku) a.v
	  and chi = b0 in
	  (Leaf {w = w; chi = chi},
	   Leaf (v0 *~$ chi, ()),       (* g *)
	   b1 -$ d10 *$ chi, (* remaining b *)
	   {u = copyL l_u; v = v1; d = d11})
      in
      let rec ulSweep fmm b bgTree f =
	match fmm with
	  Leaf a ->
	    begin match bgTree with
	      Leaf None      -> eliminateLeaf a (b -$ a.u *$ f)
	    | Leaf (Some bg) -> eliminateLeaf a (b -$ a.u *$ (f +$ bg))
	    | Node _         -> failwith "bgTree mismatch with fmm tree"
	    end
	| Node (a, lFmm, rFmm) ->
	    let (b0, b1) = partition2x1 a.ml a.mr b in
	    let (aBG, lbg, rbg) =
	      match bgTree with
		Leaf aBG             -> (aBG, Leaf None, Leaf None)
	      | Node (aBG, lbg, rbg) -> (aBG, lbg,       rbg)
	    in
	    let nf =
	      match aBG with
		None    -> f
	      | Some bg -> f +$ bg
	    in
	    print_endline "Doing leftsweep";
	    let (lWchiTree, gl, bl, al) = ulSweep lFmm b0  lbg (a.rl *$ nf) in
	    print_endline "AddBG";
	    let nrbg = addBG (lBG a.bl gl) rbg in
	    print_endline "Right sweep";
	    let (rWchiTree, gr, br, ar) = ulSweep rFmm b1 nrbg (a.rr *$ nf) in
	    let m0 = noOfRows al.d
	    and n0 = noOfCols al.d
	    and m1 = noOfRows ar.d
	    and n1 = noOfCols ar.d in
	    let d  = rawMatrix (m0+m1) (n0+n1)
	    and u  = rawMatrix (m0+m1) (noOfCols a.rl)
	    and v  = rawMatrix (n0+n1) (noOfCols a.wl)
	    and bm = rawMatrix (m0+m1) nrhs
	    and mbu = mergeBU a.bu lFmm rFmm
	    and mbl = mergeBL a.bl lFmm rFmm in
	    print_endline "Merging leaves";
	    matrix2x2     n0                        n1
	           m0 (iD al.d) (al.u |*| (mbu *$~ ar.v))
	           m1 (ar.u |*| (mbl *$~ al.v)) (iD ar.d) d;
	    matrix2x1 m0 (al.u |*| a.rl)
	              m1 (ar.u |*| a.rr) u;
	    matrix2x1 n0 (al.v |*| a.wl)
	              n1 (ar.v |*| a.wr) v;
	    matrix2x1 m0 (bl |-| al.u *$ mbu *$ getNodeData gr)
	              m1 (iD br) bm;
	    let retLeaf = eliminateLeaf {u = u; v = v; d = d} bm in
	    print_endline "merged leaf eliminated";
	    match retLeaf with
	      Leaf wCh, g, bmr, amr ->
		(Node ({wN = wCh.w; chiN = wCh.chi; nm0 = n0; nm1 = n1},
		       lWchiTree, rWchiTree),
		 Node (getNodeData g +$ a.wl *~$ getNodeData gl
			 +$ a.wr *~$ getNodeData gr, gl, gr),
		 bmr, amr)
	    | _ -> failwith "eliminateLeaf returned Node!"
      in
      let rec vSweep fmm wChiTree xp x =
	match fmm, wChiTree with
	  Leaf _, Leaf wc ->
	    matrix2x1 (noOfRows wc.chi) (iD wc.chi)
	              (noOfRows xp    ) (iD xp)     x;
	    wc.w @~* x
	| Node (a, lFmm, rFmm), Node (wc, lWchi, rWchi) ->
	    let mc = noOfRows wc.chiN
	    and mx = noOfRows xp in
	    let chip = rawMatrix (mc+mx) nrhs in
	    matrix2x1 mc (iD wc.chiN)
	              mx (iD xp)      chip;
	    wc.wN @~* chip;
	    let (xp0, xp1) = partition2x1 wc.nm0 wc.nm1 chip in
	    matrix2x1 a.nl (vSweep lFmm lWchi xp0)
	              a.nr (vSweep rFmm rWchi xp1) x
	| _ -> failwith "Mismatched wchi and fmm trees"
      in
      match fmm1D with
	Leaf a ->
	  x <-$ b;
	  a.d |/| x
	| _ ->
	    let (wChiTree, _, _, _) = ulSweep fmm1D b (Leaf None)
		(zeros 0 nrhs) in
	    vSweep fmm1D wChiTree (zeros 0 nrhs) x


  end (* module Fmm1Der (Blop : BLOP) *)


module DFmm1D = Fmm1Der (DBlop)
module SFmm1D = Fmm1Der (SBlop)
module CFmm1D = Fmm1Der (CBlop)
module ZFmm1D = Fmm1Der (ZBlop)


