(*
  File: hss.ml
  Author: Shivkumar Chandrasekaran
  Email: shiv@ece.ucsb.edu
  Date: 5.8.2002
  Purpose: Fast algorithms for hierarchically semiseparable matrices

  $Id: hss.ml,v 1.36 2006/05/30 23:53:08 shiv Exp shiv $
*)

open Bigarray
open Utils
open Nla
open Sparse

module type HSS =
  sig
    type t
    type matrix
    type transform
    type hss_leaf = {u : matrix; v : matrix; d : matrix}
    type hss_node = {
      nl : int; nr : int;
      ml : int; mr : int;
      bu : matrix; bl : matrix;
      rl : matrix; rr : matrix;
      wl : matrix; wr : matrix }
    type hss = (hss_leaf, hss_node) btree
    type ulvFactor
    val no_of_rows : hss -> int
    val no_of_cols : hss -> int
    val fastMul : hss -> matrix -> matrix -> unit
    val fastMulTransp : hss -> matrix -> matrix -> unit
    val fastSolve' : hss -> matrix -> matrix -> unit
    val fastImplicitULVfactor' : hss -> ulvFactor
    val fastImplicitULVsolve : ulvFactor -> matrix -> matrix -> unit
    val copy : ?ooc:bool -> hss -> hss
    type hss_spine_node =
	{mLeft : int; mRight : int; nLeft : int; nRight : int}
    and hss_spine_leaf = unit
    val fromFn : ?ooc:bool -> (hss_spine_leaf, hss_spine_node) btree ->
      (int -> int -> int -> int -> matrix -> unit) -> float -> hss
    val fastModelReduction : ?ooc:bool -> float -> hss -> hss
    val fastAdd : ?ooc:bool -> hss -> hss -> hss
    val fastSub : ?ooc:bool -> hss -> hss -> hss
    val toProperForm' : ?ooc:bool -> hss -> hss
    val superfastMul : ?ooc:bool -> hss -> hss -> hss
    val addDiag' : hss -> matrix -> unit
    val addBlkDiag' : matrix array -> hss -> unit
    val addLowRank : ?ooc:bool -> hss -> matrix -> matrix -> matrix -> hss
    val toND2d : ?ooc:bool -> hss -> int -> int -> hss -> hss
    val toND2d_R : ?ooc:bool -> hss -> int -> int -> hss -> hss
    val toND2dPerm : hss -> int -> int -> hss ->
      (transform * (int, int_elt, fortran_layout) Array2.t)
    val mergeSingleChildNodes : ?ooc:bool -> hss -> hss
    val hssTensorEye : ?ooc:bool -> hss -> int -> hss
    val mulDiag : ?ooc:bool -> matrix array -> hss -> hss
    val mulDiagFn : ?ooc:bool -> (int -> matrix) -> hss -> hss
    val zeroRows' : int list -> hss -> unit
    val fromLowRank' : ?ooc:bool ->
      (hss_spine_leaf, hss_spine_node) btree -> matrix -> matrix ->
	float -> hss
    val toHSSspine : hss -> (hss_spine_leaf, hss_spine_node) btree
    type sparse
    val toSparse : hss -> sparse
    val toSparse_vec : hss -> matrix -> matrix
    val fromSparse_vec : hss -> matrix -> matrix
    val fastSparseSolve : hss -> matrix -> matrix
    val fastMRsolve' : float -> hss -> matrix -> matrix -> unit
  end (* module type HSS *)


module Hsser (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix) =
  struct

    open Blop
    type t = Blop.t
    type matrix = Blop.matrix
    type transform = Blop.transform
    type sparse = Sparse.sparse

    type hss_leaf = {u : matrix; v : matrix; d : matrix}

    type hss_node = {
      nl : int; nr : int;
      ml : int; mr : int;
      bu : matrix; bl : matrix;
      rl : matrix; rr : matrix;
      wl : matrix; wr : matrix }

    type hss = (hss_leaf, hss_node) btree

    let rec copy ?(ooc=false) hss =
      match hss with
	Leaf {u = u; v = v; d = d}
	-> Leaf {u = Blop.copy ~ooc u; v = Blop.copy ~ooc v;
		 d = Blop.copy ~ooc d}
      | Node (a, ltree, rtree) ->
	  Node ({bu = Blop.copy ~ooc a.bu; bl = Blop.copy ~ooc a.bl;
		 rl = Blop.copy ~ooc a.rl; rr = Blop.copy ~ooc a.rr;
		 wl = Blop.copy ~ooc a.wl; wr = Blop.copy ~ooc a.wr;
		 nl = a.nl; nr = a.nr; ml = a.ml; mr = a.mr},
		copy ~ooc ltree, copy ~ooc rtree)

    let no_of_rows hss =
      match hss with
	  Leaf a -> noOfRows a.d
	| Node (a, _, _) -> a.ml + a.mr

    let no_of_cols hss =
      match hss with
	  Leaf a -> noOfCols a.d
	| Node (a, _, _) -> a.nl + a.nr

    (* Fast multiplication routines *)
    let fastMul hss b z =
      let getNodeData nl =
	match nl with
	    Leaf (d, _) -> d
	  | Node (d, _, _) -> d
      in
      let rec upSweep hss b =
	match hss with
	    Leaf a -> Leaf (a.v *~$ b, a.d *$ b)
	  | Node (a, l, r) ->
	      let bl, br = partition2x1 a.nl a.nr b in
	      let glTree = upSweep l bl
	      and grTree = upSweep r br in
	      let gl = getNodeData glTree
	      and gr = getNodeData grTree in
	      Node (a.wl *~$ gl +$ a.wr *~$ gr, glTree, grTree) in
      let rec downSweep hss gTree f z =
	match (hss, gTree) with
	    (Leaf a, Leaf (_, db)) ->
	      z <-$ db;
	      z <-| (a.u |*++| f)
	  | (Node (a, lHss, rHss), Node (_, lGtree, rGtree)) ->
	      let gl = getNodeData lGtree
	      and gr = getNodeData rGtree in
	      matrix2x1 a.ml (downSweep lHss lGtree (a.bu *$ gr +$ a.rl *$ f))
	                a.mr (downSweep rHss rGtree (a.bl *$ gl +$ a.rr *$ f))
		z
	  | _ -> failwith "fastMul:downSweep: Unmatched trees."
      in
      match hss with
	  Leaf a -> z <-| (a.d |*| b)
	| _ -> downSweep hss (upSweep hss b) (zeros 0 (noOfCols b)) z

    (* hss^T b -> z *)
    let fastMulTransp hss b z =
      let getNodeData nl =
	match nl with
	    Leaf (d, _) -> d
	  | Node (d, _, _) -> d
      in
      let rec upSweep hss b =
	match hss with
	    Leaf a -> Leaf (a.u *~$ b, a.d *~$ b)
	  | Node (a, l, r) ->
	      let bl, br = partition2x1 a.ml a.mr b in
	      let glTree = upSweep l bl
	      and grTree = upSweep r br in
	      let gl = getNodeData glTree
	      and gr = getNodeData grTree in
	      Node (a.rl *~$ gl +$ a.rr *~$ gr, glTree, grTree) in
      let rec downSweep hss gTree f z =
	match (hss, gTree) with
	    (Leaf a, Leaf (_, db)) ->
	      z <-$ db;
	      z <-| (a.v |*++| f)
	  | (Node (a, lHss, rHss), Node (_, lGtree, rGtree)) ->
	      let gl = getNodeData lGtree
	      and gr = getNodeData rGtree in
	      matrix2x1 a.nl (downSweep lHss lGtree (a.bl *~$ gr +$ a.wl *$ f))
	                a.nr (downSweep rHss rGtree (a.bu *~$ gl +$ a.wr *$ f))
		z
	  | _ -> failwith "fastMul:downSweep: Unmatched trees."
      in
      match hss with
	  Leaf a -> z <-| (a.d |~*| b)
	| _ -> downSweep hss (upSweep hss b) (zeros 0 (noOfCols b)) z


(* Fast solver routines *)
  (* At any node, compress left child, get its g, convert it to f for
     right child, compress right child, get g from right child, convert to
     f for left child, merge left and right child, merge their g's.
  *)
    type wChiLeaf = { w : transform; chi : matrix }
    and wChiNode =
	{ wN : transform; chiN : matrix; nm0 : int; nm1 : int }

    let fastSolve' hss b x =
      let nrhs = noOfCols b in
      assert (nrhs = noOfCols x);
      let eliminateLeaf a b =
	if noOfCols a.u >= noOfRows a.u then (* Nothing to compress *)
	  (
	    Leaf { w = eyeTransform; chi = zeros 0 nrhs },
	    zeros (noOfCols a.v) nrhs, (* g *)
	    b,                         (* remaining b *)
	    a.u,                       (* remaining u *)
	    a.v,                       (* remaining v *)
	    a.d                        (* remaining d *)
	  )
	else (* Compress *)
	  let ku = noOfCols a.u
	  and m = noOfRows a.u
	  and n = noOfCols a.d 
	  and (q, l_u) = ql' a.u in
	  q @~* b;
	  q @~* a.d;
	  let (b0, b1) = partition2x1 (m-ku) ku b
	  and (d0, d1) = partition2x1 (m-ku) ku a.d in
	  let (d00, w) = lq' d0 in
	  d1 *~@ w;
	  w @* a.v;
	  subs' d00 b0;
	  let (d10, d11) = partition1x2 (m-ku) (n-m+ku) d1
	  and (v00, v01) = partition2x1 (m-ku) (n-m+ku) a.v
	  and chi = b0 in
	  (
	    Leaf { w = w; chi = chi },
	    v00 *~$ chi,
	    b1 -$ d10 *$ chi,
	    copyL l_u,
	    v01,
	    d11
	  )
      in
      let rec ulSweep hss b f =
	match hss with
	    Leaf a -> eliminateLeaf a (b -$ a.u *$ f)
	  | Node (a, lHss, rHss) ->
	      let (b0, b1) = partition2x1 a.ml a.mr b in
	      let (lWchiTree, gl, bl, ul, vl, dl) = ulSweep lHss b0
		(a.rl *$ f) in
	      let (rWchiTree, gr, br, ur, vr, dr) = ulSweep rHss b1
		(a.bl *$ gl +$ a.rr *$ f) in
	      let m0 = noOfRows dl
	      and n0 = noOfCols dl
	      and m1 = noOfRows dr
	      and n1 = noOfCols dr in
	      let d = rawMatrix (m0+m1) (n0+n1)
	      and u = rawMatrix (m0+m1) (noOfCols a.rl)
	      and v = rawMatrix (n0+n1) (noOfCols a.wl)
	      and bm = rawMatrix (m0+m1) nrhs in
	      matrix2x2   n0                      n1
		     m0 (iD dl) (ul |*| (a.bu *$~ vr))
		     m1 (ur |*| (a.bl *$~ vl)) (iD dr) d;
	      matrix2x1 m0 (ul |*| a.rl)
		        m1 (ur |*| a.rr) u;
	      matrix2x1 n0 (vl |*| a.wl)
		        n1 (vr |*| a.wr) v;
	      matrix2x1 m0 (bl |-| ul *$ a.bu *$ gr)
		        m1 (iD br                  ) bm;
	      let retLeaf = eliminateLeaf {u = u; v = v; d = d} bm in
	      match retLeaf with
		  (Leaf wCh, g, bmr, umr, vmr, dmr) ->
		    (
		      Node
		      (
			{wN = wCh.w; chiN = wCh.chi; nm0 = n0; nm1 = n1},
			lWchiTree, rWchiTree
		      ),
		      g +$ a.wl *~$ gl +$ a.wr *~$ gr,
		      bmr,
		      umr,
		      vmr,
		      dmr
		    )
		| _ -> failwith "eliminateLeaf returned Node!"
      in
      let rec vSweep hss wChiTree xp x =
	match (hss, wChiTree) with
	    (Leaf _, Leaf {w=w; chi=chi}) ->
	      matrix2x1 (noOfRows chi) (iD chi)
	                (noOfRows xp ) (iD xp ) x;
	      w @~* x
	  | (Node (a, lHss, rHss),
	    Node ({wN=w;chiN=chi;nm0=n0;nm1=n1}, lWchi, rWchi)) ->
	      let mc = noOfRows chi
	      and mx = noOfRows xp in
	      let chip = rawMatrix (mc+mx) nrhs in
	      matrix2x1 mc (iD chi)
		        mx (iD xp ) chip;
	      w @~* chip;
	      let (xp0, xp1) = partition2x1 n0 n1 chip in
	      matrix2x1 a.nl (vSweep lHss lWchi xp0)
	                a.nr (vSweep rHss rWchi xp1) x
	  | _ -> failwith "fastSolve': Mismatched trees."
      in
      match hss with
	  Leaf a ->
	    x <-$ b;
	    a.d |/| x
	| _ ->
	    let (wChiTree, _, _, _, _, _) = ulSweep hss b
	      (zeros 0 nrhs) in
	    vSweep hss wChiTree (zeros 0 nrhs) x


    (* Fast implicit ULV factorization. *)
    type compressLeaf = { vw : transform; uq : transform; ul : matrix;
			  mu : int; ku : int; v00 : matrix;
			  d00 : factor; d10 : matrix }
    and fullLeaf = { nv : int; au : matrix }
    and ulvLeaf =
	Single of matrix (* Just a dense matrix *)
      | Full of fullLeaf (* At incompressible leaf *)
      | Compress of compressLeaf
    and ulvNode = { mleaf : ulvLeaf;
		    bm0 : int; bm1 : int; nn0 : int; nn1 : int;
		    nwl : matrix; nwr : matrix; nrl : matrix; nrr : matrix;
		    nbl : matrix; nbu : matrix;
		    bml : int; bmr : int; xnl : int; xnr : int }
    and ulvFactor = (matrix * ulvLeaf, ulvNode) Utils.btree

    let fastImplicitULVfactor' hss =
      let eliminateLeaf a =
	if noOfCols a.u >= noOfRows a.u then (* nothing to compress *)
	  (
	   Leaf (a.u, Full {nv = noOfCols a.v; au = a.u}),
	   a.u, (* remaining u *)
	   a.v, (* remaining v *)
	   a.d  (* remaining d *)
	  )
	else (* Compress *)
	  let cu = Blop.copy a.u in
	  let ku = noOfCols a.u
	  and m = noOfRows a.u
	  and n = noOfCols a.d
	  and (q, l_u) = ql' a.u in
	  q @~* a.d;
	  let (d0, d1) = partition2x1 (m-ku) ku a.d in
	  let (d00, w) = lq' d0 in
	  d1 *~@ w;
	  w @* a.v;
	  let (d10, d11) = partition1x2 (m-ku) (n-m+ku) d1
	  and (v00, v01) = partition2x1 (m-ku) (n-m+ku) a.v
	  and ul = copyL l_u in
	  (
	   Leaf (cu, Compress { vw = w; uq = q; ul = ul; mu = m; ku = ku;
				v00 = v00; d00 = d00; d10 = d10 }),
	   ul,  (* remaining u *)
	   v01, (* remaining v *)
	   d11  (* remaining d *)
	  )
      in
      let rec ulSweep hss =
	match hss with
	  Leaf a -> eliminateLeaf a (* a.u must be saved now *)
	| Node (a, lHss, rHss) ->
	    let (lULVtree, ul, vl, dl) = ulSweep lHss in
	    let (rULVtree, ur, vr, dr) = ulSweep rHss in
	    let m0 = noOfRows dl
	    and n0 = noOfCols dl
	    and m1 = noOfRows dr
	    and n1 = noOfCols dr in
	    let d = rawMatrix (m0+m1) (n0+n1)
	    and u = rawMatrix (m0+m1) (noOfCols a.rl)
	    and v = rawMatrix (n0+n1) (noOfCols a.wl) in
	    matrix2x2   n0                      n1
	           m0 (iD dl) (ul |*| (a.bu *$~ vr))
	           m1 (ur |*| (a.bl *$~ vl)) (iD dr) d;
	    matrix2x1 m0 (ul |*| a.rl)
	              m1 (ur |*| a.rr) u;
	    matrix2x1 n0 (vl |*| a.wl)
	              n1 (vr |*| a.wr) v;
	    let retLeaf = eliminateLeaf {u = u; v = v; d = d} in
	    match retLeaf with
	      (Leaf (_, ulv), umr, vmr, dmr) ->
		(
		 Node
		   (
		    { mleaf = ulv; nn0 = n0; nn1 = n1; bm0 = m0; bm1 = m1;
		      nwl = a.wl; nwr = a.wr;
		      nrl = a.rl; nrr = a.rr; nbl = a.bl; nbu = a.bu;
		      bml = a.ml; bmr = a.mr; xnl = a.nl; xnr = a.nr },
		    lULVtree, rULVtree
		   ),
		 umr,
		 vmr,
		 dmr
		)
	    | _ -> failwith "eliminateLeaf returned Node!"
      in
      match hss with
	Leaf a -> Leaf (zeros 0 0, Single a.d)
      | _ ->
	  let (ulvTree, _, _, _) = ulSweep hss in
	  ulvTree


    (* Fast solve using the implicit ULV factor computed in
       fastImplicitULVfactor *)
   let fastImplicitULVsolve ulv b x =
     let nrhs = noOfCols b in
     assert (nrhs = noOfCols x);
     let eliminateLeaf a b =
       match a with
	 Full l -> (* Nothing to compress *)
	   (
	    Leaf { w = eyeTransform; chi = zeros 0 nrhs },
	    zeros l.nv nrhs, (* g *)
	    b,               (* remaining b *)
	    l.au             (* remaining u *)
	   )
       | Compress l -> (* Compress *)
	   l.uq @~* b;
	   let (b0, b1) = partition2x1 (l.mu-l.ku) l.ku b in
	   subs' l.d00 b0;
	   let chi = b0 in
	   (
	    Leaf { w = l.vw; chi = chi },
	    l.v00 *~$ chi,      (* g *)
	    b1 -$ l.d10 *$ chi, (* remaining b *)
	    l.ul                (* remaining u *)
	   )
       | Single _ -> failwith "Should be caught earlier"
     in
     let rec ulSweep ulv b f =
       match ulv with
	 Leaf (cu, a) -> eliminateLeaf a (b -$ cu *$ f)
       | Node (a, lULV, rULV) ->
	   let (b0, b1) = partition2x1 a.bml a.bmr b in
	   let (lWchiTree, gl, bl, ul) = ulSweep lULV b0
	       (a.nrl *$ f) in
	   let (rWchiTree, gr, br, ur) = ulSweep rULV b1
	       (a.nbl *$ gl +$ a.nrr *$ f) in
	   let bm = rawMatrix (a.bm0 + a.bm1) nrhs in
	   matrix2x1 a.bm0 (bl |-| ul *$ a.nbu *$ gr)
	             a.bm1 (iD br                   ) bm;
	   let retLeaf = eliminateLeaf a.mleaf bm in
	   match retLeaf with
	     (Leaf wCh, g, bmr, umr) ->
	       (
		Node
		  ({wN = wCh.w; chiN = wCh.chi; nm0 = a.nn0; nm1 = a.nn1},
		   lWchiTree, rWchiTree),
		g +$ a.nwl *~$ gl +$ a.nwr *~$ gr,
		bmr,
		umr
	       )
	   | _ -> failwith "eliminateLeaf returned Node!"
     in
     let rec vSweep ulv wChiTree xp x =
       match (ulv, wChiTree) with
	 (Leaf _, Leaf {w=w; chi=chi}) ->
	   matrix2x1 (noOfRows chi) (iD chi)
	             (noOfRows xp ) (iD xp ) x;
	   w @~* x
       | (Node (a, lULV, rULV),
	  Node ({wN=w; chiN=chi; nm0=n0; nm1 = n1}, lWchi, rWchi)) ->
	    let mc = noOfRows chi
	    and mx = noOfRows xp in
	    let chip = rawMatrix (mc+mx) nrhs in
	    matrix2x1 mc (iD chi)
	              mx (iD xp ) chip;
	    w @~* chip;
	    let (xp0, xp1) = partition2x1 n0 n1 chip in
	    matrix2x1 a.xnl (vSweep lULV lWchi xp0)
	              a.xnr (vSweep rULV rWchi xp1) x
       | _ -> failwith "fastImplicitULVsolve: Mismatched trees"
     in
     match ulv with
       Leaf (_, Single d) ->
	 x <-$ b;
	 d |/| x
       | _ ->
	   let (wChiTree, _, _, _) = ulSweep ulv b (zeros 0 nrhs) in
	   vSweep ulv wChiTree (zeros 0 nrhs) x
	 
	     

    (* numerical-rank SVD *)
    let svdE' ?(ooc=false) tol a =
      let k = min (noOfRows a) (noOfCols a) in
      let (u, s, vt) = svd' ~ooc a in
      let r = ref 0 in
      while !r < k && s.{!r+1} > tol do
	r := !r + 1
      done;
      assert (!r <= k && (!r = 0 || s.{!r} > tol));
      (u $ (All, 1 <-> !r), Array1.sub s 1 !r, vt $ (1 <-> !r, All))

    (* numerical-rank left-SVD *)
    let svdLE' ?(ooc=false) tol a =
      let k = min (noOfRows a) (noOfCols a) in
      let (u, s) = svdL' ~ooc a in
      let r = ref 0 in
      while !r < k && s.{!r+1} > tol do
	r := !r + 1
      done;
      assert (!r <= k && (!r = 0 || s.{!r} > tol));
      (u $ (All, 1 <-> !r), Array1.sub s 1 !r)

    type hss_spine_node =
	{mLeft : int; mRight : int; nLeft : int; nRight : int}
    and hss_spine_leaf = unit

    (* Will not work on Leaf ()! *)
    let fromFn ?(ooc=false) hssSpine fn tol =
      let handle_ooc a =
	if ooc then to_ooc a else a in
      let m = match hssSpine with
	  Leaf () -> 0
	| Node (a, _, _) -> a.mLeft + a.mRight
      and n = match hssSpine with
	  Leaf () -> 0
	| Node (a, _, _) -> a.nLeft + a.nRight in
      let rec ffn ml ms nl ns hssSpine =
	match hssSpine with
	    Leaf () ->
	      let d = rawMatrix ~ooc ms ns in
	      d <-| fn (ml+1) (ml+ms) (nl+1) (nl+ns);
	      let (uh, sh, vht) =
		let h = rawMatrix ms (n - ns) in
		matrix1x2 nl (n - nl - ns)
		  (fn (ml+1) (ml+ms) 1 nl)
		  (fn (ml+1) (ml+ms) (nl+ns+1) n) h;
		svdE' ~ooc tol h
	      and (ug, sg, vgt) =
		let g = rawMatrix (m - ms) ns in
		matrix2x1 ml    (fn 1 ml (nl+1) (nl+ns))
		  (m - ml - ms) (fn (ml+ms+1) m (nl+1) (nl+ns)) g;
		svdE' ~ooc tol g in
	      rowScale_real' sh vht;
	      colScale_real' ug sg;
	      let vg = transp ~ooc vgt in
	      (Leaf {d = d; u = uh; v = vg}, vht, ug, vg)
	  | Node (a, lSpine, rSpine) ->
	      let (lHss, lsvht, lusg, lv) = ffn ml a.mLeft nl
		(a.nLeft) lSpine
	      and (rHss, rsvht, rusg, rv) = ffn (ml+a.mLeft)
		a.mRight (nl+a.nLeft) a.nRight rSpine in
	      let hk1 = noOfRows lsvht
	      and hk2 = noOfRows rsvht
	      and vk1 = noOfCols lusg
	      and vk2 = noOfCols rusg in
	      let (lsvht1, lsvht2, lsvht3) =
		partition1x3 nl a.nRight (n - nl - ns) lsvht
	      and (rsvht1, rsvht2, rsvht3) =
		partition1x3 nl a.nLeft	(n - nl - ns) rsvht
	      and (lusg1, _, lusg3) =
		partition3x1 ml
		       a.mRight
		   (m - ml - ms) lusg
	      and (rusg1, _, rusg3) =
		partition3x1 ml
		        a.mLeft
		   (m - ml - ms) rusg in
	      let (rs, sh, vht) =
		let h = rawMatrix (hk1 + hk2) (n - ns) in
		let (h1, h2) = partition2x1 hk1 hk2 h in
		matrix1x2  nl (n - nl - ns)
		   (iD lsvht1)  (iD lsvht3) h1;
		matrix1x2 nl (n - nl - ns)
		  (iD rsvht1)   (iD rsvht3) h2;
		svdE' ~ooc tol h
	      and (ug, sg, wst) =
		let g = rawMatrix (m - ms) (vk1 + vk2) in
		let (g1, g2) = partition1x2 vk1 vk2 g in
		matrix2x1   ml  (iD lusg1)
		  (m - ml - ms) (iD lusg3) g1;
		matrix2x1   ml  (iD rusg1)
		  (m - ml - ms) (iD rusg3) g2;
		svdE' ~ooc tol g
	      in
	      let (rl, rr) = partition2x1 hk1 hk2 rs
	      and (wlt, wrt) = partition1x2 vk1 vk2 wst
	      and vg = rawMatrix ~ooc ns (noOfRows wst) in
	      matrix2x1 a.nLeft  (lv |*~| wlt)
		        a.nRight (rv |*~| wrt) vg;
	      let bu = handle_ooc (lsvht2 *$ rv)
	      and bl = handle_ooc (rsvht2 *$ lv) in
	      rowScale_real' sh vht;
	      colScale_real' ug sg;
	      (Node ({ml = a.mLeft; mr = a.mRight; nl = a.nLeft; nr =
		      a.nRight; rl = rl; rr = rr; wl = transp ~ooc wlt; wr =
		      transp ~ooc wrt; bu = bu; bl = bl}, lHss, rHss),
	       vht, ug, vg)
      in
      match hssSpine with
	Leaf _ -> failwith "fromFn: will not work on leaf"
      | _ ->
	  let (hss, _, _, _) = ffn 0 m 0 n hssSpine in hss

    let fastModelReduction ?(ooc=false) tol hss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let compr tr b r uh ss =
	let n1 = (if tr then noOfRows else noOfCols) b
	and n2 = noOfCols uh in
	let a = rawMatrix (noOfRows r) (n1 + n2) in
	let (a1, a2) = partition1x2 n1 n2 a in
	a1 <-| (if tr then transp' else iD) b;
	a2 <-| (r |*| uh);
	colScale_real' a2 ss;
	svdLE' ~ooc tol a
      in
      let rec fmr hss uh ss vh gs =
	match hss with
	    Leaf a ->
	      Leaf {u = handle_ooc (a.u *$ uh); d = handle_ooc a.d;
		    v = handle_ooc (a.v *$ vh)}
	  | Node (a, lHss, rHss) ->
	      let (uhl, shl) = compr false a.bu a.rl uh ss
	      and (uhr, shr) = compr false a.bl a.rr uh ss
	      and (vhl, ghl) = compr true a.bl a.wl vh gs
	      and (vhr, ghr) = compr true a.bu a.wr vh gs in
	      Node (
		{
		  bu = handle_ooc (uhl *~$ a.bu *$ vhr);
		  bl = handle_ooc (uhr *~$ a.bl *$ vhl);
		  rl = handle_ooc (uhl *~$ a.rl *$ uh);
		  rr = handle_ooc (uhr *~$ a.rr *$ uh);
		  wl = handle_ooc (vhl *~$ a.wl *$ vh);
		  wr = handle_ooc (vhr *~$ a.wr *$ vh);
		  nl = a.nl; nr = a.nr; ml = a.ml; mr = a.mr
		},
		fmr lHss uhl shl vhl ghl,
		fmr rHss uhr shr vhr ghr
	      )
      in
      fmr hss (rawMatrix 0 0)
	(Array1.create precision_real fortran_layout 0)
	(rawMatrix 0 0) (Array1.create precision_real fortran_layout
	  0)

    let rec fastAdd ?(ooc=false) aHss bHss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let b1x2 c d =
	let n1 = noOfCols c
	and n2 = noOfCols d in
	let bd = rawMatrix ~ooc (noOfRows c) (n1+n2) in
	matrix1x2 n1 n2
	  (iD c) (iD d) bd;
	bd
      and bdiag c d =
	let m1 = noOfRows c
	and m2 = noOfRows d
	and n1 = noOfCols c
	and n2 = noOfCols d in
	let bd = zeros ~ooc (m1+m2) (n1+n2) in
	matrix2x2 n1 n2
	  m1 (iD c) ooo
	  m2 ooo (iD d) bd;
	bd
      in
      match (aHss, bHss) with
	  (Leaf a, Leaf b) ->
	    Leaf {
	      d = handle_ooc (a.d +$ b.d);
	      u = b1x2 a.u b.u;
	      v = b1x2 a.v b.v
	    }
	| (Node (a, aL, aR), Node (b, bL, bR)) ->
	    Node (
	      {
		bu = bdiag a.bu b.bu;
		bl = bdiag a.bl b.bl;
		rl = bdiag a.rl b.rl;
		rr = bdiag a.rr b.rr;
		wl = bdiag a.wl b.wl;
		wr = bdiag a.wr b.wr;
		nl = a.nl; nr = a.nr; ml = a.ml; mr = a.mr
	      },
	      fastAdd aL bL,
	      fastAdd aR bR
	    )
	| _ -> failwith "fastAdd: non-conforming"

    let rec fastSub ?(ooc=false) aHss bHss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let b1x2 c d =
	let n1 = noOfCols c
	and n2 = noOfCols d in
	let bd = rawMatrix ~ooc (noOfRows c) (n1+n2) in
	matrix1x2 n1 n2
	  (iD c) (iD d) bd;
	bd
      and bdiag ?(fl=false) c d =
	let m1 = noOfRows c
	and m2 = noOfRows d
	and n1 = noOfCols c
	and n2 = noOfCols d in
	let bd = zeros ~ooc (m1+m2) (n1+n2) in
	matrix2x2 n1 n2
	  m1 (iD c) ooo
	  m2 ooo ((if fl then (|--|) else iD) d) bd;
	bd
      in
      match (aHss, bHss) with
	  (Leaf a, Leaf b) ->
	    Leaf {
	      d = handle_ooc (a.d -$ b.d);
	      u = b1x2 a.u b.u;
	      v = b1x2 a.v b.v
	    }
	| (Node (a, aL, aR), Node (b, bL, bR)) ->
	    Node (
	      {
		bu = bdiag ~fl:true a.bu b.bu;
		bl = bdiag ~fl:true a.bl b.bl;
		rl = bdiag a.rl b.rl;
		rr = bdiag a.rr b.rr;
		wl = bdiag a.wl b.wl;
		wr = bdiag a.wr b.wr;
		nl = a.nl; nr = a.nr; ml = a.ml; mr = a.mr
	      },
	      fastSub aL bL,
	      fastSub aR bR
	    )
	| _ -> failwith "fastSub: non-conforming"

    let toProperForm' ?(ooc=false) hss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let qlf' a =
	let m = noOfRows a
	and n = noOfCols a in
	let k = min m n in
	let (q, l) = ql' a in
	let qf = zeros ~ooc m k in
	for i = 1 to k do
	  (one => (m - k + i, i)) qf
	done;
	q @* qf;
	(qf, copyL ~ooc l)
      in
      let mf rl rr rhl rhr =
	let m1 = noOfRows rhl
	and m2 = noOfRows rhr in
	let rc = rawMatrix ~ooc (m1+m2) (noOfCols rl) in
	matrix2x1
	  m1 (rhl |*| rl)
	  m2 (rhr |*| rr) rc;
	let (rcq, rh) = qlf' rc in
	let (rnewl, rnewr) = partition2x1 m1 m2 rcq in
	(rnewl, rnewr, rh)
      in
      let rec tpf hss =
	match hss with
	    Leaf a ->
	      let (unew, rh) = qlf' a.u
	      and (vnew, wh) = qlf' a.v in
	      (Leaf {d = handle_ooc a.d; u = unew; v = vnew}, rh, wh)
	  | Node (a, lHss, rHss) ->
	      let (lHssPF, rhl, whl) = tpf lHss
	      and (rHssPF, rhr, whr) = tpf rHss in
	      let (rlnew, rrnew, rh) = mf a.rl a.rr rhl rhr
	      and (wlnew, wrnew, wh) = mf a.wl a.wr whl whr in
	      (
		Node (
		  {
		    bu = handle_ooc (rhl *$ a.bu *$~ whr);
		    bl = handle_ooc (rhr *$ a.bl *$~ whl);
		    rl = rlnew; rr = rrnew;
		    wl = wlnew; wr = wrnew;
		    nl = a.nl; nr = a.nr; ml = a.ml; mr = a.mr
		  },
		  lHssPF, rHssPF
		),
		rh, wh
	      )
      in
      let (hssC, _, _) = tpf hss in hssC

    let superfastMul ?(ooc=false) ahss bhss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let size a = (noOfRows a, noOfCols a) in
      let getNodeData a =
	match a with
	    Leaf a -> a
	  | Node (a, _, _) -> a
      in
      let rec upSweep ahss bhss =
	match (ahss, bhss) with
	    (Leaf a, Leaf b) -> Leaf (handle_ooc (a.v *~$ b.u))
	  | (Node (a, lahss, rahss), Node (b, lbhss, rbhss)) ->
	      let lGtree = upSweep lahss lbhss
	      and rGtree = upSweep rahss rbhss in
	      let gl = getNodeData lGtree
	      and gr = getNodeData rGtree in
	      Node (
		handle_ooc (a.wl *~$ gl *$ b.rl +$ a.wr *~$ gr *$ b.rr),
		lGtree,
		rGtree
	      )
	  | _ -> failwith "superfastMul:upSweep: mismatched trees."
      in
      let rec downSweep ahss bhss gTree f =
	match (ahss, bhss, gTree) with
	    (Leaf a, Leaf b, Leaf _) ->
	      let kua = noOfCols a.u
	      and kub = noOfCols b.u
	      and kva = noOfCols a.v
	      and kvb = noOfCols b.v in
	      let u = rawMatrix ~ooc (noOfRows a.d) (kua+kub)
	      and v = rawMatrix ~ooc (noOfCols b.d) (kva+kvb) in
	      matrix1x2 kua kub
		(iD a.u) (a.d |*| b.u) u;
	      matrix1x2 kva kvb
		(b.d |~*| a.v) (iD b.v) v;
	      Leaf {
		d = handle_ooc (a.d *$ b.d +$ a.u *$ f *$~ b.v);
		u = u;
		v = v
	      }
	  | (Node (a, lahss, rahss), Node (b, lbhss, rbhss), Node (_,
	    lgtree, rgtree)) ->
	      let gl = getNodeData lgtree
	      and gr = getNodeData rgtree
	      and (mrla, nrla) = size a.rl
	      and (mrra, nrra) = size a.rr
	      and (mrlb, nrlb) = size b.rl
	      and (mrrb, nrrb) = size b.rr
	      and (mwla, nwla) = size a.wl
	      and (mwra, nwra) = size a.wr
	      and (mwlb, nwlb) = size b.wl
	      and (mwrb, nwrb) = size b.wr
	      and (mbua, nbua) = size a.bu
	      and (mbla, nbla) = size a.bl
	      and (mbub, nbub) = size b.bu
	      and (mblb, nblb) = size b.bl in
	      let rl = zeros ~ooc (mrla+mrlb) (nrla+nrlb)
	      and rr = zeros ~ooc (mrra+mrrb) (nrra+nrrb)
	      and wl = zeros ~ooc (mwla+mwlb) (nwla+nwlb)
	      and wr = zeros ~ooc (mwra+mwrb) (nwra+nwrb)
	      and bu = zeros ~ooc (mbua+mbub) (nbua+nbub)
	      and bl = zeros ~ooc (mbla+mblb) (nbla+nblb) in
	      matrix2x2 nrla              nrlb
		mrla (iD a.rl) (a.bu |*| (gr *$ b.rr))
		mrlb ooo                     (iD b.rl) rl;
	      matrix2x2 nrra              nrrb
		mrra (iD a.rr) (a.bl |*| (gl *$ b.rl))
		mrrb ooo                     (iD b.rr) rr;
	      matrix2x2        nwla              nwlb
		mwla (iD a.wl)                      ooo
		mwlb (b.bl |~*| (gr *~$ a.wr)) (iD b.wl) wl;
	      matrix2x2       nwra              nwrb
		mwra (iD a.wr)                      ooo
		mwrb (b.bu |~*| (gl *~$ a.wl)) (iD b.wr) wr;
	      matrix2x2 nbua           nbub
		mbua (iD a.bu) (a.rl |*| (f *$~ b.wr))
		mbub ooo                     (iD b.bu) bu;
	      matrix2x2 nbla           nblb
		mbla (iD a.bl) (a.rr |*| (f *$~ b.wl))
		mblb ooo                     (iD b.bl) bl;
	      let fl = handle_ooc (a.bu *$ gr *$ b.bl +$ a.rl *$ f *$~ b.wl)
	      and fr = handle_ooc (a.bl *$ gl *$ b.bu +$ a.rr *$ f *$~ b.wr) in
	      let lhss = downSweep lahss lbhss lgtree fl
	      and rhss = downSweep rahss rbhss rgtree fr in
	      Node (
		{
		  bu = bu; bl = bl; rl = rl; rr = rr;
		  wl = wl; wr = wr; nl = b.nl; nr = b.nr;
		  ml = a.ml; mr = a.mr
		},
		lhss,
		rhss
	      )
	  | _ -> failwith "superfastMul:downSweep: mismatched trees."
      in
      let gtree = upSweep ahss bhss in
      downSweep ahss bhss gtree (zeros 0 0)


    let rec addDiag' hss diag =      
      match hss with
	Leaf {d=d} ->
	  for i = 1 to min (noOfRows d) (noOfCols d) do
	    (addn (diag $@ (i,1)) (d $@ (i,i)) => (i,i)) d
	  done
      | Node ({ml=ml; mr=mr}, lhss, rhss) ->
	  let (diag1, diag2) = partition2x1 ml mr diag in
	  addDiag' lhss diag1;
	  addDiag' rhss diag2

    (* Add block diagonal. Diagonal of hss is modified in place. *)
    let addBlkDiag' ds hss =
      let rec addExactDiag i j ds od =
	match ds with
	  [] -> ()
	| di :: ds ->
	    let mi = noOfRows di
	    and ni = noOfCols di in
	    let odi = od $ (i <-> i+mi-1, j <-> j+ni-1) in
	    di |++| odi;
	    addExactDiag (i+mi) (j+ni) ds od
      in
      let splitFor m ds =
	let na = ref 0 in
	let rec sf acc ds =
	  if !na < m then
	    match ds with
	      [] -> failwith "addBlkDiag': non-conforming"
	    | d0 :: dr ->
		na := !na + noOfCols d0;
		sf (d0 :: acc) dr
	  else if !na = m then (List.rev acc, ds)
	  else
	    failwith "addBlkDiag': non-conforming"
	in
	sf [] ds
      in
      let rec abd ds hss =
	match hss with
	  Leaf a ->
	    let m = noOfRows a.d
	    and n = noOfCols a.d in
	    assert (List.fold_left (fun na a -> na + noOfCols a) 0 ds = n &&
		    List.fold_left (fun ma a -> ma + noOfRows a) 0 ds = m);
	    addExactDiag 1 1 ds a.d
	| Node (a, lhss, rhss) ->
	    let (dsl, dsr) = splitFor a.ml ds in
	    abd dsl lhss;
	    abd dsr rhss
      in
      abd (Array.to_list ds) hss

    let addLowRank ?(ooc=false) hss a x b =
      let handle_ooc a = if ooc then to_ooc a else a in
      let k1 = noOfCols a
      and k2 = noOfCols b in
      let eye1 = eye k1 k1
      and eye2 = eye k2 k2 in
      assert (k1 = noOfRows x && k2 = noOfCols x);
      let umerge b r w =
	let ku = noOfRows b
	and kv = noOfCols b
	and mr = noOfRows r
	and nr = noOfCols r
	and mw = noOfRows w
	and nw = noOfCols w in
	let bn = rawMatrix ~ooc (ku+k1) (kv+k2)
	and rn = rawMatrix ~ooc (mr+k1) (nr+k1)
	and wn = rawMatrix ~ooc (mw+k2) (nw+k2) in
	matrix2x2   kv         k2
	      ku (iD b) (const zero)
	      k1 (const zero) (iD x) bn;
	matrix2x2    nr         k1
	      mr (iD r)    (const zero)
	      k1 (const zero) (iD eye1) rn;
	matrix2x2    nw         k2
	      mw (iD w)    (const zero)
	      k2 (const zero) (iD eye2) wn;
	(rn, wn, bn) in
      let rec alr hss a b =
	match hss with
	  Leaf {d=d; u=u; v=v} ->
	    let m = noOfRows d
	    and n = noOfCols d
	    and ku = noOfCols u
	    and kv = noOfCols v in
	    let un = rawMatrix ~ooc m (ku+k1)
	    and vn = rawMatrix ~ooc n (kv+k2) in
	    matrix1x2 ku    k1
	           (iD u) (iD a) un;
	    matrix1x2 kv    k2
	           (iD v) (iD b) vn;
	    Leaf {d = handle_ooc (d +$ a *$ x *$~ b); u = un; v = vn}
	| Node (n, lhss, rhss) ->
	    let (rl, wr, bu) = umerge n.bu n.rl n.wr
	    and (rr, wl, bl) = umerge n.bl n.rr n.wl
	    and (a1, a2) = partition2x1 n.ml n.mr a
	    and (b1, b2) = partition2x1 n.nl n.nr b in
	    Node ( {nl=n.nl; nr=n.nr; ml=n.ml; mr=n.mr;
		    bu=bu; bl=bl; rl=rl; rr=rr; wl=wl; wr=wr},
		   alr lhss a1 b1, alr rhss a2 b2 ) in
      match alr hss a b with
	Leaf z -> Leaf {z with u = zeros (noOfRows z.d) 0;
			v = zeros (noOfCols z.d) 0}
      | Node (z,l,r) -> Node ({ml=z.ml; mr=z.mr; nl=z.nl; nr=z.nr;
			       bu=z.bu; bl=z.bl;
			       rl = rawMatrix (noOfRows z.rl) 0;
			       rr = rawMatrix (noOfRows z.rr) 0;
			       wl = rawMatrix (noOfRows z.wl) 0;
			       wr = rawMatrix (noOfRows z.wr) 0}, l, r)

(*
    (* Same as addLowRank, except it checks for zero sub-blocks in a. *)
    let addLowRankEff tol hss a b =
      let k = noOfCols a in
      assert (k = noOfCols b);
      let eyek = eye k k in
      let umerge b r rn w wn =
	let ku = noOfRows b
	and kv = noOfCols b
	and mr = noOfRows r
	and nr = noOfCols r
	and mw = noOfRows w
	and nw = noOfCols w
	and nrn = noOfCols rn
	and mrn = noOfRows rn
	and nwn = noOfCols wn
	and mwn = noOfRows wn in
	let bn = rawMatrix (ku+k) (kv+k)
	and rn = rawMatrix (mr+k) (nr+k)
	and wn = rawMatrix (mw+k) (nw+k) in
	matrix2x2      kv         mwn
	      ku (iD b)    (const zero)
	     mrn (const zero) (iD (eye mrn mwn)) bn;
	matrix2x2    nr         nrn
	      mr (iD r)    (const zero)
	     mrn (const zero) (iD rn) rn;
	matrix2x2    nw         nwn
	      mw (iD w)    (const zero)
	     mwn (const zero) (iD wn) wn;
	(rn, wn, bn) in
      let rec alr hss a b =
	match hss with
	  Leaf {d=d; u=u; v=v} ->
	    let m = noOfRows d
	    and n = noOfCols d
	    and ku = noOfCols u
	    and kv = noOfCols v
	    and (ua, sa, vat) = svdE' tol (Blop.copy a)
	    and (ub, sb, vbt) = svdE' tol (Blop.copy b) in
	    let kua = noOfCols ua
	    and kva = noOfRows vat
	    and kub = noOfCols ub
	    and kvb = noOfRows vbt in
	    rowScale_real' sa vat;
	    rowScale_real' sb vbt;
	    let un = rawMatrix m (ku+kua)
	    and vn = rawMatrix n (kv+kub) in
	    matrix1x2 ku kua
	      (iD u) (iD ua) un;
	    matrix1x2 kv kva
	      (iD v) (iD ub) vn;
	    (Leaf {d = d +$ a *$~ b; u = un; v = vn}, vat, vbt)
	| Node (n, lhss, rhss) ->
	    let (a1, a2) = partition2x1 n.ml n.mr a
	    and (b1, b2) = partition2x1 n.nl n.nr b in
	    let (lhssN, rln, wln) = alr lhss a1 b1
	    and (rhssN, rrn, wrn) = alr rhss a2 b2 in
	    let (rl, wr, bu) = umerge n.bu n.rl rln n.wr wrn
	    and (rr, wl, bl) = umerge n.bl n.rr rrn n.wl wln
	    and r = noOfCols rln
	    and w = noOfCols wln in
	    assert (r = noOfCols rrn && w = noOfCols wrn);
	    ( Node ( {nl=n.nl; nr=n.nr; ml=n.ml; mr=n.mr;
		      bu=bu; bl=bl; rl=rl; rr=rr; wl=wl; wr=wr},
		     lhssN, rhssN )
		, eye r r, eye w w) in
      match alr hss a b with
	(Leaf z, _, _) -> Leaf z
      | (Node (z,l,r), _, _) ->
	  Node ({ml=z.ml; mr=z.mr; nl=z.nl; nr=z.nr;
		 bu=z.bu; bl=z.bl;
		 rl = rawMatrix (noOfRows z.rl) 0;
		 rr = rawMatrix (noOfRows z.rr) 0;
		 wl = rawMatrix (noOfRows z.wl) 0;
		 wr = rawMatrix (noOfRows z.wr) 0}, l, r)
*)


    (* tensor (eye (max kx ky) (max kx ky)) a *)
    let blkDiag ?(ooc=false) a kx ky = 
      let m = noOfRows a
      and n = noOfCols a in
      let b = zeros ~ooc (kx*m) (ky*n) in
      if kx * ky > 0 then begin
	assert (kx = ky);
	for i = 1 to kx do
	  (b $ ((i-1)*m+1 <-> i*m, (i-1)*n+1 <-> i*n)) <-$ a
	done;
	b
      end else
	b


    (* Given Hss * rectangle it returns P_ND * (I tensor Hss) * P_ND^T.
       This gives preference to the number of columns in the array and uses
       a vecC operation at the lowest levels. *)
    let toND2d ?(ooc=false) hss nx ny hssy =
      let makeVertCut ny hssy =
	match hssy with
	  Leaf d -> failwith "makeVertCut: Cutting leaf"
	| Node (d, lhssy, rhssy) ->
	    assert (d.nl+d.nr=ny && d.nl * d.nr > 0);
	    (d.nl, d.nr, lhssy, rhssy) in
      let rec thn2d hss nx ny hssy =
	let rec vertCut hssy =
	  match (hss, hssy) with
	    (_, Leaf _) -> (false, hssy)
	  | (Leaf _, Node (z, lhssy, rhssy)) ->
	      if z.nl * z.nr > 0 then (true, hssy)
	      else if z.nl = 0 then vertCut rhssy
	      else vertCut lhssy
	  | (Node _, Node (z, lhssy, rhssy)) ->
	      if nx <= ny then
		if z.nl * z.nr > 0 then (true, hssy)
		else if z.nl = 0 then vertCut rhssy
		else vertCut lhssy
	      else
		(false, hssy) in
	let (vc, nhssy) = vertCut hssy in
	if vc then begin
	  let (ny1, ny2, lhssy, rhssy) = makeVertCut ny nhssy in
	  let (lHss, kul, kvl) = thn2d hss nx ny1 lhssy
	  and (rHss, kur, kvr) = thn2d hss nx ny2 rhssy in
	  let r = eye ~ooc (kul+kur) (kul+kur)
	  and w = eye ~ooc (kvl+kvr) (kvl+kvr) in
	  let (rl, rr) = partition2x1 kul kur r
	  and (wl, wr) = partition2x1 kvl kvr w in
	  (Node ({ml = no_of_rows lHss; nl = nx*ny1;
		  mr = no_of_rows rHss; nr = nx*ny2;
		  bu = zeros ~ooc kul kvr; bl = zeros ~ooc kur kvl;
		  rl = rl; rr = rr; wl = wl; wr = wr}, lHss, rHss),
	   kul+kur, kvl+kvr)
	end else begin
	  match hss with
	    Leaf z ->
	      let ei = eye ny ny in
	      (Leaf {d = tensor ~ooc ei z.d; u = tensor ~ooc ei z.u;
		     v = tensor ~ooc ei z.v},
	       ny * noOfCols z.u, ny * noOfCols z.v)
	  | Node (z, lHss, rHss) ->
	      assert (z.nl+z.nr=nx);
	      let (lHssN, kul, kvl) = thn2d lHss z.nl ny nhssy
	      and (rHssN, kur, kvr) = thn2d rHss z.nr ny nhssy
	      and mu = noOfRows z.bu
	      and ml = noOfRows z.bl
	      and nbu = noOfCols z.bu
	      and nbl = noOfCols z.bl in
	      let nux = if mu > 0 then kul / mu else 0
	      and nlx = if ml > 0 then kur / ml else 0
	      and nuy = if nbu > 0 then kvr / nbu else 0
	      and nly = if nbl > 0 then kvl / nbl else 0 in
	      assert (nux * mu = kul && nlx * ml = kur && nuy * nbu = kvr &&
		      nly * nbl = kvl);
	      assert (if nux * nuy > 0 then nux = nuy else true);
	      assert (if nlx * nly > 0 then nlx = nly else true);
	      (Node ({ml = z.ml*ny; mr = z.mr*ny; nl = z.nl*ny; nr = z.nr*ny;
		      bu = blkDiag ~ooc z.bu nux nuy;
		      bl = blkDiag ~ooc z.bl nlx nly;
		      rl = blkDiag ~ooc z.rl nux (max nux nlx);
		      rr = blkDiag ~ooc z.rr nlx (max nux nlx);
		      wl = blkDiag ~ooc z.wl nly (max nuy nly);
		      wr = blkDiag ~ooc z.wr nuy (max nuy nly)},
		     lHssN, rHssN),
	       max nux nlx * noOfCols z.rl, max nuy nly * noOfCols z.wl)
	end in
      let (hss, _, _) = thn2d hss nx ny hssy in
      match hss with
	Leaf _ -> hss
      | Node (z, lhss, rhss) ->
	  Node ({ml=z.ml; mr=z.mr; nl=z.nl; nr=z.nr;
		 bu=z.bu; bl=z.bl;
		 rl = rawMatrix (noOfRows z.rl) 0;
		 rr = rawMatrix (noOfRows z.rr) 0;
		 wl = rawMatrix (noOfRows z.wl) 0;
		 wr = rawMatrix (noOfRows z.wr) 0}, lhss, rhss)

    (* Inverse permutation *)
    let invPermute p =
      let n = Array1.dim p in
      let pi = Array1.create int fortran_layout n in
      for i = 1 to n do
	pi.{p.{i}} <- i
      done;
      pi
	

    (* Returns the associated permutation transform for toND2d. *)
    let toND2dPerm hss nx ny hssy =
      let p2d = Array2.create int fortran_layout nx ny in
      let numCols m n sr sc sn =
	for j = sc to sc+n-1 do
	  for i = sr to sr+m-1 do
	    p2d.{i,j} <- sn+(j-sc)*m+(i-sr)
	  done
	done;
	sn+m*n in
      let makeVertCut ny hssy =
	match hssy with
	  Leaf d -> failwith "makeVertCut: Cutting leaf"
	| Node (d, lhssy, rhssy) ->
	    assert (d.nl+d.nr=ny && d.nl * d.nr > 0);
	    (d.nl, d.nr, lhssy, rhssy) in
      let rec number hss nx ny hssy sr sc sn =
	let rec vertCut hssy =
	  match (hss, hssy) with
	    (_, Leaf _) -> (false, hssy)
	  | (Leaf _, Node (z, lhssy, rhssy)) ->
	      if z.nl * z.nr > 0 then (true, hssy)
	      else if z.nl = 0 then vertCut rhssy
	      else vertCut lhssy
	  | (Node _, Node (z, lhssy, rhssy)) ->
	      if nx <= ny then
		if z.nl * z.nr > 0 then (true, hssy)
		else if z.nl = 0 then vertCut rhssy
		else vertCut lhssy
	      else
		(false, hssy) in
	let (vc, nhssy) = vertCut hssy in
	if vc then begin
	  let (ny1, ny2, lhssy, rhssy) = makeVertCut ny nhssy in
	  let sn1 = number hss nx ny1 lhssy sr sc sn in
	  number hss nx ny2 rhssy sr (sc+ny1) sn1
	end else begin
	  match hss with
	    Leaf z -> numCols nx ny sr sc sn
	  | Node (z, lHss, rHss) ->
	      assert (z.nl+z.nr=nx);
	      let sn1 = number lHss z.nl ny nhssy sr sc sn in
	      number rHss z.nr ny nhssy (sr+z.ml) sc sn1
	end in
      let toPerm p2d =
	let p = Array1.create int fortran_layout (nx * ny) in
	for j = 1 to ny do
	  for i = 1 to nx do
	    p.{(j-1)*nx+i} <- p2d.{i,j}
	  done
	done;
	permutation (invPermute p) in
      let x = number hss nx ny hssy 1 1 1 in
      assert (x = nx * ny + 1);
      (toPerm p2d, p2d)


(* We need a permutation to take us from column ordering to row ordering. *)
    let cToRperm nx ny =
      let p = Array1.create int fortran_layout (nx * ny) in
      for j = 1 to ny do
	for i = 1 to nx do
	  p.{(j-1)*nx+i} <- (i-1)*ny+j
	done
      done;
      permutation p


    (* Given Hss * rectangle it returns P_ND * (Hss tensor I) * P_ND^T.
       This gives preference to the number of rows in the array and uses
       a vecR operation at the lowest levels. *)
    let toND2d_R ?(ooc=false) hss nx ny hssy =
      let rec makeVertCut ny hssy =
	match hssy with
	  Leaf d -> failwith "makeVertCut_R: cutting leaf"
	| Node (d, lhssy, rhssy) ->
	    assert (d.nl+d.nr=ny && d.nl * d.nr > 0);
	    (d.nl, d.nr, lhssy, rhssy) in
      let rec thn2d hss nx ny hssy =
	let rec vertCut hssy =
	  match (hss, hssy) with
	    (_, Leaf _) -> (false, hssy)
	  | (Leaf _, Node (z, lhssy, rhssy)) ->
	      if z.nl * z.nr > 0 then (true, hssy)
	      else if z.nl = 0 then vertCut rhssy
	      else vertCut lhssy
	  | (Node _, Node (z, lhssy, rhssy)) ->
	      if nx < ny then (* Differs from toND2d *)
		if z.nl * z.nr > 0 then (true, hssy)
		else if z.nl = 0 then vertCut rhssy
		else vertCut lhssy
	      else
		(false, hssy) in
	let (vc, nhssy) = vertCut hssy in
	if vc then begin
	  let (ny1, ny2, lhssy, rhssy) = makeVertCut ny nhssy in
	  let (lHss, kul, kvl) = thn2d hss nx ny1 lhssy
	  and (rHss, kur, kvr) = thn2d hss nx ny2 rhssy in
	  let r = eye ~ooc (kul+kur) (kul+kur)
	  and w = eye ~ooc (kvl+kvr) (kvl+kvr) in
	  let (rl, rr) = partition2x1 kul kur r
	  and (wl, wr) = partition2x1 kvl kvr w in
	  (Node ({ml = no_of_rows lHss; nl = nx*ny1;
		  mr = no_of_rows rHss; nr = nx*ny2;
		  bu = zeros ~ooc kul kvr; bl = zeros ~ooc kur kvl;
		  rl = rl; rr = rr; wl = wl; wr = wr}, lHss, rHss),
	   kul+kur, kvl+kvr)
	end else begin
	  match hss with
	    Leaf z ->
	      let ei = eye ny ny in
	      assert (nx = noOfCols z.d);
	      let pl = cToRperm nx ny
	      and pr = cToRperm (noOfRows z.d) ny in
	      let d = tensor ~ooc ei z.d
	      and u = tensor ~ooc ei z.u
	      and v = tensor ~ooc ei z.v in
	      d *@ pl;
	      pr @~* d;
	      pr @~* u;
	      pl @~* v;
	      (Leaf {d = d; u = u; v = v}, ny * noOfCols z.u,
	       ny * noOfCols z.v)
	  | Node (z, lHss, rHss) ->
	      assert (z.nl+z.nr=nx);
	      let (lHssN, kul, kvl) = thn2d lHss z.nl ny nhssy
	      and (rHssN, kur, kvr) = thn2d rHss z.nr ny nhssy
	      and mu = noOfRows z.bu
	      and ml = noOfRows z.bl
	      and nbu = noOfCols z.bu
	      and nbl = noOfCols z.bl in
	      let nux = if mu > 0 then kul / mu else 0
	      and nlx = if ml > 0 then kur / ml else 0
	      and nuy = if nbu > 0 then kvr / nbu else 0
	      and nly = if nbl > 0 then kvl / nbl else 0 in
	      assert (nux * mu = kul && nlx * ml = kur && nuy * nbu = kvr &&
		      nly * nbl = kvl);
	      assert (if nux * nuy > 0 then nux = nuy else true);
	      assert (if nlx * nly > 0 then nlx = nly else true);
	      (Node ({ml = z.ml*ny; mr = z.mr*ny; nl = z.nl*ny; nr = z.nr*ny;
		      bu = blkDiag ~ooc z.bu nux nuy;
		      bl = blkDiag ~ooc z.bl nlx nly;
		      rl = blkDiag ~ooc z.rl nux (max nux nlx);
		      rr = blkDiag ~ooc z.rr nlx (max nux nlx);
		      wl = blkDiag ~ooc z.wl nly (max nuy nly);
		      wr = blkDiag ~ooc z.wr nuy (max nuy nly)},
		     lHssN, rHssN),
	       max nux nlx * noOfCols z.rl, max nuy nly * noOfCols z.wl)
	end in
      let (hss, _, _) = thn2d hss nx ny hssy in
      match hss with
	Leaf _ -> hss
      | Node (z, lhss, rhss) ->
	  Node ({ml=z.ml; mr=z.mr; nl=z.nl; nr=z.nr;
		 bu=z.bu; bl=z.bl;
		 rl = rawMatrix (noOfRows z.rl) 0;
		 rr = rawMatrix (noOfRows z.rr) 0;
		 wl = rawMatrix (noOfRows z.wl) 0;
		 wr = rawMatrix (noOfRows z.wr) 0}, lhss, rhss)


    (* Merge succesive nodes that are single children of their parents. *)
    let rec mergeSingleChildNodes ?(ooc=false) hss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let msch r w hss =
	match hss with
	  Leaf z -> Leaf {d = handle_ooc z.d; u = handle_ooc (z.u *$ r);
			  v = handle_ooc (z.v *$ w)}
	| Node (a, lHss, rHss) ->
	    mergeSingleChildNodes ~ooc
	      (Node ({ml = a.ml; mr = a.mr; nl = a.nl; nr = a.nr;
		      bu = a.bu; bl = a.bl;
		      rl = handle_ooc (a.rl *$ r);
		      rr = handle_ooc (a.rr *$ r);
		      wl = handle_ooc (a.wl *$ w);
		      wr = handle_ooc (a.wr *$ w)},
		     lHss, rHss)) in
      match hss with
	Leaf z -> Leaf z
      | Node (a, lHss, rHss) ->
	  if a.ml = 0 && a.nl = 0 then
	    msch a.rr a.wr rHss
	  else if a.mr = 0 && a.nr = 0 then
	    msch a.rl a.wl lHss
	  else
	    Node (a, mergeSingleChildNodes ~ooc lHss,
		  mergeSingleChildNodes ~ooc rHss)


    (* hss tensor (eye n n). *)
    let hssTensorEye ?(ooc=false) hss n =
      let ei = eye n n in
      let rec hte hss =
	match hss with
	  Leaf z -> Leaf {d = tensor ~ooc z.d ei; u = tensor ~ooc z.u ei;
			  v = tensor ~ooc z.v ei}
	| Node (z, lHss, rHss) ->
	    Node ({ml = z.ml * n; mr = z.mr * n; nl = z.nl * n; nr = z.nr * n;
		   rl = tensor ~ooc z.rl ei; rr = tensor ~ooc z.rr ei;
		   wl = tensor ~ooc z.wl ei; wr = tensor ~ooc z.wr ei;
		   bu = tensor ~ooc z.bu ei; bl = tensor ~ooc z.bl ei},
		  hte lHss, hte rHss) in
      hte hss


    (* Multiplies block diagonal into hss. Must be conformally partitioned. *)
    let mulDiag ?(ooc=false) ds hss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let mulExactDiag ds u =
	let ms = List.map noOfCols ds in
	let m = List.fold_left (fun m di -> m + noOfRows di) 0 ds
	and us = partitionInfx1 ms u in
	let du = rawMatrix ~ooc m (noOfCols u) in
	matrixInfx1 ms (List.map2 (fun di ui -> di |*| ui) ds us) du;
	du
      in
      let splitFor m ds =
	let na = ref 0 in
	let rec sf acc ds =
	  if !na < m then
	    match ds with
	      [] -> failwith "mulDiag: non-conforming"
	    | d0 :: dr ->
		na := !na + noOfCols d0;
		sf (d0 :: acc) dr
	  else if !na = m then (List.rev acc, ds)
	  else
	    failwith "mulDiag: non-conforming"
	in
	sf [] ds
      in
      let rec muld ds hss =
	match hss with
	  Leaf a ->
	    let m = noOfRows a.d in
	    assert (List.fold_left (fun na a -> na + noOfCols a) 0 ds = m);
	    Leaf {d = mulExactDiag ds a.d; u = mulExactDiag ds a.u;
		  v = handle_ooc a.v}
	| Node (a, lhss, rhss) ->
	    let (dsl, dsr) = splitFor a.ml ds in
	    Node (a, muld dsl lhss, muld dsr rhss)
      in
      muld (Array.to_list ds) hss


    (* Multiplies block diagonal into hss. Must be conformally partitioned. *)
    let mulDiagFn ?(ooc=false) dFn hss =
      let handle_ooc a = if ooc then to_ooc a else a in
      let mulExactDiag ds u =
	let ms = List.map noOfCols ds in
	let m = List.fold_left (fun m di -> m + noOfRows di) 0 ds
	and us = partitionInfx1 ms u in
	let du = rawMatrix ~ooc m (noOfCols u) in
	matrixInfx1 ms (List.map2 (fun di ui -> di |*| ui) ds us) du;
	du
      in
      let rec gather_from i m acc =
	if m <= 0 then (List.rev acc, i)
	else
	  let di = dFn i in
	  gather_from (i+1) (m - noOfCols di) (di :: acc) in
      let rec muld i hss =
	match hss with
	  Leaf a ->
	    let m = noOfRows a.d in
	    let (ds, i_new) = gather_from i m [] in
	    assert (List.fold_left (fun na a -> na + noOfCols a) 0 ds = m);
	    (Leaf {d = mulExactDiag ds a.d; u = mulExactDiag ds a.u;
		   v = handle_ooc a.v}, i_new)
	| Node (a, lhss, rhss) ->
	    let (new_lhss, i_new) = muld i lhss in
	    let (new_rhss, i_next) = muld i_new rhss in
	    (Node (a, new_lhss, new_rhss), i_next)
      in
      let (new_hss, _) = muld 0 hss in
      new_hss


    (* Zeros out the given list of rows. *)
    let zeroRows' rows hss =
      let rec zeroD rows a =
	match rows with
	  [] -> ()
	| i :: is ->
	    const zero (a $ (i <-> i, All));
	    zeroD is a
      in
      let rec zr rows hss =
	match rows with
	  [] -> ()
	| _ ->
	    match hss with
	      Leaf a ->
		zeroD rows a.d;
		zeroD rows a.u
	    | Node (a, lhss, rhss) ->
		let (lrows, rrows) = List.partition (fun i -> i <= a.ml) rows
		in
		zr lrows lhss;
		zr (List.map (fun i -> i - a.ml) rrows) rhss
      in
      zr rows hss


    (* Wrapper for svdE that returns only 2 factors, the 1st is orthonormal. *)
    let lowRankFact' ?(ooc=false) tol a =
      let (u, s, vt) = svdE' ~ooc tol a in
      rowScale_real' s vt;
      (u, vt)

    (* lowRankFact' of (q1*r1; q2*r2) *)
    let lowRankFact2x1' tol ua vat ub vbt =
      let k1 = noOfRows vat
      and k2 = noOfRows vbt
      and n = noOfCols vat
      and m1 = noOfRows ua
      and m2 = noOfRows ub in
      assert (noOfCols vbt = n);
      let c = rawMatrix (k1 + k2) n in
      matrix2x1 k1 (iD vat)
	        k2 (iD vbt) c;
      let (u, vt) = lowRankFact' tol c in
      let r = noOfCols u
      and (u1, u2) = partition2x1 k1 k2 u in
      let uf = rawMatrix (m1+m2) r in
      matrix2x1 m1 (ua |*| u1)
	        m2 (ub |*| u2) uf;
      (uf, vt)


    (* Computes the minimal HSS form of a *$~ b efficiently. See fromFn. 
       a & b are destroyed. *)
    let fromLowRank' ?(ooc=false) hssSpine a b tol =
      let handle_ooc a = if ooc then to_ooc a else a in
      let rec fab hssSpine a b =
	match hssSpine with
	  Leaf () ->
	    let d = handle_ooc (a *$~ b) in
	    let (qa, ra) = lowRankFact' ~ooc tol a
	    and (qb, rb) = lowRankFact' ~ooc tol b in
	    (Leaf {d=d; u=qa; v=qb}, qa, ra, qb, rb)
	| Node (n, lSpine, rSpine) ->
	    let (a1, a2) = partition2x1 n.mLeft n.mRight a
	    and (b1, b2) = partition2x1 n.nLeft n.nRight b in
	    let (lhss, ul, rl, vl, wl) = fab lSpine a1 b1
	    and (rhss, ur, rr, vr, wr) = fab rSpine a2 b2 in
	    let (qa, ra) = lowRankFact2x1' tol ul rl ur rr
	    and (qb, rb) = lowRankFact2x1' tol vl wl vr wr in
	    let (qal, qar) = partition2x1 n.mLeft n.mRight qa
	    and (qbl, qbr) = partition2x1 n.nLeft n.nRight qb in
	    (Node ({rl = handle_ooc (ul *~$ qal);
		    rr = handle_ooc (ur *~$ qar);
		    wl = handle_ooc (vl *~$ qbl);
		    wr = handle_ooc (vr *~$ qbr);
		    bu = handle_ooc (rl *$~ wr);
		    bl = handle_ooc (rr *$~ wl);
		    ml = n.mLeft; mr = n.mRight;
		    nl = n.nLeft; nr = n.nRight
		  }, lhss, rhss), qa, ra, qb, rb)
      in
      let (hss, _, _, _, _) = fab hssSpine a b in
      match hss with
	Leaf z as f -> f
      | Node (a, lhss, rhss) ->
	  Node ({ml=a.ml; mr=a.mr; nl=a.nl; nr=a.nr; bu=a.bu; bl=a.bl;
		 rl = zeros (noOfRows a.rl) 0;
		 rr = zeros (noOfRows a.rr) 0;
		 wl = zeros (noOfRows a.wl) 0;
		 wr = zeros (noOfRows a.wr) 0}, lhss, rhss)

    let rec toHSSspine hss =
      match hss with
	Leaf _ -> Leaf ()
      | Node (a, lHss, rHss) ->
	  Node ({mLeft=a.ml; mRight=a.mr; nLeft=a.nl; nRight=a.nr},
		toHSSspine lHss, toHSSspine rHss)


    (* Converts hss into sparse form in nested dissection ordering. *)
    let toSparse hss =
      let nes = ref (zeros 0 0) in
      let neye m =
	if (noOfRows !nes = m) & (noOfCols !nes = m) then Some !nes
	else
	  let ne = eye m m in
	  scale' (fromFloat (-1.0)) ne;
	  nes := ne;
	  Some ne
      in
      let n =
	let rec nnodes hss =
	  match hss with
	    Leaf _ -> 3
	  | Node (_, lhss, rhss) -> 2 + nnodes lhss + nnodes rhss
	in
	nnodes hss
      in
      let sp_hss = Sparse.empty n n in
      let rec fill_sp_hss hss s =
	match hss with
	  Leaf a ->
	    let (sx, sg, sf, sn) = (s, s+1, s+2, s+3) in
	    Sparse.set sp_hss sx sx (Some a.d);            (* d x   *)
	    Sparse.set sp_hss sx sf (Some a.u);            (* u f   *)
	    Sparse.set sp_hss sg sx (Some (transp a.v));   (* v^H x *)
	    Sparse.set sp_hss sg sg (neye (noOfCols a.v)); (* -g    *)
	    Sparse.set sp_hss sf sf (neye (noOfCols a.u)); (* -f    *)
	    (sg, sf, sn)
	| Node (a, lhss, rhss) ->
	    let (sgl, sfl, st) = fill_sp_hss lhss s in
	    let (sgr, sfr, sr) = fill_sp_hss rhss st in
	    let (sg, sf, sn) = (sr, sr+1, sr+2) in
	    Sparse.set sp_hss sg  sgl (Some (transp a.wl));  (* wl^H gl *)
	    Sparse.set sp_hss sg  sgr (Some (transp a.wr));  (* wr^H gr *)
	    Sparse.set sp_hss sg  sg (neye (noOfCols a.wl)); (* -g      *)
	    Sparse.set sp_hss sf  sf (neye (noOfCols a.rr)); (* -f      *)
	    Sparse.set sp_hss sfl sf  (Some a.rl);           (* rl f    *)
	    Sparse.set sp_hss sfl sgr (Some a.bu);           (* bu gr   *)
	    Sparse.set sp_hss sfr sf  (Some a.rr);           (* rr f    *)
	    Sparse.set sp_hss sfr sgl (Some a.bl);           (* bl gl   *)
	    (sg, sf, sn)
      in
      let (_,_, l) = fill_sp_hss hss 0 in
      assert (l = n);
      sp_hss


    (* Converts the rhs b to match toSparse hss ordering. *)
    let toSparse_vec hss b =
      let rec make_vec hss b =
	match hss with
	  Leaf a ->
	    let m0 = noOfRows a.d
	    and m1 = noOfCols a.v
	    and m2 = noOfCols a.u in
	    let (b0, b1) = partition2x1 m0 (noOfRows b - m0) b in
	    (matrix2x1
	           m0 (iD b0)
	       (m1+m2) ooo,
	     m0+m1+m2, b1)
	| Node (a, lhss, rhss) ->
	    let (ltransf, m0, bt) = make_vec lhss b in
	    let (rtransf, m1, br) = make_vec rhss bt in
	    let m2 = noOfCols a.wl + noOfCols a.rr in
	    (matrix3x1
	       m0 ltransf
	       m1 rtransf
	       m2 ooo,
	     m0+m1+m2, br)
      in
      let (transf, n, bl) = make_vec hss b in
      assert (noOfRows bl = 0);
      let bv = zeros n (noOfCols b) in
      transf bv;
      bv


    (* Extracts the actual soln of hss x = b from sparse solution. *)
    let fromSparse_vec hss xgf =
      let rec make_vec hss xgf =
	match hss with
	  Leaf a ->
	    let m0 = noOfCols a.d
	    and m1 = noOfCols a.v
	    and m2 = noOfCols a.u in
	    let (x, _, xgfr) = partition3x1 m0 (m1+m2)
		(noOfRows xgf - m0 - m1 - m2) xgf in
	    (iD x, m0, xgfr)
	| Node (a, lhss, rhss) ->
	    let (ltransf, m0, xgft) = make_vec lhss xgf in
	    let (rtransf, m1, xgfr) = make_vec rhss xgft in
	    (matrix2x1
	       m0 ltransf
	       m1 rtransf,
	     m0+m1,
	     xgfr $ (noOfCols a.wl + noOfCols a.rr + 1 <-> noOfRows xgfr, All))
      in
      let (transf, n, xgfl) = make_vec hss xgf in
      assert (noOfRows xgfl = 0);
      let x = zeros n (noOfCols xgf) in
      transf x;
      x


    let fastSparseSolve hss b =
      let sp_hss = toSparse hss
      and bv = toSparse_vec hss b in
      let xgf = Sparse.linear_solve' ~use_ql:false sp_hss bv in
      fromSparse_vec hss xgf


    (* Begin: Auxiliary routines for fastMRsolve'. *)
    (* Aux. rank-revealing ql factorization. *)
    let rrqlf' tol a =
      let k = min (noOfRows a) (noOfCols a) in
      let (u, s, vt) = svd' a in
      let r = ref 0 in
      while !r < k && s.{!r+1} > tol do
	r := !r + 1
      done;
      assert (!r <= k && (!r = 0 || s.{!r} > tol));
      if !r < k then (Printf.printf "%d %d" k !r; print_newline ());
      let vt1 = vt $ (1 <-> !r, All) in
      rowScale_real' (Array1.sub s 1 !r) vt1;
      (u $ (All, 1 <-> !r), vt1)

    (* Fast Model Reduction handling of non-leaf node. *)
    type side = Left | Right
    let fmr_node tol side a fmr_u =
      let compr b r u =
	match u with
	  None -> (None, b, r)
	| Some u ->
	    let n1 = noOfCols b
	    and n2 = noOfCols u in
	    let a = rawMatrix (noOfRows r) (n1+n2) in
	    let (a1, a2) = partition1x2 n1 n2 a in
	    a1 <-| iD b;
	    a2 <-| (r |*| u);
	    let (u, br) = rrqlf' tol a in
	    let (b, r) = partition1x2 n1 n2 br in
	    (Some u, b, r)
      in
      match side with
      |	Left -> let (fmr_u, bu, rl) = compr a.bu a.rl fmr_u in
	(fmr_u, {a with bu = bu; rl = rl})
      | Right -> let (fmr_u, bl, rr) = compr a.bl a.rr fmr_u in
	(fmr_u, {a with bl = bl; rr = rr})
       
    (* toProperForm handling of leaf; v part only; not u. *)
    let tpf_leaf_v' tol a =
      let (vnew, wh) = rrqlf' tol a.v in
      ({a with v = vnew}, wh)

    (* fastSolve handling of leaf node. *)
    let fs_eliminateLeaf' tpf a tol b =
      let nrhs = noOfCols b in
      if noOfCols a.u >= noOfRows a.u then (* Nothing to compress *)
	(Leaf { w = eyeTransform; chi = zeros 0 nrhs },
	 zeros (noOfCols a.v) nrhs, (* g *)
	 b,                         (* remaining b *)
	 a,                         (* remaining u, v, d *)
	 None                       (* No tpfW *)
	)
      else (* Compress *)
	let ku = noOfCols a.u
	and m = noOfRows a.u
	and n = noOfCols a.d 
	and (q, l_u) = ql' a.u in
	q @~* b;
	q @~* a.d;
	let (b0, b1) = partition2x1 (m-ku) ku b
	and (d0, d1) = partition2x1 (m-ku) ku a.d in
	let (d00, w) = lq' d0 in
	d1 *~@ w;
	w @* a.v;
	subs' d00 b0;
	let (d10, d11) = partition1x2 (m-ku) (n-m+ku) d1
	and (v00, v01) = partition2x1 (m-ku) (n-m+ku) a.v
	and chi = b0 in
	if tpf then
	  let (new_a, tpfW) = tpf_leaf_v' tol
	      {u = copyL l_u; v = v01; d = d11} in
	  (Leaf { w = w; chi = chi },
	   v00 *~$ chi,                      (* g *)
	   b1 -$ d10 *$ chi,                 (* remaining b *)
	   new_a,                            (* remaining u, compr. v, d *)
	   Some tpfW)
	else
	  (Leaf { w = w; chi = chi },
	   v00 *~$ chi,                      (* g *)
	   b1 -$ d10 *$ chi,                 (* remaining b *)
	   {u = copyL l_u; v = v01; d = d11},(* remaining u, v, d *)
	   None)


    (* fastSolve' with fast model reduction, proper form
       reduction. fmr moves down the f lines, tpf moves up the g
       lines. *)
    let fastMRsolve' tol hss b x =
      let nrhs = noOfCols b in
      assert (nrhs = noOfCols x);
      let eliminateLeaf tpf a b fmrU =
	(* apply fmr first; tpf is done internally *)
	let a =
	  match fmrU with
	    None -> a
	  | Some u ->
	      if noOfCols u < noOfRows u then begin
		Printf.printf "For u: %d %d" (noOfRows u) (noOfCols u);
		print_newline ();
	      end;
	      {a with u = a.u *$ u}
	in
	fs_eliminateLeaf' tpf a tol b
      in
      let rec ulSweep hss b f fmrU =
	(* f travels ahead of the fmr wave-front that is moving down. *)
	(* g travels ahead of the tpf wave-front that is moving up. *)
	match hss with
	    Leaf a -> eliminateLeaf true a (b -$ a.u *$ f) fmrU
	  | Node (a, lHss, rHss) ->
	      (* first send f down *)
	      let rlf = a.rl *$ f
	      and rrf = a.rr *$ f in
	      (* next send compressing U to the left, and recompress *)
	      let (fmrLu, a) = fmr_node tol Left a fmrU in
	      let (b0, b1) = partition2x1 a.ml a.mr b in
	      let (lWchiTree, gl, bl, al, tpfL) = ulSweep lHss b0 rlf fmrLu in
	      (* send left g up and to the right *)
	      let wlgl = a.wl *~$ gl
	      and blgl = a.bl *$  gl in
	      (* apply reducing proper form factor to wl and bl *)
	      let a = begin match tpfL with
		None -> a
	      |	Some tpfl -> {a with wl = tpfl *$ a.wl ; bl = a.bl *$~ tpfl}
	      end in
	      (* need to compress bl again and propagate *)
	      (* send original U to right and recompress *)
	      let (fmrRu, a) = fmr_node tol Right a fmrU in
	      let (rWchiTree, gr, br, ar, tpfR) = ulSweep rHss b1
		  (blgl +$ rrf) fmrRu in
	      (* first send g up and to the left *)
	      let ng = wlgl +$ a.wr *~$ gr
	      and nbugr = a.bu *$ gr in
	      (* apply reducing proper form factor to wr and br *)
	      let a = begin match tpfR with
		None -> a
	      |	Some tpfr -> {a with wr = tpfr *$ a.wr ; bu = a.bu *$~ tpfr}
	      end in
	      (* need to compress bu again and propagate *)
	      let (fmrLuN, aN) = fmr_node tol Left a None in
	      begin match fmrLuN with
		None -> ()
	      | Some u ->
		  if noOfCols u < noOfRows u then begin
		    Printf.printf "Yup %d %d" (noOfRows u) (noOfCols u);
		    print_newline ()
		  end
	      end;
	      (* merge remaining two leaves *)
	      let m0 = noOfRows al.d
	      and n0 = noOfCols al.d
	      and m1 = noOfRows ar.d
	      and n1 = noOfCols ar.d in
	      let d = rawMatrix (m0+m1) (n0+n1)
	      and u = rawMatrix (m0+m1) (noOfCols a.rl)
	      and v = rawMatrix (n0+n1) (noOfCols a.wl)
	      and bm = rawMatrix (m0+m1) nrhs in
	      matrix2x2   n0                      n1
		     m0 (iD al.d) (al.u |*| (a.bu *$~ ar.v))
		     m1 (ar.u |*| (a.bl *$~ al.v)) (iD ar.d) d;
	      matrix2x1 m0 (al.u |*| a.rl)
		        m1 (ar.u |*| a.rr) u;
	      matrix2x1 n0 (al.v |*| a.wl)
		        n1 (ar.v |*| a.wr) v;
	      matrix2x1 m0 (bl |-| al.u *$ nbugr)
		        m1 (iD br               ) bm;
	      (* ask fo explicit tpf reduction: true *)
	      let retLeaf = eliminateLeaf true {u = u; v = v; d = d} bm None in
	      match retLeaf with
		  (Leaf wCh, g, bmr, amr, tpfm) ->
		    (Node ({wN = wCh.w; chiN = wCh.chi; nm0 = n0; nm1 = n1},
			   lWchiTree, rWchiTree),
		     g +$ ng,
		     bmr,
		     amr,
		     tpfm)
		| _ -> failwith "eliminateLeaf returned Node!"
      in
      let rec vSweep hss wChiTree xp x =
	match (hss, wChiTree) with
	    (Leaf _, Leaf {w=w; chi=chi}) ->
	      matrix2x1 (noOfRows chi) (iD chi)
	                (noOfRows xp ) (iD xp ) x;
	      w @~* x
	  | (Node (a, lHss, rHss),
	    Node ({wN=w;chiN=chi;nm0=n0;nm1=n1}, lWchi, rWchi)) ->
	      let mc = noOfRows chi
	      and mx = noOfRows xp in
	      let chip = rawMatrix (mc+mx) nrhs in
	      matrix2x1 mc (iD chi)
		        mx (iD xp ) chip;
	      w @~* chip;
	      let (xp0, xp1) = partition2x1 n0 n1 chip in
	      matrix2x1 a.nl (vSweep lHss lWchi xp0)
	                a.nr (vSweep rHss rWchi xp1) x
	  | _ -> failwith "fastSolve': Mismatched trees."
      in
      match hss with
	  Leaf a ->
	    x <-$ b;
	    a.d |/| x
	| _ ->
	    let (wChiTree, _, _, _, _) = ulSweep hss b
	      (zeros 0 nrhs) None in
	    vSweep hss wChiTree (zeros 0 nrhs) x


end (* module Hsser (Blop : BLOP) (Sparse : SPARSE) *)


module DHss = Hsser (DBlop) (DSparse)
module SHss = Hsser (SBlop) (SSparse)
module ZHss = Hsser (ZBlop) (ZSparse)
module CHss = Hsser (CBlop) (CSparse)

