(*
  File: interpolation.ml
  Author: Shivkumar Chandrasekaran
  Email: shiv@ece.ucsb.edu
  Date: 4.19.2007
  Purpose: Investigate interpolation and approximation
  
  $Id: interpolation.ml,v 1.15 2008/02/21 21:37:31 shiv Exp shiv $
*)


open Bigarray
open Utils
open Nla
open Chub

open DBlop
open DChub
  

(* Plotting routine *)
let plot ?(n=100) ?yrange f a b =
  let tmpDataFile = Filename.temp_file "camlFloat" "plotData" in
  let outc = open_out_bin tmpDataFile in
  for i = 1 to n do
    let xi = a +. (b -. a) *. float i /. float (n+1) in
    let fi = f xi in
    Printf.fprintf outc "%g  %g\n" xi fi
  done;
  close_out outc;
  let tmpFile = Filename.temp_file "camlFloat" "plotCmd" in
  let outc2 = open_out tmpFile in
  let ranges = match yrange with
  | None -> Printf.sprintf " [%g : %g] " a b
  | Some (yl, yr) -> Printf.sprintf " [%g : %g] [%g : %g] " a b yl yr
  in
  Printf.fprintf outc2 "plot %s '%s' with linespoints \n" ranges tmpDataFile;
  close_out outc2;
  let err = Sys.command ("gnuplot -persist " ^ tmpFile) in
  if err <> 0 then
    prerr_string ("gnuplot " ^ tmpFile ^ " : " ^ string_of_int err);
  Sys.remove tmpDataFile;
  Sys.remove tmpFile
    
(* Same as plot, but for a list of functions instead *)
let plotn ?(n=100) fs a b =
  let tempDataFiles = List.map
      (fun _ -> Filename.temp_file "camlFloat" "plotData") fs in
  let gen f file =
    let outc = open_out file in
    for i = 1 to n do
      let xi = a +. (b -. a) *. float i /. float (n+1) in
      let fi = f xi in
      Printf.fprintf outc "%g  %g\n" xi fi
    done;
    close_out outc
  in
  List.iter2 gen fs tempDataFiles;
  let tmpFile = Filename.temp_file "camlFloat" "plotCmd" in
  let outc2 = open_out tmpFile in
  Printf.fprintf outc2 "plot [%g:%g] " a b;
  for i = 0 to List.length fs - 1 do
    Printf.fprintf outc2 " '%s' with lines" (List.nth tempDataFiles i);
    if i = List.length fs - 1 then
      Printf.fprintf outc2 "\n"
    else
      Printf.fprintf outc2 ","
  done;
  close_out outc2;
  let err = Sys.command ("gnuplot -persist " ^ tmpFile) in
  if err <> 0 then
    prerr_string ("gnuplot " ^ tmpFile ^ " : " ^ string_of_int err);
  List.iter Sys.remove tempDataFiles;
  Sys.remove tmpFile


(* Same as plot; but postscript output *)
let plotf ?(n=100) ?yrange ~fname ?(title="") f a b =
  let tmpDataFile = Filename.temp_file "camlFloat" "plotData" in
  let outc = open_out_bin tmpDataFile in
  for i = 1 to n do
    let xi = a +. (b -. a) *. float i /. float (n+1) in
    let fi = f xi in
    Printf.fprintf outc "%g  %g\n" xi fi
  done;
  close_out outc;
  let tmpFile = Filename.temp_file "camlFloat" "plotCmd" in
  let outc2 = open_out tmpFile in
  Printf.fprintf outc2 "set terminal postscript\n";
  Printf.fprintf outc2 "set output '%s'\n" fname;
  let ranges = match yrange with
  | None -> Printf.sprintf " [%g : %g] " a b
  | Some (yl, yr) -> Printf.sprintf " [%g : %g] [%g : %g] " a b yl yr
  in
  if title = "" then
    Printf.fprintf outc2 "plot %s '%s' with lines notitle\n"
      ranges tmpDataFile
  else
    Printf.fprintf outc2 "plot %s '%s' with lines title '%s' \n"
      ranges tmpDataFile title;
  close_out outc2;
  let err = Sys.command ("gnuplot " ^ tmpFile) in
  if err <> 0 then
    prerr_string ("gnuplot " ^ tmpFile ^ " : " ^ string_of_int err);
  Sys.remove tmpDataFile;
  Sys.remove tmpFile
    

(* Not fully correct; reflection on circle ignored. *)
let mesh_norm zs =
  let n = Array.length zs in
  let thetas = Array.map acos zs in
  let acc = ref pi in
  for i = 0 to n-2 do
    let len = mod_float (abs_float (thetas.(i+1) -. thetas.(i))) pi in
    if len < !acc then acc := len;
  done;
  int_of_float (ceil (pi /. !acc))

let interpolate ?(orderConst=2.0) ?(sobolev=2.0) f a b xs =
  let n = Array.length xs in
  let m = int_of_float (orderConst *. float (mesh_norm xs)) in
  let r = fun2mat (fun i j -> f xs.(i-1)) n 1
  and zs = Array.copy xs in
  fromInterval' a b zs;
  let s = Array1.create precision fortran_layout m in
  Printf.printf "N=%i M(N)=%i" n m; print_newline ();
  for i = 1 to m do
    s.{i} <- 1.0 /. (float i ** sobolev)
  done;
  let v = chebVand zs m in
  colScale' v s;
  let c = v /$ r in
  rowScale' s c;
  Printf.printf "cond = %g\n" (cond v);
  let p x = ((chebVand [|fromInterval a b x|] m) *$ c) $@ (1,1) in
  (p, c)

let rec h_lp x =
  if x < 0.0 then h_lp (abs_float x)
  else if x <= 0.5 then
    1.0
  else if x >= 1.0 then
    0.0
  else
    1.0 -. exp (~-. (exp (1.0 /. (x -. 1.0)) /. (x -. 0.5)))

let hankelRank ?(filter=false) ?(orderConst=2.0) ?(sobolev=2.0) ?(disp=1)
    a b xs l epsilon =
  let n = Array.length xs
  and mm = int_of_float (orderConst *. float (mesh_norm xs))
  and zs = Array.copy xs in
  let m = 2 * mm in
  fromInterval' a b zs;
  let s = Array1.create precision fortran_layout m in
  if filter then
    for i = 1 to m do
      s.{i} <-  (sqrt (h_lp (float i /. float m))) /. (float i ** sobolev)
    done
  else
    for i = 1 to m do
      s.{i} <-  1.0 /. (float i ** sobolev)
    done;
  let v = chebVand zs m in
  colScale' v s;
  let a = v *$~ v in
  let anrm = norm1 a in
  let sing = singValues' (a $ (1 <-> l, l+disp <-> n)) in
  let i = ref 1
  and fini = ref false in
  while not !fini && !i <= l do
    if sing.{!i} < anrm *. epsilon then
      fini := true
    else
      incr i
  done;
  Printf.printf "N=%i M=%i l=%i rank=%i\n" n m l !i
    
let error ?(nrm=1) ?(n=100) a b f p =
  let acc =
    if nrm = 1 then
      (fun acc err -> acc +. err /. float n)
    else if nrm = 2 then
      (fun acc err -> sqrt (acc ** 2.0 +. err ** 2.0))
    else if nrm >= max_int then
      max
    else
      failwith "Unknown norm"
  in
  let err = ref 0.0 in
  for i = 1 to n do
    let xi = a +. (b -. a) *. float i /. float (n+1) in
    err := acc !err (abs_float (f xi -. p xi))
  done;
  !err
    
let uniform ?(ends=false) a b n =
  if ends then
    Array.init n (fun i -> a +. (b -. a) *. float i /. float (n - 1))
  else
    Array.init n (fun i -> a +. (b -. a) *. float (i+1) /. float (n+1))


    
let cluster bs n =
  let m = Array.length bs - 1 in
  Array.concat (Array.to_list
		  (Array.init m (fun i -> uniform bs.(i) bs.(i+1) (n/m))))
    
let cardinalBasis ?(orderRate=2.0) ?(preserve=2) ?(sobolev=2.0) a b xs =
  let n = Array.length xs in
  let m = int_of_float (orderRate *. float n)
  and zs = Array.copy xs in
  fromInterval' a b zs;
  let v = chebVand xs m in
  let s = Array1.create precision fortran_layout m in
  if preserve >= 1 then
    s.{1} <- (2.0 ** 23.0) /. sqrt (float m)
  else
    s.{1} <- 1.0 /. sqrt (float m);
  for i = 2 to preserve do
    s.{i} <- (2.0 ** 23.0) /. (sqrt (float m /. 2.0));
  done;
  for i = (max 1 preserve) + 1 to m do
    s.{i} <- 1.0 /. (sqrt (float m /. 2.0) *. (float (i-1) ** sobolev))
  done;
  colScale' v s;
  let c = v /$ (eye n n) in
  rowScale' s c;
  let f i x = ((chebVand [|fromInterval a b x|] m) *$
	       (c $ (All, i <-> i))) $@ (1,1) in
  f

let uniformGap a b c d n1 n2 =
  Array.init (n1 + n2) (fun i ->
    if i < n1 then begin
      a +. (b -. a) *. float (i+1) /. float (n1+1)
    end else begin
      c +. (d -. c) *. float ((i-n1)+1) /. float (n2+1)
    end)
    
(* int T_n(x) dx = (1/2)(T_{n+1}/(n+1) - T_{n-1}/(n-1)).
   T_n(1) = 1 T_n(-1) = (-1)^n *)
let intChebs n =
  let intCheb n =
    if n = 0 then 2.0
    else if n = 1 then 0.0
    else
      0.5 *. (((cheb (n+1) 1.0 -. cheb (n+1) (-1.0)) /. float (n+1))
		-. ((cheb (n-1) 1.0 -. cheb (n-1) (-1.0)) /. float (n-1)))
  in
  fun2mat (fun _ i -> intCheb (i-1)) 1 n

let quadratureWts ?(orderConst=1.5) ?(sobolev=1.5) a b xs =
  let n = Array.length xs in
  let m = int_of_float (orderConst *. float (mesh_norm xs)) in
  let zs = Array.copy xs in
  fromInterval' a b zs;
  let s = Array1.create precision fortran_layout m in
  for i = 1 to m do
    s.{i} <- 1.0 /. (float i ** sobolev)
  done;
  let v = chebVand zs m in
  colScale' v s;
  let c = v /$ (eye n n) in
  rowScale' s c;
  (intChebs m) *$ c

let integrate ?(orderConst=1.5) ?(sobolev=1.5) a b xs f =
  ((quadratureWts ~orderConst ~sobolev a b xs) *$
   (discretize xs 1 (fun x _ -> f x))) $@ (1,1)

let kernel_matrix ?(orderConst=1.5) ?(sobolev=1.5) kernel a b xs ys =
  let m = Array.length xs
  and n = Array.length ys in
  let k = fun2mat (fun i j -> kernel xs.(i-1) ys.(j-1)) m n in
  let wx = quadratureWts ~orderConst ~sobolev a b xs
  and wy = quadratureWts ~orderConst ~sobolev a b ys in
  let sx = Array1.create precision fortran_layout m
  and sy = Array1.create precision fortran_layout n in
  for i = 1 to m do
    sx.{i} <- wx $@ (1,i)
  done;
  for i = 1 to n do
    sy.{i} <- wy $@ (1,i)
  done;
  (sx, k, sy)

(* Diagonal, congruent transformation *)
let dicot' w k =
  let m = noOfRows k in
  assert (m = noOfCols k && m = Array1.dim w);
  let ws = Array1.create precision fortran_layout m in
  for i = 1 to m do
    assert (w.{i} >= 0.0);
    ws.{i} <- sqrt w.{i}
  done;
  rowScale' ws k;
  colScale' k ws


(* 1D IE solver: sigma * u(xs) + \int_a^b k(xs,y) u(y) dy = f(xs) *)
let ie1D ?(orderConst=2.0) ?(sobolev=1.5) ?(tol=1e-8) ?(sigma=1.0) k f a b xs =
  let n = Array.length xs in
  let m = int_of_float (orderConst *. float (mesh_norm xs)) in
  Printf.printf "N=%i M(N)=%i" n m; print_newline ();
  let v1 =
    let zs = Array.copy xs in
    fromInterval' a b zs;
    let v1 = chebVand zs m in
    scale_real' sigma v1;
    v1
  in
  let integrand i j =
    (chebQuadAdaptive tol a b [|{Complex.re=xs.(i-1); Complex.im=0.0}|] 40 1
       (fun y _ -> k xs.(i-1) y *. cheb (j-1) y)) $@ (1,1)
  in
  let v = v1 +$ fun2mat integrand n m in
  let s = Array1.create precision fortran_layout m in
  for i = 1 to m do
    s.{i} <- 1.0 /. (float i ** sobolev)
  done;
  colScale' v s;
  let r = fun2mat (fun i j -> f xs.(i-1)) n 1 in
  let c = v /$ r in
  rowScale' s c;
  let p x =  ((chebVand [|fromInterval a b x|] m) *$ c) $@ (1,1) in
  (p, c)


let hankelRankIE1d ?(orderConst=2.0) ?(sobolev=1.5) ?(tol=1e-8) ?(sigma=1.0)
    k f a b xs =
  let n = Array.length xs in
  let m = int_of_float (orderConst *. float (mesh_norm xs)) in
  Printf.printf "N=%i M(N)=%i" n m; print_newline ();
  let v1 =
    let zs = Array.copy xs in
    fromInterval' a b zs;
    let v1 = chebVand zs m in
    scale_real' sigma v1;
    v1
  in
  let integrand i j =
    (chebQuadAdaptive tol a b [|{Complex.re=xs.(i-1); Complex.im=0.0}|] 40 1
       (fun y _ -> k xs.(i-1) y *. cheb (j-1) y)) $@ (1,1)
  in
  let v = v1 +$ fun2mat integrand n m in
  let s = Array1.create precision fortran_layout m in
  for i = 1 to m do
    s.{i} <- 1.0 /. (float i ** sobolev)
  done;
  colScale' v s;
  let r = fun2mat (fun i j -> f xs.(i-1)) n 1 in
  let c = v /$ r in
  rowScale' s c;
  let p x =  ((chebVand [|fromInterval a b x|] m) *$ c) $@ (1,1) in
  (p, c)
      
    (* Minimum 2-norm solver with linear constraints.
       min_x ||Cx|| s.t. Ax = b. A and b are destroyed. *)
let minConstrainedNorm' c a b =
  let n = noOfCols a
  and m = noOfRows a in
  assert (m <= n);
  assert (m = noOfRows b && n = noOfRows c && n = noOfCols c);
  let (l, q1T, q2T) =
    let (l, q) = lq' a in
    let qm = eye n n in
    q @* qm;
    let (q1T, q2T) = partition2x1 m (n-m) qm in
    (l, q1T, q2T)
  in
  subs' l b;
  let h = q1T *~$ b in
  h -$ q2T *~$ ((c *$~ q2T) /$ (c *$ h))
    
    
    (* Both of these are inefficient right now. *)
let bigArrayFromList xs =
  Array1.of_array precision fortran_layout (Array.of_list xs)
    
let bigArrayToList s =
  Array.to_list (Array.init (Array1.dim s) (fun i -> s.{i+1}) )
    
let basis ?(orderRate=2.0) ?(preserve=2) ?(sobolev=2.0)
    a b xs =
  let n = List.length xs in
  if n > 0 then begin
    let preserve = min n preserve in
    let m = int_of_float (float n *. orderRate) in
    let zs = Array.of_list xs in
    fromInterval' a b zs;
    let s = Array1.create precision fortran_layout m in
    let l = b -. a in
    if preserve >= 1 then
      s.{1} <- (l *. 2.0 ** 23.0) /. sqrt (float m)
    else
      s.{1} <- l /. sqrt (float m);
    for i = 2 to preserve do
      s.{i} <- (l *. 2.0 ** 23.0) /. (sqrt (float m /. 2.0));
    done;
    for i = (max 1 preserve) + 1 to m do
      s.{i} <- l /. (sqrt (float m /. 2.0) *. (float (i-1) ** sobolev))
    done;
    let v = chebVand zs m in
    colScale' v s;
    (bigArrayToList s, iD v)
  end else
    ([], ooo)
      
type mlTree =
  | Node of float * float * float * int * int * mlTree * mlTree
  | Leaf of float * float
	
let basisML ?(orderRate=2.0) ?(preserve=2) ?(sobolev=2.0)
    ?(coarseScale=2) a b xs =
  let rec gg a b xs =
    if List.length xs <= coarseScale then begin
      let (s, v) = basis ~orderRate ~preserve ~sobolev a b xs in
      (s, v, Leaf (a, b))
    end else begin
      let mid = (a +. b) /. 2.0 in
      let (xs1, xs2) = List.partition (fun x -> x <= mid) xs in
      let (s1, basisL, treeL) = gg a mid xs1
      and (s2, basisR, treeR) = gg mid b xs2
      and (s,  basisP) = basis ~orderRate ~preserve ~sobolev a b xs in
      let n1 = List.length s1
      and n2 = List.length s2
      and n  = List.length s
      and m1 = List.length xs1
      and m2 = List.length xs2 in
      ( s @ s1 @ s2,
	matrix1x2 n            (n1  +  n2)
	  basisP (matrix2x2 n1     n2
		    m1 basisL (const zero)
		    m2 (const zero) basisR),
	Node (a, mid, b, n1, n2, treeL, treeR) )
    end
  in
  let (s, fnV, tree) = gg a b (Array.to_list xs)
  and m = Array.length xs in
  let n = List.length s in
  let v = rawMatrix m n in
  fnV v;
  (v, bigArrayFromList s, tree)
    
    
    (* Multiply multi-level Chebyshev expansion coefficients *)
let rec mlVand tree x c =
  match tree with
  | Node (a, mid, b, n1, n2, treeL, treeR) ->
      let nt = noOfRows c in
      let n = nt - n1 - n2 in
      let (cp, cc) = partition2x1 n (n1+n2) c in
      let (cl, cr) = partition2x1 n1 n2    cc in
      addn
	((chebVand [|fromInterval a b x|] n *$ cp) $@ (1,1))
	(if x <= mid then
	  mlVand treeL x cl
	else
	  mlVand treeR x cr)
  | Leaf (a, b) ->
      (chebVand [|fromInterval a b x|] (noOfRows c) *$ c) $@ (1,1)
							       
							       
let interpolateML ?(orderRate=2.0) ?(preserve=2) ?(sobolev=2.0)
    ?(coarseScale=2) f a b xs =
  let (vFn, s, tree) = basisML ~orderRate ~coarseScale
      ~preserve ~sobolev a b xs in
  let n = Array.length xs in
  let v = rawMatrix n (Array1.dim s) in
  iD vFn v;
  let r = fun2mat (fun i j -> f xs.(i-1)) n 1 in
  let c = v /$ r in
  rowScale' s c;
  (tree, c)
    (* let p x = mlVand{L} {level} tree x c in p *)
    
    
    (* mlVand but using only one level *)
let rec mlVandL level tree x c =
  match tree with
  | Node (a, mid, b, n1, n2, treeL, treeR) ->
      let nt = noOfRows c in
      let n = nt - n1 - n2 in
      let (cp, cc) = partition2x1 n (n1+n2) c in
      let (cl, cr) = partition2x1 n1 n2    cc in
      if level <= 0 then
	(chebVand [|fromInterval a b x|] n *$ cp) $@ (1,1)
      else if x <= mid then
	mlVandL (level - 1) treeL x cl
      else
	mlVandL (level - 1) treeR x cr
  | Leaf (a, b) ->
      if level <= 0 then
	(chebVand [|fromInterval a b x|] (noOfRows c) *$ c) $@ (1,1)
      else
	zero
	  
	  
	  (* like basis, but also returns boundary values *)
let basisD ?(orderRate=2.0) ?(preserve=2) ?(sobolev=2.0)
    a b xs =
  let n = List.length xs in
  if n > 0 then begin
    let preserve = min n preserve in
    let m = int_of_float (float n *. orderRate) in
    let zs = Array.of_list xs in
    fromInterval' a b zs;
    let s = Array1.create precision fortran_layout m in
    let l = b -. a in
    if preserve >= 1 then
      s.{1} <- (l *. 2.0 ** 23.0) /. sqrt (float m)
    else
      s.{1} <- l /. sqrt (float m);
    for i = 2 to preserve do
      s.{i} <- (l *. 2.0 ** 23.0) /. (sqrt (float m /. 2.0));
    done;
    for i = (max 1 preserve) + 1 to m do
      s.{i} <- l /. (sqrt (float m /. 2.0) *. (float (i-1) ** sobolev))
    done;
    let v = chebVand zs m in
    colScale' v s;
    (diag s, iD v, chebVand [|(-1.0)|] m, chebVand [|1.0|] m)
  end else
    (zeros 0 0, ooo, zeros 0 m, zeros 0 m)
      
      
      (* like basisML but takes into account delta functions in derivative *)
let basisMLd ?(orderRate=2.0) ?(preserve=2) ?(sobolev=2.0)
    ?(coarseScale=2) a b xs =
  let rec gg a b xs =
    if List.length xs <= coarseScale then begin
      let (s, v, lVal, rVal) =
	basisD ~orderRate ~preserve ~sobolev a b xs in
      (s, v, Leaf (a, b), lVal, rVal)
    end else begin
      let mid = (a +. b) /. 2.0 in
      let (xs1, xs2) = List.partition (fun x -> x <= mid) xs in
      let (s1, basisL, treeL, lValL, rValL) = gg a mid xs1
      and (s2, basisR, treeR, lValR, rValR) = gg mid b xs2
      and (s,  basisP, lVal, rVal) =
	basisD ~orderRate ~preserve ~sobolev a b xs in
      let n1 = List.length s1
      and n2 = List.length s2
      and n  = List.length s
      and m1 = List.length xs1
	  and m2 = List.length xs2 in
      ( diagD [s; s1; s2],
	matrix1x2 n            (n1  +  n2)
	  basisP (matrix2x2 n1     n2
		    m1 basisL (const zero)
		    m2 (const zero) basisR),
	Node (a, mid, b, n1, n2, treeL, treeR) )
    end
  in
  let (s, fnV, tree) = gg a b (Array.to_list xs)
  and m = Array.length xs in
  let n = List.length s in
  let v = rawMatrix m n in
  fnV v;
  (v, bigArrayFromList s, tree)
    
    
let basisMatrix ?(orderRate=1.5) ?(preserve=2) ?(sobolev=2.0) a b xs =
  let n = List.length xs in
  let m = int_of_float (orderRate *. float n)
  and zs = Array.of_list xs in
  fromInterval' a b zs;
  let s = Array1.create precision fortran_layout m in
  if n > 0 then begin
    let preserve = min preserve m in
    if preserve >= 1 then
      s.{1} <- (2.0 ** 23.0) /. sqrt (float m)
    else
      s.{1} <- 1.0 /. sqrt (float m);
    for i = 2 to preserve do
      s.{i} <- (2.0 ** 23.0) /. (sqrt (float m /. 2.0));
    done;
    for i = (max 1 preserve) + 1 to m do
      s.{i} <- 1.0 /. (sqrt (float m /. 2.0) *. (float (i-1) ** sobolev))
    done;
    let v = chebVand zs m in
    colScale' v s;
    (v, bigArrayToList s)
  end else
    (zeros 0 0, [])
      
      
      (* single level multiple overlapping intervals *)
let basisOL ?(orderRate=1.5) ?(preserve=2) ?(sobolev=2.0) l dl a b xs =
  let index x =
    let rec ii xs n =
      match xs with
      | [] -> failwith "index not found"
      | y :: xs ->
	  if y = x then
	    n
	  else
	    ii xs (n+1)
    in
    ii xs 1
  in
  let segments =
    let intervals =
      let rec mkIntervals intervals a =
	if (b -. a) <= l then
	  List.rev ((a,b) :: intervals)
	else
	  mkIntervals ((a, a +. l) :: intervals) (a +. dl)
      in
      mkIntervals [] a
    in
    List.map2 (fun (a, b) ys ->
      if List.length ys > 0 then
	(a, b, ys, index (List.hd ys),
	 int_of_float (orderRate *. float (List.length ys)))
      else
	(a, b, ys, (-1), 0))
      intervals
      (List.map (fun (a,b) -> List.filter (fun x -> a <= x & x <= b) xs)
	 intervals)
  in
  let m = List.length xs in
  let n = List.fold_left (fun r (_, _, _, _, ni) -> r + ni) 0 segments
  in
  let v = zeros m n in
  let (_, s, _) =
    List.fold_left
      (fun (v, s, ni) (a, b, ys, mi, nc) ->
	let (vi, si) = basisMatrix ~orderRate ~preserve ~sobolev a b ys in
	let mc = noOfRows vi in
	assert (mi + mc - 1 <= m);
	assert (ni + nc - 1 <= n);
	assert (nc = noOfCols vi);
	(v $ (mi <-> mi + mc - 1, ni <-> ni + nc - 1)) <-$ vi;
	(v, List.rev_append si s, ni + nc))
      (v, [], 1)
      segments
  in
  let s = bigArrayFromList (List.rev s) in
  (v, s, segments)
    
    
let rec olVand segments x c =
  match segments with
  | [] -> zero
  | (a, b, ys, ni, p) :: segments ->
      let k = noOfRows c in
      let (c1, c2) = partition2x1 p (k-p) c in
      addn
	(if a <= x & x <= b then begin
	  (chebVand [|fromInterval a b x|] p *$ c1) $@ (1,1)
	end else zero)
	(olVand segments x c2)
	
	
let interpolateOL ?(orderRate=1.5) ?(preserve=2) ?(sobolev=2.0)
    l dl f a b xs =
  let (v, s, segments) = basisOL ~orderRate ~preserve ~sobolev
      l dl a b (Array.to_list xs)
  and n = Array.length xs in
  let r = fun2mat (fun i _ -> f xs.(i-1)) n 1 in
  let c = v /$ r in
  rowScale' s c;
  let p x = olVand segments x c in
      (p, c, segments)
    
    
    
let cbInterp f c m xs =
  let n = Array.length xs in
  let v = fun2mat (fun i j -> c j xs.(i-1)) n m in
  let d = v /$ (fun2mat (fun i j -> f xs.(i-1)) n 1) in
  let p x =
    let acc = ref 0.0 in
    for i = 1 to m do
      acc := !acc +. (c i x) *. (d $@ (i,1))
    done;
    !acc
  in
  p


let lstSqrsInterp ?(orderRate=0.5) ?(qOrderRate=2.0) ?(qSobolev=1.5)
    ?(qPreserve= 0) a b xs f =
  let n = Array.length xs in
  let m = int_of_float (orderRate *. sqrt (float n)) in
  let r = fun2mat (fun i j -> f xs.(i-1)) n 1 in
  let zs = Array.copy xs in
  fromInterval' a b zs;
  let v = chebVand zs m in
  let wts = quadratureWts ~orderRate:qOrderRate ~sobolev:qSobolev
      ~preserve:qPreserve a b xs in
  let s = Array1.create precision fortran_layout n in
  for i = 1 to n do
    s.{i} <- wts $@ (1,i)
  done;
  rowScale' s v;
  rowScale' s r;
  let c = v /$ r in
  Printf.printf "resid = %g; cond = %g\n" (norm1 (v *$ c -$ r)) (cond v);
  let p x = ((chebVand [|fromInterval a b x|] m) *$ c) $@ (1,1) in
  (p, c)


(* 2D interpolation *)
type point2D = {x : float; y : float}

let tuple2D f z = {x = f z.x; y = f z.y}

let mesh_norm zs =
  let dist theta psi = mod_float (abs_float (theta -. psi)) pi in
  let n = Array.length zs in
  let ts = Array.map (tuple2D acos) zs in
  let acc = ref pi in
  for i = 0 to n-2 do
    for j = i + 1 to n - 1 do
      let len = sqrt (dist ts.(i).x ts.(j).x ** 2.0 +.
		      dist ts.(i).y ts.(j).y ** 2.0) in
      if len < !acc then acc := len;
    done;
  done;
  int_of_float (ceil (pi /. !acc))
  
let radial_set n =
  let acc = ref 0 in
  for i = 0 to n-1 do
    for j = 0 to n-1 do
      if i * i + j * j <= n * n then incr acc
    done
  done;
  let bs = Array.make !acc (-1,-1) in
  let ii = ref 0 in
  for i = 0 to n - 1 do
    for j = 0 to n - 1 do
      if i * i + j * j <= n * n then begin
	bs.(!ii) <- (i, j);
	incr ii
      end
    done
  done;
  bs

let fromRect zl zu z = {x = fromInterval zl.x zu.x z.x;
			y = fromInterval zl.y zu.y z.y}

let fromRect' zl zu zs =
  for i = 0 to Array.length zs - 1 do
    zs.(i) <- fromRect zl zu zs.(i)
  done

let cheb2D (i, j) z = cheb i z.x *. cheb j z.y

let chebVand2D zs ijs =
  let n = Array.length ijs in
  let m = Array.length zs in
  let v = rawMatrix m n in
  for i = 0 to m - 1 do
    for j = 0 to n - 1 do
      (cheb2D ijs.(j) zs.(i) => (i+1, j+1)) v
    done
  done;
  v

type region = {zl : point2D; zu : point2D; contains : point2D -> bool}

let union region1 region2 =
  {
   zl = {x = min region1.zl.x region2.zl.x;
	 y = min region1.zl.y region2.zl.y};
   zu = {x = max region1.zl.x region2.zl.x;
	 y = max region1.zl.y region2.zl.y};
   contains = fun z ->
     region1.contains z || region2.contains z
  }

let intersection region1 region2 =
  {
   zl = {x = max region1.zl.x region2.zl.x;
	 y = max region1.zl.y region2.zl.y};
   zu = {x = min region1.zl.x region2.zl.x;
	 y = min region1.zl.y region2.zl.y};
   contains = fun z ->
     region1.contains z && region2.contains z
  }

let difference region1 region2 =
  {
   zl = region1.zl;
   zu = region1.zu;
   contains = fun z ->
     region1.contains z && not (region2.contains z)
  }

let complement region =
  {
   zl = region.zl;
   zu = region.zu;
   contains = fun z ->
     let zl = region.zl
     and zu = region.zu in
     zl.x <= z.x && z.x <= zu.x && zl.y <= z.y && z.y <= zu.y &&
     not (region.contains z)
  }

let change_bbox region zl zu =
  {
   zl = zl;
   zu = zu;
   contains = fun z ->
     zl.x <= z.x && z.x <= zu.x && zl.y <= z.y && z.y <= zu.y &&
     region.contains z
  }

let rotate theta region =
  let c = cos theta
  and s = sin theta in
  let rotated z =   { x = c *. z.x -. s *. z.y;
		      y = s *. z.x +. c *. z.y }
  and unrotated z = { x = c *. z.x +. s *. z.y;
		      y = c *. z.y -. s *. z.x } in
  let zl' = rotated zl
  and zu' = rotated zu in
  {
   zl = {x = min zl'.x zu'.x; y = min zl'.y zu'.y};
   zu = {x = max zl'.x zu'.x; y = max zl'.y zu'.y};
   contains = fun z -> region.contains (unrotated z)
  }

let translate d region =
  {
   zl = {x = zl.x +. d.x; y = zl.y +. d.y};
   zu = {x = zu.x +. d.x; y = zu.y +. d.y};
   contains = fun z -> region.contains {x = z.x -. d.x; y = z.y -. d.y}
  }

let scale s region =
  {
   zl = {x = zl.x *. s.x; y = zl.y *. s.y};
   zu = {x = zu.x *. s.x; y = zu.y *. s.y};
   contains = fun z -> region.contains {x = z.x /. s.x; y = z.y /. s.y}
  }

let make_triangle ?(fst={x=0.0; y=0.0}) scnd thrd =
  {
   zl = {x = min (min fst.x scnd.x) thrd.x;
	 y = min (min fst.y scnd.y) thrd.y};
   zu = {x = max (max fst.x scnd.x) thrd.x;
	 y = max (max fst.y scnd.y) thrd.y};
   contains =
   let v1 = {x = scnd.x -. fst.x; y = scnd.y -. fst.y}
   and v2 = {x = thrd.x -. fst.x; y = thrd.y -. fst.y} in
   let det = v1.x *. v2.y -. v2.x *. v1.y in
   assert (det != 0.0);
   fun z ->
     let zx = z.x -. fst.x
     and zy = z.y -. fst.y in
     let l1 = (zx *. v2.y -. v2.x *. zy) /. det
     and l2 = (v1.x *. zy -. zx *. v1.y) /. det in
     l1 > 0.0 && l2 > 0.0 && l1 +. l2 <= 1.0
  }

let make_rect zl zu =
  {
   zl = zl;
   zu = zu;
   contains = fun z ->
     zl.x <= z.x && z.x <= zu.x && zl.y <= z.y && z.y <= zu.y
  }

let make_circle ?(center={x=0.0; y=0.0}) radius =
  {
   zl = {x = center.x -. radius; y = center.y -. radius};
   zu = {x = center.x +. radius; y = center.y +. radius};
   contains = fun z ->
     (z.x -. center.x) ** 2.0 +. (z.y -. center.y) ** 2.0 <= radius ** 2.0
  }

let make_ellipse ?(center={x=0.0; y=0.0}) ?(semi_maj_axis={x=1.0; y=0.0})
    semi_major semi_minor =
  let nrm = sqrt (semi_maj_axis.x ** 2.0 +. semi_maj_axis.y ** 2.0) in
  let cosT = semi_maj_axis.x /. nrm
  and sinT = semi_maj_axis.y /. nrm in
  {
   zl = {x = center.x -. semi_major; y = center.y -. semi_major};
   zu = {x = center.x +. semi_major; y = center.y +. semi_major};
   contains = fun z ->
     let zx = z.x -. center.x
     and zy = z.y -. center.y in
     let rx = cosT *. zx +. sinT *. zy
     and ry = sinT *. zx -. cosT *. zy in
     (rx /. semi_major) ** 2.0 +. (ry /. semi_minor) ** 2.0 <= 1.0
  }

let make_annulus ?(center={x=0.0; y=0.0}) inner_radius outer_radius =
  {
   zl = {x = center.x -. outer_radius; y = center.y -. outer_radius};
   zu = {x = center.x +. outer_radius; y = center.y +. outer_radius};
   contains = fun z ->
     let r = sqrt ((z.x -. center.x) ** 2.0 +. (z.y -. center.y) ** 2.0) in
     inner_radius <= r && r <= outer_radius
  }

let make_net zl zu mx my =
  {
   zl = zl;
   zu = zu;
   contains =
     let xs = uniform zl.x zu.x mx
     and ys = uniform zl.y zu.y my in
     let is_mem x xs =
       let found = ref false
       and i = ref 0
       and n = Array.length xs in
       while not !found && !i < n do
	 if xs.(!i) = x then found := true
	 else incr i
       done;
       !found
     in
     fun z ->
       is_mem z.x xs || is_mem z.y ys
  }

let uniform2D region nx ny =
  let xs = uniform region.zl.x region.zu.x nx
  and ys = uniform region.zl.y region.zu.y ny in
  let acc = ref [] in
  for i = 0 to nx - 1 do
    for j = 0 to ny - 1 do
      let zij =  {x = xs.(i); y = ys.(j)} in
      if region.contains zij then
	acc := zij :: !acc
    done
  done;
  Array.of_list !acc

let interpolate2D ?(orderConst=2.0) ?(sobolev=2.0) f region ws =
  let n = Array.length ws in
  let r = fun2mat (fun i j -> f ws.(i-1)) n 1 in
  let zs = Array.copy ws in
  fromRect' region.zl region.zu zs;
  let mr = int_of_float (orderConst *. float (mesh_norm zs)) in
  let mijs = radial_set mr in
  let m = Array.length mijs in
  Printf.printf "N=%i M(N)=%i #=%i" n mr m; print_newline ();
  let s = Array1.create precision fortran_layout m in
  let len (i,j) = 1.0 +. sqrt (float (i * i + j * j)) in
  for i = 0 to m-1 do
    s.{i+1} <- 1.0 /. (len mijs.(i) ** sobolev);
  done;
  let v = chebVand2D zs mijs in
  colScale' v s;
  let c = v /$ r in
  rowScale' s c;
  Printf.printf "cond = %g\n" (cond v);
  let p z =
    let w = fromRect region.zl region.zu z in
    ((chebVand2D (Array.make 1 w) mijs) *$ c) $@ (1,1)
  in
  (p, c)
    
(* Pick relatively prime (after +1) pairs *)
let mesh2D mx my nx ny =
  let xx = uniform (-1.0) 1.0 mx
  and yy = uniform (-1.0) 1.0 my
  and xs = uniform (-1.0) 1.0 nx
  and ys = uniform (-1.0) 1.0 ny in
  let vs = Array.init mx (fun i -> Array.init ny (fun j ->
    {x = xx.(i); y = ys.(j)}))
  and hs = Array.init my (fun j -> Array.init nx (fun i ->
    {x = xs.(i); y = yy.(j)}))
  in
  let vs = Array.concat (Array.to_list vs)
  and hs = Array.concat (Array.to_list hs) in
  Array.append vs hs


(* Surface plotting routine *)
let splot ?(n=10) ?(pm3d=false) ?(fname="") f region =
  let zl = region.zl
  and zu = region.zu
  and tmpDataFile = Filename.temp_file "camlFloat" "plotData" in
  let outc = open_out_bin tmpDataFile in
  for i = 1 to n do
    let xi = zl.x +. (zu.x -. zl.x) *. float i /. float (n+1) in
    for j = 1 to n do
      let yj = zl.y +. (zu.y -. zl.y) *. float j /. float (n+1) in
      let zij =  {x=xi; y=yj} in
      if region.contains zij then begin
	let fij = f zij in
	Printf.fprintf outc "%g %g %g\n" xi yj fij
      end;
    done;
    Printf.fprintf outc "\n"
  done;
  close_out outc;
  let tmpFile = Filename.temp_file "camlFloat" "plotCmd" in
  let outc2 = open_out tmpFile in
  if not (fname = "") then begin
    print_endline "hello";
    Printf.fprintf outc2 "set terminal postscript color\n";
    Printf.fprintf outc2 "set output '%s' \n" fname;
  end;
  let withpm3d = if pm3d then " with pm3d notitle " else " notitle " in
  Printf.fprintf outc2 "splot [%g:%g] [%g:%g] '%s' %s \n"
    zl.x zu.x zl.y zu.y tmpDataFile withpm3d;
  close_out outc2;
  let err = Sys.command ("gnuplot -persist " ^ tmpFile) in
  if err <> 0 then
    prerr_string ("gnuplot " ^ tmpFile ^ " : " ^ string_of_int err);
  Sys.remove tmpDataFile;
  Sys.remove tmpFile
    

let error2D ?(nrm=1) ?(n=10) region f p =
  let acc =
    if nrm = 1 then
      (fun acc err -> acc +. err /. float (n * n))
    else if nrm = 2 then
      (fun acc err -> sqrt (acc ** 2.0 +. err ** 2.0))
    else if nrm >= max_int then
      max
    else
      failwith "Unknown norm"
  in
  let err = ref 0.0 in
  let zl = region.zl
  and zu = region.zu in
  for i = 1 to n do
    let xi = zl.x +. (zu.x -. zl.x) *. float i /. float (n+1) in
    for j = 1 to n do
      let yj = zl.y +. (zu.y -. zl.y) *. float j /. float (n+1) in
      let zij = {x = xi; y = yj} in
      if region.contains zij then
	err := acc !err (abs_float (f zij -. p zij))
    done
  done;
  !err

let to_complex z = {Complex.re = z.x; Complex.im = z.y}
let from_complex z = {x = z.Complex.re; y = z.Complex.im}

(* Pi_i (zi + unit(z-zi)*|zu-zl|*(|z - zi|/|zu-zl|)^ai) *)
let pinchAt zl zu zas =
  let l = Complex.norm (Complex.sub (to_complex zu) (to_complex zl)) in
  let doPinches z =
    let zc = ref (to_complex z) in
    for i = 0 to Array.length zas - 1 do
      let (zi, ai) = zas.(i) in
      let zic = to_complex zi in
      let d = Complex.sub !zc zic in
      let dr = Complex.norm d
      and dt = Complex.arg d in
      zc := Complex.add zic (Complex.polar (l *. ((dr /. l) ** ai)) dt)
    done;
    from_complex !zc
  in
  doPinches

(* Convert a function in polar coordinates to cartesian coordinates *)
let polar_function f z =
  let r = sqrt (z.x ** 2.0 +. z.y ** 2.0)
  and t = atan2 z.y z.x in
  f r t

(* Useful for plotting *)
let extend_function f value region =
  fun z ->
    if region.contains z then
      f z
    else
      value


