(*
  File: nla.ml
  Author: Shivkumar Chandrasekaran
  Email: shiv@ece.ucsb.edu
  Date: 1.17.2002
  Purpose: Uniform interface to single/double Lapack and Blas

  $Id: nla.ml,v 1.14 2008/01/19 03:14:21 shiv Exp shiv $
*)

(** Interface to real numerical linear algebra libraries. *)
module type NLA =
sig
  open Bigarray
  type t
  type precision_elt
  type precision_elt_real
  val precision : (t, precision_elt) kind
  val precision_real : (float, precision_elt_real) kind
  val machine_epsilon : float
  val re : t -> float
  val im : t -> float
  val conjugate : t -> t
  val zero : t
  val one : t
  val minus_one : t
  val fromInt : int -> t
  val toInt : t -> int
  val fromFloat : float -> t
  val randn : unit -> t
  val printfn : t -> unit
  val sprintfn : t -> string
  val addn : t -> t -> t
  val subn : t -> t -> t
  val muln : t -> t -> t
  val mulr : float -> t -> t
  val lange : int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (float, precision_elt_real, fortran_layout) Array1.t -> float
  val lacpy : int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int -> int
  val gelqf : int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val unmlq : int -> int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val geqlf : int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val unmql : int -> int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val geqrf : int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val unmqr : int -> int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val gerqf : int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val unmrq : int -> int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val gels : int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t -> int -> int
  val trtrs : int -> int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int -> int
  val gesvd : int -> int -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (float, precision_elt_real, fortran_layout) Array1.t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array1.t -> int ->
    (float, precision_elt_real, fortran_layout) Array1.t -> int
  val cblas_trans : int
  val cblas_gemm : int -> int -> int -> int -> int -> int -> t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int -> t ->
    (t, precision_elt, fortran_layout) Array2.t -> int -> int -> int
  val cblas_swap : int -> (t, precision_elt, fortran_layout) Array2.t
    -> int -> int -> (t, precision_elt, fortran_layout) Array2.t ->
    int -> int -> int
  val getrf : int -> int -> (t, precision_elt, fortran_layout) Array2.t
    -> int -> int -> (int, int_elt, fortran_layout) Array1.t -> int
end

module Double_Real_NLA =
struct
  open Bigarray
  type t = float
  type precision_elt = float64_elt
  type precision_elt_real = float64_elt
  let precision = float64
  and precision_real = float64
  let machine_epsilon = epsilon_float
  let zero = 0.0
  and one = 1.0
  and minus_one=(-1.0)
  let re x = x
  let im _ = 0.0
  let conjugate x = x
  let fromInt n = float n
  let toInt x = int_of_float x
  let fromFloat x = x
  let randn () = Random.float 2.0 -. 1.0
  let printfn r = Printf.printf "% 4.2g\t" r
  let sprintfn r = Printf.sprintf "% 4.2g\t" r
  let addn = ( +. )
  let subn = ( -. )
  let muln = ( *. )
  let mulr = ( *. )
  let lange = Blop_c_mod.my_dlange
  let lacpy = Blop_c_mod.my_dlacpy
  let gelqf = Blop_c_mod.my_dgelqf
  let unmlq = Blop_c_mod.my_dormlq
  let geqlf = Blop_c_mod.my_dgeqlf
  let unmql = Blop_c_mod.my_dormql
  let geqrf = Blop_c_mod.my_dgeqrf
  let unmqr = Blop_c_mod.my_dormqr
  let gerqf = Blop_c_mod.my_dgerqf
  let unmrq = Blop_c_mod.my_dormrq
  let gels  = Blop_c_mod.my_dgels
  let trtrs = Blop_c_mod.my_dtrtrs
  let gesvd x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
    x14 x15 x16 x17 = Blop_c_mod.my_dgesvd x1 x2 x3 x4 x5 x6 x7 x8
    x9 x10 x11 x12 x13 x14 x15 x16
  let cblas_trans = 112
  let cblas_gemm = Blop_c_mod.my_cblas_dgemm
  let cblas_swap = Blop_c_mod.my_cblas_dswap
  let getrf = Blop_c_mod.my_dgetrf
end

module Single_Real_NLA =
struct
  open Bigarray
  type t = float
  type precision_elt = float32_elt
  type precision_elt_real = float32_elt
  let precision = float32
  and precision_real = float32
  let machine_epsilon = 5.96046447754e-08
  let zero = 0.0
  and one = 1.0
  and minus_one=(-1.0)
  let re x = x
  let im _ = 0.0
  let conjugate x = x
  let fromInt n = float n
  let toInt x = int_of_float x
  let fromFloat x = x
  let randn () = Random.float 2.0 -. 1.0
  let printfn r = Printf.printf "% 4.2g\t" r
  let sprintfn r = Printf.sprintf "% 4.2g\t" r
  let addn = ( +. )
  let subn = ( -. )
  let muln = ( *. )
  let mulr = ( *. )
  let lange = Blop_c_mod.my_slange
  let lacpy = Blop_c_mod.my_slacpy
  let gelqf = Blop_c_mod.my_sgelqf
  let unmlq = Blop_c_mod.my_sormlq
  let geqlf = Blop_c_mod.my_sgeqlf
  let unmql = Blop_c_mod.my_sormql
  let geqrf = Blop_c_mod.my_sgeqrf
  let unmqr = Blop_c_mod.my_sormqr
  let gerqf = Blop_c_mod.my_sgerqf
  let unmrq = Blop_c_mod.my_sormrq
  let gels  = Blop_c_mod.my_sgels
  let trtrs = Blop_c_mod.my_strtrs
  let gesvd x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
    x14 x15 x16 x17 = Blop_c_mod.my_sgesvd x1 x2 x3 x4 x5 x6 x7 x8
    x9 x10 x11 x12 x13 x14 x15 x16
  let cblas_trans = 112
  let cblas_gemm = Blop_c_mod.my_cblas_sgemm
  let cblas_swap = Blop_c_mod.my_cblas_sswap
  let getrf = Blop_c_mod.my_sgetrf
end

module Double_Complex_NLA =
struct
  open Bigarray
  type t = Complex.t
  type precision_elt = complex64_elt
  type precision_elt_real = float64_elt
  let precision = complex64
  and precision_real = float64
  let machine_epsilon = epsilon_float
  let zero = Complex.zero
  and one = Complex.one
  and minus_one = {Complex.re=(-1.0);Complex.im=0.0}
  let re {Complex.re=rep} = rep
  and im {Complex.im=imp} = imp
  let conjugate z = Complex.conj z
  let fromInt n = {Complex.re=float n; Complex.im=0.0}
  let toInt z = int_of_float (re z)
  let fromFloat x = {Complex.re=x; Complex.im=0.0}
  let randn () =
    {Complex.re=Random.float 2.0 -. 1.0; Complex.im=Random.float 2.0 -. 1.0}
  let printfn z = Printf.printf "% 4.2g%+4.2gi\t" (re z) (im z)
  let sprintfn z = Printf.sprintf "% 4.2g%+4.2gi\t" (re z) (im z)
  let addn = Complex.add
  let subn = Complex.sub
  let muln = Complex.mul
  let mulr r z =
    {Complex.re= r *. z.Complex.re; Complex.im = r *. z.Complex.im}
  let lange = Blop_c_mod.my_zlange
  let lacpy = Blop_c_mod.my_zlacpy
  let gelqf = Blop_c_mod.my_zgelqf
  let unmlq = Blop_c_mod.my_zunmlq
  let geqlf = Blop_c_mod.my_zgeqlf
  let unmql = Blop_c_mod.my_zunmql
  let geqrf = Blop_c_mod.my_zgeqrf
  let unmqr = Blop_c_mod.my_zunmqr
  let gerqf = Blop_c_mod.my_zgerqf
  let unmrq = Blop_c_mod.my_zunmrq
  let gels  = Blop_c_mod.my_zgels
  let trtrs = Blop_c_mod.my_ztrtrs
  let gesvd = Blop_c_mod.my_zgesvd
  let cblas_trans = 113
  let cblas_gemm = Blop_c_mod.my_cblas_zgemm
  let cblas_swap = Blop_c_mod.my_cblas_zswap
  let getrf = Blop_c_mod.my_zgetrf
end

module Single_Complex_NLA =
struct
  open Complex
  open Bigarray
  type t = Complex.t
  type precision_elt = complex32_elt
  type precision_elt_real = float32_elt
  let precision = complex32
  and precision_real = float32
  let machine_epsilon = 5.96046447754e-08
  let zero = Complex.zero
  and one = Complex.one
  and minus_one = {Complex.re=(-1.0);Complex.im=0.0}
  let re {Complex.re=rep} = rep
  and im {Complex.im=imp} = imp
  let conjugate z = conj z
  let fromInt n = {Complex.re=float n; Complex.im=0.0}
  let toInt z = int_of_float (re z)
  let fromFloat x = {Complex.re=x; Complex.im=0.0}
  let randn () =
    {Complex.re=Random.float 2.0 -. 1.0; Complex.im=Random.float 2.0 -. 1.0}
  let printfn z = Printf.printf "% 4.2g%+4.2gi\t" (re z) (im z)
  let sprintfn z = Printf.sprintf "% 4.2g%+4.2gi\t" (re z) (im z)
  let addn = Complex.add
  let subn = Complex.sub
  let muln = Complex.mul
  let mulr r z =
    {Complex.re= r *. z.Complex.re; Complex.im = r *. z.Complex.im}
  let lange = Blop_c_mod.my_clange
  let lacpy = Blop_c_mod.my_clacpy
  let gelqf = Blop_c_mod.my_cgelqf
  let unmlq = Blop_c_mod.my_cunmlq
  let geqlf = Blop_c_mod.my_cgeqlf
  let unmql = Blop_c_mod.my_cunmql
  let geqrf = Blop_c_mod.my_cgeqrf
  let unmqr = Blop_c_mod.my_cunmqr
  let gerqf = Blop_c_mod.my_cgerqf
  let unmrq = Blop_c_mod.my_cunmrq
  let gels  = Blop_c_mod.my_cgels
  let trtrs = Blop_c_mod.my_ctrtrs
  let gesvd = Blop_c_mod.my_cgesvd
  let cblas_trans = 113
  let cblas_gemm = Blop_c_mod.my_cblas_cgemm
  let cblas_swap = Blop_c_mod.my_cblas_cswap
  let getrf = Blop_c_mod.my_cgetrf
end



(* A signature that is common to real and complex blops. *)
module type BLOP =
sig
  open Bigarray
  type t
  type precision_elt
  type precision_elt_real
  val precision : (t, precision_elt) kind
  val precision_real : (float, precision_elt_real) kind
  val machine_epsilon : float

  (* Generic arithmetic *)
  val re : t -> float
  val im : t -> float
  val conjugate : t -> t
  val zero : t
  val one : t
  val minus_one : t
  val fromInt : int -> t
  val toInt : t -> int
  val fromFloat : float -> t
  val randn : unit -> t
  val printfn : t -> unit
  val sprintfn : t -> string
  val addn : t -> t -> t
  val subn : t -> t -> t
  val muln : t -> t -> t
  val mulr : float -> t -> t

  type matrix
  type transformer = matrix -> unit
  type transform
  type factor

  val ( $@ ) : matrix -> int * int -> t
  val ( => ) : t -> int * int -> matrix -> unit

  val noOfRows : matrix -> int
  val noOfCols : matrix -> int

  val printMatrixHeight : int ref
  val printMatrixWidth : int ref
  val printMatrix : matrix -> unit

  val rawMatrix : ?ooc:bool -> int -> int -> matrix
  val consts : ?ooc:bool -> t -> int -> int -> matrix
  val const : t -> matrix -> unit
  val zeros : ?ooc:bool -> int -> int -> matrix
  val ones : ?ooc:bool -> int -> int -> matrix
  val randomMatrix : ?ooc:bool -> int -> int -> matrix
  val fun2mat : ?ooc:bool -> (int -> int -> t) -> int -> int -> matrix
  val fun2mat' : (int -> int -> t) -> matrix -> unit
  val eye : ?ooc:bool -> int -> int -> matrix
  val matrix : t array array -> matrix
  val mmapdMatrix : ?shared:bool -> string -> int -> int -> matrix
  val is_ooc : matrix -> bool
  val is_in : matrix -> bool
  val to_ooc : matrix -> matrix
  val from_ooc : matrix -> matrix

  type range =
      Range of int * int
    | All
  val ( $ ) : matrix -> range * range -> matrix
  val ( <-> ) : int -> int -> range

  val partition2x1 : int -> int -> matrix -> matrix * matrix
  val partition1x2 : int -> int -> matrix -> matrix * matrix
  val partition2x2 :
    int -> int -> int -> int -> matrix -> matrix * matrix * matrix * matrix
  val partition1x3 : int -> int -> int -> matrix -> matrix * matrix * matrix
  val partition3x1 : int -> int -> int -> matrix -> matrix * matrix * matrix
  val partition3x3 :
    int -> int -> int -> int -> int -> int -> matrix ->
	matrix * matrix * matrix *
	matrix * matrix * matrix *
	matrix * matrix * matrix
  val partitionInfx1 : int list -> matrix -> matrix list
  val partition1xInf : int list -> matrix -> matrix list
  val partitionNx1 : int array -> matrix -> matrix array
  val partition1xN : int array -> matrix -> matrix array

  val ooo : matrix -> unit
  val matrix2x1 :
    int -> transformer -> int -> transformer -> matrix -> unit
  val matrix3x1 : int -> transformer -> int -> transformer ->
    int -> transformer -> matrix -> unit
  val matrix1x2 :
    int -> int -> transformer -> transformer -> matrix -> unit
  val matrix1x3 : int -> int -> int ->
    transformer -> transformer -> transformer -> matrix -> unit
  val matrix2x2 : int -> int ->
	int -> transformer -> transformer ->
	int -> transformer -> transformer -> matrix -> unit
  val matrix2x3 : int -> int -> int ->
	int -> transformer -> transformer -> transformer ->
	int -> transformer -> transformer -> transformer -> matrix -> unit
  val matrix4x2 : int -> int ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer -> matrix -> unit
  val matrixInfx1 : int list -> transformer list -> matrix ->
    unit
  val matrixNx1 : int array -> transformer array -> matrix -> unit

  val copy : ?ooc:bool -> matrix -> matrix
  val iD : matrix -> matrix -> unit
  val ( $=> ) : matrix -> range * range -> matrix -> unit
  val ( <-| ) : matrix -> transformer -> unit
  val ( <-$ ) : matrix -> matrix -> unit

  val ( |+| ) : matrix -> matrix -> matrix -> unit
  val ( |-| ) : matrix -> matrix -> matrix -> unit
  val ( |++| ) : matrix -> matrix -> unit
  val ( |--| ) : matrix -> matrix -> unit
  val ( +$ ) : matrix -> matrix -> matrix
  val ( -$ ) : matrix -> matrix -> matrix

  val transp : ?ooc:bool -> matrix -> matrix
  val transp' : matrix -> matrix -> unit
  val conjg' : matrix -> unit

  val ( |*| ) : matrix -> matrix -> matrix -> unit
  val ( |*~| ) : matrix -> matrix -> matrix -> unit
  val ( |~*| ) : matrix -> matrix -> matrix -> unit
  val ( *$ ) : matrix -> matrix -> matrix
  val ( *~$ ) : matrix -> matrix -> matrix
  val ( *$~ ) : matrix -> matrix -> matrix
  val ( |*++| ) : matrix -> matrix -> matrix -> unit
  val ( |*--| ) : matrix -> matrix -> matrix -> unit
  val ( |*~++| ) : matrix -> matrix -> matrix -> unit
  val ( |*~--| ) : matrix -> matrix -> matrix -> unit
  val ( |~*++| ) : matrix -> matrix -> matrix -> unit
  val ( |~*--| ) : matrix -> matrix -> matrix -> unit

  val norm1 : matrix -> float
  val normInf : matrix -> float
  val normF : matrix -> float
  val normMax : matrix -> float

  val eyeTransform : transform
  val lq' : matrix -> factor * transform
  val ql' : matrix -> transform * factor
  val qr' : matrix -> transform * factor
  val rq' : matrix -> factor * transform
  val reflectors : transform ->
    (t, precision_elt, fortran_layout) Array1.t * matrix
  val plu' : matrix -> transform * factor * factor
  val copyL : ?ooc:bool -> factor -> matrix
  val iDL : factor -> matrix -> unit
  val copyU : ?ooc:bool -> factor -> matrix
  val iDU : factor -> matrix -> unit
  val ( @* ) : transform -> matrix -> unit
  val ( @~* ) : transform -> matrix -> unit
  val ( *@ ) : matrix -> transform -> unit
  val ( *~@ ) : matrix -> transform -> unit
  val subs' : factor -> matrix -> unit
  val subsT' : factor -> matrix -> unit
  val ( |/| ) : matrix -> matrix -> unit
  val ( |~/| ) : matrix -> matrix -> unit
  val ( /$ ) : matrix -> matrix -> matrix
  val ( /~$ ) : matrix -> matrix -> matrix
  val pluSolve' : matrix -> matrix -> unit

  val scale' : t -> matrix -> unit
  val rowScale' : (t, precision_elt, fortran_layout) Array1.t ->
    matrix -> unit
  val colScale' : matrix -> (t, precision_elt, fortran_layout)
    Array1.t -> unit
  val scale_real' : float -> matrix -> unit
  val rowScale_real' : (float, precision_elt_real, fortran_layout) Array1.t ->
    matrix -> unit
  val colScale_real' : matrix -> (float, precision_elt_real, fortran_layout)
    Array1.t -> unit

  val svd' :
      ?ooc:bool -> matrix ->
	matrix *
	(float, precision_elt_real, fortran_layout) Array1.t *
	matrix
  val svdL' :
      ?ooc:bool -> matrix ->
	matrix *
	(float, precision_elt_real, fortran_layout) Array1.t
  val svdR' :
      ?ooc:bool -> matrix ->
	(float, precision_elt_real, fortran_layout) Array1.t *
	  matrix
  val singValues' : matrix ->
    (float, precision_elt_real, fortran_layout) Array1.t
  val cond : matrix -> float
  val isLowerTriangular : matrix -> factor
  val isUpperTriangular : matrix -> factor
  val permutation : (int, int_elt, fortran_layout) Array1.t -> transform
  val tensor : ?ooc:bool -> matrix -> matrix -> matrix
  val vecC' : matrix -> matrix -> unit
  val unvecC' : matrix -> matrix -> unit
  val vecR' : matrix -> matrix -> unit
  val unvecR' : matrix -> matrix -> unit
end


module Bloper (LnB : NLA) =
struct
  open Bigarray
  open LnB

  type t = LnB.t
  type precision_elt = LnB.precision_elt
  type precision_elt_real = LnB.precision_elt_real
  let precision = LnB.precision
  and precision_real = LnB.precision_real
  let zero = LnB.zero
  and one = LnB.one
  and minus_one = LnB.minus_one
  let re = LnB.re
  let im = LnB.im
  let conjugate = LnB.conjugate
  let fromInt = LnB.fromInt
  let toInt = LnB.toInt
  let fromFloat = LnB.fromFloat
  let printfn = LnB.printfn
  let sprintfn = LnB.sprintfn
  let randn = LnB.randn
  let addn = LnB.addn
  let subn = LnB.subn
  let muln = LnB.muln
  let mulr = LnB.mulr

  let machine_epsilon = LnB.machine_epsilon

  type weakBigarray2 = { ooc_fname : string;
			 mutable dirty : bool;
			 warray : 
			   (t, precision_elt, fortran_layout) Array2.t Weak.t
		       }

  type supArray2 =
      Reg of (t, precision_elt, fortran_layout) Array2.t
    | Wek of weakBigarray2

  type matrix = Dense of supArray2 * int * int * int * int

  type transform =
      QofLQ of (t, precision_elt, fortran_layout) Array1.t * matrix * int
    | QofQL of (t, precision_elt, fortran_layout) Array1.t  * matrix * int
    | QofQR of (t, precision_elt, fortran_layout) Array1.t  * matrix * int
    | QofRQ of (t, precision_elt, fortran_layout) Array1.t  * matrix * int
    | Permutation of (int, int_elt, fortran_layout) Array1.t
    | Product of transform list

  type factor =
    | LofQL of supArray2 * int * int * int * int
    | LofLQ of supArray2 * int * int * int * int
    | LofPLU of supArray2 * int * int * int * int
    | UofPLU of supArray2 * int * int * int * int
    | RofQR of supArray2 * int * int * int * int
    | RofRQ of supArray2 * int * int * int * int

  type transformer = matrix -> unit
	
  let noOfRows (Dense (_, sr, lr, _, _)) = lr - sr + 1

  let noOfCols (Dense (_, _, _, sc, lc)) = lc - sc + 1

  (* Finalizing the bigarray *)
  let finalize_wba_ba2 wba ba2 =
    if wba.dirty then begin
      let fd_o = open_out_bin wba.ooc_fname in
      Marshal.to_channel fd_o ba2 [];
      close_out fd_o;
      wba.dirty <- false
    end

  (* Finalize the weakBigarray2 itself *)
  let finalize_wba wba =
    Sys.remove wba.ooc_fname

  (* Wrapper for Array2.create that tries compact in case of no memory. *)
  let myCreate prec lay m n =
    try
      Array2.create prec lay m n
    with
      Out_of_memory ->
	prerr_endline "Out of memory...";
	Gc.compact ();
	Gc.minor ();
	Array2.create prec lay m n

  let wba2create ooc m n =
    if ooc then begin
      let barr = myCreate precision fortran_layout m n in
      let (fname,fd) = Filename.open_temp_file ~mode:[Open_binary]
	  "nla" ".wba" in
      (* Marshal.to_channel fd barr []; *) (* not needed? *)
      close_out fd; (* Close it now that it has been created *)
      let wba = { ooc_fname = fname; dirty = true; warray = Weak.create 1} in
      Gc.finalise (finalize_wba_ba2 wba) barr;
      Weak.set wba.warray 0 (Some barr);
      Gc.finalise finalize_wba wba;
      Wek wba
    end else
      Reg (myCreate precision fortran_layout m n)

  let rawMatrix ?(ooc = false) m n =
    Dense (wba2create ooc m n, 1, m, 1, n)

  let sup2big a =
    match a with
      Reg aa -> aa
    | Wek wba ->
	match Weak.get wba.warray 0 with
	  Some aa -> aa
	| None ->
	    if wba.dirty then begin
	      prerr_endline "Warn: GC being called.";
	      (* Finalisation function was not called. Try forcing it. *)
	      Gc.major ();
	      if wba.dirty then
		Gc.full_major ();
	      if wba.dirty then
		Gc.compact ()
	    end;
	    assert (not wba.dirty);
	    let fd_in = open_in_bin wba.ooc_fname in
	    let aa = Marshal.from_channel fd_in in
	    close_in fd_in;
	    Weak.set wba.warray 0 (Some aa);
	    Gc.finalise (finalize_wba_ba2 wba) aa;
	    aa

  let get_array2 (Dense (a, _, _, _, _)) = sup2big a

  let get_all2 (Dense (_, sr, lr, sc, lc) as a) =
    let aa = get_array2 a in
    (aa, sr, lr, sc, lc)

  (* a2 ia anchored here to ensure that it doesn't get GC'd before the
     dirty bit is set. *)
  let mark_dirty (Dense (a, _, _, _, _)) a2 =
    match a with
      Reg _ -> ()
    | Wek wba ->
	wba.dirty <- true (* rest is not needed? *) (*
	match Weak.get wba.warray 0 with
	  None ->
	    prerr_endline "Loosing data....";
	    wba.dirty <- true
	| Some barr ->
	    assert (a2 == barr);
	    let fd = open_out_bin wba.ooc_fname in
	    Marshal.to_channel fd barr [];
	    close_out fd;
	    wba.dirty <- false *)

  let mmapdMatrix ?(shared=true) fname m n =
    let flags = [Unix.O_RDWR; Unix.O_CREAT]
    and perm = 0000400 lor 0000200 in
    let fd = Unix.openfile fname flags perm in
    let a = Array2.map_file fd precision fortran_layout shared m n in
    let ma = Array2.dim1 a
    and na = Array2.dim2 a in
    Dense (Reg a, 1, ma, 1, na)

  let matrix aa =
    let a = Array2.of_array precision fortran_layout aa in
    Dense (Reg a, 1, Array2.dim1 a, 1, Array2.dim2 a)

  let consts ?(ooc=false) x m n =
    let aa = rawMatrix ~ooc m n in
    Array2.fill (get_array2 aa) x;
    mark_dirty aa (get_array2 aa);
    aa

  let const x aa =
    let (a, sr, lr, sc, lc) = get_all2 aa in
    for j = sc to lc do
      for i = sr to lr do
	a.{i,j} <- x
      done
    done;
    mark_dirty aa a

  let zeros ?(ooc=false) m n = consts ~ooc zero m n

  let ones ?(ooc=false) m n = consts ~ooc one m n

  let ( $@ ) a (i, j) =
    let (aa, sr, _, sc, _) = get_all2 a in
    Array2.get aa (sr+i-1) (sc+j-1)

  type range = Range of int * int
               | All

  let extractRange ix f =
    match ix with
	Range (x, y) -> (x, y)
      | All -> f

  let ( $ ) (Dense (a, sr, lr, sc, lc)) (r, c) =
    let (i1, i2) = extractRange r (1, lr - sr + 1)
    and (j1, j2) = extractRange c (1, lc - sc + 1) in
    assert (1 <= i1 && i1 - 1 <= i2 && i2 <= lr
    && 1 <= j1 && j1 - 1 <= j2 && j2 <= lc);
    Dense (a, sr + i1 - 1, sr + i2 - 1, sc + j1 - 1, sc + j2 - 1)
	
  let ( <-> ) i j = Range (i,j)

  let fun2mat' fn a =
    let (d, sr, lr, sc, lc) = get_all2 a in
    let m = lr - sr + 1
    and n = lc - sc + 1 in
    for j = 1 to n do
      for i = 1 to m do
	d.{sr+i-1, sc+j-1} <- fn i j
      done
    done;
    mark_dirty a d

  let fun2mat ?(ooc=false) fn m n =
    let a = rawMatrix ~ooc m n in
    fun2mat' fn a;
    a

  let eye ?(ooc=false) m n =
    fun2mat ~ooc (fun i j -> if i = j then one else zero) m n

  let randomMatrix ?(ooc=false) m n =
    let genRandFloat _ _ = randn () in
      fun2mat ~ooc genRandFloat m n

  let printMatrixHeight = ref 10
  let printMatrixWidth = ref 10
  let printMatrix a =
    let m = noOfRows a
    and n = noOfCols a in
    if m > !printMatrixHeight || n > !printMatrixWidth then begin
      Printf.printf "Rows = %i; Columns = %i" m n;
      print_newline ();
    end;
    for i=1 to min m !printMatrixHeight do
      for j=1 to min n !printMatrixWidth do
	printfn (a $@ (i,j))
      done;
      if n > !printMatrixWidth then print_string "...";
      print_newline ()
    done;
    if m > !printMatrixHeight then begin
      print_string "| ...";
      print_newline ()
    end
      
  let partition2x1 m0 m1 a =
    assert (0 <= m0 && 0 <= m1 && m0 + m1 = noOfRows a);
    let a0 = a $ (1 <-> m0, All) in
      (a0, a $ (m0 + 1 <-> m0 + m1, All))

  let partition3x1 m0 m1 m2 a =
    assert (m0 + m1 + m2 = noOfRows a);
    (a $ (1 <-> m0, All), a $ (m0 + 1 <-> m0 + m1, All),
     a $ (m0 + m1 + 1 <-> noOfRows a, All))

  let partition1x2 n0 n1 a =
    assert (n0 + n1 = noOfCols a);
    let a0 = a $ (All, 1 <-> n0) in
      (a0, a $ (All, n0 + 1 <-> n0 + n1))

  let partition1x3 n0 n1 n2 a =
    assert (n0 + n1 + n2 = noOfCols a);
    (a $ (All, 1 <-> n0), a $ (All, n0 + 1 <-> n0 + n1),
     a $ (All, n0 + n1 + 1 <-> noOfCols a))

  let partition2x2 n0 n1 m0 m1 a =
    assert (n0 + n1 = noOfCols a && m0 + m1 = noOfRows a);
    let a00 = a $ (1 <-> m0, 1 <-> n0)
    and a01 = a $ (1 <-> m0, n0 + 1 <-> n0 + n1)
    and a10 = a $ (m0 + 1 <-> m0 + m1, 1 <-> n0) in
      (a00, a01, a10, a $ (m0 + 1 <-> m0 + m1, n0 + 1 <-> n0 + n1))

  let partition3x3 n0 n1 n2 m0 m1 m2 a =
    assert (n0 + n1 + n2 = noOfCols a && m0 + m1 + m2 = noOfRows a);
    let a00 = a $ (1 <-> m0, 1 <-> n0)
    and a01 = a $ (1 <-> m0, n0 + 1 <-> n0 + n1)
    and a02 = a $ (1 <-> m0, n0 + n1 + 1 <-> noOfCols a)
    and a10 = a $ (m0 + 1 <-> m0 + m1, 1 <-> n0)
    and a11 = a $ (m0 + 1 <-> m0 + m1, n0 + 1 <-> n0 + n1)
    and a12 = a $ (m0 + 1 <-> m0 + m1, n0 + n1 + 1 <-> noOfCols a)
    and a20 = a $ (m0 + m1 + 1 <-> noOfRows a, 1 <-> n0)
    and a21 = a $ (m0 + m1 + 1 <-> noOfRows a, n0 + 1 <-> n0 + n1)
    and a22 = a $ (m0 + m1 + 1 <-> noOfRows a, n0 + n1 + 1 <-> noOfCols a) in
      (a00, a01, a02,
       a10, a11, a12,
       a20, a21, a22)

  let rec partitionInfx1 ms a =
    match ms with
      | [] -> []
      | [m0] -> assert (noOfRows a = m0); [a]
      | m0 :: ms ->
	  let a0, a1 = partition2x1 m0 (noOfRows a - m0) a in
	    a0 :: partitionInfx1 ms a1

  let rec partition1xInf ns a =
    match ns with
      | [] -> []
      | [n0] -> assert (noOfCols a = n0); [a]
      | n0 :: ns ->
	  let a0, a1 = partition1x2 n0 (noOfCols a - n0) a in
	    a0 :: partition1xInf ns a1

  let partitionNx1 ms b =
    let n = Array.length ms in
    let bs = Array.make n (rawMatrix 0 0)
    and cr = ref 1 in
      for i = 0 to n - 1 do
	let ncr = !cr + ms.(i) - 1 in
	  bs.(i) <- b $ (!cr <-> ncr, All);
	  cr := ncr + 1
      done;
      bs

  let partition1xN ms b =
    let n = Array.length ms in
    let bs = Array.make n (rawMatrix 0 0)
    and cc = ref 1 in
      for i = 0 to n - 1 do
	let ncc = !cc + ms.(i) - 1 in
	  bs.(i) <- b $ (All, !cc <-> ncc);
	  cc := ncc + 1
      done;
      bs

  let ooo a = ()

  let (=>) x (i,j) aa =
    let (a, sr, _, sc, _) = get_all2 aa in
    a.{sr+i-1,sc+j-1} <- x;
    mark_dirty aa a

  let matrix2x1 m0 a0
                m1 a1 a =
    assert (m0 + m1 = noOfRows a);
    a0 (a $ (1 <-> m0, All)); (* a is modified *)
    a1 (a $ (m0 + 1 <-> m0 + m1, All)) (* a is modified *)

  let matrix3x1 m0 a0
                m1 a1
                m2 a2 a =
    assert (m0 + m1 + m2 = noOfRows a);
    a0 (a $ (1 <-> m0, All)); (* a is modified *)
    a1 (a $ (m0 + 1 <-> m0 + m1, All)); (* a is modified *)
    a2 (a $ (m0 + m1 + 1 <-> m0 + m1 + m2, All))

  let matrix1x2 n0 n1
                a0 a1 a =
    assert (n0 + n1 = noOfCols a);
    a0 (a $ (All, 1 <-> n0)); (* a is modified *)
    a1 (a $ (All, n0 + 1 <-> n0 + n1)) (* a is modified *)

  let matrix1x3 n0 n1 n2
                a0 a1 a2 a =
    assert (n0 + n1 + n2 = noOfCols a);
    a0 (a $ (All, 1 <-> n0)); (* a is modified *)
    a1 (a $ (All, n0 + 1 <-> n0 + n1)); (* a is modified *)
    a2 (a $ (All, n0 + n1 + 1 <-> n0 + n1 + n2))

  let matrix2x2 n0  n1
            m0 a00 a01
	    m1 a10 a11 a =
    assert (m0 + m1 = noOfRows a && n0 + n1 = noOfCols a);
    a00 (a $ (1 <-> m0, 1 <-> n0)); (* a is modified in sequence *)
    a01 (a $ (1 <-> m0, n0 + 1 <-> n0 + n1));
    a10 (a $ (m0 + 1 <-> m0 + m1, 1 <-> n0));
    a11 (a $ (m0 + 1 <-> m0 + m1, n0 + 1 <-> n0 + n1))

  let matrix2x3 n0  n1 n2
            m0 a00 a01 a02
	    m1 a10 a11 a12 a =
    assert (m0 + m1 = noOfRows a && n0 + n1 + n2 = noOfCols a);
    a00 (a $ (1 <-> m0, 1 <-> n0)); (* a is modified in sequence *)
    a01 (a $ (1 <-> m0, n0 + 1 <-> n0 + n1));
    a02 (a $ (1 <-> m0, n0 + n1 + 1 <-> n0 + n1 + n2));
    a10 (a $ (m0 + 1 <-> m0 + m1, 1 <-> n0));
    a11 (a $ (m0 + 1 <-> m0 + m1, n0 + 1 <-> n0 + n1));
    a12 (a $ (m0 + 1 <-> m0 + m1, n0 + n1 + 1 <-> n0 + n1 + n2))

  let matrix4x2 n0   n1
             m0 a00 a01
             m1 a10 a11
             m2 a20 a21
             m3 a30 a31 a =
    assert (m0+m1+m2+m3 = noOfRows a && n0+n1 = noOfCols a);
    a00 (a $ (1 <-> m0, 1 <-> n0)); (* a is modified in sequence *)
    a01 (a $ (1 <-> m0, n0 + 1 <-> n0 + n1));
    a10 (a $ (m0 + 1 <-> m0 + m1, 1 <-> n0));
    a11 (a $ (m0 + 1 <-> m0 + m1, n0 + 1 <-> n0 + n1));
    a20 (a $ (m0 + m1 + 1 <-> m0 + m1 + m2, 1 <-> n0));
    a21 (a $ (m0 + m1 + 1 <-> m0 + m1 + m2, n0 + 1 <-> n0 + n1));
    a30 (a $ (m0 + m1 + m2 + 1 <-> m0 + m1 + m2 + m3, 1 <-> n0));
    a31 (a $ (m0 + m1 + m2 + 1 <-> m0 + m1 + m2 + m3, n0 + 1 <-> n0 + n1))
    
  let rec matrixInfx1 ms bs b =
    match ms, bs with
      | [], [] -> ()
      | [m0], [b0] -> assert (noOfRows b = m0); b0 b
      | m0 :: ms, b0 :: bs ->
	  let b0', b1' = partition2x1 m0 (noOfRows b - m0) b in
	    b0 b0';
	    matrixInfx1 ms bs b1'
      | _ ->
	  invalid_arg "matrixInfx1 requires lists to be of the same length."

  let matrixNx1 ms bs b =
    assert (Array.length ms = Array.length bs);
    let n = Array.length ms
    and sr = ref 1 in
    for i = 0 to n-1 do
      bs.(i) (b $ (!sr <-> !sr + ms.(i) - 1, All));
      sr := !sr + ms.(i)
    done

  let getArrayParams aa =
    let (a, sr, lr, sc, lc) = get_all2 aa in
    let lda = Array2.dim1 a in
    (a, (sc - 1 ) * lda + sr - 1, lda)

(* {'F', 'L', 'U'} from blop_c.c *)
  let lacpy_full = 0
  and lacpy_lower = 1
  and lacpy_upper = 2

  let setTo a b =
    let m = noOfRows a
    and n = noOfCols a in
    assert (m = noOfRows b && n = noOfCols b);
    if min m n > 0 then
      let (aa, dispa, lda) = getArrayParams a
      and (bb, dispb, ldb) = getArrayParams b in
      ignore (lacpy lacpy_full m n aa dispa lda bb dispb ldb);
      mark_dirty b bb
	    
  let copy ?(ooc=false) a =
    let b = rawMatrix ~ooc (noOfRows a) (noOfCols a) in
      setTo a b;
      b

  let to_ooc (Dense (aa, sr, lr, sc, lc) as a) =
    match aa with
      Reg barr ->
	copy ~ooc:true a (* otherwise it is a bug! *)
(*	let (fname,fd) = Filename.open_temp_file ~mode:[Open_binary]
	    "nla" ".wba" in
	Marshal.to_channel fd barr [];
	close_out fd; (* Close it now that it has been created *)
	let wba = {ooc_fname = fname; dirty = false; warray = Weak.create 1} in
	Gc.finalise (finalize_wba_ba2 wba) barr;
	Weak.set wba.warray 0 (Some barr);
	Gc.finalise finalize_wba wba;
	Dense (Wek wba, sr, lr, sc, lc) *)
    | Wek wba -> a

  let from_ooc (Dense (aa, sr, lr, sc, lc) as a) =
    match aa with
      Reg _ -> a
    | Wek _ -> copy a (* otherwise bug *)

  (* same as setTo except that only the lower triangular (or trapezoidal)
     portion of a is copied into b. *)
  let setToL a b =
    let m = noOfRows b
    and n = noOfCols b
    and (bb, dispb, ldb) = getArrayParams b in
    if min m n > 0 then begin
      (match a with
	(LofQL (aaa, sra, lra, sca, lca)) ->
	  let aa = sup2big aaa in
	  assert (m = lra - sra + 1 && n = lca - sca + 1);
	  let lda = Array2.dim1 aa in
	  let dispa = lda * (sca - 1) + (sra - 1) in
	  if m = n then
	    ignore (lacpy lacpy_lower m n aa dispa lda bb dispb ldb)
	  else if m > n then begin
	    failwith "This should not happen"
	  end else begin
	    ignore (lacpy lacpy_lower m m aa (dispa + lda * (n - m))
		      lda bb (dispb + ldb * (n - m)) ldb);
	    ignore (lacpy lacpy_full m (n-m) aa dispa lda bb dispb ldb)
	  end
      | (LofLQ (aaa, sra, lra, sca, lca)) ->
	  let aa = sup2big aaa in
	  assert (m = lra - sra + 1 && n = lca - sca + 1);
	  let lda = Array2.dim1 aa in
	  let dispa = lda * (sca - 1) + (sra - 1) in
	  ignore (lacpy lacpy_lower m n aa dispa lda bb dispb ldb)
      | LofPLU (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  assert (m = lrl - srl + 1 && n = lcl - scl + 1);
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  ignore (lacpy lacpy_lower m n ll displ ldl bb dispb ldb);
	  for i = 1 to min m n do
	    (one => (i,i)) b
	  done
      | UofPLU _ | RofQR _ | RofRQ _ -> failwith "setToL: upper");
      mark_dirty b bb
    end

  let setToU a b =
    let m = noOfRows b
    and n = noOfCols b
    and (bb, dispb, ldb) = getArrayParams b in
    if min m n > 0 then begin
      (match a with
	UofPLU (uuu, sru, lru, scu, lcu) ->
	  let uu = sup2big uuu in
	  assert (m = lru - sru + 1 && n  = lcu - scu + 1);
	  let ldu = Array2.dim1 uu in
	  let dispu = (scu-1)*ldu+(sru-1) in
	  ignore (lacpy lacpy_upper m n uu dispu ldu bb dispb ldb)
      | RofQR (rrr, srr, lrr, scr, lcr) ->
	  let rr = sup2big rrr in
	  assert (m = lrr - srr + 1 && n = lcr - scr + 1);
	  let ldr = Array2.dim1 rr in
	  let dispr = (scr-1)*ldr+(srr-1) in
	  if m = n then
	    ignore (lacpy lacpy_upper m n rr dispr ldr bb dispb ldb)
	  else begin
	    assert (m < n);
	    ignore (lacpy lacpy_upper m m rr dispr ldr bb dispb ldb);
	    ignore (lacpy lacpy_full m (n-m) rr (dispr+ldr*m) ldr
		      bb (dispb+ldb*m) ldb)
	  end
      | RofRQ (rrr, srr, lrr, scr, lcr) ->
	  let rr = sup2big rrr in
	  assert (m = lrr - srr + 1 && n = lcr - scr + 1);
	  let ldr = Array2.dim1 rr in
	  let dispr = (scr-1)*ldr+(srr-1) in
	  if m = n then
	    ignore (lacpy lacpy_upper m n rr dispr ldr bb dispb ldb)
	  else if m < n then begin
	    failwith "This should not happen"
	  end else begin
	    ignore (lacpy lacpy_full (m-n) n rr dispr ldr bb dispb ldb);
	    ignore (lacpy lacpy_upper n n rr (dispr+m-n) ldr
		      bb (dispb+m-n) ldb)
	  end
      | LofQL _ | LofLQ _ | LofPLU _ -> failwith "setToU: lower");
      mark_dirty b bb
    end


  let noOfRows_f f =
    match f with
	LofLQ (_, sr, lr, _, _) -> lr - sr + 1
      | LofQL (_, sr, lr, _, _) -> lr - sr + 1
      | LofPLU (_, sr, lr, _, _) -> lr - sr + 1
      | UofPLU (_, sr, lr, _, _) -> lr - sr + 1
      | RofQR (_, sr, lr, _, _) -> lr - sr + 1
      | RofRQ (_, sr, lr, _, _) -> lr - sr + 1

  let noOfCols_f f =
    match f with
	LofQL (_, _, _, sc, lc) -> lc - sc + 1
      | LofLQ (_, _, _, sc, lc) -> lc - sc + 1
      | LofPLU (_, _, _, sc, lc) -> lc - sc + 1
      | UofPLU (_, _, _, sc, lc) -> lc - sc + 1
      | RofQR (_, _, _, sc, lc) -> lc - sc + 1
      | RofRQ (_, _, _, sc, lc) -> lc - sc + 1

  (* Same as copy except that only the lower triangular (or trapezoidal)
     portion of a is copied. *)
  let copyL ?(ooc=false) a =
    let b = zeros ~ooc (noOfRows_f a) (noOfCols_f a) in
      setToL a b;
      b

  let copyU ?(ooc=false) a =
    let b = zeros ~ooc (noOfRows_f a) (noOfCols_f a) in
    setToU a b;
    b

  let iD a b = setTo a b

  let iDL a b = setToL a b

  let iDU a b = setToU a b

  let ( $=> ) a rc = fun b -> setTo a (b $ rc)

  let ( <-$ ) a b = setTo b a

  let ( <-| ) a op = op a

  let (|+|) a b c =
    let (aa, sra, _, sca, _) = get_all2 a
    and (bb, srb, _, scb, _) = get_all2 b
    and (cc, src, _, scc, _) = get_all2 c in
    let m = noOfRows a
    and n = noOfCols a in
    assert (m = noOfRows b && m = noOfRows c &&
	    n = noOfCols b && n = noOfCols c);
    for j = 1 to n do
      for i = 1 to m do
	cc.{src + i - 1, scc + j - 1} <-
	  addn aa.{sra + i - 1, sca + j - 1} bb.{srb + i - 1, scb + j - 1}
      done
    done;
    mark_dirty c cc

  let (|-|) a b c =
    let (aa, sra, _, sca, _) = get_all2 a
    and (bb, srb, _, scb, _) = get_all2 b
    and (cc, src, _, scc, _) = get_all2 c in
    let m = noOfRows a
    and n = noOfCols a in
    assert (m = noOfRows b && m = noOfRows c &&
	    n = noOfCols b && n = noOfCols c);
    for j = 1 to n do
      for i = 1 to m do
	cc.{src + i - 1, scc + j - 1} <-
	  subn aa.{sra + i - 1, sca + j - 1} bb.{srb + i - 1, scb + j - 1}
      done
    done;
    mark_dirty c cc

  let (|++|) a b =
    let (aa, sra, _, sca, _) = get_all2 a
    and (bb, srb, _, scb, _) = get_all2 b in
    let m = noOfRows a
    and n = noOfCols a in
    assert (m = noOfRows b && n = noOfCols b);
    for j = 1 to n do
      for i = 1 to m do
	bb.{srb + i - 1, scb + j - 1} <-
	  addn bb.{srb + i - 1, scb + j - 1} aa.{sra + i - 1, sca + j - 1}
      done
    done;
    mark_dirty b bb

  (* a |--| b <=> b <- b - a *)
  let (|--|) a b =
    let (aa, sra, _, sca, _) = get_all2 a
    and (bb, srb, _, scb, _) = get_all2 b in
    let m = noOfRows a
    and n = noOfCols a in
    assert (m = noOfRows b && n = noOfCols b);
    for j = 1 to n do
      for i = 1 to m do
	bb.{srb + i - 1, scb + j - 1} <-
	  subn bb.{srb + i - 1, scb + j - 1} aa.{sra + i - 1, sca + j - 1}
      done
    done;
    mark_dirty b bb

  let is_ooc (Dense (a, _, _, _, _))  =
    match a with
      Reg _ -> false
    | Wek _ -> true

  let is_in (Dense (a, _, _, _, _)) =
    match a with
      Reg _ -> true
    | Wek wba -> Weak.check wba.warray 0

  let (+$) a b =
    let ooc = is_ooc a & is_ooc b in
    let c = rawMatrix ~ooc (noOfRows a) (noOfCols a) in
      (a |+| b) c;
      c

  let (-$) a b =
    let ooc = is_ooc a & is_ooc b in
    let c = rawMatrix ~ooc (noOfRows a) (noOfCols a) in
      (a |-| b) c;
      c

  let transp' a b =
    let (aa, sra, lra, sca, lca) = get_all2 a
    and (bb, srb, lrb, scb, lcb) = get_all2 b in
    let m = noOfRows a
    and n = noOfCols a in
    assert (m = noOfCols b && n = noOfRows b);
    if min m n > 0 then begin
      for j = 1 to n do
	for i = 1 to m do
	  bb.{srb+(j-1), scb+(i-1)} <- conjugate aa.{sra+(i-1),sca+(j-1)}
	done
      done;
      mark_dirty b bb
    end
    
  let transp ?(ooc=false) a =
    let b = rawMatrix ~ooc (noOfCols a) (noOfRows a) in
    transp' a b;
    b

  let rawArray1 n = Array1.create precision fortran_layout n
  and rawArray1_real n = Array1.create precision_real fortran_layout n

(* {'O','F','I','M'} from blop_c.c *)
  let lange_one = 0
  and lange_frobenius = 1
  and lange_infty = 2
  and lange_max = 3

  let norm1 a =
    let (aa, dispa, lda) = getArrayParams a in
    lange lange_one (noOfRows a) (noOfCols a) aa dispa lda (rawArray1_real 1)

  let normInf a =
    let (aa, dispa, lda) = getArrayParams a
    and m = noOfRows a in
    lange lange_infty m (noOfCols a) aa dispa lda (rawArray1_real m)

  let normF a =
    let (aa, dispa, lda) = getArrayParams a in
    lange lange_frobenius (noOfRows a) (noOfCols a) aa dispa lda
      (rawArray1_real 1)

  let normMax a =
    let (aa, dispa, lda) = getArrayParams a in
    lange lange_max (noOfRows a) (noOfCols a) aa dispa lda (rawArray1_real 1)

(*
	CBLAS parameters. Keep in syncs with system cblas.h file.
   enum CBLAS_ORDER {CblasRowMajor=101, CblasColMajor=102 };
   enum CBLAS_TRANSPOSE {CblasNoTrans=111, CblasTrans=112, CblasConjTrans=113,
                         AtlasConj=114};
   enum CBLAS_UPLO  {CblasUpper=121, CblasLower=122};
   enum CBLAS_DIAG  {CblasNonUnit=131, CblasUnit=132};
   enum CBLAS_SIDE  {CblasLeft=141, CblasRight=142};
*)
  let cblasRowMajor  = 101
  and cblasColMajor  = 102
  and cblasNoTrans   = 111
  and cblasTrans     = 112
  and cblasConjTrans = 113
  and atlasConj      = 114
  and cblasUpper     = 121
  and cblasLower     = 122
  and cblasNonUnit   = 131
  and cblasUnit      = 132
  and cblasLeft      = 141
  and cblasRight     = 142

  let ( |*| ) a b =
    fun c ->
      assert (noOfRows c = noOfRows a && noOfCols a = noOfRows b &&
	      noOfCols b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfCols a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblasNoTrans cblasNoTrans
		  (noOfRows a) (noOfCols b) (noOfCols a) one aa dispa lda bb
		  dispb ldb zero cc dispc ldc);
	mark_dirty c cc
      else const zero c
	  
  let ( |*~| ) a b =
    fun c ->
      assert (noOfRows c = noOfRows a && noOfCols a = noOfCols b  &&
	      noOfRows b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfCols a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblasNoTrans cblas_trans
		  (noOfRows a) (noOfRows b) (noOfCols a) one aa dispa lda bb
		  dispb ldb zero cc dispc ldc);
	mark_dirty c cc
      else const zero c

  let ( |~*| ) a b =
    fun c ->
      assert (noOfRows c = noOfCols a && noOfRows a = noOfRows b &&
	      noOfCols b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfRows b) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblas_trans cblasNoTrans
		  (noOfCols a) (noOfCols b) (noOfRows a) one aa dispa lda bb
		  dispb ldb zero cc dispc ldc);
	mark_dirty c cc
      else const zero c

  let ( *$ ) a b =
    let ooc = is_ooc a & is_ooc b in
    let c = rawMatrix ~ooc (noOfRows a) (noOfCols b) in
    (a |*| b) c;
    c

  let ( *~$ ) a b =
    let ooc = is_ooc a & is_ooc b in
    let c = rawMatrix ~ooc (noOfCols a) (noOfCols b) in
      (a |~*| b) c;
      c

  let ( *$~ ) a b =
    let ooc = is_ooc a & is_ooc b in
    let c = rawMatrix ~ooc (noOfRows a) (noOfRows b) in
      (a |*~| b) c;
      c

  let ( |*++| ) a b =
    fun c ->
      assert (noOfRows c = noOfRows a && noOfCols a = noOfRows b &&
	      noOfCols b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfCols a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblasNoTrans cblasNoTrans
		  (noOfRows a) (noOfCols b) (noOfCols a) one aa dispa lda bb
		  dispb ldb one cc dispc ldc);
	mark_dirty c cc
	  
  (* (a |*--| b) c <=> c <- c - a * b *)
  let ( |*--| ) a b =
    fun c ->
      assert (noOfRows c = noOfRows a && noOfCols a = noOfRows b &&
	      noOfCols b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfCols a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblasNoTrans cblasNoTrans
		  (noOfRows a) (noOfCols b) (noOfCols a) minus_one aa dispa lda
		  bb dispb ldb one cc dispc ldc);
	mark_dirty c cc

  let ( |*~++| ) a b =
    fun c ->
      assert (noOfRows c = noOfRows a && noOfCols a = noOfCols b &&
	      noOfRows b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfCols a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblasNoTrans cblas_trans
		  (noOfRows a) (noOfRows b) (noOfCols a) one aa dispa lda bb
		  dispb ldb one cc dispc ldc);
	mark_dirty c cc

  let ( |*~--| ) a b =
    fun c ->
      assert (noOfRows c = noOfRows a && noOfCols a = noOfCols b &&
	      noOfRows b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfCols a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblasNoTrans cblas_trans
		  (noOfRows a) (noOfRows b) (noOfCols a) minus_one aa dispa lda
		  bb dispb ldb one cc dispc ldc);
	mark_dirty c cc

  let ( |~*++| ) a b =
    fun c ->
      assert (noOfRows c = noOfCols a && noOfRows a = noOfRows b &&
	      noOfCols b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfRows a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblas_trans cblasNoTrans
		  (noOfCols a) (noOfCols b) (noOfRows a) one aa dispa lda bb
		  dispb ldb one cc dispc ldc);
	mark_dirty c cc
	  
  let ( |~*--| ) a b =
    fun c ->
      assert (noOfRows c = noOfCols a && noOfRows a = noOfRows b &&
	      noOfCols b = noOfCols c);
      if min (min (noOfRows c) (noOfCols c)) (noOfRows a) > 0 then
	let (aa, dispa, lda) = getArrayParams a
	and (bb, dispb, ldb) = getArrayParams b
	and (cc, dispc, ldc) = getArrayParams c in
	ignore (cblas_gemm cblasColMajor cblas_trans cblasNoTrans
		  (noOfCols a) (noOfCols b) (noOfRows a) minus_one aa dispa lda
		  bb dispb ldb one cc dispc ldc);
	mark_dirty c cc

  let eyeTransform = Product []

  let permutation ns = Permutation ns

  (* a is modified in place *)
  let lq' (Dense (aaa, sr, lr, sc, lc) as a) =
    let (aa, dispa, lda) = getArrayParams a
    and m = noOfRows a
    and n = noOfCols a in
    if min m n > 0 then
      let tau = rawArray1 (min m n)
      and lWork = -1
      and work = rawArray1 1 in
      let info = gelqf m n aa dispa lda tau work lWork in
      assert (info = 0);
      let lWork = toInt work.{1} in
      assert (lWork >= max 1 m);
      let info = gelqf m n aa dispa lda tau (rawArray1 lWork) lWork in
      assert (info = 0);
      mark_dirty a aa;
      if n > m then
	(LofLQ (aaa, sr, lr, sc, lc + m - n), QofLQ (tau, a, 0))
      else
	(LofLQ (aaa, sr, lr, sc, lc), QofLQ (tau, a, 0))
    else if n > m then
      (LofLQ (aaa, sr, lr, sc, lc + m - n), Product [])
    else
      (LofLQ (aaa, sr, lr, sc, lc), Product [])

  (* extract the individual reflectors of q *)
  let reflectors q =
    match q with
    | QofLQ (tau, a, k) ->
	let m = noOfRows a
	and n = noOfCols a in
	let vs = zeros n m in (* transposed *)
	for i = 1 to m do
	  (fromInt 1 => (i,i)) vs;
	  (vs $ (i+1 <-> n, i <-> i)) <-| transp' (a $ (i <-> i, i+1 <-> n))
	done;
	(tau, vs)
    | Product [] -> (rawArray1 0, zeros 0 0)
    | _ -> failwith "Not implemented yet! Do you need it?"

  (* a is modified in place *)
  let ql' (Dense (aaa, sr, lr, sc, lc) as a) =
    let (aa, dispa, lda) = getArrayParams a
    and m = noOfRows a
    and n = noOfCols a in
    if min m n > 0 then
      let tau = rawArray1 (min m n)
      and lWork = -1
      and work = rawArray1 1 in
      let info = geqlf m n aa dispa lda tau work lWork in
      assert (info = 0);
      let lWork = toInt work.{1} in
      assert (lWork >= max 1 n);
      let info = geqlf m n aa dispa lda tau (rawArray1 lWork) lWork in
      assert (info = 0);
      mark_dirty a aa;
      if m > n then
	(QofQL (tau, a, 0), LofQL (aaa, sr + m - n, lr, sc, lc))
      else
	(QofQL (tau, a, 0), LofQL (aaa, sr, lr, sc, lc))
    else if m > n then
      (Product [], LofQL (aaa, sr + m - n, lr, sc, lc))
    else
      (Product [], LofQL (aaa, sr, lr, sc, lc))

  (* a is modified in place *)
  let qr' (Dense (aaa, sr, lr, sc, lc) as a) =
    let (aa, dispa, lda) = getArrayParams a
    and m = noOfRows a
    and n = noOfCols a in
    if min m n > 0 then
      let tau = rawArray1 (min m n)
      and lWork = -1
      and work = rawArray1 1 in
      let info = geqrf m n aa dispa lda tau work lWork in
      assert (info = 0);
      let lWork = toInt work.{1} in
      assert (lWork >= max 1 n);
      let info = geqrf m n aa dispa lda tau (rawArray1 lWork) lWork in
      assert (info = 0);
      mark_dirty a aa;
      if m > n then
	(QofQR (tau, a, 0), RofQR (aaa, sr, lr - (m - n), sc, lc))
      else
	(QofQR (tau, a, 0), RofQR (aaa, sr, lr, sc, lc))
    else if m > n then
      (Product [], RofQR (aaa, sr, lr - (m - n), sc, lc))
    else
      (Product [], RofQR (aaa, sr, lr, sc, lc))

  (* a is modified in place *)
  let rq' (Dense (aaa, sr, lr, sc, lc) as a) =
    let (aa, dispa, lda) = getArrayParams a
    and m = noOfRows a
    and n = noOfCols a in
    if min m n > 0 then
      let tau = rawArray1 (min m n)
      and lWork = -1
      and work = rawArray1 1 in
      let info = gerqf m n aa dispa lda tau work lWork in
      assert (info = 0);
      let lWork = toInt work.{1} in
      assert (lWork >= max 1 n);
      let info = gerqf m n aa dispa lda tau (rawArray1 lWork) lWork in
      assert (info = 0);
      mark_dirty a aa;
      if m <= n then
	(RofRQ (aaa, sr, lr, sc+(n-m), lc), QofRQ (tau, a, 0))
      else
	(RofRQ (aaa, sr, lr, sc, lc), QofRQ (tau, a, 0))
    else if m <= n then
      (RofRQ (aaa, sr, lr, sc+(n-m), lc), Product [])
    else
      (RofRQ (aaa, sr, lr, sc, lc), Product [])

(* Switch from PLU to standard permutation form. *)
  let toStdPermute ipiv =
    let n = Array1.dim ipiv in
    let p = Array1.create int fortran_layout n in
    for i = 1 to n do
      p.{i} <- i
    done;
    for i = 1 to n do
      let pi = p.{i}
      and pj = p.{ipiv.{i}} in
      p.{i} <- pj;
      p.{ipiv.{i}} <- pi
    done;
    p
      
(* Inverse permutation *)
  let invPermute p =
    let n = Array1.dim p in
    let pi = Array1.create int fortran_layout n in
    for i = 1 to n do
      pi.{p.{i}} <- i
    done;
    pi

(* a is modified in place *)
  let plu' (Dense (aaa, sr, lr, sc, lc) as a) =
    let (aa, dispa, lda) = getArrayParams a
    and m = noOfRows a
    and n = noOfCols a in
    if min m n > 0 then
      let ipiv = Array1.create int fortran_layout m in
      let info = getrf m n aa dispa lda ipiv in
      assert (info = 0);
      mark_dirty a aa;
      (Permutation (invPermute (toStdPermute ipiv)),
       LofPLU (aaa, sr, lr, sc, lc),
       UofPLU (aaa, sr, lr, sc, lc))
    else
      (Product [], LofPLU (aaa, sr, lr, sc, lc), UofPLU (aaa, sr, lr, sc, lc))
	
(* {'L','R'}, {'N','T'} from blop_c.c *)
  let orm_left = 0
  and orm_right = 1
  and orm_noTrans = 0
  and orm_trans = 1

  let rec ( @* ) t b =
    if min (noOfRows b) (noOfCols b) > 0 then
      match t with
	QofLQ (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfCols a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lWork = - 1
	  and work = rawArray1 1 in
	  let info = unmlq orm_left orm_noTrans m n
	      (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	      ldb work lWork in
	  assert (info = 0);
	  let lWork = toInt work.{1} in
	  assert (lWork >= max 1 n);
	  let info = unmlq orm_left orm_noTrans m n
	      (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	      ldb (rawArray1 lWork) lWork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQL (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfRows a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and ma = noOfRows a
	  and na = noOfCols a
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let dispa =
	    if ma >= na then dispa else dispa + lda * (na - ma) in
	  let info = unmql orm_left orm_noTrans m n (min ma na) aa
	    dispa lda tau bb dispb ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmql orm_left orm_noTrans m n (min ma na) aa
	    dispa lda tau bb dispb ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQR (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfRows a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and ma = noOfRows a
	  and na = noOfCols a
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmqr orm_left orm_noTrans m n (min ma na) aa
	    dispa lda tau bb dispb ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmqr orm_left orm_noTrans m n (min ma na) aa
	    dispa lda tau bb dispb ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofRQ (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfCols a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let ma = noOfRows a
	  and na = noOfCols a in
	  let dispa =
	    if na >= ma then dispa else dispa + ma - na in
	  let info = unmrq orm_left orm_noTrans m n (min ma na) aa
	    dispa lda tau bb dispb ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmrq orm_left orm_noTrans m n (min ma na) aa
	    dispa lda tau bb dispb ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | Permutation ms ->
	  (* Code from Lapack _lapmt *)
	  let m = noOfRows b  in
	  assert (Array1.dim ms = m);
	  let (bb, dispb, ldb) = getArrayParams b
	  and n = noOfCols b in
	  for i = 1 to m do
	    ms.{i} <- -ms.{i}
	  done;
	  for i = 1 to m do
	    if ms.{i} < 0 then
	      let j = ref i in
	      ms.{!j} <- -ms.{!j};
	      let jin = ref ms.{!j} in
	      while ms.{!jin} < 0 do
		ignore (cblas_swap n bb (dispb+ !j-1) ldb bb
		  (dispb+ !jin-1) ldb);
		ms.{!jin} <- -ms.{!jin};
		j := !jin;
		jin := ms.{!jin}
	      done
	  done;
	  mark_dirty b bb
      | Product ts ->
	  let rec mp ts =
	    match ts with
	    | [] -> ()
	    | t :: ts -> t @* b; mp ts in
	  mp (List.rev ts)
	    
  let rec ( @~* ) t b =
    if min (noOfRows b) (noOfCols b) > 0 then
      match t with
	QofLQ (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfCols a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmlq orm_left orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmlq orm_left orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQL (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfRows a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and ma = noOfRows a
	  and na = noOfCols a
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let dispa =
	    if ma >= na then dispa else dispa + lda * (na - ma) in
	  let info = unmql orm_left orm_trans m n (min ma na) aa
	    dispa lda tau bb dispb ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmql orm_left orm_trans m n (min ma na) aa
	    dispa lda tau bb dispb ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQR (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfRows a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and ma = noOfRows a
	  and na = noOfCols a
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmqr orm_left orm_trans m n (min ma na) aa
	    dispa lda tau bb dispb ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmqr orm_left orm_trans m n (min ma na) aa
	    dispa lda tau bb dispb ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofRQ (tau, a, disp) ->
	  let b' = b $ (1 + disp <-> disp + noOfCols a, All) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let ma = noOfRows a
	  and na = noOfCols a in
	  let dispa =
	    if na >= ma then dispa else dispa + ma - na in
	  let info = unmrq orm_left orm_trans m n (min ma na) aa
	    dispa lda tau bb dispb ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 n);
	  let info = unmrq orm_left orm_trans m n (min ma na) aa
	    dispa lda tau bb dispb ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | Permutation ms ->
	  (* Code from Lapack _lapmt *)
	  let m = noOfRows b  in
	  assert (Array1.dim ms = m);
	  let (bb, dispb, ldb) = getArrayParams b
	  and n = noOfCols b in
	  for i = 1 to m do
	    ms.{i} <- -ms.{i}
	  done;
	  for i = 1 to m do
	    if ms.{i} < 0 then begin
	      ms.{i} <- -ms.{i};
	      let j = ref ms.{i} in
	      while !j <> i do
		ignore (cblas_swap n bb (dispb+ i-1) ldb bb
		  (dispb+ !j-1) ldb);
		ms.{!j} <- -ms.{!j};
		j := ms.{!j}
	      done
	    end
	  done;
	  mark_dirty b bb
      | Product ts ->
	  let rec mp ts =
	    match ts with
	    | [] -> ()
	    | t :: ts -> t @~* b; mp ts in
	  mp ts

  let rec ( *@ ) b t =
    if min (noOfRows b) (noOfCols b) > 0 then
      match t with
	QofLQ (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfCols a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmlq orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmlq orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQL (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfRows a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let ma = noOfRows a
	  and na = noOfCols a in
	  let dispa =
	    if ma >= na then dispa else dispa + lda * (na - ma) in
	  let info = unmql orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmql orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQR (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfRows a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmqr orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmqr orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofRQ (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfCols a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let ma = noOfRows a
	  and na = noOfCols a in
	  let dispa =
	    if na >= ma then dispa else dispa + ma - na in
	  let info = unmrq orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmrq orm_right orm_noTrans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | Permutation ns ->
	  (* Code from Lapack _lapmt *)
	  let n = noOfCols b  in
	  assert (Array1.dim ns = n);
	  let (bb, dispb, ldb) = getArrayParams b in
	  for i = 1 to n do
	    ns.{i} <- -ns.{i}
	  done;
	  for i = 1 to n do
	    if ns.{i} < 0 then begin
	      ns.{i} <- -ns.{i};
	      let j = ref ns.{i} in
	      while !j <> i do
		ignore (cblas_swap n bb (dispb+ldb*(i-1)) 1 bb
		  (dispb+ldb*(!j-1)) 1);
		ns.{!j} <- -ns.{!j};
		j := ns.{!j}
	      done
	    end
	  done;
	  mark_dirty b bb
      | Product ts ->
	  let rec mp ts =
	    match ts with
	    | [] -> ()
	    | t :: ts -> b *@ t; mp ts in
	  mp ts
  
  let rec ( *~@ ) b t =
    if min (noOfRows b) (noOfCols b) > 0 then
      match t with
	QofLQ (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfCols a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmlq orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmlq orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQL (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfRows a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmql orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let ma = noOfRows a
	  and na = noOfCols a in
	  let dispa =
	    if ma >= na then dispa else dispa + lda * (na - ma) in
	  let info = unmql orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofQR (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfRows a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let info = unmqr orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmqr orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | QofRQ (tau, a, disp) ->
	  let b' = b $ (All, 1 + disp <-> disp + noOfCols a) in
	  let m = noOfRows b'
	  and n = noOfCols b'
	  and (aa, dispa, lda) = getArrayParams a
	  and (bb, dispb, ldb) = getArrayParams b'
	  and lwork = -1
	  and work = rawArray1 1 in
	  let ma = noOfRows a
	  and na = noOfCols a in
	  let dispa =
	    if na >= ma then dispa else dispa + ma - na in
	  let info = unmrq orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb work lwork in
	  assert (info = 0);
	  let lwork = toInt work.{1} in
	  assert (lwork >= max 1 m);
	  let info = unmrq orm_right orm_trans m n
	    (min (noOfRows a) (noOfCols a)) aa dispa lda tau bb dispb
	    ldb (rawArray1 lwork) lwork in
	  assert (info = 0);
	  mark_dirty b bb
      | Permutation ns ->
	  (* Code from Lapack _lapmt *)
	  let n = noOfCols b  in
	  assert (Array1.dim ns = n);
	  let (bb, dispb, ldb) = getArrayParams b
	  and m = noOfRows b in
	  for i = 1 to n do
	    ns.{i} <- -ns.{i}
	  done;
	  for i = 1 to n do
	    if ns.{i} < 0 then
	      let j = ref i in
	      ns.{!j} <- -ns.{!j};
	      let jin = ref ns.{!j} in
	      while ns.{!jin} < 0 do
		ignore (cblas_swap m bb (dispb+ldb*(!j-1)) 1 bb
		  (dispb+ldb*(!jin-1)) 1);
		ns.{!jin} <- -ns.{!jin};
		j := !jin;
		jin := ns.{!jin}
	      done
	  done;
	  mark_dirty b bb
      | Product ts ->
	  let rec mp ts =
	    match ts with
	    | [] -> ()
	    | t :: ts -> b *~@ t; mp ts in
	  mp (List.rev ts)

(* {'L','U'}, {'N','T'/'C'}, {'N','U'} from blop_c.c *)
  let trtrs_lower = 0
  and trtrs_upper = 1
  and trtrs_noTrans = 0
  and trtrs_trans = 1
  and trtrs_nonUnitDiag = 0
  and trtrs_unitDiag = 1

  (* b <- l \ b where l is triangular *)
  let subs' l b =
    let m = noOfRows b
    and n = noOfCols b
    and ml = noOfRows_f l
    and nl = noOfRows_f l
    and (bb, dispb, ldb) = getArrayParams b in
    if min m n > 0 then begin
      (match l with
	LofQL (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  assert (m = min ml nl);
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  if ml >= nl then
	    let info = trtrs trtrs_lower trtrs_noTrans
		trtrs_nonUnitDiag nl n ll (displ+ml-nl) ldl bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting fat L."
      | RofQR (rrr, srr, lrr, scr, lcr) ->
	  let rr = sup2big rrr in
	  assert (m = min ml nl);
	  let ldr = Array2.dim1 rr in
	  let dispr = (scr-1)*ldr+(srr-1) in
	  if ml >= nl then
	    let info = trtrs trtrs_upper trtrs_noTrans
		trtrs_nonUnitDiag nl n rr dispr ldr bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting fat R."
      | RofRQ (rrr, srr, lrr, scr, lcr) ->
	  let rr = sup2big rrr in
	  assert (m = min ml nl);
	  let ldr = Array2.dim1 rr in
	  let dispr = (scr-1)*ldr+(srr-1) in
	  if nl >= ml then
	    let info = trtrs trtrs_upper trtrs_noTrans
		trtrs_nonUnitDiag ml n rr dispr ldr bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting fat R."
      | LofLQ (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  assert (m = min ml nl);
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  if nl >= ml then
	    let info = trtrs trtrs_lower trtrs_noTrans
		trtrs_nonUnitDiag ml n ll displ ldl bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting skinny L."
      | LofPLU (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  assert (m = ml);
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  let info = trtrs trtrs_lower trtrs_noTrans
	      trtrs_unitDiag (min ml nl) n ll displ ldl bb dispb ldb in
	  assert (info = 0);
	  if ml > nl then
	    let lm = Dense (lll, srl, lrl, scl, lcl) in
	    (b $ (nl+1 <-> ml, All)) <-|
	    ((lm $ (nl+1 <-> ml, 1 <-> nl)) |*--| (b $ (1 <-> nl, All)))
	  else
	    ()
      | UofPLU (uuu, sru, lru, scu, lcu) ->
	  let uu = sup2big uuu in
	  assert (nl = m);
	  let ldu = Array2.dim1 uu in
	  let dispu = (scu-1)*ldu+(sru-1) in
	  let info = trtrs trtrs_upper trtrs_noTrans trtrs_nonUnitDiag
	      (min ml nl) n uu dispu ldu bb dispb ldb in
	  assert (info = 0);
	  if ml < nl then
	    const zero (b $ (ml+1 <-> m, All))
	  else
	    ());
      mark_dirty b bb
    end
	

  (* b <- l^T \ b where l is triangular *)
  let subsT' l b =
    let m = noOfRows b
    and n = noOfCols b
    and ml = noOfRows_f l
    and nl = noOfRows_f l
    and (bb, dispb, ldb) = getArrayParams b in
    assert (m = min ml nl);
    if min m n > 0 then begin
      (match l with
	LofQL (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  if ml >= nl then
	    let info = trtrs trtrs_lower trtrs_trans
		trtrs_nonUnitDiag nl n ll (displ+ml-nl) ldl bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting fat L."
      | RofQR (rrr, srr, lrr, scr, lcr) ->
	  let rr = sup2big rrr in
	  assert (m = min ml nl);
	  let ldr = Array2.dim1 rr in
	  let dispr = (scr-1)*ldr+(srr-1) in
	  if ml >= nl then
	    let info = trtrs trtrs_upper trtrs_trans
		trtrs_nonUnitDiag nl n rr dispr ldr bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting fat R."
      | RofRQ (rrr, srr, lrr, scr, lcr) ->
	  let rr = sup2big rrr in
	  assert (m = min ml nl);
	  let ldr = Array2.dim1 rr in
	  let dispr = (scr-1)*ldr+(srr-1) in
	  if nl >= ml then
	    let info = trtrs trtrs_upper trtrs_trans
		trtrs_nonUnitDiag ml n rr dispr ldr bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting fat R."
      | LofLQ (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  if nl >= ml then
	    let info = trtrs trtrs_lower trtrs_trans
		trtrs_nonUnitDiag ml n ll displ ldl bb dispb ldb in
	    assert (info = 0)
	  else failwith "Inverting skinny L."
      | LofPLU (lll, srl, lrl, scl, lcl) ->
	  let ll = sup2big lll in
	  assert (m = ml);
	  let ldl = Array2.dim1 ll in
	  let displ = (scl-1)*ldl+(srl-1) in
	  if ml > nl then
	    let lm = Dense (lll, srl, lrl, scl, lcl) in
	    (b $ (1 <-> nl, All)) <-|
	    ((lm $ (nl+1 <-> ml, 1 <-> nl)) |~*--| (b $ (nl+1 <-> m, All)));
	      let info = trtrs trtrs_lower trtrs_trans trtrs_unitDiag
		  nl n ll displ ldl bb dispb ldb in
	      assert (info = 0)
	  else
	    let info = trtrs trtrs_lower trtrs_trans trtrs_unitDiag
		ml n ll displ ldl bb dispb ldb in
	    assert (info = 0)
      | UofPLU (uuu, sru, lru, scu, lc) ->
	  let uu = sup2big uuu in
	  assert (ml >= nl && m = nl);
	  let ldu = Array2.dim1 uu in
	  let dispu = (scu-1)*ldu+(sru-1) in
	  let info = trtrs trtrs_upper trtrs_trans trtrs_nonUnitDiag
	      (min ml nl) n uu dispu ldu bb dispb ldb in
	  assert (info = 0));
      mark_dirty b bb
    end

  let pluSolve' a b =
    let (p, l, u) = plu' a in
    p @~* b;
    subs' l b;
    subs' u b

(* {'N','T'} from blop_c.c *)
  let gels_noTrans = 0
  and gels_trans = 1

  (* a |/| b <=> b <- a \ b
     If a is fat b must be large enough to contain the solution and the first
     part of b must contain the actual b. If a is skinny the first part of b
     will contain the solution. a is destroyed.
  *)
  let ( |/| ) a b =
    let m = noOfRows a
    and n = noOfCols a
    and nrhs = noOfCols b
    and (aa, dispa, lda) = getArrayParams a
    and (bb, dispb, ldb) = getArrayParams b in
    assert (noOfRows b = max m n);
    if min (min m n) nrhs > 0 then
      let lwork = -1
      and work = rawArray1 1 in
      let info = gels gels_noTrans m n nrhs aa dispa lda bb dispb
	ldb work lwork in
      assert (info = 0);
      let lwork = toInt work.{1} in
      assert (lwork >= max 1 (min m n + max (min m n) nrhs));
      let info = gels gels_noTrans m n nrhs aa dispa lda bb dispb
	ldb (rawArray1 lwork) lwork in
      assert (info = 0);
      mark_dirty a aa;
      mark_dirty b bb

  (* a |~/| b <=> b <- a' \ b
     If a' is fat b must be large enough to contain the solution and the first
     part of b must contain the actual b. If a' is skinny the first part of b
     will contain the solution. a is destroyed.
  *)
  let ( |~/| ) (Dense (_,_,_,_,_) as a) (Dense (_,_,_,_,_) as b) =
    let m = noOfRows a
    and n = noOfCols a
    and nrhs = noOfCols b
    and (aa, dispa, lda) = getArrayParams a
    and (bb, dispb, ldb) = getArrayParams b in
    assert (noOfRows b = max m n);
    if min (min m n) nrhs > 0 then
      let lwork = -1
      and work = rawArray1 1 in
      let info = gels gels_trans m n nrhs aa dispa lda bb dispb
	ldb work lwork in
      assert (info = 0);
      let lwork = toInt work.{1} in
      assert (lwork >= max 1 (min m n + max (min m n) nrhs));
      let info = gels gels_trans m n nrhs aa dispa lda bb dispb ldb
	(rawArray1 lwork) lwork in
      assert (info = 0);
      mark_dirty a aa;
      mark_dirty b bb

  (* a /$ b <=> a \ b *)
  let ( /$ ) a b =
    assert (noOfRows a = noOfRows b);
    let ma = noOfRows a
    and na = noOfCols a
    and nrhs = noOfCols b
    and aC = copy ~ooc:(is_ooc a) a in
    let bC = rawMatrix ~ooc:(is_ooc b) (max ma na) nrhs in
      (b $=> (1 <-> ma, All)) bC;
      aC |/| bC;
      bC $ (1 <-> na, All)

  (* a /~$ b <=> a' \ b *)
  let ( /~$ ) a b =
    assert (noOfCols a = noOfRows b);
    let ma = noOfRows a
    and na = noOfCols a
    and nrhs = noOfCols b
    and aC = copy ~ooc:(is_ooc a) a in
    let bC = rawMatrix ~ooc:(is_ooc b) (max ma na) nrhs in
      (b $=> (1 <-> na, All)) bC;
      aC |~/| bC;
      bC $ (1 <-> ma, All)

(* {'A', 'S', 'O', 'N'} form blop_c.c *)
  let svd_all = 0
  and svd_skinny = 1
  and svd_overwrite = 2
  and svd_noComp = 3

  (* svd a => (u, s, vt) such that a = u * diag(s) * vt *)
  let svd' ?(ooc=false) a =
    let m = noOfRows a
    and n = noOfCols a
    and (aa, dispa, lda) = getArrayParams a in
    let k = min m n in
    let u = rawMatrix ~ooc m k
    and vt = rawMatrix ~ooc k n
    and s = rawArray1_real k in
    if k > 0 then
      let (uu, dispu, ldu) = getArrayParams u
      and (vvt, dispvt, ldvt) = getArrayParams vt
      and lwork = -1
      and work = rawArray1 1
      and rwork = rawArray1_real (5 * min m n * min m n + 5 * min m n)
	in
      let info = gesvd svd_skinny svd_skinny m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt work lwork rwork in
      assert (info = 0);
      let lwork = toInt work.{1} in
      (* assert (lwork >= max (3 * min m n + max m n) (5 * min m n)); *)
      assert (lwork >= 2 * min m n + max m n); (* complex *)
      let info = gesvd svd_skinny svd_skinny m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt (rawArray1 lwork) lwork rwork in
      if info <> 0 then
	Printf.printf "svd' info=%i" info;
      assert (info = 0);
      mark_dirty u uu;
      mark_dirty vt vvt;
      (u, s, vt)
    else
      (u, s, vt)

  let svdL' ?(ooc=false) a =
    let m = noOfRows a
    and n = noOfCols a
    and (aa, dispa, lda) = getArrayParams a in
    let k = min m n in
    let u = rawMatrix ~ooc m k
    and vt = rawMatrix 1 1
    and s = rawArray1_real k in
    if k > 0 then
      let (uu, dispu, ldu) = getArrayParams u
      and (vvt, dispvt, ldvt) = getArrayParams vt
      and lwork = -1
      and work = rawArray1 1
      and rwork = rawArray1_real (5 * min m n * min m n + 5 * min m n)
      in
      let info = gesvd svd_skinny svd_noComp m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt work lwork rwork in
      assert (info = 0);
      let lwork = toInt work.{1} in
(*      assert (lwork >= 3 * min m n * min m n +
      max (max m n) (4 * min m n * min m n + 4 * min m n)); *)
      let info = gesvd svd_skinny svd_noComp m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt (rawArray1 lwork) lwork rwork in
      assert (info = 0);
      mark_dirty u uu;
      (u, s)
    else
      (u, s)

  let svdR' ?(ooc=false) a =
    let m = noOfRows a
    and n = noOfCols a
    and (aa, dispa, lda) = getArrayParams a in
    let k = min m n in
    let u = rawMatrix 1 1
    and vt = rawMatrix ~ooc k n
    and s = rawArray1_real k in
    if k > 0 then
      let (uu, dispu, ldu) = getArrayParams u
      and (vvt, dispvt, ldvt) = getArrayParams vt
      and lwork = -1
      and work = rawArray1 1
      and rwork = rawArray1_real (5 * min m n * min m n + 5 * min m n)
	in
      let info = gesvd svd_noComp svd_skinny m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt work lwork rwork in
      assert (info = 0);
      let lwork = toInt work.{1} in
      assert (lwork >= 3 * min m n * min m n +
		max (max m n) (4 * min m n * min m n + 4 * min m n));
      let info = gesvd svd_noComp svd_skinny m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt (rawArray1 lwork) lwork rwork in
      assert (info = 0);
      mark_dirty vt vvt;
      (s, vt)
    else
      (s, vt)

  let singValues' a =
    let m = noOfRows a
    and n = noOfCols a
    and (aa, dispa, lda) = getArrayParams a in
    let k = min m n in
    let u = rawMatrix 1 1
    and vt = rawMatrix 1 1
    and s = rawArray1_real k in
    if k > 0 then
      let (uu, dispu, ldu) = getArrayParams u
      and (vvt, dispvt, ldvt) = getArrayParams vt
      and lwork = -1
      and work = rawArray1 1
      and rwork = rawArray1_real (5 * min m n * min m n + 5 * min m n)
	in
      let info = gesvd svd_noComp svd_noComp m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt work lwork rwork in
      assert (info = 0);
      let lwork = toInt work.{1} in
(*      assert (lwork >= 3 * min m n * min m n +
		max (max m n) (4 * min m n * min m n + 4 * min m n)); *)
      let info = gesvd svd_noComp svd_noComp m n aa dispa lda s
	uu dispu ldu vvt dispvt ldvt (rawArray1 lwork) lwork rwork in
      assert (info = 0);
      s
    else
      s

  let cond a =
    let s = singValues' (copy a) in
    s.{1} /. s.{min (noOfRows a) (noOfCols a)}

  let rowScale' s (Dense (aaa, sr, lr, sc, lc) as a) =
    assert (Array1.dim s = noOfRows a);
    let m = noOfRows a
    and n = noOfCols a in
    let aa = sup2big aaa in
    if n > 0 then begin
      for j = 1 to n do
	for i = 1 to m do
	  aa.{sr+i-1, sc+j-1} <- muln s.{i} aa.{sr+i-1,sc+j-1}
	done
      done;
      mark_dirty a aa
    end
    

  let colScale' (Dense (aaa, sr, lr, sc, lc) as a) s =
    assert (Array1.dim s = noOfCols a);
    let m = noOfRows a
    and n = noOfCols a in
    let aa = sup2big aaa in
    if m > 0 then begin
      for j = 1 to n do
	for i = 1 to m do
	  aa.{sr+i-1, sc+j-1} <- muln aa.{sr+i-1,sc+j-1} s.{j}
	done
      done;
      mark_dirty a aa
    end

  let scale' s (Dense (aaa, sr, lr, sc, lc) as a) =
    let aa = sup2big aaa in
    for j = sc to lc do
      for i = sr to lr do
	aa.{i,j} <- muln s aa.{i,j}
      done
    done;
    mark_dirty a aa

  let rowScale_real' s (Dense (aaa, sr, lr, sc, lc) as a) =
    assert (Array1.dim s = noOfRows a);
    let m = noOfRows a
    and n = noOfCols a in
    let aa = sup2big aaa in
    if n > 0 then begin
      for j = 1 to n do
	for i = 1 to m do
	  aa.{sr+i-1, sc+j-1} <- mulr s.{i} aa.{sr+i-1,sc+j-1}
	done
      done;
      mark_dirty a aa
    end

  let colScale_real' (Dense (aaa, sr, lr, sc, lc) as a) s =
    assert (Array1.dim s = noOfCols a);
    let m = noOfRows a
    and n = noOfCols a in
    let aa = sup2big aaa in
    if m > 0 then begin
      for j = 1 to n do
	for i = 1 to m do
	  aa.{sr+i-1, sc+j-1} <- mulr s.{j} aa.{sr+i-1,sc+j-1}
	done
      done;
      mark_dirty a aa
    end

  let scale_real' s (Dense (aaa, sr, lr, sc, lc) as a) =
    let aa = sup2big aaa in
    for j = sc to lc do
      for i = sr to lr do
	aa.{i,j} <- mulr s aa.{i,j}
      done
    done;
    mark_dirty a aa

  let conjg' (Dense (aaa, sr, lr, sc, lc) as a) =
    let aa = sup2big aaa in
    for j = sc to lc do
      for i = sr to lr do
	aa.{i,j} <- conjugate aa.{i,j}
      done
    done;
    mark_dirty a aa

  let isLowerTriangular (Dense (l, sr, lr, sc, lc)) =
    LofQL (l, sr, lr, sc, lc)

  let isUpperTriangular (Dense (r, sr, lr, sc, lc)) =
    RofQR (r, sr, lr, sc, lc)


(* Multilinear algebra *)

  let tensor ?(ooc=false) a b =
    let ma = noOfRows a
    and na = noOfCols a
    and mb = noOfRows b
    and nb = noOfCols b in
    let c = rawMatrix ~ooc (ma * mb) (na * nb) in
    for j = 1 to na do
      for i = 1 to ma do
	let cij = c $ ((i-1) * mb + 1 <-> i * mb, (j-1) * nb + 1 <-> j * nb) in
	cij <-$ b;
	scale' (a $@ (i,j)) cij
      done
    done;
    c
      
  let vecC' a b =
    let m = noOfRows a
    and n = noOfCols a in
    assert (m * n = noOfRows b && noOfCols b = 1);
    for j = 1 to n do
      (b $ ((j-1)*m+1<->j*m,1<->1)) <-$ (a $ (All, j<->j))
    done

  let unvecC' a b =
    let m = noOfRows b
    and n = noOfCols b in
    assert (m * n = noOfRows a && noOfCols a = 1);
    for j = 1 to n do
      (b $ (All, j<->j)) <-$ (a $ ((j-1)*m+1<->j*m,1<->1))
    done

  let vecR' a b =
    let m = noOfRows a
    and n = noOfCols a in
    assert (noOfRows b = m * n && noOfCols b = 1);
    for i = 1 to m do
      transp' (a $ (i<->i,All)) (b $ ((i-1)*n+1<->i*n, 1<->1))
    done;
    conjg' b

  let unvecR' a b =
    let m = noOfRows b
    and n = noOfCols b in
    assert (noOfRows a = m * n && noOfCols a = 1);
    for i = 1 to m do
      transp' (b $ (i<->i,All)) (a $ ((i-1)*n+1<->i*n, 1<->1))
    done;
    conjg' b

end (* module Real_Bloper *)


(* Double precision real blop *)
module DBlop = Bloper (Double_Real_NLA)

(* Single precision real blop *)
module SBlop = Bloper (Single_Real_NLA)

(* Double Complex Blop *)
module ZBlop = Bloper (Double_Complex_NLA)

(* Single Complex Blop *)
module CBlop = Bloper (Single_Complex_NLA)
