(*
  File: testChub.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 12.23.2003
  Email: shiv@ece.ucsb.edu
  Purpose: Chebyshev stuff

   $Id$
*)

open Bigarray;;
open Utils;;
open Nla;;
open Hss;;
open Chub;;
open NestedDissection;;


module Tester (Blop : BLOP)
    (Hss : HSS with type matrix = Blop.matrix and type t = Blop.t)
    (Chub : CHUB with type matrix = Blop.matrix and type t = Blop.t
    and type hss = Hss.hss) =
  struct

    open Blop
    open Hss
    open Chub

    let test1 () =
      let p = 10 in
      let z = chebZeros p in
      let zp = Array.init p (fun i -> cheb p z.(i)) in
      let acc = ref 0.0 in
      for i = 0 to p-1 do
	if abs_float zp.(i) > !acc then acc := abs_float zp.(i)
      done;
      !acc <= 4.0 *. float p *. machine_epsilon

    let test2 () =
      let pmax = 5
      and ok = ref true in
      for p = 1 to pmax do
	let f x _ = fromFloat (x ** float p)
	and fp x _ = fromFloat (float p *. x ** float (p-1))
	and n = pmax * 2 in
	let fx = discretize (chebZeros n) 1 f
	and fpx = discretize (chebZeros (n-1)) 1 fp
	and fp = rawMatrix (n-1) 1 in
	toCheb' fx;
	spectralDerivative fx fp;
	let fxp = fromCheb fp in
	ok := !ok && normInf (fxp -$ fpx) <= 8.0 *.
	  float pmax *. machine_epsilon *. norm1 fpx
      done;
      !ok

    let test3 () =
      let ns = [| 1;2;3;4;5 |]
      and ok = ref true in
      for i = 0 to Array.length ns - 1 do
	let n = ns.(i) in
	let fx = randomMatrix n 1
	and fp = rawMatrix n 1
	and fxi = rawMatrix (n+1) 1 in
	spectralIntegral fx fxi;
	spectralDerivative fxi fp;
	ok := !ok && normInf (fp -$ fx) <= machine_epsilon *. norm1 fx
      done;
      !ok
            
    let test4 () =
      let pmax = 5
      and ok = ref true in
      for p = 1 to pmax do
	let f x _ = fromFloat (x ** float p)
	and n = pmax * 2 in
	let fx = discretize (chebZeros n) 1 f in
	let fxc = Blop.copy fx
	and fp = rawMatrix n 1
	and fxi = rawMatrix (n+1) 1 in
	toCheb' fx;
	spectralIntegral fx fxi;
	spectralDerivative fxi fp;
	let fxp = fromCheb fp in
	ok := !ok && normInf (fxp -$ fxc) <= 8.0 *.
	  float pmax *. machine_epsilon *. norm1 fxc
      done;
      !ok

    let test5 () =
      let pmax = 5
      and f x j = fromFloat (x ** float (j+1))
      and fp x j = fromFloat (float (j+1) *. x ** float j) in
      let n = pmax * 2 in
      let fx = discretize (chebZeros n) pmax f
      and fpx = discretize (chebZeros n) pmax fp
      and fp = rawMatrix n pmax in
      psDerivative' fx fp;
      normInf (fp -$ fpx) <= 6.0 *. float pmax *. machine_epsilon *. norm1 fpx

    let test6 () =
      let pmax = 5
      and f x j = fromFloat (x ** float (j+1))
      and fp x j = fromFloat (float (j+1) *. x ** float j) in
      let n =  2 * pmax * (int_of_float (-. (log machine_epsilon)))
      and tol = sqrt machine_epsilon in
      let p = min (int_of_float (-. (log tol))) n
      and fx = discretize (chebZeros n) pmax f
      and fpx = discretize (chebZeros n) pmax fp
      and fp = rawMatrix n pmax in
      let hssps = hssPSderivative tol p n in
      fastMul hssps fx fp;
      normInf (fp -$ fpx) <= float n *. float pmax *. tol
	*. norm1 fpx

    let test7 () =
      let pmax = 5
      and ok = ref true
      and f x j = fromFloat (x ** float (j+1)) in
      let n = pmax * 2 in
      let fx = discretize (chebZeros n) pmax f in
      for k = n-1 downto 1 do
	let fxk = rawMatrix k pmax in
	projectTo' k (Blop.copy fx) fxk;
	let fkx = discretize (chebZeros k) pmax f in
	ok := !ok && normInf (fxk -$ fkx) <= machine_epsilon *.
	    float pmax *. norm1 fx
      done;
      !ok

    (* Testing tpBVP_ps solver. u'(x) + lambda u(x) = 0; u(-1) = I *)
    let test8 () =
      let n = 3
      and nrhs = 3
      and p = 100
      and lambda = 1E1 in
      let tol = sqrt (float p) *. float p *. sqrt machine_epsilon in
      let a1 x = eye n n
      and a0 x = let d = eye n n in scale_real' lambda d; d
      and f  x = zeros n nrhs
      and bcl  = eye n n
      and bcr  = zeros n n
      and z    = eye n nrhs
      and u x  =
	let ux = fromFloat (exp (-. lambda *. (1.0 +. x)))
	and um = eye n nrhs in
	for i = 1 to min n nrhs do
	  (ux => (i,i)) um
	done;
	um in
      let uh = tpBVP_ps_1 tol p a1 a0 f bcl bcr z
      and ua = (* exact solution *)
	let uam = rawMatrix (n * p) nrhs
	and xs = chebZeros p in
	matrixNx1 (Array.make p n)
	  (Array.init p (fun i -> iD (u xs.(i)))) uam;
	uam in
      norm1 (ua -$ uh) <= tol *. norm1 ua

    (* Testing twoD_elliptic_ps solver.
     axx uxx + axy uxy + ayy uyy + ax ux + ay uy + a u = f in [-1,1]x[-1,1] and
     bc u + bcx ux + bcy uy = z on boundary. *)
    let test9 () =
      let nrhs = 1
      and aspectRatio = 1.00 in
      let ok = ref true
      and eye1 = eye 1 1
      and eyeA2 = consts (fromFloat (aspectRatio ** (-2.0))) 1 1
      and zero1 = zeros 1 1
      and zeroNrhs = zeros 1 nrhs in
      let a   x y = zero1
      and ax  x y = zero1
      and ay  x y = zero1
      and axx x y = eye1
      and axy x y = zero1
      and ayy x y = eyeA2
      and f   x y = zeroNrhs
      and bc  x y = eye1
      and bcx x y = zero1
      and bcy x y = zero1 in
	(* onesNrhs in *)
      for p = 200 to 200 do
	let ooc = true in
	let px = p
	and py = int_of_float (aspectRatio *. float p) in
	let zp l x y =
	  let zt = {Complex.re=x; Complex.im=y} in
	  let zpt = Complex.pow zt {Complex.re=float l; Complex.im=0.0} in
	  fromFloat zpt.Complex.re (* im *) in
	let z x y = matrix [| (Array.init nrhs (fun j -> zp j x y)) |] in
	let tol = (float (px * py) ** 1.0) *. sqrt machine_epsilon in
	Printf.printf "tol = %g \n" tol;
	let um = twoD_elliptic_ps ~ooc tol px py axx axy ayy ax ay a f
	    bc bcx bcy z
	in
	let ua =
	  let xs = chebZeros px
	  and ys = chebZeros py in
	  let ufn i j =
	    let jx = (i - 1) / py in
	    let iy = i - 1 - jx * py in
	    (z xs.(jx) ys.(iy)) $@ (1,j) in
	  fun2mat ~ooc ufn (px * py) nrhs in
	Printf.printf "norm = %g %g\n" (normInf um)
	  (norm1 (um -$ ua) /. norm1 ua);
	ok := !ok && norm1 (um -$ ua) /. norm1 ua <= tol
      done;
      !ok


    let test10 () =
      let n = 15 in
      let nrhs = n-1 in
      let tt = chebVand (chebZeros n) nrhs in
      let si = rawMatrix n nrhs
      and sd = rawMatrix n nrhs in
      fPSint' (Blop.copy tt) si;
      psDerivative' si sd;
      norm1 (tt -$ sd) <= (float (n * n)) *. machine_epsilon *. norm1 tt


    let test11 () =
      for n = 4 to 40 do
	let sif = ones (n+1) n
	and sib = ones (n+1) n in
	fPSint' (eye n n) sif;
	bPSint' (eye n n) sib;
	let pD = rawMatrix (n+1) n in
	projectTo' (n+1) (eye n n) pD;
	let lop = chebVand [| -1.0 |] (n+1)
	and rop = chebVand [|  1.0 |] (n+1) in
	let pc = eye (n+1) (n+1) +$
	  (consts (fromFloat 0.5) (n+1) 1) *$ (lop +$ rop) in
	let pci = pc /$ eye (n+1) (n+1) in
(*	Printf.printf "n = %i cond = %g\n" n (cond (pci *$ (sif +$ sib)));
	Printf.printf "n = %i cond = %g\n" n (cond ((sif +$ sib) *$ pD)); *)
	let pc = pci *$ (sif +$ sib) *$~ pD in
	let s = singValues' pc in
	Printf.printf "n = %i cond = %g\n" n (s.{1} /. s.{n});
      done;
      true


    let test12 () =
      for n = 4 to 40 do
	let sif = ones (n+1) n
	and sib = ones (n+1) n
	and xp1np1 = Array1.create precision_real fortran_layout (n+1)
	and xp1n   = Array1.create precision_real fortran_layout n
	and xm1np1 = Array1.create precision_real fortran_layout (n+1)
	and xm1n   = Array1.create precision_real fortran_layout n
	and znp1 = chebZeros (n+1)
	and zn = chebZeros n in
	for i = 1 to n do
	  xp1n.{i} <- zn.(i-1) +. 1.0;
	  xm1n.{i} <- zn.(i-1) -. 1.0
	done;
	for i = 1 to n+1 do
	  xp1np1.{i} <- znp1.(i-1) +. 1.0;
	  xm1np1.{i} <- znp1.(i-1) -. 1.0
	done;
	fPSint' (eye n n) sif;
	bPSint' (eye n n) sib;
	rowScale_real' xm1np1 sif;
	colScale_real' sif xp1n;
	rowScale_real' xp1np1 sib;
	colScale_real' sib xm1n;
	let pD = rawMatrix (n) (n+1) in
	projectTo' (n) (eye (n+1) (n+1)) pD;
	let lop = chebVand [| -1.0 |] (n+1)
	and rop = chebVand [|  1.0 |] (n+1) in
	let omx = fun2mat (fun i _ -> fromFloat (1.0 -. znp1.(i-1))) (n+1) 1
	and opx = fun2mat (fun i _ -> fromFloat (1.0 +. znp1.(i-1))) (n+1) 1 in
	let r2cor = omx *$ lop +$ opx *$ rop in
	scale_real' 0.5 r2cor;
	let cor = rawMatrix (n+1) (n+2) in
	(cor $ (All,1<->1)) <-$ omx;
	(cor $ (All,2<->n+1)) <-| (sif |+| sib);
	(cor $ (All,n+2<->n+2)) <-$ opx;
(*	Printf.printf "n = %i cond = %g\n" n (cond (sif +$ sib)); *)
(*	Printf.printf "n = %i cond = %g\n" n (cond ((sif +$ sib) *$ pD)); *)
	let pc = cor in
	let s = singValues' pc in
	Printf.printf "n = %i cond = %g\n" n (s.{1} /. s.{n+1});
      done;
      true


    let power x n = fromFloat (x ** float n)
    let intPower x n = fromFloat (x ** float (n + 1) /. float (n + 1))

    let test13 () =
      let ok = ref true in
      for p = 3 to 6 do
	let n = p - 2 in
	let z = chebZeros p in
	toInterval' 0.0 1.0 z;
	let a = discretize z n power
	and b = rawMatrix (p+1) n in
	fPSint' a b;
	scale_real' 0.5 b;
	let z = chebZeros (p+1) in
	toInterval' 0.0 1.0 z;
	let c = discretize z n intPower in
	ok := !ok && norm1 (c -$ b) <= machine_epsilon *. norm1 c *. float p
      done;
      !ok


    let test14 () =
      let p = 6 in
      let n = p - 2 in
      let a = chebQuad 0.0 1.0 p n power in
      let b = fun2mat (fun i j -> intPower 1.0 (j-1)) 1 n in
      norm1 (a -$ b) <= machine_epsilon *. norm1 b *. float p

    let test15 () =
      let ok = ref true in
      for p = 3 to 6 do
	let n = p - 2 in
	let a = chebQuadPW [| 0.0; 0.5; 1.0 |] p n power in
	let b = fun2mat (fun i j -> intPower 1.0 (j-1)) 1 n in
	ok := !ok && norm1 (a -$ b) <= machine_epsilon *. norm1 b *. float p
      done;
      !ok

    let ( +: ) x y = {Complex.re = x; Complex.im = y}

    let test16 () =
      let ok = ref true
      and tol = 1E-6 in
      for p = 3 to 6 do
	let n = p - 2 in
	let a = chebQuadAdaptive tol 0.0 1.0 [| 0.5 +: 0.0 |] p n power in
	let b = fun2mat (fun i j -> intPower 1.0 (j-1)) 1 n in
	ok := !ok && norm1 (a -$ b) <= tol *. norm1 b *. float p
      done;
      !ok

    (* (x - z)^n *)
    let shiftedPower z x n =
      fromFloat ((Complex.pow
		    {Complex.re = x -. z.Complex.re; Complex.im = z.Complex.im}
		    {Complex.re = float n; Complex.im = 0.0}).Complex.re)

    let intShiftedPower z x n =
      fromFloat ((Complex.pow
		    {Complex.re = x -. z.Complex.re; Complex.im = z.Complex.im}
		    {Complex.re = float (n+1); Complex.im = 0.0}).Complex.re /.
		 float (n + 1))

    let test17 () =
      let tol = 1E-6 in
      let alpha = 0.05 in
      let z = 0.0 +: alpha in
      let s = [| z |] in
      let p = 10 in
      let n = p - 2 in
      let a = chebQuadAdaptive tol 0.0 1.0 s p n (shiftedPower z) in
      let b = (discretize [|1.0|] n (intShiftedPower z)) -$
	(discretize [|0.0|] n (intShiftedPower z)) in
      norm1 (a -$ b) <= tol *. norm1 b *. float p

    let test18 () =
      let tol = 1E-6 in
      let alpha = 0.05 in
      let z = 0.0 +: alpha in
      let s = [| z |] in
      let p = 10 in
      let n = p - 2 in
      let a = chebQuadAdaptive tol 0.0 1.0 s p n
	  (fun x n -> shiftedPower z x (-n)) in
      let b = (discretize [|1.0|] n (fun x n -> intShiftedPower z x (-n))) -$
	(discretize [|0.0|] n (fun x n -> intShiftedPower z x (-n))) in
      norm1 (a -$ b) <= tol *. norm1 b *. float p

    let test19 () =
      let y = 0.5
      and p = 10
      and tol = sqrt machine_epsilon in
      let a = chebQuadAdaptive tol 0.0 1.0 [| y+:0.0 |] p 1
	  (fun x _ -> fromFloat (log (abs_float (x -. y)))) in
      let b = matrix
	  [| [| fromFloat (y *. log y +. (1.0 -. y) *. log (1.0 -. y) -. 1.0)
	      |] |] in
      norm1 (a -$ b) <= tol *. norm1 a

    let test20 () =
      let p = 10 in
      let (v11, v21) = translators p in
      let zp = chebZeros p
      and z2p = Array.make (2 * p) 0.0 in
      for i = 0 to p - 1 do
	z2p.(i) <- toInterval (-1.0) 0.0 zp.(i);
	z2p.(i+p) <- toInterval 0.0 1.0 zp.(i)
      done;
      let t1 = chebVand zp p
      and t2 = chebVand z2p p in
      let (t21, t22) = partition2x1 p p t2 in
      norm1 (t1 *$ v11 -$ t21) +. norm1 (t1 *$ v21 -$ t22) <=
      machine_epsilon *. float p *. norm1 t2

    let test21 () =
      let p = 10 in
      let (v12, v22) = innovators p in
      let zp = chebZeros p
      and z2p = Array.make (2 * p) 0.0 in
      for i = 0 to p - 1 do
	z2p.(i) <- toInterval (-1.0) 0.0 zp.(i);
	z2p.(i+p) <- toInterval 0.0 1.0 zp.(i)
      done;
      let t1 = chebVand zp p
      and t2 = chebVand z2p p
      and ti = rawMatrix (2 * p) (2 * p) in
      let (ti1, ti2) = partition1x2 p p ti in
      ti1 <-$ t2;
      matrix2x1 p (t1 |*| v12)
	        p (t1 |*| v22) ti2;
      cond ti <= float p

    (* Two D tests *)
    let cheb2D i j x y = fromFloat (cheb i x *. cheb j y)

    let test22 () =
      let px = 10
      and py = 13 in
      let zy = chebZeros py
      and zx = chebZeros px in
      let ok = ref true in
      for i = 0 to py - 1 do
	for j = 0 to px - 1 do
	  let a = discretize2D zx zy (cheb2D j i) in
	  let ac = Blop.copy a in
	  toCheb2D' ac;
	  let aij = ac $@ (i+1, j+1) in
	  (zero => (i+1, j+1)) ac;
	  let err = norm1 ac in
	  (aij => (i+1, j+1)) ac;
	  fromCheb2D' ac;
	  let err = norm1 (a -$ ac)  +. err in
	  ok := !ok && err <= norm1 a *. machine_epsilon *. float (px + py)
	done
      done;
      !ok

    (* int_ax^bx int_ay^by fx(x) fy(y) dx dy *)
    let integrateTensor px py ax bx ay by fx fy =
      let ix = chebQuad ax bx px 1 fx
      and iy = chebQuad ay by py 1 fy in
      muln (ix $@ (1,1)) (iy $@ (1,1))

    let absn r =
      let x = re r
      and y = im r in
      let d = (x *. x) +. (y *. y) in
      sqrt d

    let test23 () =
      let px = 10
      and py = 13
      and ax = 0.0
      and bx = 1.0
      and ay = 0.0
      and by = 1.0
      and ok = ref true in
      for j = 0 to px - 2 do
	for i = 0 to py - 2 do
	  let ixy = chebQuad2D ax bx ay by px py (cheb2D j i) in
	  let axy = integrateTensor px py ax bx ay by
	      (fun x _ -> fromFloat (cheb j x))
	      (fun y _ -> fromFloat (cheb i y)) in
	  ok := !ok && absn (subn ixy axy) <= machine_epsilon *. float (px + py)
	      *. (1.0 +. absn axy)
	done
      done;
      !ok

    (* int_ax^bx int_ay^by (d^2 f / dx dy) dx dy =
       f(bx,by) - f(ax,by) - f(bx,ay) + f(ax,ay) *)
    let intDXY ax bx ay by f =
      subn (addn (f bx by) (f ax ay))
	   (addn (f ax by) (f bx ay))

    let test24 () =
      let px = 10
      and py = 13
      and ax = 0.0
      and bx = 1.0
      and ay = 0.0
      and by = 1.0
      and tol = sqrt machine_epsilon in
      let f x y = fromFloat (sqrt (x *. x +. y *. y)) in
      let dxyf x y = fromFloat ((-1.0) *. x *. y /. ((x*.x+.y*.y) ** 1.5)) in
      let q = chebQuadAdaptive2D tol ax bx ay by [|0.0 +: 0.0|] px py dxyf in
      let qa = intDXY ax bx ay by f in
      absn (subn q qa) <= tol *. (1.0 +. absn qa)


    let test25 () =
      let ss = [| -1.5 +: 0.0; -1.25 +: 0.0; -1.125 +: 0.0; -1.01 +: 0.0;
		  -1.001 +: 0.0; -1.0001 +: 0.0 |] in
      let sn = Array.length ss in
      let tol = sqrt machine_epsilon
      and n = 10 in
      let w s x = log (Complex.norm (Complex.sub (x +: 0.0) s)) in
      let ok = ref true in
      for i = 0 to sn-1 do
	let s = ss.(i) in
	let q = quadWtsSingularKernel tol n (w s) [|s|] in
	let a = transp (chebQuadAdaptive tol (-1.0) 1.0 [|s|] (2 * n) n
			  (fun x i -> fromFloat (w s x *. cheb (i-1) x))) in
	Printf.printf "%g +: %g -> %g\n" s.Complex.re s.Complex.im (norm1 q);
	ok := !ok && norm1(chebVand (chebZeros n) n *$ q -$ a) <=
	  tol *. norm1 a /. norm1 q
      done;
      !ok


    (* l2 norm *)
    let l2norm f =
      let p = noOfRows f in
      let zw = chebQuadWts p in
      let zwm = fun2mat
	  (fun i j -> if i = j then
	    fromFloat (sqrt (absn (zw $@ (1,i))))
	  else fromFloat 0.0) p p in
      normF (zwm *$ f)


    (* Testing boundedness of Cauchy p.v. integral. *)
    let test26 () =
      let f y x i =
	if x = y then
	  fromFloat 0.0
	else begin
	  fromFloat ((cheb (i-1) x -. cheb (i-1) y) /. (y -. x))
	end
      in
      let p = 20
      and n = 10 in
      let zp = chebZeros p
      and tol = sqrt (sqrt machine_epsilon)
      and a = rawMatrix p n in
      for i = 1 to p do
	(a $ (i <-> i, All)) <-$ chebQuadAdaptive tol (-1.0) 1.0
	    [| zp.(i-1) +: 0.0 |] p n (f zp.(i-1))
      done;
      let ok = ref true in
      for i = 1 to n do
	(* Printf.printf "%i %g\n" (i-1) (l2norm (a $ (All, i <-> i))); *)
	ok := !ok && l2norm (a $ (All, i <-> i)) <= 4.0
      done;
      !ok


    (* Test plotting. I need to close_process to get either gnuplot or
       aquaterm to display the plots. Not sure what is going on.
       Figured it out. Need "\n" after "plot sin(x)"! *)
    let test27' () =
      let (gin, gout) = Unix.open_process "gnuplot" in
      output_string gout "plot sin(x)";
      flush gout;
      begin
	match Unix.close_process (gin, gout) with
	  Unix.WEXITED n -> Printf.printf "WEXITED %i\n" n
	| Unix.WSIGNALED n -> Printf.printf "WSIGNALED %i\n" n
	| Unix.WSTOPPED n -> Printf.printf "WSTOPPED %i\n" n
      end;
      true


    (* Plot a real-array versus the real part of a matrix, with one plot
       per column. *)
    let plot ~xaxis:xs ~yaxis:ys =
      let dfile = Filename.temp_file "plot" ".txt" in
      let dout = open_out dfile in
      for j = 1 to noOfCols ys do
	for i = 1 to Array.length xs do
	  Printf.fprintf dout "%g %g\n" xs.(i-1) (re (ys $@ (i,j)));
	done;
	Printf.fprintf dout "\n";
      done;
      close_out dout;
      let (gin, gout) = Unix.open_process "gnuplot" in
      output_string gout ("plot " ^ "\"" ^ dfile ^ "\"");
      flush gout;
      ignore (Unix.close_process (gin, gout));
      Sys.remove dfile



    (* Check smoothness of convolution with log outside interval. *)
    let test27 () =
      let logK y x =
	if x = y then 0.0
	else log (abs_float (x -. y))
      in
      let p = 20
      and n = 5
      and m = 100
      and tol = 1E-3 in
      let ys = Array.init m (fun i -> (-2.0) +. 4.0 *. float i /. float m) in
      let vs = rawMatrix m n in
      for i = 1 to m do
	(vs $ (i<->i,All)) <-$ chebQuadAdaptive tol (-1.0) 1.0
	    [| ys.(i-1) +: 0.0 |] p n
	    (fun x j -> fromFloat (logK ys.(i-1) x *. cheb j x))
      done;
      plot ~xaxis:ys ~yaxis:vs;
      true
      

    let tests = [| test1; test2; test3; test4; test5; test6; test7; test8;
		   test9; test10; test11; test12; test13; test14; test15;
		   test16; test17; test18; test19; test20; test21; test22;
		   test23; test24; test25; test26; test27 |]
      
    let run_tests () =
      let suc = ref 0
      and fal = ref 0
      and abo = ref 0
      and start = 27
      and fin = 27 in
      for i = max 0 (start - 1) to min (fin - 1) (Array.length tests - 1) do
	print_string "Test ";
	print_int (i + 1);
	begin
	  try
	    if tests.(i) () then
	      (print_string " : Pass";
	       suc := !suc + 1)
	    else
	      (print_string " : Fail";
	       fal := !fal + 1)
	  with
	    exn ->
	      (print_string " : Abort";
	       abo := !abo + 1; raise exn)
	end;
	print_newline ()
      done;
      print_string "Passed: ";
      print_int !suc;
      print_string "\nFailed: ";
      print_int !fal;
      print_string "\nAborted: ";
      print_int !abo;
      print_newline ()
	
  end;; (* module Tester *)


module Dtests = Tester (DBlop) (DHss) (DChub);;
module Stests = Tester (SBlop) (SHss) (SChub);;
module Ztests = Tester (ZBlop) (ZHss) (ZChub);;
module Ctests = Tester (CBlop) (CHss) (CChub);;


let main () =
  begin
    print_string "Double Precision Tests:\n";
    Dtests.run_tests ();
    print_newline ();
    
    print_string "Single Precision Tests:\n";
    Stests.run_tests ();
    print_newline ();
    
    print_string "Double Complex Tests:\n";
    Ztests.run_tests ();
    print_newline ();
    
    print_string "Single Complex Tests:\n";
    Ctests.run_tests ();
    print_newline ();
  end;;


main ();;
