(*
  File: testFmm1D.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 7.29.2005
  Email: shiv@ece.ucsb.edu
  Purpose: test fmm1D.ml

   $Id$
*)

open Utils
open Nla
open Fmm1D


module Tester (Blop : BLOP)
    (Fmm1D : FMM1D with type matrix = Blop.matrix and type t = Blop.t) =
struct

  open Blop
  open Fmm1D

  type spine_leaf = { m : int; n : int; k : int; l : int }
  type spine_node = { ku : int; lv : int }
  type spine = (spine_leaf, spine_node) btree

  let randomFmm1D ?(ooc = false) spine =
    let rec rbu lfmm rfmm =
      match lfmm, rfmm with
	Leaf l, Leaf r -> Pu (randomMatrix ~ooc (noOfCols l.u) (noOfCols r.v))
      |	Leaf l, Node (r, lf, _) ->
	  Su { bu11 = zeros ~ooc 0 (noOfRows r.wl);
	       bu12 = zeros ~ooc 0 (noOfRows r.wr);
	       bu21 = rbu lfmm lf;
	       bu22 = randomMatrix ~ooc (noOfCols l.u) (noOfRows r.wr)
	     }
      |	Node (l, _, rf), Leaf r ->
	  Su { bu11 = randomMatrix ~ooc (noOfRows l.rl) (noOfCols r.v);
	       bu12 = zeros ~ooc (noOfRows l.rl) 0;
	       bu21 = rbu rf rfmm;
	       bu22 = zeros ~ooc (noOfRows l.rr) 0
	     }
      |	Node (l, _, rf), Node (r, lf, _) ->
	  Su { bu11 = randomMatrix ~ooc (noOfRows l.rl) (noOfRows r.wl);
	       bu12 = randomMatrix ~ooc (noOfRows l.rl) (noOfRows r.wr);
	       bu21 = rbu rf lf;
	       bu22 = randomMatrix ~ooc (noOfRows l.rr) (noOfRows r.wr)
	     }
    in
    let rec rbl lfmm rfmm =
      match lfmm, rfmm with
	Leaf l, Leaf r -> Pl (randomMatrix ~ooc (noOfCols r.u) (noOfCols l.v))
      |	Leaf l, Node (r, lf, _) ->
	  Sl { bl11 = zeros ~ooc (noOfRows r.rl) 0;
	       bl12 = rbl lfmm lf;
	       bl21 = zeros ~ooc (noOfRows r.rr) 0;
	       bl22 = randomMatrix ~ooc (noOfCols r.rr) (noOfCols l.v)
	     }
      |	Node (l, _, rf), Leaf r ->
	  Sl { bl11 = randomMatrix ~ooc (noOfCols r.u) (noOfRows l.wl);
	       bl12 = rbl rf rfmm;
	       bl21 = zeros ~ooc 0 (noOfRows l.wl);
	       bl22 = zeros ~ooc 0 (noOfRows l.wr)
	     }
      |	Node (l, _, rf), Node (r, lf, _) ->
	  Sl { bl11 = randomMatrix ~ooc (noOfRows r.rl) (noOfRows l.wl);
	       bl12 = rbl rf lf;
	       bl21 = randomMatrix ~ooc (noOfRows r.rr) (noOfRows l.wl);
	       bl22 = randomMatrix ~ooc (noOfRows r.rr) (noOfRows l.wr)
	     }
    in
    let rec rfmm spine =
      match spine with
	Leaf a ->
	  (a.k, a.l, a.m, a.n,
	   Leaf { u = randomMatrix ~ooc a.m a.k;
		  v = randomMatrix ~ooc a.n a.l;
		  d = randomMatrix ~ooc a.m a.n
		}
	     )
      |	Node (a, lSpine, rSpine) ->
	  let (k0, l0, m0, n0, lFmm) = rfmm lSpine
	  and (k1, l1, m1, n1, rFmm) = rfmm rSpine in
	  (a.ku, a.lv, m0+m1, n0+n1,
	   Node ({ ml = m0; mr = m1; nl = n0; nr = n1;
		   rl = randomMatrix ~ooc k0 a.ku;
		   rr = randomMatrix ~ooc k1 a.ku;
		   wl = randomMatrix ~ooc l0 a.lv;
		   wr = randomMatrix ~ooc l1 a.lv;
		   bu = rbu lFmm rFmm;
		   bl = rbl lFmm rFmm
		 },
		 lFmm, rFmm))
    in
    let (_, _, _, _, fmm) = rfmm spine in
    fmm


  let spine1 = Node ({ku=0; lv=0},
		     Leaf {k=1; l=1; m=3; n=3},
		     Leaf {k=1; l=1; m=4; n=4})
  and spine2 = Node ({ku=0; lv=0},
		     Leaf {k=2; l=2; m=4; n=4},
		     Leaf {k=3; l=3; m=5; n=5})
  and spine3 = Node ({ku=0; lv=0},
		     Node ({ku=1; lv=1},
			   (Leaf {k=1; l=1; m=4; n=4}),
			   (Leaf {k=1; l=1; m=3; n=3})),
		     Node ({ku=1; lv=1},
			   (Leaf {k=1; l=1; m=4; n=4}),
			   (Leaf {k=1; l=1; m=3; n=3})))

  let spines = [| spine1; spine2; spine3 |]


  let test1 () =
    let ok = ref true in
    for i = 2 to 2 do
      let fmm1D = randomFmm1D spines.(i) in
      let m = no_of_rows fmm1D
      and n = no_of_cols fmm1D in
      let x = eye n n
      and xc = zeros n n
      and b = zeros m n in
      fastMul fmm1D x b;
      print_endline "multiply done";
      fastSolve' fmm1D (copy b) xc;
      printMatrix (x -$ xc);
      Printf.printf "\nerror = %g\n" (norm1 (x -$ xc) /. norm1 x);
      ok := !ok && norm1 (x -$ xc) /. norm1 x <= 500. *. machine_epsilon
    done;
    !ok


  let tests = [| test1 |]
      
  let run_tests () =
    let suc = ref 0
    and fal = ref 0
    and abo = ref 0
    and start = 1
    and fin = 2000 in
    for i = max 0 (start - 1) to min (fin - 1) (Array.length tests - 1) do
      print_string "Test ";
      print_int (i + 1);
      begin
	try
	  if tests.(i) () then
	    (print_string " : Pass";
	     suc := !suc + 1)
	  else
	    (print_string " : Fail";
	     fal := !fal + 1)
	with
	  exn ->
	    (print_string " : Abort";
	     abo := !abo + 1; raise exn)
      end;
      print_newline ()
    done;
    print_string "Passed: ";
    print_int !suc;
    print_string "\nFailed: ";
    print_int !fal;
    print_string "\nAborted: ";
    print_int !abo;
    print_newline ()

end (* module Tester *)


module Dtests = Tester (DBlop) (DFmm1D)
module Stests = Tester (SBlop) (SFmm1D)
module Ztests = Tester (ZBlop) (ZFmm1D)
module Ctests = Tester (CBlop) (CFmm1D)


let main () =
  begin
    print_string "Single Precision Tests:\n";
    Stests.run_tests ();
    print_newline ();

    print_string "Double Precision Tests:\n";
    Dtests.run_tests ();
    print_newline ();
    
    print_string "Double Complex Tests:\n";
    Ztests.run_tests ();
    print_newline ();
    
    print_string "Single Complex Tests:\n";
    Ctests.run_tests ();
    print_newline ();
  end;;


main ();;

