(*
  File: testHss.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 5.14.2002
  Email: shiv@ece.ucsb.edu
  Purpose: test hss.ml and nestedDissection.ml

   $Id: testHss.ml,v 1.1 2005/01/10 22:39:03 shiv Exp shiv $
*)

open Bigarray;;
open Utils;;
open Nla;;
open Sparse;;
open Hss;;
open NestedDissection;;


module Tester (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix)
    (Hss : HSS with type matrix = Blop.matrix and type t = Blop.t
    and type transform = Blop.transform and type sparse = Sparse.sparse) =
struct

  open Blop
  open Hss

  type leaf_split = { m : int; n : int; k : int; l : int }
  and node_split = { ku : int; lv : int }
  and hss_spine = (leaf_split, node_split) btree

  let randomHss ?(ooc=false) spine =
    let rec rec1 spine =
      match spine with
	  Node (a, lSpine, rSpine) ->
	    let (lHss, k0, l0, m0, n0) = rec1 lSpine
	    and (rHss, k1, l1, m1, n1) = rec1 rSpine in
	    (
	      Node
	      (
		{
		  ml = m0; mr = m1; nl = n0; nr = n1;
		  bu = randomMatrix ~ooc k0 l1;
		  bl = randomMatrix ~ooc k1 l0;
		  rl = randomMatrix ~ooc k0 a.ku;
		  rr = randomMatrix ~ooc k1 a.ku;
		  wl = randomMatrix ~ooc l0 a.lv;
		  wr = randomMatrix ~ooc l1 a.lv;
		},
		lHss,
		rHss
	      ),
	      a.ku, a.lv, m0+m1, n0+n1
	    )
	| Leaf a ->
	    (
	      Leaf
	      {
		u = randomMatrix ~ooc a.m a.k;
		v = randomMatrix ~ooc a.n a.l;
		d = randomMatrix ~ooc a.m a.n
	      },
	      a.k, a.l, a.m, a.n
	    )
    in
    let (hss, _, _, _, _) = rec1 spine in hss

  let rec to_hss_spine nd_spine =
    match nd_spine with
	Leaf _ -> Leaf ()
      | Node ((m1, m2), lSpine, rSpine) ->
	  Node ({mLeft = m1; mRight = m2; nLeft = m1; nRight = m2},
	  to_hss_spine lSpine, to_hss_spine rSpine)

  let rec get_hss_spine hss =
    match hss with
	Leaf a -> Leaf {m = noOfRows a.d; n = noOfCols a.d; k =
      noOfCols a .u; l = noOfCols a.v }
      | Node (a, lHss, rHss) -> Node ({ku = noOfCols a.rl; lv =
	  noOfCols a.wl}, get_hss_spine lHss, get_hss_spine rHss)

  let rec print_spine hss_spine =
    match hss_spine with
	Leaf a -> Printf.printf "Leaf {m=%i; n=%i; k=%i; l=%i}" a.m
	  a.n a.k a.l
      | Node (a, lSpine, rSpine) ->
	  Printf.printf "Node ({ku=%i; lv=%i},\n" a.ku a.lv;
	  print_spine lSpine; Printf.printf ",\n";
	  print_spine rSpine; Printf.printf ")"

  let print_hss_spine hss =
    let rec pp d hss =
      Format.open_box d;
      begin
	match hss with
	    Leaf a ->
	      for i=0 to d-1 do Format.print_string "|" done;
	      Format.print_string "Leaf";
	      Format.print_break 1 2;
	      Format.printf "m = %i" (noOfRows a.d);
	      Format.print_break 1 2;
	      Format.printf "n = %i" (noOfCols a.d);
	      Format.print_break 1 2;
	      Format.printf "ku = %i" (noOfCols a.u);
	      Format.print_break 1 2;
	      Format.printf "kv = %i" (noOfCols a.v)
	  | Node (a, lHss, rHss) ->
	      for i=0 to d-1 do Format.print_string "|" done;
	      Format.print_string "Node";
	      Format.print_break 1 2;
	      Format.printf "ku = %i" (noOfCols a.rl);
	      Format.print_break 1 2;
	      Format.printf "kv = %i" (noOfCols a.wl);
	      Format.print_newline ();
	      pp (d+1) lHss;
	      Format.print_newline ();
	      pp (d+1) rHss
      end;
      Format.close_box ()
    in
    pp 0 hss;
    Format.print_newline ()

  let rec checkHss hss =
    let rec chk hss =
      match hss with
	  Leaf a ->
	    let m = noOfRows a.d
	    and n = noOfCols a.d
	    and ku = noOfCols a.u
	    and kv = noOfCols a.v in
	    assert (m = noOfRows a.u);
	    assert (n = noOfRows a.v);
	    (ku, kv)
	| Node (a, lhss, rhss) ->
	    let (kul, kvl) = chk lhss
	    and (kur, kvr) = chk rhss in
	    let ku = noOfCols a.rl
	    and kv = noOfCols a.wl in
	    assert (kul = noOfRows a.rl);
	    assert (kur = noOfRows a.rr);
	    assert (ku = noOfCols a.rr);
	    assert (kvl = noOfRows a.wl);
	    assert (kvr = noOfRows a.wr);
	    assert (kv = noOfCols a.wr);
	    assert (kul = noOfRows a.bu);
	    assert (kvr = noOfCols a.bu);
	    assert (kur = noOfRows a.bl);
	    assert (kvl = noOfCols a.bl);
	    (ku, kv)
    in
    let _ = chk hss in ()

  let kernel1D alpha omega xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    assert (alpha >= -1.0);
    let kernelFn x y =
      fromFloat (
	let d = abs_float (x -. y) in
	if d > 0.0 then
	  (* sin *) cos (omega *. d) *. (d ** alpha)
	else if alpha = -1.0 then 1.0 +. omega
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let kernel2D alpha omega xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    let sq x = x *. x in
    let dist x y = sqrt (sq (x.(0) -. y.(0)) +. sq (x.(1) -. y.(1))) in
    let kernelFn x y =
      fromFloat (
	let d = dist x y in
	if d > 0.0 then
	  let d = log d in sin (omega *. d) *. (abs_float d ** alpha)
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let kernel3D alpha omega xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    let sq x = x *. x in
    let dist x y = sqrt (sq (x.(0) -. y.(0)) +. sq (x.(1) -. y.(1))
    +. sq (x.(2) -. y.(2))) in
    let kernelFn x y =
      fromFloat (
	let d = dist x y in
	if d > 0.0 then sin (omega *. d) *. (d ** alpha)
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let pie = 2.0 *. acos 0.0

  let elasticG3D nu g xs x y i j =
    let r = Array.make 3 0.0 in
    r.(0) <- xs.(y).(0) -. xs.(x).(0);
    r.(1) <- xs.(y).(1) -. xs.(x).(1);
    r.(2) <- xs.(y).(2) -. xs.(x).(2);
    let rl = sqrt (r.(0) *. r.(0) +. r.(1) *. r.(1) +. r.(2) *. r.(2))
    in
    if rl > 0.0 then
      ((if i <> j then (3.0 -. 4.0 *. nu) /. rl else 0.0) +. (r.(i)
      *. r.(j) /. (rl *. rl *. rl))) /. (16.0 *. pie *. (1.0 -. nu)
      *. g)
    else
      float (Array.length xs)

  let elasticH3D nu xs ns x y i j =
    let r = Array.make 3 0.0 in
    r.(0) <- xs.(y).(0) -. xs.(x).(0);
    r.(1) <- xs.(y).(1) -. xs.(x).(1);
    r.(2) <- xs.(y).(2) -. xs.(x).(2);
    let rl = sqrt (r.(0) *. r.(0) +. r.(1) *. r.(1) +. r.(2) *. r.(2))
    in
    let rl2 = rl *. rl in
    let rl3 = rl2 *. rl in
    let rl4 = rl3 *. rl in
    if rl > 0.0 then
      (1.0 -. 2.0 *. nu) *. ((r.(i) *. ns.(y).(j)
      -. r.(j)*.ns.(y).(i)) /. rl3 -. ((if i <> j
      then (1.0 /. rl2) else 0.0) +. (3.0 *. r.(i) *. r.(j)
      /. ((1.0 -. 2.0 *. nu) *. rl4)))
      *. ((r.(0)*.ns.(y).(0) +. r.(1)*.ns.(y).(1) +. r.(2)*.ns.(y).(2))
      /. rl)) /. (8.0 *. pie *. (1.0 -. nu))
    else
      float (Array.length xs)

  let elasticC3D nu g xs ns ds x y i j =
    assert (0 <= x && x < Array.length xs);
    assert (0 <= y && y < Array.length xs);
    assert (0 <= i && i < 3);
    assert (0 <= j && j < 3);
    assert (3*y+j < Array.length ds);
    fromFloat (elasticH3D nu xs ns x y i j +.
      elasticG3D nu g xs x y i j *. ds.(3*y+j))

  let kernel3Delastic nu g xs ns ds sr lr sc lc a =
    let kFn i j =
      let ai = sr + i - 1
      and aj = sc + j - 1 in
      let xx = (ai - 1) / 3
      and xy = (aj - 1) / 3
      and xi = (ai - 1) mod 3
      and yj = (aj - 1) mod 3 in
      elasticC3D nu g xs ns ds xx xy xi yj in
    fun2mat' kFn a

  let spine1 = Node ({ku=0; lv=0},
    (Leaf {k=1; l=1; m=3; n=3}),
    (Leaf {k=1; l=1; m=4; n=4}))
  and spine2 = Node ({ku=0; lv=0},
    (Leaf {k=2; l=2; m=4; n=4}),
    (Leaf {k=3; l=3; m=5; n=5}))
  and spine3 = Node ({ku=0; lv=0},
    Node ({ku=2; lv=3},
      (Leaf {k=1; l=1; m=4; n=4}),
      (Leaf {k=1; l=1; m=3; n=3})),
    Node ({ku=2; lv=1},
      (Leaf {k=1; l=1; m=4; n=4}),
      (Leaf {k=1; l=1; m=3; n=3})))
  and spine4 = Node ({ku=0; lv=0},
  Node ({ku=2; lv=3},
    (Leaf {k=1; l=1; m=2; n=2}),
    (Leaf {k=1; l=1; m=3; n=3})),
  Node ({ku=2; lv=1},
    (Leaf {k=1; l=1; m=4; n=4}),
    Node ({ku=2; lv=1},
      Node ({ku=2; lv=3},
        (Leaf {k=1; l=1; m=3; n=3}),
        (Leaf {k=1; l=1; m=3; n=3})),
      Node ({ku=2; lv=1},
        (Leaf {k=1; l=1; m=2; n=2}),
        (Leaf {k=1; l=1; m=3; n=3})))))
  and spine5 = Node ({ku=0; lv=0},
  Node ({ku=2; lv=3},
    (Leaf {k=1; l=1; m=2; n=2}),
    Node ({ku=3; lv=4},
      Node ({ku=2; lv=3},
        (Leaf {k=1; l=1; m=2; n=2}),
        (Leaf {k=1; l=1; m=3; n=3})),
      Node ({ku=2; lv=1},
        (Leaf {k=1; l=1; m=4; n=4}),
        Node ({ku=2; lv=1},
          Node ({ku=2; lv=3},
            (Leaf {k=1; l=1; m=3; n=3}),
            (Leaf {k=1; l=1; m=3; n=3})),
          Node ({ku=2; lv=1},
            (Leaf {k=1; l=1; m=2; n=2}),
            (Leaf {k=1; l=1; m=3; n=3})))))),
  Node ({ku=2; lv=1},
    (Leaf {k=1; l=1; m=4; n=4}),
    Node ({ku=2; lv=1},
      Node ({ku=2; lv=3},
        (Leaf {k=1; l=1; m=3; n=3}),
        (Leaf {k=1; l=1; m=3; n=3})),
      Node ({ku=2; lv=1},
        (Leaf {k=1; l=1; m=2; n=2}),
        (Leaf {k=1; l=1; m=3; n=3})))))
  and spine6 = Node ({ku=0; lv=0},
  Node ({ku=2; lv=3},
    Node ({ku=3; lv=3},
      Node ({ku=2; lv=3},
        (Leaf {k=1; l=1; m=2; n=2}),
        (Leaf {k=1; l=1; m=3; n=3})),
      Node ({ku=2; lv=1},
        (Leaf {k=1; l=1; m=4; n=4}),
        Node ({ku=2; lv=1},
          Node ({ku=2; lv=3},
            (Leaf {k=1; l=1; m=3; n=3}),
            (Leaf {k=1; l=1; m=3; n=3})),
          Node ({ku=2; lv=1},
            (Leaf {k=1; l=1; m=2; n=2}),
            (Leaf {k=1; l=1; m=3; n=3}))))),
    Node ({ku=3; lv=4},
      Node ({ku=2; lv=3},
        (Leaf {k=1; l=1; m=2; n=2}),
        (Leaf {k=1; l=1; m=3; n=3})),
      Node ({ku=2; lv=1},
        (Leaf {k=1; l=1; m=4; n=4}),
        Node ({ku=2; lv=1},
          Node ({ku=2; lv=3},
            (Leaf {k=1; l=1; m=3; n=3}),
            (Leaf {k=1; l=1; m=3; n=3})),
          Node ({ku=2; lv=1},
            (Leaf {k=1; l=1; m=2; n=2}),
            (Leaf {k=1; l=1; m=3; n=3})))))),
  Node ({ku=2; lv=1},
    (Leaf {k=1; l=1; m=4; n=4}),
    Node ({ku=2; lv=1},
      Node ({ku=2; lv=3},
        (Leaf {k=1; l=1; m=3; n=3}),
        (Leaf {k=2; l=3; m=5; n=5})),
      Node ({ku=2; lv=1},
        (Leaf {k=1; l=1; m=2; n=2}),
        (Leaf {k=1; l=1; m=3; n=3})))))
  and spine7 = Node ({ku=0; lv=0},
  Node ({ku=2; lv=2},
    Leaf {m=1; n=1; k=1; l=1},
    Leaf {m=4; n=4; k=2; l=2}),
  Node ({ku=2; lv=2},
    Leaf {m=5; n=5; k=2; l=3},
    Node ({ku=2; lv=2},
      Leaf {m=1; n=1; k=1; l=1},
      Node ({ku=2; lv=2},
        Leaf {m=0; n=0; k=0; l=0},
        Node ({ku=2; lv=2},
          Leaf {m=0; n=0; k=0; l=0},
          Node ({ku=2; lv=2},
            Leaf {m=1; n=1; k=1; l=1},
            Node ({ku=2; lv=2},
              Leaf {m=2; n=2; k=2; l=2},
              Leaf {m=1; n=1; k=1; l=1})))))))
  and spine8 = Node ({ku=0; lv=0},
  Node ({ku=2; lv=2},
    Node ({ku=2; lv=2},
      Leaf {m=0; n=0; k=0; l=0},
      Node ({ku=2; lv=2},
        Leaf {m=1; n=1; k=1; l=1},
        Node ({ku=2; lv=2},
          Leaf {m=1; n=1; k=1; l=1},
          Leaf {m=1; n=1; k=1; l=1}))),
    Leaf {m=1; n=1; k=1; l=1}),
  Node ({ku=2; lv=2},
    Node ({ku=2; lv=2},
      Node ({ku=2; lv=2},
        Leaf {m=1; n=1; k=1; l=1},
        Node ({ku=2; lv=2},
          Leaf {m=0; n=0; k=0; l=0},
          Node ({ku=2; lv=2},
            Leaf {m=1; n=1; k=1; l=1},
            Leaf {m=1; n=1; k=1; l=1}))),
      Leaf {m=1; n=1; k=1; l=1}),
    Node ({ku=2; lv=2},
      Leaf {m=0; n=0; k=0; l=0},
      Node ({ku=2; lv=2},
        Node ({ku=2; lv=2},
          Leaf {m=1; n=1; k=1; l=1},
          Leaf {m=1; n=1; k=1; l=1}),
        Leaf {m=0; n=0; k=0; l=0}))))

  let spines = [| spine1; spine2; spine3; spine4; spine5; spine6; spine7;
	spine8 |]

  let test1 () =
    let nrhss = [| 1; 2; 3; 4 |]
    and ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for nrhsi = 0 to Array.length nrhss - 1 do
	for ooci = 0 to 1 do
	  let ooc = ooci = 1 in
	  let nrhs = nrhss.(nrhsi)
	  and hss = randomHss ~ooc spines.(spinei) in
	  let m = no_of_rows hss
	  and n = no_of_cols hss in
	  let x = randomMatrix ~ooc n nrhs
	  and b = rawMatrix ~ooc m nrhs
	  and xc = rawMatrix ~ooc n nrhs in
	  fastMul hss x b;
	  fastSolve' hss b xc;
	  ok := !ok && (norm1 (x -$ xc) <= norm1 xc *. machine_epsilon
			  *. float (m * n * nrhs) *. 10.0)
	done
      done
    done;
    !ok

  let test2 () =
    let ns = [| 10; 20; 30; 40 |]
    and nrhss = [| 1; 2; 3; 4 |]
    and alphas = [| -1.0; -0.5; 0.0; 0.5 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0; 3.0
      *. acos 0.0 |]
    and ps = [| 1; 2; 3; 4 |]
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n (fun _ -> Random.float 1.0) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissect1D xs 0.0 1.0) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel1D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		fastSolve' (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test3 () =
    let ns = [| 20; 40 |]
    and nrhss = [| 1; 4 |]
    and alphas = [| 1.0; 0.5; 2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 1; 4 |]
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n (fun _ -> Array.init 2
		    (fun _ -> Random.float 1.0)) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissectND xs [| 0.0; 0.0 |]
					  [| 1.0; 1.0 |] ) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel2D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		fastSolve' (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test4 () =
    let ns = [| 100; 200 |]
    and nrhss = [| 1 |]
    and alphas = [| -1.0; -0.5; -2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 4; 8 |]
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n (fun _ -> Array.init 3
		    (fun _ -> Random.float 1.0)) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissectND xs [| 0.0; 0.0; 0.0 |]
					  [| 1.0; 1.0; 1.0 |] ) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel3D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		fastSolve' (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test5 () =
    let ns = [| 20; 40; 80 |]
    and nrhss = [| 1; 4 |]
    and alphas = [| 1.0; 0.5; 2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 4; 8 |]
    and twoPi = 4.0 *. acos 0.0
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n
		    (fun _ ->
		      let theta = Random.float twoPi in
		      let x = Array.make 2 0.0 in
		      x.(0) <- cos theta;
		      x.(1) <- sin theta;
		      x) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissectND xs [| -1.0; -1.0 |]
					  [| 1.0; 1.0 |] ) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel2D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		fastSolve' (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test6 () =
    let ns = [| 100; 200 |]
    and nrhss = [| 1 |]
    and alphas = [| -1.0; -0.5; -2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 8; 12 |]
    and pie = 2.0 *. acos 0.0
    and twoPi = 4.0 *. acos 0.0
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n
		    (fun _ ->
		      let theta = Random.float pie
		      and phi = Random.float twoPi
		      and x = Array.make 3 0.0 in
		      x.(0) <- sin theta *. sin phi;
		      x.(1) <- sin theta *. cos phi;
		      x.(2) <- cos theta;
		      x) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissectND xs [| -1.0; -1.0; -1.0 |]
					  [| 1.0; 1.0; 1.0 |] ) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel3D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		fastSolve' (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test7 () =
    let ns = [| 40 |]
    and nrhss = [| 1; 4 |]
    and ps = [| 30; 10 |]
    and nus = [| 0.4; 0.1 |]
    and lambdas = [| 10.0; 100.0 |]
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for nrhsi = 0 to Array.length nrhss - 1 do
	for pi = 0 to Array.length ps - 1 do
	  for nui = 0 to Array.length nus - 1 do
	    for lambdai = 0 to Array.length lambdas - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi) in
		let tol = sqrt (machine_epsilon *. 1000000.0 *. float n)
		and p = ps.(pi) (* leaf size *)
		and k = 3 (* vector size of unknowns in PDE. *)
		and nu = nus.(nui)
		and g = 1.0
		and lambda = lambdas.(lambdai)
		and twoPie = pie *. 2.0 in
		Printf.printf "n=%i nrhs=%i p=%i nu=%g lambda=%g" n nrhs
		  p nu lambda; print_newline ();
		let m = n * k in
		let xsa = Array.init n
		    (fun _ ->
		      let theta = Random.float pie
		      and phi = Random.float twoPie
		      and x = Array.make 3 0.0 in
		      x.(0) <- sin theta *. sin phi;
		      x.(1) <- sin theta *. cos phi;
		      x.(2) <- cos theta;
		      x) in
		let xs = Array.to_list xsa in
		let ds = Array.init m
		    (fun _ -> let heads = Random.bool () in
		    if heads then lambda else 1.0 /. lambda) in
		let oTree = prune_tree
		    (dissectND xs [| -1.0; -1.0; -1.0 |] [| 1.0; 1.0; 1.0 |] )
		    p in
		let iTree = inflate_partition_tree oTree k in
		let ps = Array.of_list (breadthFirstLeafData oTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree
		and normals = xs in
		let hss = fromFn ~ooc hss_spine
		    (kernel3Delastic nu g xs normals ds) tol in
		let b = randomMatrix ~ooc m nrhs
		and x = rawMatrix ~ooc m nrhs in
		fastSolve' (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc m nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= 3.0 *. machine_epsilon *.
				(norm1 x *. float (m * nrhs * m) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok
	      
  let test8 () =
    let tol = 0.001
    and ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hss = randomHss ~ooc spines.(spinei) in
	checkHss hss;
	let hssCompr = fastModelReduction ~ooc tol hss in
	checkHss hss;
	let m = no_of_rows hss
	and n = no_of_cols hss in
	let ei = eye n n
	and hssc = rawMatrix ~ooc m n
	and hssComprc = rawMatrix ~ooc m n in
	fastMul hss ei hssc;
	fastMul hssCompr ei hssComprc;
	ok := !ok && (norm1 (hssc -$ hssComprc) <= tol)
      done
    done;
    !ok

  let test9 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let ahss = randomHss ~ooc spines.(spinei)
	and bhss = randomHss ~ooc spines.(spinei) in
	let chss = fastAdd ~ooc ahss bhss
	and m = no_of_rows ahss
	and n = no_of_cols ahss in
	let ac = rawMatrix ~ooc m n
	and bc = rawMatrix ~ooc m n
	and cc = rawMatrix ~ooc m n
	and ei = eye ~ooc n n in
	fastMul ahss ei ac;
	fastMul bhss ei bc;
	fastMul chss ei cc;
	ok := !ok && (norm1 (ac +$ bc -$ cc) <= 3.0 *. norm1 cc *.
			machine_epsilon)
      done
    done;
    !ok

  let test10 () =
    let ok = ref true in
    for spinei = 0 to 0 (*Array.length spines - 1*) do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hss = randomHss ~ooc spines.(spinei) in
	checkHss hss;
	let hssp = toProperForm' ~ooc (Hss.copy ~ooc hss) in
	checkHss hssp;
	let m = no_of_rows hss
	and n = no_of_cols hss in
	let ei = eye ~ooc n n
	and hssc = rawMatrix ~ooc m n
	and hsspc = rawMatrix ~ooc m n in
	fastMul hss ei hssc;
	fastMul hssp ei hsspc;
	ok := !ok && (norm1 (hssc -$ hsspc) <= machine_epsilon *. norm1 hssc)
      done
    done;
    !ok

  let test11 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let ahss = randomHss ~ooc spines.(spinei)
	and bhss = randomHss ~ooc spines.(spinei) in
	let chss = fastSub ~ooc ahss bhss
	and m = no_of_rows ahss
	and n = no_of_cols ahss in
	let ac = rawMatrix ~ooc m n
	and bc = rawMatrix ~ooc m n
	and cc = rawMatrix ~ooc m n
	and ei = eye ~ooc n n in
	fastMul ahss ei ac;
	fastMul bhss ei bc;
	fastMul chss ei cc;
	ok := !ok && (norm1 (ac -$ bc -$ cc) <= 3.0 *. norm1 cc *.
			machine_epsilon)
      done
    done;
    !ok

  let test12 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let ahss = randomHss ~ooc spines.(spinei)
	and bhss = randomHss ~ooc spines.(spinei) in
	let chss = superfastMul ~ooc ahss bhss in
	let m = no_of_rows ahss
	and n = no_of_cols ahss in
	checkHss chss;
	let ac = rawMatrix ~ooc m n
	and bc = rawMatrix ~ooc m n
	and cc = rawMatrix ~ooc m n
	and ei = eye ~ooc n n in
	fastMul ahss ei ac;
	fastMul bhss ei bc;
	fastMul chss ei cc;
	ok := !ok && (norm1 (ac *$ bc -$ cc) <= float m *. norm1 cc *.
			machine_epsilon)
      done
    done;
    !ok

  let test13 () =
    let ok = ref true
    and km = 5 in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hss = randomHss ~ooc spines.(spinei) in
	let m = no_of_rows hss
	and n = no_of_cols hss in
	let k = min km (min m n) in
	let a = randomMatrix ~ooc m k
	and b = randomMatrix ~ooc n k
	and x = eye k k in
	let abhss = toProperForm' ~ooc (addLowRank ~ooc hss a x b) in
	checkHss abhss;
	let c = rawMatrix ~ooc m n
	and d = rawMatrix ~ooc m n
	and ei = eye ~ooc n n in
	fastMul hss ei d;
	fastMul abhss ei c;
	ok := !ok && (norm1 (d +$ a *$ x *$~ b -$ c) <= float m *.
			(norm1 d +. norm1 a *. norm1 b) *. machine_epsilon)
      done
    done;
    !ok

  let test14 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hss = randomHss ~ooc spines.(spinei) in
	let m = no_of_rows hss in
	let ndHss = toND2d ~ooc hss m m hss
	and (perm, _) = toND2dPerm hss m m hss in
	checkHss ndHss;
	let c = randomMatrix ~ooc m m
	and cc = rawMatrix ~ooc (m * m) 1
	and d = rawMatrix ~ooc m m
	and d2 = rawMatrix ~ooc (m * m) 1
	and dd = rawMatrix ~ooc (m * m) 1 in
	vecC' c cc;
	perm @* cc;
	fastMul hss c d;
	vecC' d d2;
	perm @* d2;
	fastMul ndHss cc dd;
	ok := !ok &&
	  (norm1 (d2 -$ dd) <= float m *. norm1 d *. machine_epsilon)
      done
    done;
    !ok

  let transp_no_conjg ?(ooc=false) a =
    let b = transp ~ooc a in
    conjg' b;
    b

  let test15 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hss = randomHss ~ooc spines.(spinei) in
	let m = no_of_rows hss in
	let ndHss = toND2d_R ~ooc hss m m hss in
	let (perm, _) = toND2dPerm hss m m hss in
	checkHss ndHss;
	let c = randomMatrix ~ooc m m
	and cc = rawMatrix ~ooc (m * m) 1
	and d = rawMatrix ~ooc m m
	and d2 = rawMatrix ~ooc (m * m) 1
	and dd = rawMatrix ~ooc (m * m) 1 in
	vecC' c cc;
	perm @* cc;
	fastMul hss (transp_no_conjg ~ooc c) d;
	vecC' (transp_no_conjg d) d2;
	perm @* d2;
	fastMul ndHss cc dd;
	ok := !ok &&
	  (norm1 (d2 -$ dd) <= float m *. norm1 d *. machine_epsilon)
      done
    done;
    !ok

  let test16 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hssi = randomHss ~ooc spines.(spinei) in
	let mi = no_of_cols hssi in
	for spinej = 0 to Array.length spines - 1 do
	  let hssj = randomHss ~ooc spines.(spinej) in
	  let mj = no_of_cols hssj in
	  let ndHssi = toND2d ~ooc hssi mi mj hssj
	  and ndHssj = toND2d_R ~ooc hssj mj mi hssi in
	  let (perm, _) = toND2dPerm hssi mi mj hssj in
	  checkHss ndHssi;
	  checkHss ndHssj;
	  let c = randomMatrix ~ooc mi mj
	  and cc = rawMatrix ~ooc (mi * mj) 1
	  and d = rawMatrix ~ooc mi mj
	  and d1 = rawMatrix ~ooc mj mi
	  and d2 = rawMatrix ~ooc (mi * mj) 1
	  and dd = rawMatrix ~ooc (mi * mj) 1 in
	  fastMul hssi c d;
	  fastMul hssj (transp_no_conjg ~ooc d) d1;
	  vecC' (transp_no_conjg d1) d2;
	  perm @* d2;
	  vecC' c cc;
	  perm @* cc;
	  fastMul ndHssi cc dd;
	  fastMul ndHssj dd cc;
	  ok := !ok && (norm1 (d2 -$ cc) <= float mi *. float mj *. norm1 d2 *.
			  machine_epsilon)
	done
      done
    done;
    !ok

  let test17 () =
    let ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for ooci = 0 to 1 do
	let ooc = ooci = 1 in
	let hssi = mergeSingleChildNodes ~ooc
	    (randomHss ~ooc spines.(spinei)) in
	let mi = no_of_cols hssi in
	for spinej = 0 to Array.length spines - 1 do
	  Printf.printf "spines %i %i\n" spinei spinej;
	  let hssj = mergeSingleChildNodes ~ooc
	      (randomHss ~ooc spines.(spinej)) in
	  let mj = no_of_cols hssj in
	  let ndHssi = toND2d ~ooc hssi mi mj hssj
	  and ndHssj = toND2d_R ~ooc hssj mj mi hssi in
	  let (perm, _) = toND2dPerm hssi mi mj hssj in
	  checkHss ndHssi;
	  checkHss ndHssj;
	  let ndHssij = superfastMul ~ooc ndHssj ndHssi in
	  checkHss ndHssij;
	  let c = randomMatrix ~ooc mi mj
	  and cc = rawMatrix ~ooc (mi * mj) 1
	  and d = rawMatrix ~ooc mi mj
	  and d1 = rawMatrix ~ooc mj mi
	  and d2 = rawMatrix ~ooc (mi * mj) 1
	  and dd = rawMatrix ~ooc (mi * mj) 1 in
	  fastMul hssi c d;
	  fastMul hssj (transp_no_conjg ~ooc d) d1;
	  vecC' (transp_no_conjg ~ooc d1) d2;
	  perm @* d2;
	  vecC' c cc;
	  perm @* cc;
	  fastMul ndHssij cc dd;
	  ok := !ok && (norm1 (d2 -$ dd) <= float mi *. float mj *. norm1 d2 *.
			  machine_epsilon)
	done
      done
    done;
    !ok

  let spy sp =
    let m = Sparse.no_of_row_partitions sp
    and n = Sparse.no_of_col_partitions sp in
    Printf.printf "m = %i n = %i\n" m n;
    for i = 0 to m-1 do
      for j = 0 to n-1 do
	match Sparse.get sp i j with
	  None -> print_string "  "
	| Some _ -> print_string "* "
      done;
      print_newline ()
    done
      
  let test18 () =
    let ok = ref true
    and nrhs = 1 in
    for spinei = 0 to Array.length spines - 1 do
      let hss = randomHss spines.(spinei) in
      let m = no_of_rows hss
      and n = no_of_cols hss in
      let b = randomMatrix m nrhs
      and x = rawMatrix n nrhs in
      fastSolve' (Hss.copy hss) (Blop.copy b) x;
      let x_sp = fastSparseSolve hss b in
      ok := !ok & (norm1 (x_sp -$ x) /. norm1 x) <=
	float (n * m) *. machine_epsilon
    done;
    !ok

  let test19 () =
    let nrhss = [| 1; 2; 3; 4 |]
    and ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for nrhsi = 0 to Array.length nrhss - 1 do
	for ooci = 0 to 1 do
	  let ooc = ooci = 1 in
	  let nrhs = nrhss.(nrhsi)
	  and hss = randomHss ~ooc spines.(spinei) in
	  let m = no_of_rows hss
	  and n = no_of_cols hss in
	  let x = randomMatrix ~ooc n nrhs
	  and b = rawMatrix ~ooc m nrhs
	  and xc = rawMatrix ~ooc n nrhs in
	  fastMul hss x b;
	  let ulv = fastImplicitULVfactor' hss in
	  fastImplicitULVsolve ulv b xc;
	  ok := !ok && (norm1 (x -$ xc) <= norm1 xc *. machine_epsilon
			  *. float (m * n * nrhs) *. 10.0)
	done
      done
    done;
    !ok

  let test20 () =
    let ns = [| 10; 20; 30; 40 |]
    and nrhss = [| 1; 2; 3; 4 |]
    and alphas = [| -1.0; -0.5; 0.0; 0.5 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0; 3.0
      *. acos 0.0 |]
    and ps = [| 1; 2; 3; 4 |]
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n (fun _ -> Random.float 1.0) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissect1D xs 0.0 1.0) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel1D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		let ulv = fastImplicitULVfactor' (Hss.copy ~ooc hss) in
		fastImplicitULVsolve ulv (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  (* only in real single precision is this slightly different from fastSolve *)
  let test21 () =
    let nrhss = [| 1; 2; 3; 4 |]
    and ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for nrhsi = 0 to Array.length nrhss - 1 do
	for ooci = 0 to 0 do
	  let ooc = ooci = 1 in
	  let nrhs = nrhss.(nrhsi)
	  and hss = randomHss ~ooc spines.(spinei) in
	  let m = no_of_rows hss
	  and n = no_of_cols hss in
	  let x = randomMatrix ~ooc n nrhs
	  and b = rawMatrix ~ooc m nrhs
	  and xc = rawMatrix ~ooc n nrhs
	  and tol = sqrt machine_epsilon in
	  fastMul hss x b;
	  fastMRsolve' tol hss b xc;
	  ok := !ok && (norm1 (x -$ xc) <= norm1 xc *. tol
			  *. float (m * n * nrhs) *. 10.0)
	done
      done
    done;
    !ok

  (* *NO* extra compression takes places! *)
  let test22 () =
    let ns = [| 5000; 10000 |] (* [| 10; 20; 30; 40 |] *)
    and nrhss = [| 1 |] (* [| 1; 2; 3; 4 |] *)
    and alphas = [| -1.0 |] (* [| -1.0; -0.5; 0.0; 0.5 |] *)
    and omegas = [| acos 0.0 |] (* [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0; 3.0
      *. acos 0.0 |] *)
    and ps = [| 10; 20 |] (* [| 1; 2; 3; 4 |] *)
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 0 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n (fun _ -> Random.float 1.0) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree
		   (* (dissect1DWithSeparators ~width xs 0.0 1.0) p in *)
		    (dissect1D xs 0.1 1.0) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel1D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		print_endline "Starting...";
		fastMRsolve' (10.0 *. tol) (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		print_endline "...Done";
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= tol *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test23 () =
    let ns = [| 20; 40 |]
    and nrhss = [| 1; 4 |]
    and alphas = [| 1.0; 0.5; 2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 1; 4 |]
    and ok = ref true in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      for ooci = 0 to 1 do
		let ooc = ooci = 1 in
		let n = ns.(ni)
		and nrhs = nrhss.(nrhsi)
		and tol = sqrt machine_epsilon
		and alpha = alphas.(alphai)
		and p = ps.(pi)
		and omega = omegas.(omegai) in
		let xsa = Array.init n (fun _ -> Array.init 2
		    (fun _ -> Random.float 1.0)) in
		let xs = Array.to_list xsa in
		let iTree = prune_tree (dissectND xs [| 0.0; 0.0 |]
					  [| 1.0; 1.0 |] ) p in
		let ps = Array.of_list (breadthFirstLeafData iTree) in
		let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
		let hss_spine = to_hss_spine iTree in
		let hss = fromFn ~ooc hss_spine (kernel2D alpha omega xs)
		    tol in
		let b = randomMatrix ~ooc n nrhs
		and x = rawMatrix ~ooc n nrhs in
		fastMRsolve' tol (Hss.copy ~ooc hss) (Blop.copy ~ooc b) x;
		let bh = rawMatrix ~ooc n nrhs in
		fastMul hss x bh;
		ok := !ok && (norm1 (bh -$ b) <= machine_epsilon *.
				(norm1 x *. float (n * nrhs * n) +. norm1 b))
	      done
	    done
	  done
	done
      done
    done;
    !ok

  let test24 () =
    let nrhss = [| 1; 2; 3; 4 |]
    and ok = ref true in
    for spinei = 0 to Array.length spines - 1 do
      for nrhsi = 0 to Array.length nrhss - 1 do
	for ooci = 0 to 1 do
	  let ooc = ooci = 1 in
	  let nrhs = nrhss.(nrhsi)
	  and hss = randomHss ~ooc spines.(spinei) in
	  let m = no_of_rows hss
	  and n = no_of_cols hss in
	  let hssC = rawMatrix m n in
	  fastMul hss (eye m n) hssC;
	  let x = randomMatrix ~ooc m nrhs
	  and b = rawMatrix ~ooc n nrhs in
	  fastMulTransp hss x b;
	  ok := !ok && (norm1 (b -$ hssC *~$ x) <= norm1 b *. machine_epsilon
			  *. float (m * n * nrhs) *. 10.0)
	done
      done
    done;
    !ok

  let tests = [| test1; test2; test3; test4; test5; test6; test7;
		 test8; test9; test10; test11; test12; test13; test14;
		 test15; test16; test17; test18; test19; test20; test21;
		 test22; test23; test24 |]
      
  let run_tests () =
    let suc = ref 0
    and fal = ref 0
    and abo = ref 0
    and start = 13
    and fin = 13 in
    for i = max 0 (start - 1) to min (fin - 1) (Array.length tests - 1) do
      print_string "Test ";
      print_int (i + 1);
      begin
	try
	  if tests.(i) () then
	    (print_string " : Pass";
	     suc := !suc + 1)
	  else
	    (print_string " : Fail";
	     fal := !fal + 1)
	with
	  exn ->
	    (print_string " : Abort";
	     abo := !abo + 1; raise exn)
      end;
      print_newline ()
    done;
    print_string "Passed: ";
    print_int !suc;
    print_string "\nFailed: ";
    print_int !fal;
    print_string "\nAborted: ";
    print_int !abo;
    print_newline ()

end;; (* module Tester *)


module Dtests = Tester (DBlop) (DSparse) (DHss);;
module Stests = Tester (SBlop) (SSparse) (SHss);;
module Ztests = Tester (ZBlop) (ZSparse) (ZHss);;
module Ctests = Tester (CBlop) (CSparse) (CHss);;


let main () =
  begin
    print_string "Single Precision Tests:\n";
    Stests.run_tests ();
    print_newline ();

    print_string "Double Precision Tests:\n";
    Dtests.run_tests ();
    print_newline ();
    
    print_string "Double Complex Tests:\n";
    Ztests.run_tests ();
    print_newline ();
    
    print_string "Single Complex Tests:\n";
    Ctests.run_tests ();
    print_newline ();
  end;;


main ();;
