(*
  File: testPoisson2D.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 10.15.2004
  Email: shiv@ece.ucsb.edu
  Purpose: Poisson discretizer

   $Id: testPoisson2D.ml,v 1.1 2005/01/15 18:12:49 shiv Exp shiv $
*)

open Bigarray;;
open Utils;;
open Nla;;
open Chub;;
open Poisson2D;;


module Tester (Blop : BLOP)
    (Chub : CHUB with type matrix = Blop.matrix and type t = Blop.t)
    (Poisson2D : POISSON2D with type matrix = Blop.matrix) =
  struct

    open Blop
    open Chub

    let saveMatrix fname a =
      let fout = open_out_bin fname in
      let m = noOfRows a
      and n = noOfCols a in
      for i = 1 to m do
	for j = 1 to n do
	  output_string fout (string_of_float (re (a $@ (i,j))) ^ " ")
	done;
	output_string fout "\n"
      done;
      close_out fout

    let norm2 a =
      try
	(singValues' (Blop.copy a)).{1}
      with
	exn -> begin
	  print_endline "singValues' failed; trying transpose";
	  let ac = copy a in
	  let e = randomMatrix (noOfRows a) (noOfCols a) in
	  scale_real' (norm1 a *. machine_epsilon *. 100.) e;
	  e |++| ac;
	  (singValues' (ac)).{1}
	  (* saveMatrix "/Users/shiv/a.txt" a; raise exn *)
	end
      

    (* Poor L2 norm estimator. It does not know singularity structure. *)
    let normL2 f =
      let px = noOfRows f
      and py = noOfCols f in
      let zla = Array1.create precision_real fortran_layout px
      and zra = Array1.create precision_real fortran_layout py in
      let lwts = chebQuadWts px
      and rwts = chebQuadWts py in
      for i = 1 to px do
	zla.{i} <- sqrt (re (lwts $@ (1,i)))
      done;
      for i = 1 to py do
	zra.{i} <- 1.0 /. sqrt (re (rwts $@ (1,i)))
      done;
      let fc = Blop.copy f in
      colScale_real' fc zra;
      rowScale_real' zla fc;
      norm2 fc
      
     
    let chebQuadWts2D px py =
      let qx = chebQuadWts px
      and qy = chebQuadWts py in
      tensor qx qy (* please check *)


    (* Two D L2 norm estimator. *)
    let normL22D mx my nx ny f =
      let m = noOfRows f
      and n = noOfCols f in
      assert (m = mx * my && n = nx * ny);
      let zla = Array1.create precision_real fortran_layout m
      and zra = Array1.create precision_real fortran_layout n in
      let lwts = chebQuadWts2D mx my
      and rwts = chebQuadWts2D nx ny in
      for i = 1 to m do
	zla.{i} <- sqrt (re (lwts $@ (1,i)))
      done;
      for i = 1 to n do
	zra.{i} <- 1.0 /. sqrt (re (rwts $@ (1,i)))
      done;
      let fc = Blop.copy f in
      colScale_real' fc zra;
      rowScale_real' zla fc;
      norm2 fc


    (* Two D rectangular surface to rectangle L2 norm estimator. *)
    let normSurfaceL22D mx my pt pr pb pl f =
      let m = noOfRows f
      and n = noOfCols f in
      assert (m = mx * my && n = pt + pr + pb + pl);
      let zla = Array1.create precision_real fortran_layout m
      and zra = Array1.create precision_real fortran_layout n in
      let lwts = chebQuadWts2D mx my
      and rwtst = chebQuadWts pt
      and rwtsr = chebQuadWts pr
      and rwtsb = chebQuadWts pb
      and rwtsl = chebQuadWts pl in
      for i = 1 to m do
	zla.{i} <- sqrt (re (lwts $@ (1, i)))
      done;
      for i = 1 to pt do
	zra.{i} <- 1.0 /. sqrt (re (rwtst $@ (1,i)))
      done;
      for i = 1 to pr do
	zra.{pt+i} <- 1.0 /. sqrt (re (rwtsr $@ (1,i)))
      done;
      for i = 1 to pb do
	zra.{pt+pr+i} <- 1.0 /. sqrt (re (rwtsb $@ (1,i)))
      done;
      for i = 1 to pl do
	zra.{pt+pr+pb+i} <- 1.0 /. sqrt (re (rwtsl $@ (1,i)))
      done;
      let fc = Blop.copy f in
      colScale_real' fc zra;
      rowScale_real' zla fc;
      norm2 fc
      
      
    let test1 () =
      for p = 3 to 10 do
	let pd = Poisson2D.make_pd p in
	let dxx = Poisson2D.partialXX (p+1) (p+1) in
	let dyy = Poisson2D.partialYY (p+1) (p+1) in
	let dx = Poisson2D.partialX (p+1) (p+1) in
	let dy = Poisson2D.partialY (p+1) (p+1) in
	let dxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$
	  (Poisson2D.projectDownX 1 p (p+1)) *$ dx *$ pd in
	let dypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$
	  (Poisson2D.projectDownY 1 (p+1) p) *$ dy *$ pd in
	let dxxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$ dxx *$ pd in
	let dyypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$ dyy *$ pd in
	let lap = dxxpd +$ dyypd in
	let p2 = (p - 1) * (p - 1)
	and lc = noOfCols pd in
	assert (lc = p2 + 4*p);
	Printf.printf
	  "p = %i norms pd = %g dxx.pd = %g dyy.pd = %g lap = %g dx = %g dy = %g\n"
	  p
	  (norm2 (pd    $ (All, 1 <-> p2)))
	  (norm2 (dxxpd $ (All, 1 <-> p2)))
	  (norm2 (dyypd $ (All, 1 <-> p2)))
	  (norm2 (lap   $ (All, 1 <-> p2)))
	  (norm2 (dxpd  $ (All, 1 <-> p2)))
	  (norm2 (dypd  $ (All, 1 <-> p2)));
	Printf.printf "p = %i pd_bc = %g dxx.pd_bc = %g dyy.pd_bc = %g lap_bc = %g dx_bc = %g dy_bc = %g\n"
	  p
	  (norm2 (pd    $ (All, p2+1 <-> lc)))
	  (norm2 (dxxpd $ (All, p2+1 <-> lc)))
	  (norm2 (dyypd $ (All, p2+1 <-> lc)))
	  (norm2 (lap   $ (All, p2+1 <-> lc)))
	  (norm2 (dxpd  $ (All, p2+1 <-> lc)))
	  (norm2 (dypd  $ (All, p2+1 <-> lc)));
	(* saveMatrix ("dxxpd" ^ (string_of_int p)) dxxpd; *)
      done;
      true


    (* Same as test1 but in l2 norm. *)
    let test2 () =
      for p = 3 to 10 do
	let pd = Poisson2D.make_pd p in
	let dxx = Poisson2D.partialXX (p+1) (p+1) in
	let dyy = Poisson2D.partialYY (p+1) (p+1) in
	let dx = Poisson2D.partialX (p+1) (p+1) in
	let dy = Poisson2D.partialY (p+1) (p+1) in
	let dxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$
	  (Poisson2D.projectDownX 1 p (p+1)) *$ dx *$ pd in
	let dypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$
	  (Poisson2D.projectDownY 1 (p+1) p) *$ dy *$ pd in
	let dxxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$ dxx *$ pd in
	let dyypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$ dyy *$ pd in
	let lap = dxxpd +$ dyypd in
	let p2 = (p - 1) * (p - 1)
	and lc = noOfCols pd in
	assert (lc = p2 + 4*p);
	Printf.printf
	  "p = %i norms pd = %g dxx.pd = %g dyy.pd = %g lap = %g dx = %g dy = %g\n"
	  p
	  (normL22D (p+1) (p+1) (p-1) (p-1) (pd    $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dxxpd $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dyypd $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (lap   $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dxpd  $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dypd  $ (All, 1 <-> p2)));
	Printf.printf "p = %i pd_bc = %g dxx.pd_bc = %g dyy.pd_bc = %g lap_bc = %g dx_bc = %g dy_bc = %g\n"
	  p
	  (normSurfaceL22D (p+1) (p+1) p p p p (pd    $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dxxpd $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dyypd $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (lap   $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dxpd  $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dypd  $ (All, p2+1 <-> lc)));
	(* saveMatrix ("dxxpd" ^ (string_of_int p)) dxxpd; *)
      done;
      true

(** Conclusion from tests 1 & 2: PD is bounded in L2 norm. However the
    first and second derivative operators are only bounded in L2 for
    the portion that maps the Laplacian to the derivative. That
    portion that maps the Dirichlet data to the derivative seems to
    grow like O(3p) for the first derivative and like O(100p) for the
    second derivative. This must be natural since we do not require
    any smoothness from the Dirichlet data (?). A proof of this can be
    inferred from exercise 1--3 of Folland's Intro. to PDE on page 91
    (section 2.F) of the second edition.  From there it is clear that
    there exists a continuous function on the circle that is not the
    restriction of a harmonic function in the unit disk with bounded
    L2 norm for its gradient. So in general this map cannot have
    bounded L2 norm. What about Neumann data? *)


    (* Same as test1 except that we use P_N instead of P_D. *)
    let test3 () =
      for p = 3 to 10 do
	let pn = Poisson2D.make_pn p in
	let dxx = Poisson2D.partialXX p p in
	let dyy = Poisson2D.partialYY p p in
	let dx = Poisson2D.partialX p p in
	let dy = Poisson2D.partialY p p in
	let dxpn = (Poisson2D.projectDownY 2 (p-2) p) *$
	  (Poisson2D.projectDownX 1 (p-1) p) *$ dx *$ pn in
	let dypn = (Poisson2D.projectDownX 2 p (p-2)) *$
	  (Poisson2D.projectDownY 1 p (p-1)) *$ dy *$ pn in
	let dxxpn = (Poisson2D.projectDownY 2 (p-2) p) *$ dxx *$ pn in
	let dyypn = (Poisson2D.projectDownX 2 p (p-2)) *$ dyy *$ pn in
	let lap = dxxpn +$ dyypn in
	let p2 = (p - 2) * (p - 2)
	and lc = noOfCols pn in
	assert (lc = p2 + 4*p + 1);
	Printf.printf
	  "p = %i norms pn = %g dxx.pn = %g dyy.pn = %g lap = %g dx = %g dy = %g\n"
	  p
	  (norm2 (pn    $ (All, 1 <-> p2)))
	  (norm2 (dxxpn $ (All, 1 <-> p2)))
	  (norm2 (dyypn $ (All, 1 <-> p2)))
	  (norm2 (lap   $ (All, 1 <-> p2)))
	  (norm2 (dxpn  $ (All, 1 <-> p2)))
	  (norm2 (dypn  $ (All, 1 <-> p2)));
	Printf.printf "p = %i pn_bc = %g dxx.pn_bc = %g dyy.pn_bc = %g lap_bc = %g dx_bc = %g dy_bc = %g\n"
	  p
	  (norm2 (pn    $ (All, p2+1 <-> lc-1)))
	  (norm2 (dxxpn $ (All, p2+1 <-> lc-1)))
	  (norm2 (dyypn $ (All, p2+1 <-> lc-1)))
	  (norm2 (lap   $ (All, p2+1 <-> lc-1)))
	  (norm2 (dxpn  $ (All, p2+1 <-> lc-1)))
	  (norm2 (dypn  $ (All, p2+1 <-> lc-1)));
	(* saveMatrix ("dxxpd" ^ (string_of_int p)) dxxpd; *)
      done;
      true

    (* Same as test3 except that we use l2 norm. *)
    let test4 () =
      for p = 3 to 10 do
	let pn = Poisson2D.make_pn p in
	let dxx = Poisson2D.partialXX p p in
	let dyy = Poisson2D.partialYY p p in
	let dx = Poisson2D.partialX p p in
	let dy = Poisson2D.partialY p p in
	let dxpn = (Poisson2D.projectDownY 2 (p-2) p) *$
	  (Poisson2D.projectDownX 1 (p-1) p) *$ dx *$ pn in
	let dypn = (Poisson2D.projectDownX 2 p (p-2)) *$
	  (Poisson2D.projectDownY 1 p (p-1)) *$ dy *$ pn in
	let dxxpn = (Poisson2D.projectDownY 2 (p-2) p) *$ dxx *$ pn in
	let dyypn = (Poisson2D.projectDownX 2 p (p-2)) *$ dyy *$ pn in
	let lap = dxxpn +$ dyypn in
	let p2 = (p - 2) * (p - 2)
	and lc = noOfCols pn in
	assert (lc = p2 + 4*p + 1);
	Printf.printf
	  "p = %i norms pn = %g dxx.pn = %g dyy.pn = %g lap = %g dx = %g dy = %g\n"
	  p
	  (normL22D p     p     (p-2) (p-2) (pn    $ (All, 1 <-> p2)))
	  (normL22D (p-2) (p-2) (p-2) (p-2) (dxxpn $ (All, 1 <-> p2)))
	  (normL22D (p-2) (p-2) (p-2) (p-2) (dyypn $ (All, 1 <-> p2)))
	  (normL22D (p-2) (p-2) (p-2) (p-2) (lap   $ (All, 1 <-> p2)))
	  (normL22D (p-2) (p-2) (p-2) (p-2) (dxpn  $ (All, 1 <-> p2)))
	  (normL22D (p-2) (p-2) (p-2) (p-2) (dypn  $ (All, 1 <-> p2)));
	Printf.printf "p = %i pn_bc = %g dxx.pn_bc = %g dyy.pn_bc = %g lap_bc = %g dx_bc = %g dy_bc = %g\n"
	  p
	  (normSurfaceL22D p     p     p p p p (pn    $ (All, p2+1 <-> lc-1)))
	  (normSurfaceL22D (p-2) (p-2) p p p p (dxxpn $ (All, p2+1 <-> lc-1)))
	  (normSurfaceL22D (p-2) (p-2) p p p p (dyypn $ (All, p2+1 <-> lc-1)))
	  (normSurfaceL22D (p-2) (p-2) p p p p (lap   $ (All, p2+1 <-> lc-1)))
	  (normSurfaceL22D (p-2) (p-2) p p p p (dxpn  $ (All, p2+1 <-> lc-1)))
	  (normSurfaceL22D (p-2) (p-2) p p p p (dypn  $ (All, p2+1 <-> lc-1)));
	(* saveMatrix ("dxxpd" ^ (string_of_int p)) dxxpd; *)
      done;
      true

(* Conclusions from test3 and test4 for make_pn1 and make_pn. make_pn1
   does not work eiether for u(0,0) or int u as the extra boundary
   condition. The one that seems to work is to require the (0,0)
   Chebyshev coefficient to be zero. But that depends on the order of
   the method p. So we now try make_pnd. *)

    (* Same as test1, except using make_pnd. *)
    let test5 () =
      for p = 3 to 10 do
	let gamma = float (4 * p * p) in
	let pd = Poisson2D.make_pnd gamma p in
	let dxx = Poisson2D.partialXX (p+1) (p+1) in
	let dyy = Poisson2D.partialYY (p+1) (p+1) in
	let dx = Poisson2D.partialX (p+1) (p+1) in
	let dy = Poisson2D.partialY (p+1) (p+1) in
	let dxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$
	  (Poisson2D.projectDownX 1 p (p+1)) *$ dx *$ pd in
	let dypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$
	  (Poisson2D.projectDownY 1 (p+1) p) *$ dy *$ pd in
	let dxxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$ dxx *$ pd in
	let dyypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$ dyy *$ pd in
	let lap = dxxpd +$ dyypd in
	let p2 = (p - 1) * (p - 1)
	and lc = noOfCols pd in
	assert (lc = p2 + 4*p);
	Printf.printf
	  "p = %i norms pd = %g dxx.pd = %g dyy.pd = %g lap = %g dx = %g dy = %g\n"
	  p
	  (norm2 (pd    $ (All, 1 <-> p2)))
	  (norm2 (dxxpd $ (All, 1 <-> p2)))
	  (norm2 (dyypd $ (All, 1 <-> p2)))
	  (norm2 (lap   $ (All, 1 <-> p2)))
	  (norm2 (dxpd  $ (All, 1 <-> p2)))
	  (norm2 (dypd  $ (All, 1 <-> p2)));
	Printf.printf "p = %i pd_bc = %g dxx.pd_bc = %g dyy.pd_bc = %g lap_bc = %g dx_bc = %g dy_bc = %g\n"
	  p
	  (norm2 (pd    $ (All, p2+1 <-> lc)))
	  (norm2 (dxxpd $ (All, p2+1 <-> lc)))
	  (norm2 (dyypd $ (All, p2+1 <-> lc)))
	  (norm2 (lap   $ (All, p2+1 <-> lc)))
	  (norm2 (dxpd  $ (All, p2+1 <-> lc)))
	  (norm2 (dypd  $ (All, p2+1 <-> lc)));
	(* saveMatrix ("dxxpd" ^ (string_of_int p)) dxxpd; *)
      done;
      true


    (* Same as test5 but in l2 norm. *)
    let test6 () =
      for p = 3 to 10 do
	let gamma = float (4 * p * p) in
	let pd = Poisson2D.make_pnd gamma p in
	let dxx = Poisson2D.partialXX (p+1) (p+1) in
	let dyy = Poisson2D.partialYY (p+1) (p+1) in
	let dx = Poisson2D.partialX (p+1) (p+1) in
	let dy = Poisson2D.partialY (p+1) (p+1) in
	let dxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$
	  (Poisson2D.projectDownX 1 p (p+1)) *$ dx *$ pd in
	let dypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$
	  (Poisson2D.projectDownY 1 (p+1) p) *$ dy *$ pd in
	let dxxpd = (Poisson2D.projectDownY 2 (p-1) (p+1)) *$ dxx *$ pd in
	let dyypd = (Poisson2D.projectDownX 2 (p+1) (p-1)) *$ dyy *$ pd in
	let lap = dxxpd +$ dyypd in
	let p2 = (p - 1) * (p - 1)
	and lc = noOfCols pd in
	assert (lc = p2 + 4*p);
	Printf.printf
	  "p = %i norms pd = %g dxx.pd = %g dyy.pd = %g lap = %g dx = %g dy = %g\n"
	  p
	  (normL22D (p+1) (p+1) (p-1) (p-1) (pd    $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dxxpd $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dyypd $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (lap   $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dxpd  $ (All, 1 <-> p2)))
	  (normL22D (p-1) (p-1) (p-1) (p-1) (dypd  $ (All, 1 <-> p2)));
	Printf.printf "p = %i pd_bc = %g dxx.pd_bc = %g dyy.pd_bc = %g lap_bc = %g dx_bc = %g dy_bc = %g\n"
	  p
	  (normSurfaceL22D (p+1) (p+1) p p p p (pd    $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dxxpd $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dyypd $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (lap   $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dxpd  $ (All, p2+1 <-> lc)))
	  (normSurfaceL22D (p-1) (p-1) p p p p (dypd  $ (All, p2+1 <-> lc)));
	(* saveMatrix ("dxxpd" ^ (string_of_int p)) dxxpd; *)
      done;
      true

(* Conclusions so far: make_pnd seems to work best, but with gamma = p^2 *)


    let tests = [| test1; test2; test3; test4; test5; test6 |]
      
    let run_tests () =
      let suc = ref 0
      and fal = ref 0
      and abo = ref 0
      and start = 5
      and fin = 6 in
      for i = max 0 (start - 1) to min (fin - 1) (Array.length tests - 1) do
	print_string "Test ";
	print_int (i + 1);
	begin
	  try
	    if tests.(i) () then
	      (print_string " : Pass";
	       suc := !suc + 1)
	    else
	      (print_string " : Fail";
	       fal := !fal + 1)
	  with
	    exn ->
	      (print_string " : Abort";
	       abo := !abo + 1; raise exn)
	end;
	print_newline ()
      done;
      print_string "Passed: ";
      print_int !suc;
      print_string "\nFailed: ";
      print_int !fal;
      print_string "\nAborted: ";
      print_int !abo;
      print_newline ()
	
  end;; (* module Tester *)


module Dtests = Tester (DBlop) (DChub) (DPoisson2D);;
module Stests = Tester (SBlop) (SChub) (SPoisson2D);;
module Ztests = Tester (ZBlop) (ZChub) (ZPoisson2D);;
module Ctests = Tester (CBlop) (CChub) (CPoisson2D);;


let main () =
  begin
    print_string "Double Precision Tests:\n";
    Dtests.run_tests ();
    print_newline ();
    
  (*  print_string "Single Precision Tests:\n";
    Stests.run_tests ();
    print_newline ();
    
    print_string "Double Complex Tests:\n";
    Ztests.run_tests ();
    print_newline ();
    
    print_string "Single Complex Tests:\n";
    Ctests.run_tests ();
    print_newline (); *)
  end;;


main ();;
