(*
  File: timeFmm.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 6.05.2006
  Email: shiv@ece.ucsb.edu
  Purpose: time fmm.ml

   $Id$
*)

open Bigarray;;
open Utils;;
open Nla;;
open Sparse;;
open Fmm;;


module Timer (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix)
    (Fmm : FMM with type matrix = Blop.matrix and type t = Blop.t
    and type transform = Blop.transform and type sparse = Sparse.sparse) =
struct

  open Blop
  open Fmm

  let kernel alpha omega xs sr m sc n a =
    assert (m = noOfRows a && n = noOfCols a);
    assert (alpha >= -1.0);
    let dim = Array.length xs.(0) in
    let kernelFn x y =
      fromFloat (
	let d =
	  let acc = ref 0.0 in
	  for i = 0 to dim-1 do
	    acc := !acc +. abs_float (x.(i) -. y.(i))
	  done;
	  !acc
	in
	if d > 0.0 then
	  (* sin *) cos (omega *. d) *. (d ** alpha)
	else if alpha = -1.0 then 1.0 +. omega
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let randomPoint n a b =
    Array.init n (fun i -> a.(i) +. Random.float (b.(i) -. a.(i)))

  let mkInt n1 n2 =
    let rec mi i acc =
      if i < n1 then acc else mi (i-1) (i :: acc)
    in
    mi n2 []

  let time1 () =
    let ms = [| 1000; 2000; 4000 |] in
    let tol = sqrt (sqrt machine_epsilon) in
    let nrhs = 1 in
    for dim = 1 to 3 do
      for mi = 0 to Array.length ms - 1 do
	let fmm, m, n, mt =
	  let a = Array.make dim 0.0
	  and b = Array.make dim 1.0
	  and mt = ms.(mi)
	  and alpha = 0.5
	  and omega = acos 0.0 in
	  let p = int_of_float (10.0 ** float dim) in
	  let xs = List.map (fun _ -> randomPoint dim a b) (mkInt 1 mt) in
	  let ndTree = nestedDissection p xs in
	  let ys = Array.of_list (pointsFromNDtree ndTree) in
	  let fmmSpine = fmmSpineFromND ndTree in
	  let (m, n) = fmmSize fmmSpine in
	  let a = rawMatrix m n in
	  kernel alpha omega ys 1 m 1 n a;
	  let aFmm = fromFn tol fmmSpine
	      (fun sr m sc n dst ->
		dst <-$ (a $ (sr <-> sr+m-1, sc <-> sc+n-1)))
	  in
	  (aFmm, m, n, mt)
	in
	let x = randomMatrix n nrhs in
	let rhs = randomMatrix m nrhs in
	Gc.compact ();
	let t0 = Sys.time () in
	fastSolve' fmm rhs x;
	let tt = Sys.time () -. t0 in
	Printf.printf "dim = %i m = %i tol = %g time = %g" dim mt tol tt;
	print_newline ()
      done
    done
(*
  Double Precision Times: shiv-pb-g4
  toProperForm' vHss
  -inline 1000 -noassert
  cos (3 pi / 4 |x-y|) /. sqrt |x-y|, 0 <= x,y <= 1
  p = 100
  dim = 1 m = 1000 tol = 0.00012207 time = 0.179068
  dim = 1 m = 2000 tol = 0.00012207 time = 0.636194
  dim = 1 m = 4000 tol = 0.00012207 time = 2.248
  dim = 2 m = 1000 tol = 0.00012207 time = 28.2931
  dim = 2 m = 2000 tol = 0.00012207 time = 226.588
  dim = 2 m = 4000 tol = 0.00012207 time = 1714.45
  dim = 3 m = 1000 tol = 0.00012207 time = 28.2424
  dim = 3 m = 2000 tol = 0.00012207 time = 228.746
  dim = 3 m = 4000 tol = 0.00012207 time = 1812.68

  Double Precision Times: gimli
  toProperForm' vHss
  -inline 1000 -noassert
  cos (pi / 2 |x-y|) *. sqrt |x-y|, 0 <= x,y <= 1
  p = 100
  dim = 1 m = 1000 tol = 0.00012207 time = 0.152066
  dim = 1 m = 2000 tol = 0.00012207 time = 0.4281
  dim = 1 m = 4000 tol = 0.00012207 time = 1.16305
  dim = 2 m = 1000 tol = 0.00012207 time = 45.6122
  dim = 2 m = 2000 tol = 0.00012207 time = 380.929
  dim = 2 m = 4000 tol = 0.00012207 time = 3153.94
  dim = 3 m = 1000 tol = 0.00012207 time = 44.4876
  dim = 3 m = 2000 tol = 0.00012207 time = 381.17
  dim = 3 m = 4000 tol = 0.00012207 time = 3317.46
*)

  let timers = [| time1 |]

  let run_timers () =
    let first_timer = 1
    and last_timer = 100 in
    (* Random.self_init (); *)
    for i = max 0 (first_timer - 1) to
      min (Array.length timers) last_timer - 1 do
	timers.(i) ()
    done;
    print_newline ()
      

end;; (* module Timer *)


module Dtimers = Timer (DBlop) (DSparse) (DFmm);;
module Stimers = Timer (SBlop) (SSparse) (SFmm);;
module Ztimers = Timer (ZBlop) (ZSparse) (ZFmm);;
module Ctimers = Timer (CBlop) (CSparse) (CFmm);;


let main () =
  begin
(*    print_string "Single Precision Times:\n";
    Stimers.run_timers ();
    print_newline ();
*)
    print_string "Double Precision Times:\n";
    Dtimers.run_timers ();
    print_newline ();
(*    
    print_string "Double Complex Times:\n";
    Ztimers.run_timers ();
    print_newline ();
    
    print_string "Single Complex Times:\n";
    Ctimers.run_timers ();
    print_newline ();
*)
  end;;


main ();;
