(*
  File: timeHss.ml
  Author: Shivkumar Chandrasekaran
  Email: shiv@ece.ucsb.edu
  Date: 7.16.2002
  Package: CamlFloat
  Purpose: Timings for hss stuff
*)

open Bigarray;;
open Nla;;
open Sparse;;
open Utils;;
open Hss;;
open NestedDissection;;

module Timer (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix) 
    (Hss : HSS with type matrix = Blop.matrix and type t = Blop.t
    and type sparse = Sparse.sparse) =
  struct

  open Blop
  open Hss

  (* Begin keep in sync with hss.ml *)
  type leaf_split = { m : int; n : int; k : int; l : int }
  and node_split = { ku : int; lv : int }
  and hss_spine = (leaf_split, node_split) btree

  let randomHss spine =
    let rec rec1 spine =
      match spine with
	  Node (a, lSpine, rSpine) ->
	    let (lHss, k0, l0, m0, n0) = rec1 lSpine
	    and (rHss, k1, l1, m1, n1) = rec1 rSpine in
	    (
	      Node
	      (
		{
		  ml = m0; mr = m1; nl = n0; nr = n1;
		  bu = randomMatrix k0 l1;
		  bl = randomMatrix k1 l0;
		  rl = randomMatrix k0 a.ku;
		  rr = randomMatrix k1 a.ku;
		  wl = randomMatrix l0 a.lv;
		  wr = randomMatrix l1 a.lv;
		},
		lHss,
		rHss
	      ),
	      a.ku, a.lv, m0+m1, n0+n1
	    )
	| Leaf a ->
	    (
	      Leaf
	      {
		u = randomMatrix a.m a.k;
		v = randomMatrix a.n a.l;
		d = randomMatrix a.m a.n
	      },
	      a.k, a.l, a.m, a.n
	    )
    in
    let (hss, _, _, _, _) = rec1 spine in hss

  let pi = 3.141592653589793238462643

  (* Left to right in natural order *)
  let chebZeros n =
    Array.init n (fun i -> cos ((float (n-i-1) +. 0.5) *. pi /. float n))

  let skewness atree =
    let rec sk atree =
      match atree with
	Leaf _ -> (0, 0)
      | Node (_, ltree, rtree) ->
	  let (lmax, lmin) = sk ltree
	  and (rmax, rmin) = sk rtree in
	  (1 + max lmax rmax, 1 + min lmin rmin)
    in
    let (maxp, minp) = sk atree in
    if maxp = minp then 1.0
    else float maxp /. float minp

  let rec to_hss_spine nd_spine =
    match nd_spine with
	Leaf _ -> Leaf ()
      | Node ((m1, m2), lSpine, rSpine) ->
	  Node ({mLeft = m1; mRight = m2; nLeft = m1; nRight = m2},
	  to_hss_spine lSpine, to_hss_spine rSpine)

  let rec get_hss_spine hss =
    match hss with
	Leaf a -> Leaf {m = noOfRows a.d; n = noOfCols a.d; k =
      noOfCols a .u; l = noOfCols a.v }
      | Node (a, lHss, rHss) -> Node ({ku = noOfCols a.rl; lv =
	  noOfCols a.wl}, get_hss_spine lHss, get_hss_spine rHss)

  let rec print_spine hss_spine =
    match hss_spine with
	Leaf a -> Printf.printf "Leaf {m=%i; n=%i; k=%i; l=%i}" a.m
	  a.n a.k a.l
      | Node (a, lSpine, rSpine) ->
	  Printf.printf "Node ({ku=%i; lv=%i},\n" a.ku a.lv;
	  print_spine lSpine; Printf.printf ",\n";
	  print_spine rSpine; Printf.printf ")"


  let kernel1D_nosc alpha xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    assert (alpha >= -1.0);
    let kernelFn x y = fromFloat (abs_float (x -. y) ** alpha) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let kernel1D alpha omega xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    assert (alpha >= -1.0);
    let kernelFn x y =
      fromFloat (
	let d = abs_float (x -. y) in
	if d > 0.0 then
	  sin (omega *. d) *. (d ** alpha)
	else if alpha = -1.0 then 1.0 +. omega
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let kernel2D alpha omega xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    let sq x = x *. x in
    let dist x y = sqrt (sq (x.(0) -. y.(0)) +. sq (x.(1) -. y.(1))) in
    let kernelFn x y =
      fromFloat (
	let d = dist x y in
	if d > 0.0 then
	  let d = log d in sin (omega *. d) *. (abs_float d ** alpha)
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let kernel3D alpha omega xs sr lr sc lc a =
    assert (lr - sr + 1 = noOfRows a && lc - sc + 1 = noOfCols a);
    let sq x = x *. x in
    let dist x y = sqrt (sq (x.(0) -. y.(0)) +. sq (x.(1) -. y.(1))
    +. sq (x.(2) -. y.(2))) in
    let kernelFn x y =
      fromFloat (
	let d = dist x y in
	if d > 0.0 then sin (omega *. d) *. (d ** alpha)
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let pie = 2.0 *. acos 0.0

  let elasticG3D nu g xs x y i j =
    let r = Array.make 3 0.0 in
    r.(0) <- xs.(y).(0) -. xs.(x).(0);
    r.(1) <- xs.(y).(1) -. xs.(x).(1);
    r.(2) <- xs.(y).(2) -. xs.(x).(2);
    let rl = sqrt (r.(0) *. r.(0) +. r.(1) *. r.(1) +. r.(2) *. r.(2))
    in
    if rl > 0.0 then
      ((if i <> j then (3.0 -. 4.0 *. nu) /. rl else 0.0) +. (r.(i)
      *. r.(j) /. (rl *. rl *. rl))) /. (16.0 *. pie *. (1.0 -. nu)
      *. g)
    else
      float (Array.length xs)

  let elasticH3D nu xs ns x y i j =
    let r = Array.make 3 0.0 in
    r.(0) <- xs.(y).(0) -. xs.(x).(0);
    r.(1) <- xs.(y).(1) -. xs.(x).(1);
    r.(2) <- xs.(y).(2) -. xs.(x).(2);
    let rl = sqrt (r.(0) *. r.(0) +. r.(1) *. r.(1) +. r.(2) *. r.(2))
    in
    let rl2 = rl *. rl in
    let rl3 = rl2 *. rl in
    let rl4 = rl3 *. rl in
    if rl > 0.0 then
      (1.0 -. 2.0 *. nu) *. ((r.(i) *. ns.(y).(j)
      -. r.(j)*.ns.(y).(i)) /. rl3 -. ((if i <> j
      then (1.0 /. rl2) else 0.0) +. (3.0 *. r.(i) *. r.(j)
      /. ((1.0 -. 2.0 *. nu) *. rl4)))
      *. ((r.(0)*.ns.(y).(0) +. r.(1)*.ns.(y).(1) +. r.(2)*.ns.(y).(2))
      /. rl)) /. (8.0 *. pie *. (1.0 -. nu))
    else
      float (Array.length xs)

  let elasticC3D nu g xs ns ds x y i j =
    assert (0 <= x && x < Array.length xs);
    assert (0 <= y && y < Array.length xs);
    assert (0 <= i && i < 3);
    assert (0 <= j && j < 3);
    assert (3*y+j < Array.length ds);
    fromFloat (elasticH3D nu xs ns x y i j +.
      elasticG3D nu g xs x y i j *. ds.(3*y+j))

  let kernel3Delastic nu g xs ns ds sr lr sc lc a =
    let kFn i j =
      let ai = sr + i - 1
      and aj = sc + j - 1 in
      let xx = (ai - 1) / 3
      and xy = (aj - 1) / 3
      and xi = (ai - 1) mod 3
      and yj = (aj - 1) mod 3 in
      elasticC3D nu g xs ns ds xx xy xi yj in
    fun2mat' kFn a
      (* End keep in sync with testHss.ml *)

  let time1 () =
    print_endline "1D Kernel timings";
    let ns = [| 10; 20; 30; 40 |]
    and nrhss = [| 1; 2; 3; 4 |]
    and alphas = [| -1.0; -0.5; 0.0; 0.5 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0; 3.0
      *. acos 0.0 |]
    and ps = [| 1; 2; 3; 4 |] in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      let n = ns.(ni)
	      and nrhs = nrhss.(nrhsi)
	      and tol = sqrt machine_epsilon
	      and alpha = alphas.(alphai)
	      and p = ps.(pi)
	      and omega = omegas.(omegai) in
	      Printf.printf "n=%i alpha=%g omega=%g nrhs=%i p=%i " n
		alpha omega nrhs p;
	      let xsa = Array.init n (fun _ -> Random.float 1.0) in
	      let xs = Array.to_list xsa in
	      let iTree = prune_tree (dissect1D xs 0.0 1.0) p in
	      let ps = Array.of_list (breadthFirstLeafData iTree) in
	      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
	      let hss_spine = to_hss_spine iTree in
	      let hss = fromFn hss_spine (kernel1D alpha omega xs) tol in
	      let b = randomMatrix n nrhs
	      and x = rawMatrix n nrhs in
	      let t1 = Sys.time () in
	      fastSolve' hss b x;
	      let tot = Sys.time () -. t1 in
	      Printf.printf "time=%g" tot; print_newline ()
	    done
	  done
	done
      done
    done

  let time2 () =
    print_endline "2D Kernel timings";
    let ns = [| 20; 40 |]
    and nrhss = [| 1; 4 |]
    and alphas = [| 1.0; 0.5; 2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 1; 4 |] in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      let n = ns.(ni)
	      and nrhs = nrhss.(nrhsi)
	      and tol = sqrt machine_epsilon
	      and alpha = alphas.(alphai)
	      and p = ps.(pi)
	      and omega = omegas.(omegai) in
	      Printf.printf "n=%i alpha=%g omega=%g nrhs=%i p=%i " n
		alpha omega nrhs p;
	      let xsa = Array.init n (fun _ -> Array.init 2
		(fun _ -> Random.float 1.0)) in
	      let xs = Array.to_list xsa in
	      let iTree = prune_tree (dissectND xs [| 0.0; 0.0 |]
		[| 1.0; 1.0 |] ) p in
	      let ps = Array.of_list (breadthFirstLeafData iTree) in
	      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
	      let hss_spine = to_hss_spine iTree in
	      let hss = fromFn hss_spine (kernel2D alpha omega xs) tol in
	      let b = randomMatrix n nrhs
	      and x = rawMatrix n nrhs in
	      let t1 = Sys.time () in
	      fastSolve' hss b x;
	      let tot = Sys.time () -. t1 in
	      Printf.printf "time=%g" tot; print_newline ()
	    done
	  done
	done
      done
    done

  let time3 () =
    print_endline "3D Kernel timings";
    let ns = [| 100; 200 |]
    and nrhss = [| 1 |]
    and alphas = [| -1.0; -0.5; -2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 4; 8 |] in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      let n = ns.(ni)
	      and nrhs = nrhss.(nrhsi)
	      and tol = sqrt machine_epsilon
	      and alpha = alphas.(alphai)
	      and p = ps.(pi)
	      and omega = omegas.(omegai) in
	      Printf.printf "n=%i alpha=%g omega=%g nrhs=%i p=%i " n
		alpha omega nrhs p;
	      let xsa = Array.init n (fun _ -> Array.init 3
		(fun _ -> Random.float 1.0)) in
	      let xs = Array.to_list xsa in
	      let iTree = prune_tree (dissectND xs [| 0.0; 0.0; 0.0 |]
		[| 1.0; 1.0; 1.0 |] ) p in
	      let ps = Array.of_list (breadthFirstLeafData iTree) in
	      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
	      let hss_spine = to_hss_spine iTree in
	      let hss = fromFn hss_spine (kernel3D alpha omega xs) tol in
	      let b = randomMatrix n nrhs
	      and x = rawMatrix n nrhs in
	      let t1 = Sys.time () in
	      fastSolve' (Hss.copy hss) (Blop.copy b) x;
	      let tot = Sys.time () -. t1 in
	      Printf.printf "time=%g" tot; print_newline ()
	    done
	  done
	done
      done
    done

  let time4 () =
    print_endline "2D Kernel on 1D curve timings";
    let ns = [| 20; 40; 80 |]
    and nrhss = [| 1; 4 |]
    and alphas = [| 1.0; 0.5; 2.0 |]
    and omegas = [| acos 0.0; 1.5 *. acos 0.0; 2.0 *. acos 0.0 |]
    and ps = [| 4; 8 |]
    and twoPi = 4.0 *. acos 0.0 in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      let n = ns.(ni)
	      and nrhs = nrhss.(nrhsi)
	      and tol = sqrt machine_epsilon
	      and alpha = alphas.(alphai)
	      and p = ps.(pi)
	      and omega = omegas.(omegai) in
	      Printf.printf "n=%i alpha=%g omega=%g nrhs=%i p=%i " n
		alpha omega nrhs p;
	      let xsa = Array.init n
		(fun _ ->
		  let theta = Random.float twoPi in
		  let x = Array.make 2 0.0 in
		  x.(0) <- cos theta;
		  x.(1) <- sin theta;
		  x) in
	      let xs = Array.to_list xsa in
	      let iTree = prune_tree (dissectND xs [| -1.0; -1.0 |]
		[| 1.0; 1.0 |] ) p in
	      let ps = Array.of_list (breadthFirstLeafData iTree) in
	      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
	      let hss_spine = to_hss_spine iTree in
	      let hss = fromFn hss_spine (kernel2D alpha omega xs) tol in
	      let b = randomMatrix n nrhs
	      and x = rawMatrix n nrhs in
	      let t1 = Sys.time () in
	      fastSolve' (Hss.copy hss) (Blop.copy b) x;
	      let tot = Sys.time () -. t1 in
	      Printf.printf "time=%g" tot; print_newline ()
	    done
	  done
	done
      done
    done

  let time5 () =
    print_endline "3D kernel 2D surface timings";
    let ns = [| 51200; 102400 |]
    and nrhss = [| 1 |]
    and alphas = [| -1.0 |]
    and omegas = [| 2.0 *. acos 0.0 |]
    and ps = [| 60; 70 |]
    and pie = 2.0 *. acos 0.0
    and twoPi = 4.0 *. acos 0.0 in
    for ni = 0 to Array.length ns - 1 do
      for alphai = 0 to Array.length alphas - 1 do
	for omegai = 0 to Array.length omegas - 1 do
	  for nrhsi = 0 to Array.length nrhss - 1 do
	    for pi = 0 to Array.length ps - 1 do
	      let n = ns.(ni)
	      and nrhs = nrhss.(nrhsi)
	      and tol = sqrt machine_epsilon
	      and alpha = alphas.(alphai)
	      and p = ps.(pi)
	      and omega = omegas.(omegai) in
	      Printf.printf "n=%i alpha=%g omega=%g nrhs=%i p=%i " n
		alpha omega nrhs p;
	      let xsa = Array.init n
		(fun _ ->
		  let theta = Random.float pie
		  and phi = Random.float twoPi
		  and x = Array.make 3 0.0 in
		  x.(0) <- sin theta *. sin phi;
		  x.(1) <- sin theta *. cos phi;
		  x.(2) <- cos theta;
		  x) in
	      let xs = Array.to_list xsa in
	      let iTree = prune_tree (dissectND xs [| -1.0; -1.0; -1.0 |]
		[| 1.0; 1.0; 1.0 |] ) p in
	      let ps = Array.of_list (breadthFirstLeafData iTree) in
	      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
	      let hss_spine = to_hss_spine iTree in
	      let hss = fromFn hss_spine (kernel3D alpha omega xs) tol in
	      let b = randomMatrix n nrhs
	      and x = rawMatrix n nrhs in
	      let t1 = Sys.time () in
	      fastSolve' (Hss.copy hss) (Blop.copy b) x;
	      let tot = Sys.time () -. t1 in
	      Printf.printf "time=%g" tot; print_newline ()
	    done
	  done
	done
      done
    done

  let time6 () =
    print_endline "3D elastic BEM timings";
    let ns = [| 300; 200; 100 |]
    and nrhss = [| 1; 4 |]
    and ps = [| 30; 12; 14 |]
    and nus = [| 0.4; 0.1; 0.0; 0.1 |]
    and lambdas = [| 10.0; 100.0; 1000.0 |] in
    for ni = 0 to Array.length ns - 1 do
      for nrhsi = 0 to Array.length nrhss - 1 do
	for pi = 0 to Array.length ps - 1 do
	  for nui = 0 to Array.length nus - 1 do
	    for lambdai = 0 to Array.length lambdas - 1 do
	      let n = ns.(ni)
	      and nrhs = nrhss.(nrhsi) in
	      let tol = sqrt (machine_epsilon *. 1000000.0 *. float n)
	      and p = ps.(pi) (* leaf size *)
	      and k = 3 (* vector size of unknowns in PDE. *)
	      and nu = nus.(nui)
	      and g = 1.0
	      and lambda = lambdas.(lambdai)
	      and twoPie = pie *. 2.0 in
	      Printf.printf "n=%i nrhs=%i p=%i nu=%g lambda=%g " n nrhs
		p nu lambda;
	      let m = n * k in
	      let xsa = Array.init n
		(fun _ ->
		  let theta = Random.float pie
		  and phi = Random.float twoPie
		  and x = Array.make 3 0.0 in
		  x.(0) <- sin theta *. sin phi;
		  x.(1) <- sin theta *. cos phi;
		  x.(2) <- cos theta;
		  x) in
	      let xs = Array.to_list xsa in
	      let ds = Array.init m
		(fun _ -> let heads = Random.bool () in
		if heads then lambda else 1.0 /. lambda) in
	      let oTree = prune_tree
		(dissectND xs [| -1.0; -1.0; -1.0 |] [| 1.0; 1.0; 1.0 |] ) p in
	      let iTree = inflate_partition_tree oTree k in
	      let ps = Array.of_list (breadthFirstLeafData oTree) in
	      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
	      let hss_spine = to_hss_spine iTree
	      and normals = xs in
	      let hss = fromFn hss_spine
		(kernel3Delastic nu g xs normals ds) tol in
	      let b = randomMatrix m nrhs
	      and x = rawMatrix m nrhs in
	      let t1 = Sys.time () in
	      fastSolve' (Hss.copy hss) (Blop.copy b) x;
	      let tot = Sys.time () -. t1 in
	      Printf.printf "time=%g" tot; print_newline ()
	    done
	  done
	done
      done
    done

  (* Timings for fast adaptive solver HSS paper. *)
  let time7 () =
    print_endline "1D Kernel timings";
    let ns = [| 256; 512; 1024; 2048; 4096; 8192; 8192*2; 8192*4; 8192*8;
		8192*16; 8192*32; 8192*64; 8192*128; 8192*256 |]
    and nrhs = 1
    and alpha = 0.5 in
    for ni = 0 to Array.length ns - 1 do
      let n = ns.(ni)
      and tol = sqrt machine_epsilon in
      let p = 5 + int_of_float (log (float n) /. log 2.) in
      (*let xsa = Array.init n (fun _ -> Random.float 2.0 -. 1.0) in*)
      let xsa = chebZeros n in
      let xs = Array.to_list xsa in
      let iTree = prune_tree (dissect1D xs (-1.0) 1.0) p in
      let ps = Array.of_list (breadthFirstLeafData iTree) in
      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
      let hss_spine = to_hss_spine iTree in
      let skew = skewness hss_spine in
      Printf.printf "n=%i p=%i skew=%g " n p skew;
      let hss = fromFn hss_spine (kernel1D_nosc alpha xs) tol in
      let b = randomMatrix n nrhs
      and x = rawMatrix n nrhs in
      let bc = Blop.copy b
      and hssc = Hss.copy hss in
      let t1 = Sys.time () in
      fastSolve' hssc bc x;
      let tot = Sys.time () -. t1 in
      let r = rawMatrix n nrhs in
      fastMul hss x r;
      let berr = norm1 (r -$ b) /. (float n *. norm1 x +. norm1 b) in
      Printf.printf "time=%g berr=%g" tot berr;
      print_newline ()
    done

  (* Timings for fast sparse HSS solver paper. *)
  let time8 () =
    print_endline "1D Kernel timings";
    let ns = [| 256; 512; 1024; 2048; 4096; 8192; 8192*2; 8192*4; 8192*8;
		8192*16; 8192*32; 8192*64; 8192*128; 8192*256 |]
    and nrhs = 1
    and alpha = 0.5 in
    for ni = 0 to Array.length ns - 1 do
      let n = ns.(ni)
      and tol = sqrt machine_epsilon in
      let p = 5 + int_of_float (log (float n) /. log 2.) in
      (*let xsa = Array.init n (fun _ -> Random.float 2.0 -. 1.0) in*)
      let xsa = chebZeros n in
      let xs = Array.to_list xsa in
      let iTree = prune_tree (dissect1D xs (-1.0) 1.0) p in
      let ps = Array.of_list (breadthFirstLeafData iTree) in
      let xs = Array.init n (fun i -> xsa.(ps.(i)-1)) in
      let hss_spine = to_hss_spine iTree in
      let skew = skewness hss_spine in
      Printf.printf "n=%i p=%i skew=%g " n p skew;
      let hss = fromFn hss_spine (kernel1D_nosc alpha xs) tol in
      let b = randomMatrix n nrhs in
      let t1 = Sys.time () in
      let x = fastSparseSolve hss b in
      let tot = Sys.time () -. t1 in
      let r = rawMatrix n nrhs in
      fastMul hss x r;
      let berr = norm1 (r -$ b) /. (float n *. norm1 x +. norm1 b) in
      Printf.printf "time=%g berr=%g" tot berr;
      print_newline ()
    done

  let timers = [| time1; time2; time3; time4; time5; time6; time7; time8 |]

  let run_timers () =
    let first_timer = 8
    and last_timer = 10 in
    (* Random.self_init (); *)
    for i = max 0 (first_timer - 1) to
      min (Array.length timers) last_timer - 1 do
	timers.(i) ()
    done;
    print_newline ()
      
end;; (* module Timer *)


module Dtests = Timer (DBlop) (DSparse) (DHss);;
module Stests = Timer (SBlop) (SSparse) (SHss);;
module Ztests = Timer (ZBlop) (ZSparse) (ZHss);;
module Ctests = Timer (CBlop) (CSparse) (CHss);;

let main () =
  print_string "Double Precision Times:\n";
  Dtests.run_timers ();
  print_newline ();

  print_string "Single Precision Times:\n";
  Stests.run_timers ();
  print_newline ();

  print_string "Double Complex Times:\n";
  Ztests.run_timers ();
  print_newline ();

  print_string "Single Complex Times:\n";
  Ctests.run_timers ();;

main ();;
